(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

(* opens Ocaml *)
open String
open Printf
open Format

(* opens Coq Stdlib *)
open Datatypes
open BinNums

(* opens CompCert *)
open Camlcoq
open Errors
open RTL
open Maps
open Heaps
open Registers
open Values

(** Converts a "de-augmented node" to a string.
   -1 = EN (unique entry node)
    0 = EX (unique exit node)
   >0 = string_of_int
**)
let toDeaugString (i : int) : string =
  if i = -1 then "EN" else if i = 0 then "EX" else (string_of_int i)

(** "De-augments" a node (node - 2) and converts it to a string.
   -1 = EN (unique entry node)
    0 = EX (unique exit node)
**)
let deaugPosToString (p : positive) : string =
  toDeaugString (Camlcoq.P.to_int p - 2)

(* Fonction auxiliaire *)
let rec split_aux (c : char) (s : string) : string list =
  try
    let i = index s c in
      (sub s 0 i) :: (split_aux c (sub s i (String.length s - i)))
  with
    | Not_found -> [s]

let split (c : char) (s : string) : string list =
  split_aux c s

(**** Fonctions de conversion vers String ****)

let rec customListToString_aux (separator : string) (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> separator ^ (aToString e) ^ (customListToString_aux separator aToString r)

(** Conversion 'a list -> string **)
let customListToString (leftpar : string) (rightpar : string) (separator : string) (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> leftpar ^ rightpar
  | e :: r -> leftpar ^ (aToString e) ^ (customListToString_aux separator aToString r) ^ rightpar

let rec listToString_aux (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> ", " ^ (aToString e) ^ (listToString_aux aToString r)

(** Conversion 'a list -> string **)
let listToString (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> "[]"
(*  | [e] -> "[" ^ (aToString e) ^ "]"*)
  | e :: r -> "[" ^ (aToString e) ^ (listToString_aux aToString r) ^ "]"

(* Version qui n'ajoute pas de virgules et utilise '(' à la place de '[' *)
let rec cleanListToString' (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> " " ^ (aToString e) ^ (cleanListToString' aToString r)

let cleanListToString (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> "[]"
  | [e] -> "(" ^ (aToString e) ^ ")"
  | e :: r -> "(" ^ (aToString e) ^ (cleanListToString' aToString r) ^ ")"

(** Conversion 'a * 'b -> string **)
let pairToString (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  "(" ^ (aToString (fst p)) ^ ", " ^ (bToString (snd p)) ^ ")"

let customPairToString (leftpar : string) (separator : string) (rightpar : string) (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  leftpar ^ (aToString (fst p)) ^ separator ^ (bToString (snd p)) ^ rightpar

(** Conversion 'a * 'b -> string (format DOT) **)
let pairToDotString (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  (aToString (fst p)) ^ " -> " ^ (bToString (snd p)) ^ ";\n"

(** Conversion 'a option -> string **)
let optionToString (aToString : 'a -> string) (o : 'a option) : string =
  match o with
    | None -> "None"
    | Some a -> "Some (" ^ (aToString a) ^ ")"

(** Conversion Datatypes.nat -> string **)
let natToString (n : nat) : string =
  string_of_int (Camlcoq.Nat.to_int n)

(** Conversion Coq.BinPos -> string **)
let binPosToString (p : positive) : string =
  Int32.to_string (Camlcoq.P.to_int32 p)
let (~$) = binPosToString

(** Conversion reg -> string
    (conversion d'un BinPos avec un 'x' au début) **)
let regToString (r : positive) : string =
  "x" ^ Int32.to_string (Camlcoq.P.to_int32 r)

(** Conversion regMem -> string
    (conversion d'un BinPos avec un 'x' au début, sauf pour la mémoire) **)
let regMemToString (r : positive) : string =
  if (Camlcoq.P.to_int32 r) = 1l then
    "Mem"
  else
    "x" ^ (Int32.to_string (Int32.sub (Camlcoq.P.to_int32 r) 1l))

(** Conversion PSet -> string **)
let pSetToString (pset : unit PTree.t) : string =
  listToString (fun p -> binPosToString (fst p)) (PTree.elements pset)

(** Conversion PSet -> string 'spéciale' **)
let pSetToSpecialString (pToString : positive -> string) (pset : unit PTree.t) : string =
  listToString (fun p -> pToString (fst p)) (PTree.elements pset)

(** Conversion Errors.errcode -> string **)
let errcodeToString (e : errcode) : string =
  match e with
  | MSG cl -> "MSG(" ^ (camlstring_of_coqstring cl) ^ ")"
  | CTX p -> "CTX(" ^ (binPosToString p) ^ ")"
  | POS p -> "POS(" ^ (binPosToString p) ^ ")"

let errorToString (el : errcode list) : string =
  listToString errcodeToString el

(** Conversion 'a Errors.res -> string **)
let resToString (aToString : 'a -> string) (r : 'a res) : string =
  match r with
    | Error msg -> "Error" ^ (listToString errcodeToString msg)
    | OK a -> "OK (" ^ (aToString a) ^ ")"

(** Convertion L.t -> string **)
type l_t = unit PTree.t option
let lToString (l : l_t) : string =
  optionToString pSetToString l

(** Conversion 'a PTree.t -> string **)
let pTreeToString (aToString : 'a -> string) (t : 'a PTree.t) : string =
  let els = PTree.elements t in
(*  let compare p1 p2 = (Int32.to_int (Camlcoq.camlint_of_positive (fst p1))) - (Int32.to_int (Camlcoq.camlint_of_positive (fst p2))) in
  let sorted_els = List.sort compare els in*)
  "{" ^ (listToString (fun p -> (binPosToString (fst p)) ^ " -> " ^ (aToString (snd p))) els) ^ "}"

(** Conversion 'a PMap.t -> string **)
let pMapToString (aToString : 'a -> string) (m : 'a PMap.t) : string =
  "PMap:<" ^ (aToString (fst m)) ^ "," ^ (pTreeToString aToString (snd m)) ^ ">"

(** Conversion PHeap.t -> string **)
let rec pHeapToString_aux (h : PHeap.R.heap) : string =
  match h with
  | PHeap.R.Empty -> ""
  | PHeap.R.Node(l,x,r) -> (pHeapToString_aux l) ^ (if l = PHeap.R.Empty then "" else ", ") ^ (binPosToString x) ^ (if r = PHeap.R.Empty then "" else ", ") ^ (pHeapToString_aux r)

let pHeapToString (h : PHeap.R.heap) : string =
"[>" ^ (pHeapToString_aux h) ^ "|]"

(** Conversion compute_postdoms -> string **)
let postdomsToString (m : unit PTree.t option PMap.t) : string =
  pMapToString (optionToString pSetToString) m

(** Conversion 'a array.t -> string **)
let arrayToString (aToString: 'a -> string) (a : 'a array) : string =
  Array.fold_left (fun s0 e -> s0 ^ (aToString e) ^ "; ") "" a

let gotoNextToString (p : positive) (d : positive) : string =
  if (Camlcoq.P.to_int32 d) = Int32.sub (Camlcoq.P.to_int32 p) (Int32.of_int 1) then "" else "; goto " ^ (binPosToString d)

(** Convertit une instruction RTL (format différent de celui de PrintRTL ; n'utilise pas de pretty-printer) vers une string **)
let rtlInstructionToString (p : positive) (regStringFun : positive -> string) (i : RTL.instruction) =
  match i with
  | Inop s -> "nop" ^ (gotoNextToString p s)
  | Iop(op, args, res, s) -> (regStringFun res) ^ " := op" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Iload(chunk, addr, args, dst, s) ->
      (regStringFun dst) ^ " := load" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Istore(chunk, addr, args, src, s) ->
      "store(" ^ (regStringFun src) ^ ", " ^ (listToString regStringFun args) ^ ")" ^ (gotoNextToString p s)
  | Icall(sg, fn, args, res, s) ->
      (regStringFun res) ^ " := call" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Itailcall(sg, fn, args) ->
      "tailcall" ^ (listToString binPosToString args)
  | Ibuiltin(ef, args, res, s) ->
      (binPosToString res) ^ " := builtin" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Icond(cond, args, s1, s2) ->
      "if cond" ^ (listToString regStringFun args) ^ " goto " ^ (binPosToString s1) ^ " else goto " ^ (binPosToString s2)
  | Ijumptable(arg, tbl) ->
      "jumptable: cond[" ^ (regStringFun arg) ^ "] -> jumps " ^ (listToString binPosToString tbl)
  | Ireturn None ->
      "return"
  | Ireturn (Some arg) ->
      "return " ^ (regStringFun arg)

let valToString (v : coq_val) : string =
  match v with
    | Vint n -> sprintf "Vint(%ld)" (camlint_of_coqint n)
    | Vlong n -> sprintf "Vlong(%ld)" (camlint_of_coqint n)
    | Vfloat f -> sprintf "Vfloat(%F)" (camlfloat_of_coqfloat f)
    | Vsingle f -> sprintf "Vsingle(%F)" (camlfloat_of_coqfloat32 f)
    | Vptr(b, ofs) -> sprintf "Vptr(?)"
    | Vundef -> "Vundef"

(** Conversion Regset.t -> string **)
let regsetToString (s : Regset.t) : string =
  listToString regMemToString (Regset.elements s)

(** Conversion regset (val Regmap.t) -> string **)
let regmapToString (s : coq_val PMap.t) : string =
  pMapToString valToString s

(** Conversion string -> Errors.Error **)
let stringToError (msg : string) : 'a res =
  Error [MSG (coqstring_of_camlstring msg)]

(** Conversion d'une fonction vers une string, en spécifiant le domaine à afficher **)
let fToString (aToString : 'a -> string) (bToString : 'b -> string) (dom : 'a list) (f : 'a -> 'b) : string =
  listToString (fun s -> s) (List.map (fun a -> (aToString a) ^ " -> " ^ (bToString (f a))) dom)

(* Fonction d'arité 2 *)
let f2ToString (aToString : 'a -> string) (bToString : 'b -> string) (cToString : 'c -> string) (doma : 'a list) (domb : 'b list) (f : 'a -> 'b -> 'c) : string =
  listToString (fun s -> s) (List.map2 (fun a b -> (pairToString aToString bToString (a, b)) ^ " -> " ^ (cToString (f a b))) doma domb)

let nToString (n : coq_N) : string =
  match n with
    | N0 -> "0"
    | Npos p -> Z.to_string (Zpos p)

let compare_pos p1 p2 = match BinPos.Pos.compare p1 p2 with | Lt -> -1 | Eq -> 0 | Gt -> 1

let compare_pos_rev p1 p2 = compare_pos p2 p1

let compare_pos_pair (p1 : positive * 'a) (p2 : positive * 'b) =
  compare_pos (fst p1) (fst p2)

let compare_pos_pair_rev (p1 : positive * 'a) (p2 : positive * 'b) =
  compare_pos_pair p2 p1

let countersToString (cs : nat PMap.t) : string =
  let els = List.sort compare_pos_pair_rev (PTree.elements (snd cs)) in
  customListToString "[" "]" "," (fun (node, counter) -> sprintf "%s#%s" (binPosToString node) (natToString counter)) els

(** Conversion 'a PMap.t -> string **)
let pMapToString (aToString : 'a -> string) (m : 'a PMap.t) : string =
  "PMap:<" ^ (aToString (fst m)) ^ "," ^ (pTreeToString aToString (snd m)) ^ ">"
