(** This module offers an abstract domain of polyhedra
which takes as inputs linear integer expressions ([LinExpr.ZExpr]). *)

Require String.
Require Import Itv.
Require Import PolC.

Import AssignD.
Import ZCond.

Module CorePolLinZ.

  Definition t:= PolLinQ.t.

  Definition gamma (a:t) m
    := PolLinQ.gamma a (Mem.lift ZtoQ.ofZ m).

(* NO MORE NEEDED FOR BoundVarD !
  Instance mdo: MayDependOn := {| mayDependOn:=PolLinQ._mayDependOn |}.
  Hint Immediate PolLinQ.gamma_mdo: polC.

  Instance gamma_mdo: mdoExt gamma iff.
  Proof.
    assert (mdoExt gamma Program.Basics.impl); auto with progvar.
    constructor 1.
    - unfold gamma, Basics.impl. intros; erewrite eval_mdo_compat; eauto with polC.
    unfold bExt, Mem.lift in * |- *. intros; rewrite H; auto.
    - unfold gamma, Basics.impl. intros; erewrite eval_diff_compat; eauto with polC.
    unfold bExt, Mem.lift in * |- *. intros; rewrite H0; auto.
  Qed.
*)
   Instance gamma_ext: Proper (Logic.eq ==> pointwise_relation PVar.t Logic.eq ==> iff) gamma.
   Proof.
     unfold gamma.
     intros x y H m1 m2 H0. subst; rewrite H0; tauto.
   Qed.

  Definition isIncl:=PolLinQ.isIncl.
  Global Opaque PolLinQ.isIncl.

  Lemma isIncl_correct: forall a1 a2, If isIncl a1 a2 THEN forall m, gamma a1 m -> gamma a2 m.
  Proof.
    unfold gamma. VPLAsimplify.
  Qed.

  Definition top:= PolLinQ.top.

  Lemma top_correct: forall m, gamma top m.
    simpl; auto.
  Qed.
  Hint Resolve top_correct: vpl.

  Definition bottom: t := PolLinQ.bottom.

  Lemma bottom_correct: forall m, ~(gamma bottom m).
  Proof.
    simpl; auto.
  Qed.

  Definition isBottom : t -> imp bool
    := fun p => PolLinQ.isBottom p.

  Lemma isBottom_correct : forall a, If isBottom a THEN forall m, gamma a m -> False.
  Proof.
    unfold isBottom.
    unfold gamma.
    VPLAsimplify.
    intro.
    eapply (PolLinQ.isBottom_correct a true Hexta).
  Qed.

  Definition widen:= PolLinQ.widen.

  Definition join:= PolLinQ.join.

  Lemma join_correct a1 a2: WHEN p <- join a1 a2 THEN forall m, gamma a1 m \/ gamma a2 m -> gamma p m.
  Proof.
    unfold gamma; VPLAsimplify.
  Qed.

  Definition assume (c:ZCond.t) (a:t) : imp t :=
    PolLinQ.assume (QCond.lift c) a.
  
  Lemma assume_correct: forall c a, WHEN p <- assume c a THEN forall m, sat c m -> gamma a m -> gamma p m.
  Proof.
    unfold assume, gamma. xasimplify ltac:idtac.
    intros; eapply PolLinQ.assume_correct; eauto.
    autorewrite with vpl; auto.
  Qed.

  Definition project (a:t) (x:PVar.t) : imp t :=
    PolLinQ.project a x.

  Lemma project_correct: forall a x, WHEN p <- project a x THEN forall m v, gamma a m -> gamma p (Mem.assign x v m).
  Proof.
    unfold project, gamma. xasimplify ltac:idtac. 
    intros; eapply PolLinQ.gamma_ext; [ auto | idtac | eapply PolLinQ.project_correct; eauto].
    autorewrite with vpl; eauto.
  Qed.

  Definition rename (x y: PVar.t) (a:t): imp t :=
    PolLinQ.rename x y a.

  Lemma rename_correct: forall x y a, WHEN p <- (rename x y a) THEN
    forall m, gamma a (Mem.assign x (m y) m) -> gamma p m.
  Proof.
    intros x y a.
    unfold rename, gamma. xasimplify ltac:idtac. 
    intros; eapply PolLinQ.rename_correct; eauto.
    eapply PolLinQ.gamma_ext; [ auto | idtac | eauto].
    autorewrite with vpl; auto.
  Qed.

  Definition pr: t -> String.string
    := fun ab => PolLinQ.pr ab.

End CorePolLinZ.


Module PolZItv <: HasGetItv ZNum ZTerm ZItv CorePolLinZ.

  Import CorePolLinZ.

  Definition get_itv: ZTerm.t -> t -> imp ZItv.t
    := fun a ab =>
      let qa := QTerm.lift a in
        BIND qitv <- PolLinQ.get_itv qa ab -;
        pure (proj1_sig (ZItv.fromQItv qitv)).

  Lemma get_itv_correct: forall te a, WHEN i <- (get_itv te a) THEN
    forall m, gamma a m -> ZItv.sat i (ZTerm.eval te m).
  Proof.
    unfold get_itv, gamma. xasimplify ltac:idtac. 
    intros; apply (proj2_sig (ZItv.fromQItv _)).
    rewrite <- QTerm.lift_correct.
    eapply PolLinQ.get_itv_correct; eauto.
  Qed.

End PolZItv.

Module ZAtomicRequire <: HasAtomicRequire ZNum ZCond CorePolLinZ.

Import CorePolLinZ.

Definition atomicRequire (cmp:AffCond.cmpT) (tl tr:ZTerm.t) (a:t) : imp (option bool) :=
  PolLinQ.atomicRequire cmp (QTerm.lift tl) (QTerm.lift tr) a.

Local Hint Resolve PolLinQ.atomicRequire_correct: vpl.

Local Opaque ZCond.sat QCond.sat.

Lemma atomicRequire_correct: 
  forall t1 t2 cmp a, WHEN o <- atomicRequire cmp t1 t2 a THEN
                      (WHEN b <- o THEN If b THEN forall m, gamma a m -> ZCond.sat (ZCond.Atom cmp t1 t2) m)%option.
Proof.
  VPLAsimplify.
  intros.
  rewrite <- QCond.lift_correct.
  simpl; auto.
Qed.

End ZAtomicRequire.


Module PolLinZ <: AbstractDomain ZNum ZCond 
  := CorePolLinZ <+ ZAtomicRequire <+ AffCondRequire ZNum ZCond <+ PolZItv.

Module FullPolLinZ <: FullItvAbstractDomain ZNum ZCond ZItv
  := MakeAssign ZNum ZCond PolLinZ PolLinZ <+ AssignLiftItv ZNum ZCond ZItv PolLinZ PolLinZ PolZItv 
                                           <+ AssignSimplePrinting ZNum ZCond PolLinZ PolLinZ PolLinZ
  .


