(** This module implements the [AbIdealEnv.ab_ideal_env] interface
of abtract domains on ideal numbers. *)

Require Import ZArith_base.
Require PolZ.
Import NumC ProgVar Itv.
Require Import AdomLib Util.
Require Import IdealExpr.
Require Import IdealIntervals.
Require Import AbIdealEnv.
Require Import String.
Require Import ToString.

Local Open Scope Z_scope.

Local Notation zmemT := (Mem.t ZNum.t).
Local Notation var := positive.

Definition liftN (dummy: ZNum.t) (n: ideal_num): ZNum.t
  := match n with
         | INf _ => dummy
         | INz z => z
     end.

Definition liftRho (rho: var -> ideal_num) (dummy: ZNum.t): zmemT
  := fun x => liftN dummy (rho x).

Definition cmpTrans : Integers.comparison -> PolZ.Cmp.t
  := fun cmp =>
       match cmp with
         | Integers.Ceq => PolZ.Cmp.Eq
         | Integers.Cne => PolZ.Cmp.Neq
         | Integers.Clt => PolZ.Cmp.Lt
         | Integers.Cle => PolZ.Cmp.Le
         | Integers.Cgt => PolZ.Cmp.Gt
         | Integers.Cge => PolZ.Cmp.Ge
       end.

(* XXX: revamp the proof *)
Lemma CmpTransTrue : forall (cmp : Integers.comparison),
                   forall z1 z2 : Z,
                     ZIntervals.Zcmp cmp z1 z2 = true -> PolZ.Cmp.denote (cmpTrans cmp) z1 z2.
Proof.
  intros cmp z1 z2 hyp.
  assert (fact : match z1 ?= z2 with
                   | Eq => z1 = z2
                   | Lt => z1 < z2
                   | Gt => z1 > z2
                 end).
  apply Zcompare_elim; trivial.
  destruct cmp; simpl; simpl in hyp; destruct (z1 ?= z2);
  try discriminate; try assumption.
  intro contr.
  subst.
  apply Z.lt_irrefl in fact.
  contradiction.
  intro contr.
  subst.
  apply Z.gt_lt in fact.
  apply Z.lt_irrefl in fact.
  contradiction.
  subst.
  apply Z.le_refl.
  apply Z.lt_le_incl.
  assumption.
  apply Z.gt_lt in fact.
  assumption.
  subst.
  apply Z.le_refl.
  apply Z.gt_lt in fact.
  apply Z.lt_le_incl in fact.
  assumption.
Qed.

(* XXX: revamp the proof *)
Lemma CmpTransFalse : forall (cmp: Integers.comparison),
                      forall z1 z2 : Z,
                        ZIntervals.Zcmp cmp z1 z2 = false ->
                        ~ PolZ.Cmp.denote (cmpTrans cmp) z1 z2.
Proof.
  intros cmp z1 z2 hyp contr.
  assert (fact : match z1 ?= z2 with
                   | Eq => z1 = z2
                   | Lt => z1 < z2
                   | Gt => z1 > z2
                 end).
  apply Zcompare_elim; trivial.
  destruct cmp; simpl in hyp; simpl in contr; destruct (z1 ?= z2); try discriminate.
  subst.
  apply Z.lt_irrefl in fact.
  contradiction.
  subst.
  apply Z.gt_lt, Z.lt_irrefl in fact.
  contradiction.
  contradiction (contr fact).
  subst.
  apply Z.lt_irrefl in contr.
  contradiction.
  apply Z.lt_asymm in contr.
  apply Z.gt_lt in fact.
  contradiction (contr fact).
  apply Z.gt_lt in fact.
  apply Zle_lt_or_eq in contr.
  destruct contr.
  apply (Z.lt_trans z1 z2 z1) in fact.
  apply Z.lt_irrefl in fact.
  contradiction.
  assumption.
  subst.
  apply Z.lt_irrefl in fact.
  contradiction.
  subst.
  apply Z.lt_irrefl in contr.
  contradiction.
  apply (Z.lt_trans z1 z2 z1) in contr.
  apply Z.lt_irrefl in contr.
  contradiction.
  assumption.
  apply Zlt_not_le in fact.
  contradiction (fact contr).
Qed.

Lemma CmpTransPreservation : forall (cmp: Integers.comparison) (z1 z2 : Z),
                               (if ZIntervals.Zcmp cmp z1 z2 then 1 else 0) =
                               (if PolZ.Cmp.dec (cmpTrans cmp) z1 z2 then 1 else 0).
Proof.
  intros cmp z1 z2.
  assert (fact1 := CmpTransTrue cmp z1 z2).
  assert (fact2 := CmpTransFalse cmp z1 z2).
  destruct (ZIntervals.Zcmp cmp z1 z2);
  destruct (PolZ.Cmp.dec (cmpTrans cmp) z1 z2).
  reflexivity.
  contradiction n. apply fact1; reflexivity.
  refine (match fact2 _ d with end). reflexivity.
  reflexivity.
Qed.

Definition ItvTrSpec (i1: ZIntervals.itv+⊤) (i2: ZItv.t): Prop
  := forall z: Z.t, γ i1 z <-> ZItv.sat i2 z.

Definition itvTr (i1: ZIntervals.itv+⊤): {i2: ZItv.t | ItvTrSpec i1 i2}.
  refine (
      match i1
            return {i2: ZItv.t | ItvTrSpec i1 i2}
      with
        | Just (ZIntervals.ITV z1 z2) =>
          exist _ (proj1_sig (ZItv.mk (ZItv.Closed z1) (ZItv.Closed z2))) _
        | All =>
          exist _ (proj1_sig (ZItv.mk ZItv.Infty ZItv.Infty)) _
      end);
  unfold ItvTrSpec;
  intro z;
  split;
  intro h;
  simpl; try easy.
  - unfold ZIntervals.gamma in h.
    destruct (ZNum.ltLeDec z2 z1) as [h1|h1].
    + assert (h2 := h1). (* duplicate because h1 is used in goal *)
      apply Zlt_not_le in h2.
      destruct h2.
      apply Z.le_trans with (m := z);
        apply h.
    + simpl.
      assumption.
  - simpl in h.
    destruct (ZNum.ltLeDec z2 z1) as [h1|h1].
    + destruct h.
    + simpl in h.
      assumption.
Defined.

Fixpoint iexprTr (ie : iexpr var) : PolZ.Expr.t
  := match ie with
       | IEvar INTz x => PolZ.Expr.Evar x
       | IEvar INTf x => PolZ.Expr.Eunknown
       | IEZconst i => PolZ.Expr.Eitv (proj1_sig (itvTr i))
       | IEFconst _ => PolZ.Expr.Eunknown
       | IEunop op ie' =>
         let e' := iexprTr ie' in
         match op with
           | IOneg => PolZ.Expr.Eunop PolZ.Expr.Oneg e'
           | IOnot => PolZ.Expr.Eunop PolZ.Expr.Onot e'
           | _ => PolZ.Expr.Eunknown
         end
       | IEbinop op ie1 ie2 =>
         let e1 := iexprTr ie1 in
         let e2 := iexprTr ie2 in
         match op with
           | IOadd => PolZ.Expr.Ebinop PolZ.Expr.Oadd e1 e2
           | IOsub => PolZ.Expr.Ebinop PolZ.Expr.Osub e1 e2
           | IOmul => PolZ.Expr.Ebinop PolZ.Expr.Omul e1 e2
           | IOdiv => PolZ.Expr.Ebinop PolZ.Expr.Odiv e1 e2
           | IOmod => PolZ.Expr.Ebinop PolZ.Expr.Omod e1 e2
           | IOand => PolZ.Expr.Ebinop PolZ.Expr.Oand e1 e2
           | IOor => PolZ.Expr.Ebinop PolZ.Expr.Oor e1 e2
           | IOxor => PolZ.Expr.Ebinop PolZ.Expr.Oxor e1 e2
           | IOshl => PolZ.Expr.Ebinop PolZ.Expr.Oshl e1 e2
           | IOshr => PolZ.Expr.Ebinop PolZ.Expr.Oshr e1 e2
           | IOcmp c => PolZ.Expr.Ebinop (PolZ.Expr.Ocmp (cmpTrans c)) e1 e2
           | _ => PolZ.Expr.Eunknown
         end
     end.

Lemma IExprTrPreservationInF :
  forall (ie : iexpr var) (m : var -> ideal_num) (f: Floats.float),
    eval_iexpr m ie (INf f) -> iexprTr ie = PolZ.Expr.Eunknown.
Proof.
  intros ie m f hyp.
  induction ie; inversion hyp; subst.
  - destruct i; reflexivity || inversion H3.
  - reflexivity.
  - destruct i; reflexivity || inversion H3.
  - destruct i; reflexivity || inversion H5.
Qed.

Lemma IExprTrPreservationInZ :
  forall (ie : iexpr var) (m : var -> ideal_num) (z : Z) (dummy : Z),
    eval_iexpr m ie (INz z) -> PolZ.Expr.denote (liftRho m dummy) (iexprTr ie) z.
Proof.
  intros ie m z dummy.
  revert z.
  induction ie; intros z hyp; inversion hyp; subst.
  - destruct i; [|inversion H3].
    constructor.
    unfold liftRho.
    rewrite H1.
    reflexivity.
  - constructor.
    apply (proj2_sig (itvTr t)).
    assumption.
  - destruct n; [inversion H3|].
    + constructor.
    + destruct i; simpl; try constructor.
      * apply PolZ.Expr.EvalEunop with (z1 := z0).
        apply IHie.
        assumption.
        inversion H3.
        constructor.
      * apply PolZ.Expr.EvalEunop with (z1 := z0).
        apply IHie.
        assumption.
        inversion H3.
        constructor.
  - destruct n1.
    inversion H5.
    constructor.
    destruct n2.
    inversion H5.
    destruct i; simpl; try constructor;
    apply PolZ.Expr.EvalEbinop with (z1 := z0) (z2 := z1);
    try (apply IHie1; assumption) || (apply IHie2; assumption);
    inversion H5;
    try constructor.
    assumption.
    assumption.
    rewrite CmpTransPreservation.
    constructor.
Qed.

Lemma IExprTrPreservation :
  forall (dummy : Z) (ie : iexpr var) (m : var -> ideal_num) (n : ideal_num),
    eval_iexpr m ie n -> PolZ.Expr.denote (liftRho m dummy) (iexprTr ie) (liftN dummy n).
Proof.
  intros dummy ie m n hyp.
  destruct n as [f|z].
  replace (iexprTr ie) with PolZ.Expr.Eunknown.
  constructor.
  symmetry.
  apply IExprTrPreservationInF with (m := m) (f := f).
  assumption.
  apply IExprTrPreservationInZ.
  assumption.
Qed.

Definition A: Type
  := PolZ.AbDom.A.

Definition B: Type
  := var -> ideal_num.

Definition gamma: A -> B -> Prop
  := fun ab rho => PolZ.AbDom.gamma ab (liftRho rho ZNum.z).

Instance PolGammaOp: gamma_op A B
  := {
      γ := gamma
     }.

Lemma gamma_monotone: forall (ab1:A*in_chan) (ab2: A) (rho:B),
                        PolZ.AbDom.leb (fst ab1) ab2 = true ->
                        γ ab1 rho -> gamma ab2 rho.
  unfold gamma.
  intros ab1 ab2 h rho [H _]. revert H.
  apply PolZ.AbDom.gamma_monotone.
  assumption.
Qed.

Lemma gamma_top: forall x: B, gamma PolZ.AbDom.top x.
  unfold gamma.
  intro x.
  generalize (liftRho x ZNum.z).
  apply PolZ.AbDom.gamma_top.
Qed.

Lemma join_sound: forall (ab1 ab2: A*in_chan) (chan:in_chan) (x: B),
    γ ab1 x \/ γ ab2 x -> γ chan x ->
    γ (PolZ.AbDom.join (fst ab1) (fst ab2)) x.
  intros ab1 ab2 chan x H _.
  destruct H as [[]|[]];
    apply PolZ.AbDom.join_sound;
    [left|right];
    assumption.
Qed.

Definition ItvConvSpec (i1: ZItv.t) (i2: ZIntervals.itv+⊤+⊥): Prop
  := forall z: BinInt.Z, ZItv.sat i1 z -> γ i2 z.

Definition itvConv (i1: ZItv.t): {i2: ZIntervals.itv+⊤+⊥ | ItvConvSpec i1 i2}.
  refine (
      match i1
            return {i2: ZIntervals.itv+⊤+⊥ | ItvConvSpec i1 i2}
      with
        | ZItv.Bot => exist _ Bot _
        | ZItv.NotBot low up pfi1 =>
          match low
                  return forall pf: ZItv.BndOk low up,
                           {i2: ZIntervals.itv+⊤+⊥ | ItvConvSpec (ZItv.NotBot low up pf) i2}
          with
            | ZItv.Infty => fun pfl => exist _ (NotBot All) _
            | ZItv.Open b1 =>
              fun pfl =>
                match up
                      return forall pf: ZItv.BndOk (ZItv.Open b1) up,
                               {i2: ZIntervals.itv+⊤+⊥ |
                                ItvConvSpec (ZItv.NotBot (ZItv.Open b1) up pf) i2}
                with
                  | ZItv.Infty => fun pfu => exist _ (NotBot All) _
                  | ZItv.Open b2
                  | ZItv.Closed b2 =>
                    fun pfu => exist _ (AdomLib.NotBot (Just (ZIntervals.ITV b1 b2))) _
                end pfl
            | ZItv.Closed b1 =>
              fun pfl =>
                match up
                      return forall pf: ZItv.BndOk (ZItv.Closed b1) up,
                               {i2: ZIntervals.itv+⊤+⊥ |
                                ItvConvSpec (ZItv.NotBot (ZItv.Closed b1) up pf) i2}
                with
                  | ZItv.Infty => fun pfu => exist _ (NotBot All) _
                  | ZItv.Open b2
                  | ZItv.Closed b2 =>
                    fun pfu => exist _ (NotBot (Just (ZIntervals.ITV b1 b2))) _
                end pfl
          end pfi1
      end);
  unfold ItvConvSpec;
  intros z h;
  unfold γ;
  simpl;
  match goal with
    | |- γ All z => constructor
    | _ => idtac
  end;
  unfold γ;
  simpl;
  unfold ZItv.sat, ZItv.satLower, ZItv.satUpper in h;
  match goal with
    | h: False |- False => destruct h
    | _ => idtac
  end.
  - split;
      apply Zlt_le_weak, h.
  - split;
    [apply Zlt_le_weak|idtac];
    apply h.
  - split;
    [idtac|apply Zlt_le_weak];
    apply h.
  - split;
    apply h.
Defined.

Definition get_itv: iexpr var -> A -> IdealIntervals.abs+⊤+⊥
  := fun e ab =>
       match iexpr_ty e with
         | INTz =>
           let e' := iexprTr e in
           match proj1_sig (itvConv (PolZ.AbDom.get_itv e' ab)) with
             | Bot => Bot
             | NotBot (Just itv) => NotBot (Just (Az itv))
             | NotBot All => NotBot All
           end
         | INTf => NotBot All
       end.

Lemma get_itv_correct: forall (e: iexpr var) (rho: B) (ab: A),
                         gamma ab rho ->
                         (eval_iexpr rho e) ⊆ γ (get_itv e ab).
  simpl.
  intros e rho ab hab i h.
  unfold get_itv.
  pose proof (iexpr_ty_correct h). destruct H. 2:constructor.
  assert (pfe := IExprTrPreservation ZNum.z e).
  destruct (itvConv (PolZ.AbDom.get_itv (iexprTr e) ab)) as [x pfx].
  simpl.
  unfold ItvConvSpec in pfx.
    destruct x as [|[]].
    + apply pfx with (z := i).
      apply PolZ.AbDom.get_itv_correct with (rho := liftRho rho 0).
      assumption.
      apply pfe with (n := INz i).
      assumption.
    + exact I.
    + unfold γ.
      simpl.
      unfold γ.
      simpl.
      apply pfx.
      apply PolZ.AbDom.get_itv_correct with (rho := liftRho rho 0).
      assumption.
      apply pfe with (n := INz i).
      assumption.
Qed.

Definition propagateBot : A -> A+⊥ :=
  fun ab => if PolZ.AbDom.isBottom ab then Bot else NotBot ab.

Lemma propagateBot_correct : forall (ab : A) (m : B),
                               gamma ab m <-> lift_gamma gamma (propagateBot ab) m.
Proof.
  intros ab m.
  split.
  - intro hyp.
    unfold lift_gamma.
    unfold propagateBot.
    assert (pf := PolZ.AbDom.isBottom_correct ab).
    destruct (PolZ.AbDom.isBottom ab).
    apply pf with (m := liftRho m ZNum.z).
    reflexivity.
    assumption.
    assumption.
  - unfold lift_gamma.
    assert (pf := PolZ.AbDom.isBottom_correct ab).
    unfold propagateBot.
    destruct (PolZ.AbDom.isBottom ab).
    intro contr.
    destruct contr.
    intro h.
    assumption.
Qed.

Definition assume : iexpr var -> bool -> A * @in_chan var -> @in_chan var -> A+⊥
  := fun e b ab _ => propagateBot (PolZ.AbDom.assume (iexprTr e) b (fst ab)).

Lemma assume_correct: forall (c: iexpr var) (rho: B) (ab: A*in_chan) (b: bool) (chan:in_chan),
                        γ ab rho -> γ chan rho ->
                        eval_iexpr rho c (INz(if b then 1 else 0)) ->
                        lift_gamma gamma (assume c b ab chan) rho.
  intros e rho ab b chan [hab _] _ hc.
  unfold assume.
  simpl.
  apply propagateBot_correct.
  apply PolZ.AbDom.assume_correct.
  assumption.
  apply IExprTrPreservation with (n := INz (if b then 1 else 0)).
  assumption.
Qed.

Definition assign: var -> iexpr var -> A * @in_chan var -> @in_chan var -> A+⊥
  := fun x e ab _ => propagateBot (PolZ.AbDom.assign x (iexprTr e) (fst ab)).

Instance VarEqDec: EqDec var
  := { eq_dec := PosEqDec }.

Lemma assign_correct: forall (x: var) (e: iexpr var) (ρ: var -> ideal_num)
                             (n: ideal_num) (ab: A*in_chan) (c:in_chan),
                        γ ab ρ -> γ c (upd ρ x n) -> eval_iexpr ρ e n ->
                        γ (assign x e ab c) (upd ρ x n).
  unfold γ.
  unfold PolGammaOp.
  intros x e rho n ab c [h1 _] _ h2.
  unfold assign.
  assert (pf := PolZ.AbDom.assign_correct).
  specialize (pf x (iexprTr e) (liftRho rho 0) (liftN 0 n) (fst ab) h1
                 (IExprTrPreservation _ _ _ _ h2)).
  simpl.
  unfold γ.
  apply propagateBot_correct.
  unfold gamma.

  assert (pfm : forall x0, Mem.assign x (liftN 0 n) (liftRho rho 0) x0 =
                          liftRho (upd rho x n) ZNum.z x0).
  intro x0.
  unfold Mem.assign, upd, liftRho, liftN.
  destruct (eq_dec x0 x) as [eqx | eqx].
    subst.
    destruct (PVar.eq_dec x x) as [eq0 | eq0];
    destruct (eq_dec x x) as [eq1 | eq1].
    destruct n; reflexivity.
    contradiction (eq1 eq0).
    contradiction (eq0 eq1).
    destruct (rho x); reflexivity.

    destruct (PVar.eq_dec x x0) as [eq0 | eq0].
    symmetry in eq0. contradiction (eqx eq0).
    replace (if eq_dec x0 x then n else rho x0) with (rho x0).
    destruct (rho x0); reflexivity.
    destruct (eq_dec x0 x) as [eq1 | eq1].
    contradiction (eqx eq1).
    reflexivity.

  revert pf.
  generalize (PolZ.AbDom.assign x (iexprTr e) (fst ab)).
  intros ab' pfab'.
  eapply PolZ.AbDom.gamma_ext.
  reflexivity.
  symmetry. auto.
  assumption.
Qed.

Definition forget: var -> A * @in_chan var -> @in_chan var -> A+⊥
  := fun x ab _ =>
       propagateBot (PolZ.AbDom.project (fst ab) x).

Lemma forget_correct: forall (x: var) (ρ: var -> ideal_num)
                             (n: ideal_num) (ab: A * in_chan) (c:in_chan),
                        γ ab ρ -> γ c (upd ρ x n) ->
                        γ (forget x ab c) (upd ρ x n).
  intros x rho n ab c [h1 _] _.
  unfold forget.
  simpl.
  assert (pf := PolZ.AbDom.project_correct).
  specialize (pf (fst ab) x (liftRho rho 0) (liftN 0 n)).
  unfold γ.
  apply propagateBot_correct.
  unfold lift_gamma.
  unfold γ.
  unfold PolGammaOp.

  assert (pfm : forall x0, Mem.assign x (liftN 0 n) (liftRho rho 0) x0 =
                          liftRho (upd rho x n) ZNum.z x0).
  intro x0.
  unfold Mem.assign, upd, liftRho, liftN.
  destruct (eq_dec x0 x) as [eqx | eqx].
    subst.
    destruct (PVar.eq_dec x x) as [eq0 | eq0];
    destruct (eq_dec x x) as [eq1 | eq1].
    destruct n; reflexivity.
    contradiction (eq1 eq0).
    contradiction (eq0 eq1).
    destruct (rho x); reflexivity.

    destruct (PVar.eq_dec x x0) as [eq0 | eq0].
    symmetry in eq0. contradiction (eqx eq0).
    replace (if eq_dec x0 x then n else rho x0) with (rho x0).
    destruct (rho x0); reflexivity.
    destruct (eq_dec x0 x) as [eq1 | eq1].
    contradiction (eqx eq1).
    reflexivity.

  eapply PolZ.AbDom.gamma_ext.
  reflexivity.
  symmetry. auto.
  apply pf.
  assumption.
Qed.

Definition to_string: A -> string
  := fun p => PolZ.AbDom.pr p.

Instance PolToString: ToString A
  := {
      to_string := to_string
    }.

Definition extract_inchan (c:in_chan) (ab:A) : (A*in_chan) :=
  (ab, {| AbIdealEnv.get_itv expr :=
       match c.(AbIdealEnv.get_itv) expr, get_itv expr ab with
         | NotBot (Just (Az i1)), NotBot (Just (Az i2)) =>
           botlift_fun1 (@Just _ ∘ Az) (ZIntervals.meet i1 i2)
         | NotBot (Just (Af i1)), NotBot (Just (Af i2)) =>
           botlift_fun1 (@Just _ ∘ Af) (FloatIntervals.meet i1 i2)
         | NotBot All, NotBot ab | NotBot ab, NotBot All => NotBot ab
         | _, _ => Bot
       end;
          get_congr := c.(get_congr);
          get_eq_expr := c.(get_eq_expr) |}).

  Lemma extract_inchan_correct :
    forall chan ab ρ,
      ρ ∈ γ chan -> ρ ∈ γ ab -> ρ ∈ γ (extract_inchan chan ab).
  Proof.
    intros chan ab ρ H H0; split; auto.
    constructor.
    - intros. simpl.
      destruct H. specialize (get_itv_correct0 _ _ H1). destruct AbIdealEnv.get_itv. contradiction.
      pose proof (get_itv_correct _ _ _ H0 _ H1).
      destruct x as [|[]], (get_itv e ab) as [|[|[]]], a; auto; try contradiction;
      eapply botlift_fun1_spec; eauto.
      apply ZIntervals.meet_correct; eauto.
      apply FloatIntervals.meet_correct; eauto.
    - intros. apply H. auto.
    - intros. apply H. auto.
  Qed.

  Lemma extract_inchan_redundant :
    forall chan ab,
      γ chan ∩ γ (fst (extract_inchan chan ab)) ⊆ γ (snd (extract_inchan chan ab)).
  Proof. intros. destruct H. edestruct extract_inchan_correct; eauto. Qed.

Instance PolAbIdealEnv: ab_ideal_env (var := positive) A (H := PosEqDec)
  := {
      id_gamma := PolGammaOp;
      id_top c := extract_inchan c PolZ.AbDom.top;
      id_join x y c := extract_inchan c (PolZ.AbDom.join (fst x) (fst y));
      id_widen x y c := extract_inchan c (PolZ.AbDom.widen (fst x) (fst y));
      assign x e ab c := botlift_fun1 (extract_inchan c) (assign x e ab c);
      forget x ab c := botlift_fun1 (extract_inchan c) (forget x ab c);
      assume e b ab c := botlift_fun1 (extract_inchan c) (assume e b ab c)
    }.
Proof.
- intros. eapply extract_inchan_correct. auto.
  apply PolZ.AbDom.gamma_top.
- intros. eapply gamma_monotone. 2:eauto. apply H.
- intros. eapply extract_inchan_correct. auto. eapply join_sound; eauto.
- intros. eapply botlift_fun1_spec.
  intros. eapply extract_inchan_correct; eauto.
  apply assign_correct; auto.
- intros. eapply botlift_fun1_spec.
  intros. eapply extract_inchan_correct; eauto.
  apply forget_correct; auto.
- intros. eapply botlift_fun1_spec.
  intros. eapply extract_inchan_correct; eauto.
  apply assume_correct; auto.
- intros. apply extract_inchan_redundant. auto.
- intros. apply extract_inchan_redundant. auto.
- intros. apply extract_inchan_redundant. auto.
- unfold botlift_fun1, botbind. intros. destruct assign. now intuition.
  apply extract_inchan_redundant; auto.
- unfold botlift_fun1, botbind. intros. destruct forget. now intuition.
  apply extract_inchan_redundant; auto.
- unfold botlift_fun1, botbind. intros. destruct assume. now intuition.
  apply extract_inchan_redundant; auto.
Defined.
