(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8 Util.
Require Import AdomLib.
Require Import ZCongruences_defs.
Require Import IdealExpr IdealIntervals.

Set Implicit Arguments.
Unset Elimination Schemes.

Section var.

Context {var:Type}.

Record in_chan : Type :=
  { get_itv: iexpr var -> IdealIntervals.abs+⊤+⊥;
    get_congr: iexpr var → zcongr+⊥;
    get_eq_expr: var -> option (mux_iexpr var) }.

Record in_chan_gamma (chan:in_chan) (ρ:var->ideal_num) :=
  { get_itv_correct:
      ∀ e, eval_iexpr ρ e ⊆ γ (chan.(get_itv) e);
    get_congr_correct:
      ∀ e, (eval_iexpr ρ e ∘ INz) ⊆ γ (chan.(get_congr) e);
    get_eq_expr_correct:
      ∀ x e, chan.(get_eq_expr) x = Some e ->
             eval_mux_iexpr ρ e (ρ x) }.
Global Instance in_chan_gamma' : gamma_op in_chan (var -> ideal_num)
  := in_chan_gamma.

Definition in_chan_top : in_chan :=
  {| get_itv := fun e => NotBot All;
     get_congr e := top;
     get_eq_expr := fun _ => None |}.

Lemma in_chan_top_correct :
  ∀ ρ, ρ ∈ γ in_chan_top.
Proof.
  intros. constructor.
  - intros. simpl.
    pose proof (iexpr_ty_correct H). destruct iexpr_ty.
    inv H0. constructor.
    inv H0. constructor.
  - simpl. intros e a H. exists (-a). simpl. ring.
  - discriminate.
Qed.

Context `{EqDec var}.

Class ab_ideal_env (t:Type) : Type := {
  id_gamma:> gamma_op t (var->ideal_num);
  id_top: in_chan -> t * in_chan;
  id_leb: t * in_chan -> t -> bool;
  id_join: t * in_chan -> t * in_chan -> in_chan -> t * in_chan;
  id_widen: t * in_chan -> t * in_chan -> in_chan -> t * in_chan;
  assign: var -> iexpr var -> t * in_chan -> in_chan -> (t * in_chan)+⊥;
  forget: var -> t * in_chan -> in_chan -> (t *  in_chan)+⊥;
  assume: iexpr var -> bool -> t * in_chan -> in_chan -> (t * in_chan)+⊥;

  id_top_correct: forall chan ρ, ρ ∈ γ chan -> ρ ∈ γ (id_top chan);
  id_leb_correct: forall ab1 ab2 ρ,
    id_leb ab1 ab2 = true ->
    ρ ∈ γ ab1 ->
    ρ ∈ γ ab2;
  id_join_correct: forall ab1 ab2 chan ρ,
    ρ ∈ γ ab1 \/ ρ ∈ γ ab2 -> ρ ∈ γ chan ->
    ρ ∈ γ (id_join ab1 ab2 chan);
  assign_correct: forall x e ρ n ab chan,
    ρ ∈ γ ab -> (upd ρ x n) ∈ γ chan ->
    n ∈ eval_iexpr ρ e ->
    (upd ρ x n) ∈ γ (assign x e ab chan);
  forget_correct: forall x ρ n ab chan,
    ρ ∈ γ ab -> (upd ρ x n) ∈ γ chan ->
    (upd ρ x n) ∈ γ (forget x ab chan);
  assume_correct: forall c ρ ab b chan,
    ρ ∈ γ ab -> ρ ∈ γ chan ->
    (INz (if b:bool then 1 else 0)) ∈ eval_iexpr ρ c ->
    ρ ∈ γ (assume c b ab chan);

  chan_redundant (chan:in_chan) (ret:t*in_chan) :=
    γ chan ∩ γ (fst ret) ⊆ γ (snd ret);
  chan_redundant_bot (chan:in_chan) (ret:(t*in_chan)+⊥) :=
    γ chan ∩ γ (botlift_fun1 fst ret) ⊆ γ (botlift_fun1 snd ret);
  id_top_redundant : forall chan, chan_redundant chan (id_top chan);
  id_join_redundant : forall ab1 ab2 chan, chan_redundant chan (id_join ab1 ab2 chan);
  id_widen_redundant : forall ab1 ab2 chan, chan_redundant chan (id_widen ab1 ab2 chan);
  assign_redundant : forall x e ab chan, chan_redundant_bot chan (assign x e ab chan);
  forget_redundant : forall x ab chan, chan_redundant_bot chan (forget x ab chan);
  assume_redundant : forall c b ab chan, chan_redundant_bot chan (assume c b ab chan)
}.

End var.
