(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import String ToString Coqlib Utf8 Maps ShareTree Sets Util.
Require Import AdomLib.
Require Import IdealExpr AbIdealEnv.
Require Import Integers.
Require Import MSetRBT.
Require Import Hash.

Module Make(T:SHARETREE).

Module S <: SET with Definition elt := T.elt := Tree2Set(T).
Module SO := SetOp(S).

Definition var := T.elt.

Instance var_eq : EqDec var := T.elt_eq.

Section T.
  Context (MIX_var:N -> var -> N).

  Fixpoint iexpr_free_vars_acc (e:iexpr var) (acc:S.t) : S.t :=
    match e with
      | IEvar _ x => S.add x acc
      | IEZconst _ => acc
      | IEFconst _ => acc
      | IEunop _ e => iexpr_free_vars_acc e acc
      | IEbinop _ e1 e2 =>
        iexpr_free_vars_acc e1 (iexpr_free_vars_acc e2 acc)
    end.

  Fixpoint iexpr_is_free (e:iexpr var) (y:var) : bool :=
    match e with
      | IEvar _ x => if eq_dec y x then true else false
      | IEZconst _ => false
      | IEFconst _ => false
      | IEunop _ e => iexpr_is_free e y
      | IEbinop _ e1 e2 =>
        iexpr_is_free e1 y || iexpr_is_free e2 y
    end.

  Lemma iexpr_is_free_eval_iexpr_acc:
    ∀ e y acc,
      S.mem y (iexpr_free_vars_acc e acc) ->
      S.mem y acc \/ iexpr_is_free e y.
  Proof.
    induction e; simpl; intros; try discriminate; auto.
    - destruct eq_dec. auto. rewrite S.mem_add in H. intuition.
    - unfold is_true. rewrite Bool.orb_true_iff.
      apply IHe1 in H. destruct H; auto.
      apply IHe2 in H. intuition.
  Qed.

  Fixpoint mux_iexpr_free_vars_acc (e:mux_iexpr var) (acc:S.t) : S.t :=
    match e with
      | Leaf e => iexpr_free_vars_acc e acc
      | Mux c et ef =>
        mux_iexpr_free_vars_acc c
          (mux_iexpr_free_vars_acc et
            (mux_iexpr_free_vars_acc ef acc))
    end.

  Lemma iexpr_free_vars_acc_iff:
    ∀ e acc x,
      S.mem x (iexpr_free_vars_acc e acc) <->
      S.mem x acc \/ S.mem x (iexpr_free_vars_acc e S.empty).
  Proof.
    induction e; simpl; intros.
    - rewrite S.mem_add. setoid_rewrite S.mem_add. intuition.
      apply S.mem_empty in H. destruct H.
    - intuition. apply S.mem_empty in H0. destruct H0.
    - intuition. apply S.mem_empty in H0. destruct H0.
    - apply IHe.
    - rewrite IHe1, IHe2. rewrite (IHe1 (iexpr_free_vars_acc e2 S.empty)). intuition.
  Qed.

  Lemma eval_iexpr_upd_free_acc:
    ∀ x ρ n v e,
      eval_iexpr ρ e v ->
    ∀ acc,
      ~S.mem x (iexpr_free_vars_acc e acc) ->
      eval_iexpr (upd ρ x n) e v.
  Proof.
    induction 1; intros; simpl in *.
    - rewrite S.mem_add in H1.
      constructor; auto. unfold upd.
      destruct eq_dec; auto. destruct H1. auto.
    - constructor. auto.
    - constructor. auto.
    - econstructor; eauto.
    - econstructor; eauto.
      eapply IHeval_iexpr2. contradict H2. rewrite iexpr_free_vars_acc_iff. eauto.
  Qed.

  Lemma mux_iexpr_free_vars_acc_iff:
    ∀ e acc x,
      S.mem x (mux_iexpr_free_vars_acc e acc) <->
      S.mem x acc \/ S.mem x (mux_iexpr_free_vars_acc e S.empty).
  Proof.
    induction e; simpl; intros.
    - apply iexpr_free_vars_acc_iff.
    - rewrite IHe1, IHe2, IHe3.
      rewrite (IHe1 (mux_iexpr_free_vars_acc e2 (mux_iexpr_free_vars_acc e3 S.empty))).
      rewrite (IHe2 (mux_iexpr_free_vars_acc e3 S.empty)).
      intuition.
  Qed.

  Lemma eval_mux_iexpr_upd_free_acc:
    ∀ x ρ n v e,
      eval_mux_iexpr ρ e v ->
    ∀ acc,
      ~S.mem x (mux_iexpr_free_vars_acc e acc) ->
      eval_mux_iexpr (upd ρ x n) e v.
  Proof.
    induction 1; intros; simpl in *.
    - constructor. eapply eval_iexpr_upd_free_acc; eauto.
    - rewrite (mux_iexpr_free_vars_acc_iff c) in H1.
      rewrite (mux_iexpr_free_vars_acc_iff et) in H1.
      rewrite (mux_iexpr_free_vars_acc_iff ef) in H1.
      econstructor; intuition eauto.
    - rewrite (mux_iexpr_free_vars_acc_iff c) in H1.
      rewrite (mux_iexpr_free_vars_acc_iff et) in H1.
      rewrite (mux_iexpr_free_vars_acc_iff ef) in H1.
      econstructor 3; intuition eauto.
  Qed.

  Definition mux_iexpr_free_vars (e:mux_iexpr var) : S.t :=
    mux_iexpr_free_vars_acc e S.empty.

  Lemma eval_mux_iexpr_upd_free:
    ∀ x ρ n v e,
      ~S.mem x (mux_iexpr_free_vars e) ->
      eval_mux_iexpr ρ e v ->
      eval_mux_iexpr (upd ρ x n) e v.
  Proof.
    intros. eapply eval_mux_iexpr_upd_free_acc; eauto.
  Qed.

  Record t0 :=
    { eqs : T.t (mux_iexpr var);
      eqs_free : T.t S.t;
      facts : PShareTree.t (mux_iexpr var * bool);
      facts_free : T.t PSet.t
    }.
  Record t0_inv (ab:t0) : Prop :=
    { eqs_free_correct :
        ∀ x e, T.get x ab.(eqs) = Some e ->
        ∀ y, S.mem y (mux_iexpr_free_vars e) ->
        ∃ s, T.get y ab.(eqs_free) = Some s /\ S.mem x s;
      facts_free_correct :
        ∀ p e b, PShareTree.get p ab.(facts) = Some (e, b) ->
        ∀ y, S.mem y (mux_iexpr_free_vars e) ->
        ∃ s, T.get y ab.(facts_free) = Some s /\ PSet.mem p s }.

  Definition t := {ab: t0 | t0_inv ab}.

  Record gamm0 (ab:t0) (ρ:var -> ideal_num) : Prop :=
    { eqs_gamm :
        ∀ x e, T.get x ab.(eqs) = Some e -> eval_mux_iexpr ρ e (ρ x);
      facts_gamm :
        ∀ p e b, PShareTree.get p ab.(facts) = Some (e, b) ->
                 eval_mux_iexpr ρ e (INz (if b then 1 else 0)) }.
  Instance gamm0' : gamma_op t0 (var -> ideal_num) := gamm0.

  Instance gamm : gamma_op t (var -> ideal_num) :=
    fun ab => gamm0 (proj1_sig ab).

  Definition extract_inchan (ab:t) (c:in_chan) : t * in_chan :=
    (ab, {| get_itv := c.(get_itv);
            get_congr := c.(get_congr);
            get_eq_expr x := T.get x (proj1_sig ab).(eqs) |}).

  Lemma extract_inchan_correct:
    ∀ ρ ab c, ρ ∈ γ ab -> ρ ∈ γ c -> ρ ∈ γ (extract_inchan ab c).
  Proof.
    intros ρ ab c H H0; split; auto; constructor; simpl.
    - intros e a H1. apply H0; auto.
    - intros e a H1. apply H0; auto.
    - intros x e H1. apply H; auto.
  Qed.

  Definition remove_eq (x:var) (ab:t0) :=
    match T.get x ab.(eqs) with
      | Some e =>
        {| eqs := T.remove x ab.(eqs);
           eqs_free :=
             S.fold (fun x' free =>
                       match T.get x' free with
                         | None => free
                         | Some s => T.set x' (S.rem x s) free
                       end)
                    (mux_iexpr_free_vars e) ab.(eqs_free);
           facts := ab.(facts);     facts_free := ab.(facts_free)
        |}
      | None => ab
    end.

  Lemma remove_eq_inv:
    ∀ ab x, t0_inv ab -> t0_inv (remove_eq x ab).
  Proof.
    unfold remove_eq. constructor; intros.
    - destruct (T.get x ab.(eqs)) eqn:?. simpl in *.
      + rewrite T.grspec in H0.
        destruct T.elt_eq. discriminate.
        rewrite S.fold_spec.
        induction (S.elements (mux_iexpr_free_vars m)).
        * destruct H. eauto.
        * simpl.
          match goal with
            | |- context [T.get _ match ?x with _ => _ end = _] => destruct x eqn:?
          end; auto.
          rewrite T.gsspec. destruct T.elt_eq; auto.
          eexists. split. eauto.
          subst a. destruct IHl as (s & Hs & Hs').
          fold S.t in *.
          assert (s = t1) by congruence. subst. clear Heqo0.
          rewrite S.mem_rem. auto.
      + destruct H. eauto.
    - destruct T.get, H; eauto.
  Qed.

  Lemma remove_eq_gamm0:
    ∀ ab ρ x, ρ ∈ gamm0 ab -> ρ ∈ gamm0 (remove_eq x ab).
  Proof.
    unfold remove_eq. constructor; intros.
    - eapply H. destruct (T.get x ab.(eqs)); auto. simpl in H0.
      rewrite T.grspec in H0. destruct T.elt_eq. discriminate. auto.
    - destruct T.get, H; eauto.
  Qed.

  Lemma remove_eq_get:
    ∀ ab x y, T.get x (remove_eq y ab).(eqs) =
       if T.elt_eq x y then None else T.get x ab.(eqs).
  Proof.
    intros. unfold remove_eq.
    destruct (T.get y ab.(eqs)) eqn:?.
    - simpl. apply (T.grspec).
    - destruct T.elt_eq. subst. auto. auto.
  Qed.

  Definition remove_fact (p:positive) (ab:t0) :=
    match PShareTree.get p ab.(facts) with
      | Some e =>
        {| facts := PShareTree.remove p ab.(facts);
           facts_free :=
             S.fold (fun x' free =>
                       match T.get x' free with
                         | None => free
                         | Some s => T.set x' (PShareTree.remove p s) free
                       end)
                  (mux_iexpr_free_vars (fst e)) ab.(facts_free);
           eqs := ab.(eqs);    eqs_free := ab.(eqs_free)
        |}
      | None => ab
    end.

  Lemma remove_fact_inv:
    ∀ ab p, t0_inv ab -> t0_inv (remove_fact p ab).
  Proof.
    unfold remove_fact. constructor; intros.
    - destruct PShareTree.get, H; eauto.
    - destruct ((facts ab) ! p) eqn:?. simpl in *.
      + rewrite PShareTree.grspec in H0.
        destruct (PShareTree.elt_eq p0 p). discriminate.
        rewrite S.fold_spec.
        induction (S.elements (mux_iexpr_free_vars (fst p1))).
        * destruct H. eauto.
        * simpl.
          match goal with
            | |- context [T.get _ match ?x with _ => _ end = _] => destruct x eqn:?
          end; auto.
          rewrite T.gsspec. destruct T.elt_eq; auto.
          eexists. split. eauto.
          subst a. destruct IHl as (s & Hs & Hs').
          assert (Some s = Some t1) by (rewrite <- Hs, <- Heqo0; auto).
          inv H2. clear Heqo0.
          rewrite PSet.mem_rem. auto.
      + destruct H. eauto.
  Qed.

  Lemma remove_fact_gamm0:
    ∀ ab ρ p, ρ ∈ gamm0 ab -> ρ ∈ gamm0 (remove_fact p ab).
  Proof.
    unfold remove_fact. constructor; intros.
    - destruct PShareTree.get, H; eauto.
    - eapply H. destruct ((facts ab) ! p); auto. simpl in H0.
      rewrite PShareTree.grspec in H0. destruct PShareTree.elt_eq. discriminate. eauto. auto.
  Qed.

  Lemma remove_fact_get:
    ∀ ab x y, (remove_fact y ab).(facts) ! x =
       if eq_dec x y then None else ab.(facts) ! x.
  Proof.
    intros. unfold remove_fact.
    destruct ((facts ab) ! y) eqn:?.
    - simpl. apply (PShareTree.grspec).
    - destruct (eq_dec x y). subst. auto. auto.
  Qed.

  Definition forget0 (x:var) (ab:t0) : t0 :=
    let ab := remove_eq x ab in
    let ab :=
        match T.get x ab.(eqs_free) with
          | None => ab
          | Some s =>
            let ab := S.fold (fun x' ab => remove_eq x' ab) s ab in
            {| eqs := ab.(eqs);
               eqs_free := T.remove x ab.(eqs_free)
               (* Optim : we could do this rem before the
                  fold, but it is harder to prove correct *);
               facts := ab.(facts);    facts_free := ab.(facts_free)
            |}
        end
    in
    match T.get x ab.(facts_free) with
      | None => ab
      | Some s =>
        let ab := PSet.fold (fun p ab => remove_fact p ab) s ab in
        {| eqs := ab.(eqs);     eqs_free := ab.(eqs_free);
           facts := ab.(facts);
           facts_free := T.remove x ab.(facts_free)
        (* Optim : we could do this rem before the
           fold, but it is harder to prove correct *)
        |}
    end.

  Lemma forget0_inv:
    ∀ ab x, t0_inv ab -> t0_inv (forget0 x ab).
  Proof.
    unfold forget0. intros.
    apply remove_eq_inv with (x:=x) in H.
    remember (remove_eq x ab).
    remember (match T.get x t1.(eqs_free) with
                | None => t1
                | Some s =>
                  let ab := S.fold (fun x' ab => remove_eq x' ab) s t1 in
                  {| eqs := ab.(eqs); facts := ab.(facts);
                     eqs_free := T.remove x ab.(eqs_free);
                     facts_free := ab.(facts_free) |}
              end).
    assert (t0_inv t2).
    { simpl in Heqt2.
      destruct (T.get x t1.(eqs_free)) eqn:?; [|congruence].
      assert (t0_inv (S.fold (λ x' ab0, remove_eq x' ab0) t3 t1)).
      { rewrite S.fold_spec. induction (S.elements t3). auto. apply remove_eq_inv, IHl. }
      constructor; subst t2; simpl; intros.
      - rewrite T.grspec. destruct T.elt_eq.
        + exfalso. subst y.
          assert (T.get x0 t1.(eqs) = Some e).
          { revert H1. rewrite S.fold_spec. induction (S.elements t3).
            intros. apply H1. simpl. rewrite remove_eq_get.
            destruct (T.elt_eq x0 a). discriminate. auto. }
          edestruct (eqs_free_correct _ H) as [? []]; eauto. clear H H2 H3.
          assert (x1 = t3) by congruence. subst x1. clear Heqo H4.
          revert H1 H5. rewrite S.fold_spec, S.elements_spec. induction (S.elements t3).
          * auto.
          * simpl. rewrite remove_eq_get. destruct (T.elt_eq x0 a). discriminate.
            intros. apply IHl; auto. intuition.
        + eapply H0; eauto.
      - eapply H0; eauto. }
    clear Heqt2 Heqt1 t1 H.
    destruct (T.get x (facts_free t2)) eqn:?; [|easy].
    assert (t0_inv (PSet.fold (λ p ab, remove_fact p ab) t1 t2)).
    { rewrite PSet.fold_spec. induction (PSet.elements t1). auto.
      intros. apply remove_fact_inv. auto. }
    constructor; simpl; intros.
    - eapply H; eauto.
    - rewrite T.grspec. destruct T.elt_eq.
      + exfalso. subst y.
        assert (t2.(facts) ! p = Some (e, b)).
        { revert H1. rewrite PSet.fold_spec. induction (PSet.elements t1); auto.
          simpl. intros. apply IHl. rewrite remove_fact_get in H1.
          destruct (eq_dec p a). discriminate. auto. }
        edestruct (facts_free_correct _ H0) as [? []]; eauto. clear H0 H2 H3.
        assert (x0 = t1) by congruence. subst x0. clear Heqo H4.
        revert H1 H5. rewrite PSet.fold_spec, PSet.elements_spec.
        induction (PSet.elements t1).
        * auto.
        * simpl. intros. rewrite remove_fact_get in H0. destruct (eq_dec p a). discriminate.
          intuition.
      + eapply H; eauto.
  Qed.

  Lemma forget0_gamm0:
    ∀ x ρ n ab,
      t0_inv ab ->
      ρ ∈ gamm0 ab -> (upd ρ x n) ∈ gamm0 (forget0 x ab).
  Proof.
    intros. pose proof (forget0_inv _ x H). unfold forget0 in *.
    remember (remove_eq x ab).
    remember (match T.get x t1.(eqs_free) with
                | None => t1
                | Some s =>
                  let ab := S.fold (fun x' ab => remove_eq x' ab) s t1 in
                  {| eqs := ab.(eqs); facts := ab.(facts);
                     eqs_free := T.remove x ab.(eqs_free);
                     facts_free := ab.(facts_free) |}
              end).
    constructor; intros.
    - assert (∀ s0 (v := PSet.fold (λ p ab, remove_fact p ab) s0 t2),
                v.(eqs) = t2.(eqs) /\ v.(eqs_free) = t2.(eqs_free)).
      { intro. rewrite PSet.fold_spec. induction (PSet.elements s0). auto.
        simpl in *. unfold remove_fact. destruct PShareTree.get; auto. }
      apply eval_mux_iexpr_upd_free.
      + pose proof (eqs_free_correct _ H1 _ _ H2 x).
        intro. specialize (H4 H5).
        assert (∃ s, T.get x t2.(eqs_free) = Some s ∧ S.mem x0 s).
        { destruct (T.get x (facts_free t2)); auto.
          simpl in H4. destruct (H3 t3). rewrite H7 in H4. auto. }
        assert (T.get x0 t2.(eqs) = Some e).
        { destruct (T.get x (facts_free t2)); auto.
          simpl in H2. destruct (H3 t3). rewrite H7 in H2. auto. }
        clear H2 H4.
        destruct (T.get x t1.(eqs_free)) eqn:?.
        * subst t2. simpl in *.
          rewrite T.grspec in H6. destruct T.elt_eq.
          destruct H6 as [? []]. discriminate. congruence.
        * destruct H6 as [? []]. congruence.
      + assert (T.get x0 t2.(eqs) = Some e).
        { destruct (T.get x (facts_free t2)); auto. simpl in H2.
          destruct (H3 t3). congruence. }
        clear H2 H1.
        assert (T.get x0 t1.(eqs) = Some e).
        { revert H4. subst t2.
          destruct (T.get x t1.(eqs_free)) eqn:?; auto.
          simpl. rewrite S.fold_spec. induction (S.elements t2); auto.
          intros. simpl in H4. rewrite remove_eq_get in H4. destruct T.elt_eq; auto.
          discriminate. }
        subst t1. rewrite remove_eq_get in H1. unfold upd, eq_dec, var_eq.
        destruct T.elt_eq. discriminate. apply H0. auto.
    - apply eval_mux_iexpr_upd_free.
      + pose proof (facts_free_correct _ H1 _ _ _ H2 x).
        intro. specialize (H3 H4).
        destruct (T.get x t2.(facts_free)) eqn:?.
        * simpl in *.
          rewrite (T.grspec) in H3. destruct T.elt_eq.
          destruct H3 as [? []]. discriminate. congruence.
        * destruct H3 as [? []]. congruence.
      + assert (t2.(facts) ! p = Some (e, b)).
        { destruct (T.get x (facts_free t2)); auto.
          revert H2. simpl.
          rewrite PSet.fold_spec. induction (PSet.elements t3); auto.
          simpl. intros. rewrite remove_fact_get in H2.
          destruct (eq_dec p a). discriminate. auto. }
        clear H2.
        assert (t1.(facts) ! p = Some (e, b)).
        { subst t2. destruct (T.get x (eqs_free t1)); auto.
          revert H3. simpl. rewrite S.fold_spec. induction (S.elements t2); auto.
          simpl. unfold remove_eq.
          match goal with
              |- (facts match ?x with _ => _ end) ! _ = _ -> _ => destruct x
          end; auto. }
        clear H3.
        subst t1. unfold remove_eq in H2.
        destruct (T.get x (eqs ab)); eapply H0; eauto.
  Qed.

  Program Definition forget (x:var) (ab:t * @in_chan var) (chan:in_chan): (t * in_chan)+⊥ :=
    NotBot (extract_inchan (forget0 x (fst ab)) chan).
  Next Obligation. apply forget0_inv. apply proj2_sig. Qed.

  Lemma forget_correct :
    ∀ x ρ n ab chan, ρ ∈ γ ab -> (upd ρ x n) ∈ γ chan ->
                     (upd ρ x n) ∈ γ (forget x ab chan).
  Proof.
    intros. apply extract_inchan_correct; auto.
    apply forget0_gamm0. apply proj2_sig. apply H.
  Qed.

  Definition add_eq (x:var) (e:mux_iexpr var) (ab:t0) :=
    let ab := remove_eq x ab in
    {| eqs := T.set x e ab.(eqs);
       eqs_free := S.fold
                     (fun x' free =>
                        T.set x'
                         (S.add x
                             (match T.get x' free with
                                | None => S.empty
                                | Some s => s
                              end))
                          free)
                     (mux_iexpr_free_vars e) ab.(eqs_free);
       facts := ab.(facts);  facts_free := ab.(facts_free) |}.

  Lemma add_eq_inv:
    ∀ ab x e, t0_inv ab -> t0_inv (add_eq x e ab).
  Proof.
    intros.
    constructor.
    - simpl. intros.
      rewrite T.gsspec in H0.
      destruct T.elt_eq.
      + inv H0. revert H1. rewrite S.fold_spec, S.elements_spec.
        induction (S.elements (mux_iexpr_free_vars e0)).
        * easy.
        * intros. simpl.
          rewrite T.gsspec. destruct T.elt_eq; auto.
          subst. eexists. split. eauto. apply S.mem_add. auto.
          destruct H0. now intuition. auto.
      + rewrite S.fold_spec. induction (S.elements (mux_iexpr_free_vars e)).
        * eapply (remove_eq_inv ab x H); eauto.
        * simpl. rewrite T.gsspec. destruct T.elt_eq; auto. eexists. split. eauto.
          subst. destruct IHl as [? []]. apply S.mem_add.
          fold S.t in *. rewrite H2. auto.
    - simpl. intros. eapply remove_eq_inv; eauto.
  Qed.

  Lemma add_eq_gamm0:
    ∀ x e ρ ab,
      ρ ∈ gamm0 ab -> eval_mux_iexpr ρ e (ρ x) -> t0_inv ab ->
      ρ ∈ gamm0 (add_eq x e ab).
  Proof.
    intros.
    assert (
        let ab := remove_eq x ab in
        ρ ∈ gamm0 {| eqs := T.set x e ab.(eqs);
                     eqs_free := S.fold
                                 (fun x' free =>
                                    T.set x'
                                      (S.add x
                                         (match T.get x' free with
                                            | None => S.empty
                                            | Some s => s
                                          end))
                                      free)
                                 (mux_iexpr_free_vars e) ab.(eqs_free);
                     facts := ab.(facts);   facts_free := ab.(facts_free)
              |}).
    { constructor.
      - simpl. intros.
        rewrite T.gsspec in H2. destruct T.elt_eq.
        + inv H2. auto.
        + apply (remove_eq_gamm0 _ _ x H). auto.
      - simpl. intros.
        eapply (remove_eq_gamm0 _ _ x H). eauto. }
    unfold add_eq. destruct e as [[]|]; try apply remove_eq_gamm0; auto.
  Qed.

  Opaque add_eq.

  Program Definition assign (x:var) (e:iexpr var) (ab:t * in_chan) (chan:in_chan):
    (t * in_chan)+⊥ :=
    if
      match e return _ with
        | IEvar _ x' => if eq_dec x x' then true else false
        | _ => false
      end
    then NotBot (extract_inchan (fst ab) chan)
    else
      if iexpr_is_free e x then
        forget x ab chan
      else
        let ab := add_eq x (Leaf e) (forget0 x (proj1_sig (fst ab))) in
        NotBot (extract_inchan ab chan).
  Next Obligation. apply add_eq_inv, forget0_inv. apply proj2_sig. Qed.

  Lemma assign_correct:
    ∀ x e ρ n ab chan,
      ρ ∈ γ ab -> (upd ρ x n) ∈ γ chan ->
      n ∈ eval_iexpr ρ e ->
      (upd ρ x n) ∈ γ (assign x e ab chan).
  Proof.
    intros. unfold assign.
    destruct (match e with
                | IEvar _ x' => if eq_dec x x' then true else false
                |  _ => false
              end) eqn:?.
    - destruct e; try discriminate.
      destruct (eq_dec x v); try discriminate.
      subst. inv H1.
      apply extract_inchan_correct; auto.
      assert (∀ e n ρ, eval_mux_iexpr ρ e n ->
              ∀ ρ', (∀ y:var, ρ y = ρ' y) ->
                    eval_mux_iexpr ρ' e n).
      { induction 1; intros.
        - intros. constructor.
          induction H1; econstructor; eauto. congruence.
        - constructor 2; eauto.
        - constructor 3; eauto. }
      constructor; intros; eapply H1.
      + unfold upd. destruct eq_dec; [subst|]; apply (proj1 H); auto.
      + intro. unfold upd. destruct eq_dec; congruence.
      + eapply (proj1 H); eauto.
      + intro. unfold upd. destruct eq_dec; congruence.
    - destruct (iexpr_is_free e x) eqn:?.
      apply forget_correct; auto.
      apply extract_inchan_correct. 2:auto.
      apply add_eq_gamm0. apply forget0_gamm0. apply proj2_sig. apply H.
      constructor. eapply eval_iexpr_upd_free_acc.
      unfold upd. destruct eq_dec. auto. congruence.
      intro. apply iexpr_is_free_eval_iexpr_acc in H2.
      destruct H2. eapply S.mem_empty. eauto.
      rewrite Heqb0 in H2. discriminate.
      apply forget0_inv. apply proj2_sig.
  Qed.

  (* Low level function :
       - Does not compare p with counter.
       - There sould not be prior pth fact *)
  Definition add_fact_p (p:positive) (f:mux_iexpr var * bool) (ab:t0) :=
    {| eqs := ab.(eqs); eqs_free := ab.(eqs_free);
       facts := PShareTree.set p f ab.(facts);
       facts_free :=
         S.fold
           (fun x' free =>
              T.set x'
                (PSet.add p
                   (match T.get x' free with
                      | None => PSet.empty
                      | Some s => s
                    end))
                free)
           (mux_iexpr_free_vars (fst f)) ab.(facts_free)
    |}.

  Lemma add_fact_p_inv:
    ∀ p f ab, t0_inv ab -> t0_inv (add_fact_p p f ab).
  Proof.
    intros. constructor.
    - intros. eapply H; eauto.
    - simpl. intros. rewrite PShareTree.gsspec in H0. destruct peq.
      + inv H0. simpl. revert H1. rewrite S.fold_spec, S.elements_spec.
        induction (S.elements (mux_iexpr_free_vars e)).
        * easy.
        * intros. simpl. rewrite T.gsspec. destruct T.elt_eq.
          subst. eexists. split. eauto. rewrite PSet.mem_add. auto.
          destruct H0. now intuition. auto.
      + rewrite S.fold_spec. induction (S.elements (mux_iexpr_free_vars (fst f))).
        * eapply H; eauto.
        * simpl. rewrite T.gsspec. destruct T.elt_eq; auto. eexists. split. eauto.
          subst. apply PSet.mem_add. destruct IHl as [? []].
          unfold PSet.t in *. rewrite H2. auto.
  Qed.

  Lemma add_fact_p_gamm0:
    ∀ p f ρ ab,
      ρ ∈ gamm0 ab -> eval_mux_iexpr ρ (fst f) (INz (if snd f:bool then 1 else 0)) ->
      t0_inv ab ->
      ρ ∈ gamm0 (add_fact_p p f ab).
  Proof.
    intros. constructor.
    - simpl. eapply H.
    - simpl. intros.
      rewrite PShareTree.gsspec in H2. destruct peq.
      + inv H2. auto.
      + eapply H. eauto.
  Qed.

  Definition add_fact (f:mux_iexpr var * bool) (ab:t0) :=
    match fst f with
      | Leaf (IEFconst _) | Leaf (IEZconst _) =>
         ab (* Any other domain will take care of that *)
      | _ =>
        let h := MIX_mux_iexpr MIX_var 12 (fst f) in
        let h := N.land 65535 (final_MIX h) in
        let ab := add_fact_p (N.succ_pos h) f ab in
        {| eqs := ab.(eqs);       eqs_free := ab.(eqs_free);
           facts := ab.(facts);   facts_free := ab.(facts_free)
        |}
    end.

  Lemma add_fact_inv:
    ∀ f ab, t0_inv ab -> t0_inv (add_fact f ab).
  Proof.
    intros. unfold add_fact.
    remember (N.succ_pos
                (N.land 65535
                        (final_MIX (MIX_mux_iexpr MIX_var 12 (fst f))))).
    pose proof (add_fact_p_inv p f ab).
    destruct (fst f) as [[]|]; constructor; simpl; intros; (eapply H0 || eapply H); eauto.
  Qed.

  Lemma add_fact_gamm0:
    ∀ f ρ ab,
      ρ ∈ gamm0 ab -> eval_mux_iexpr ρ (fst f) (INz (if snd f:bool then 1 else 0)) ->
      t0_inv ab ->
      ρ ∈ gamm0 (add_fact f ab).
  Proof.
    intros. unfold add_fact.
    remember (N.succ_pos
                (N.land 65535
                        (final_MIX (MIX_mux_iexpr MIX_var 12 (fst f))))).
    pose proof (add_fact_p_gamm0 p f ρ ab H H0 H1).
    destruct (fst f) as [[]|]; constructor; simpl; intros;
      ((eapply H2; now eauto) || (eapply H; now eauto)).
  Qed.

  Definition assume0 (e:iexpr var) (b:bool) (ab:t0) : t0 :=
    match e with
      | IEbinop (IOcmp Ceq) (IEvar _ x) e'
      | IEbinop (IOcmp Ceq) e' (IEvar _ x) =>
        if b then
          match e' with
            | IEFconst _ | IEZconst _ => ab
            | _ => add_eq x (Leaf e') ab
          end
        else ab
      | IEbinop (IOcmp Cne) (IEvar _ x) e'
      | IEbinop (IOcmp Cne) e' (IEvar _ x) =>
        if b then ab
        else
          match e' with
            | IEFconst _ | IEZconst _ => ab
            | _ => add_eq x (Leaf e') ab
          end
      | _ => ab
    end.

  Lemma assume0_inv:
    ∀ e b ab, t0_inv ab -> t0_inv (assume0 e b ab).
  Proof.
    destruct e; simpl; intros; auto using add_fact_inv.
    assert (∀ x e, t0_inv (if b then add_eq x (Leaf e) ab else ab)).
    { destruct b; auto. intros. apply add_eq_inv. auto. }
    assert (∀ x e, t0_inv (if b then ab else add_eq x (Leaf e) ab)).
    { destruct b; auto. intros. apply add_eq_inv. auto. }
    assert (t0_inv (if b then ab else ab)). { destruct b; auto. }
    destruct i; auto; destruct c; auto;
    destruct e1, e2; auto;
    try (destruct i as [|[] []]; auto);
    try (destruct i0 as [|[] []]; auto).
  Qed.

  Lemma assume0_gamm0:
    ∀ c ρ ab b v,
      t0_inv ab ->
      ρ ∈ γ ab ->
      v <> 0 ->
      (INz (if b:bool then v else 0)) ∈ eval_iexpr ρ c ->
      ρ ∈ γ (assume0 c b ab).
  Proof.
    destruct c; simpl; eauto; intros.
    inv H2. inv H9; auto; auto.
    assert (∀ x e ρ ab, gamm0 ab ρ → eval_mux_iexpr ρ e (ρ x) → t0_inv ab → gamm0 (add_eq x e ab) ρ).
    { intros. eapply add_eq_gamm0 in H3; eauto. }
    assert (0 ∈ γ (ZIntervals.ITV 0 0)) by (compute; split; discriminate).
    destruct c; simpl in H7;
      destruct b, (Z.compare_spec i1 i2); try congruence;
      inv H6; inv H8; auto;
      (try (destruct itv as [|[] []]));
      (try (destruct itv0 as [|[] []])); auto;
      try eapply H2;
      try match goal with H : ρ ?id = _ |- context [ρ ?id] => rewrite H end;
      try match goal with H : ?x ∈ γ (ZIntervals.ITV 0 0) |- _ =>
          change (0 <= x <= 0) in H; assert (x = 0) by omega; subst; clear H end;
      try (simpl; eauto; try (econstructor; eauto); omega).
  Qed.

  Program Definition assume (e:iexpr var) (b:bool) (ab:t * @in_chan var) (chan:in_chan): (t * in_chan)+⊥ :=
    let ab := add_fact (Leaf e, b) (fst ab) in
    NotBot (extract_inchan (assume0 e b ab) chan).
  Next Obligation. apply assume0_inv. apply add_fact_inv. apply proj2_sig. Qed.

  Lemma assume_correct :
    ∀ c ρ ab b chan,
      ρ ∈ γ ab -> ρ ∈ γ chan ->
      (INz (if b:bool then 1 else 0)) ∈ eval_iexpr ρ c ->
      ρ ∈ γ (assume c b ab chan).
  Proof.
    intros. apply extract_inchan_correct; auto.
    apply assume0_gamm0 with (v:=1); auto; try discriminate.
    apply add_fact_inv, proj2_sig.
    apply add_fact_gamm0.
    apply H.
    constructor. auto.
    apply proj2_sig.
  Qed.

  Lemma mux_iexpr_eqdec : ∀ (e1 e2:mux_iexpr var), {e1 = e2} + {e1 <> e2}.
  Proof. repeat decide equality.
         apply eq_dec. apply Floats.Float.eq_dec. apply Floats.Float.eq_dec.
  Defined.

  Program Definition leb (ab1:t * @in_chan var) (ab2:t) : bool :=
    let ab1 := proj1_sig (fst ab1) in
    let ab2 := proj1_sig ab2 in
    T.shforall2 (fun _ e1' e2' =>
                   match e1', e2' return _ with
                     | _, None => true
                     | None, _ => false
                     | Some e1, Some e2 => if mux_iexpr_eqdec e1 e2 then true else false
                   end)
                _ ab1.(eqs) ab2.(eqs) &&
    PShareTree.shforall2 (fun _ f1 f2 =>
                            match f1, f2 return _ with
                              | _, None => true
                              | None, _ => false
                              | Some (e1, b1), Some (e2, b2) =>
                                Bool.eqb b1 b2 && if mux_iexpr_eqdec e1 e2 then true else false
                            end)
                         _  ab1.(facts) ab2.(facts).
  Next Obligation. destruct v; auto. destruct mux_iexpr_eqdec; congruence. Qed.
  Next Obligation. destruct v as [[]|]; auto. destruct b, mux_iexpr_eqdec; simpl; congruence. Qed.

  Lemma leb_correct:
    ∀ ab1 ab2 ρ, leb ab1 ab2 = true -> ρ ∈ γ ab1 -> ρ ∈ γ ab2.
  Proof.
    unfold leb. destruct ab1 as [[[ab1 ?] ?] ?], ab2 as [[ab2 ?] ?]; simpl.
    intros. destruct H0.
    rewrite Bool.andb_true_iff in H. destruct H. constructor; intros; simpl in *.
    - apply H0. simpl.
      rewrite T.shforall2_correct in H. specialize (H x).
      destruct T.get, T.get; try congruence.
      destruct mux_iexpr_eqdec; congruence.
    - apply H0 with (p:=p). simpl.
      rewrite PShareTree.shforall2_correct in H2. specialize (H2 p).
      destruct PShareTree.get as [[]|], PShareTree.get as [[]|]; try congruence.
      destruct b0, b1, mux_iexpr_eqdec; simpl in H2; congruence.
  Qed.

  Program Definition top (chan:in_chan) : t * in_chan :=
    extract_inchan {| eqs := T.empty _;      eqs_free := T.empty _;
                      facts := PShareTree.empty _;  facts_free := T.empty _
                   |} chan.
  Next Obligation.
    constructor.
    - intros. rewrite T.gempty in H. discriminate.
    - intros. rewrite PShareTree.gempty in H. discriminate.
  Qed.

  Lemma top_correct:
    ∀ chan ρ, ρ ∈ γ chan -> ρ ∈ γ (top chan).
  Proof.
    intros. apply extract_inchan_correct; auto. constructor; simpl; intros.
    - rewrite T.gempty in H0. discriminate.
    - rewrite PShareTree.gempty in H0. discriminate.
  Qed.

  Program Definition join0_cond (ab1:t0) (ab2:t0) (widen:bool) : option (mux_iexpr var * bool) :=
    if widen return _ then None
    else
      let candidates :=
        PShareTree.shcombine_diff (fun _ eb1 eb2 =>
          match eb1, eb2 return _ with
            | None, _ | _, None => None
            | Some (e1, b1), Some (e2, b2) =>
              if xorb b1 b2 return _ then
                if mux_iexpr_eqdec e1 e2 return _ then eb1
                else None
              else None
          end)
          _ ab1.(facts) ab2.(facts)
      in
      PShareTree.fold (fun b p ebcur =>
        match b return _ with
          | Some eb => Some eb
          | None => Some ebcur
        end) candidates None.
  Next Obligation. destruct v as [[? []]|]; auto. Qed.

  Lemma join0_cond_correct :
    ∀ ab1 ab2 w ρ e b,
      join0_cond ab1 ab2 w = Some (e, b) ->
      (ρ ∈ gamm0 ab1 -> eval_mux_iexpr ρ e (INz (if b then 1 else 0))) /\
      (ρ ∈ gamm0 ab2 -> eval_mux_iexpr ρ e (INz (if negb b then 1 else 0))).
  Proof.
    unfold join0_cond. intros ab1 ab2 w ρ. destruct w. discriminate.
    apply PTree_Properties.fold_rec. eauto. discriminate.
    intros. rewrite PShareTree.gshcombine_diff in H0.
    destruct a; inv H2. auto.
    destruct ((facts ab1) ! k) as [[]|]eqn:?; [|discriminate].
    destruct ((facts ab2) ! k) as [[]|]eqn:?; [|discriminate].
    assert (b1 = negb b0) by (destruct b0, b1; try reflexivity; try discriminate).
    subst b1. rewrite <- negb_xorb_r, xorb_nilpotent in H0.
    destruct (mux_iexpr_eqdec m0 m1); inv H0. subst.
    split; intro GAM; eapply GAM; eauto.
  Qed.

  Program Definition join0_eqs (ab1:t0) (ab2:t0) (ab:t0) (w:bool) : t0 :=
    match join0_cond ab1 ab2 w return _ with
      | Some (cond, swap) =>
        let diffs :=
          T.shcombine_diff (fun _ e1 e2 =>
            match e1, e2 return _ with
              | Some e1, None => Some (e1, false)
              | Some e1, Some e2 =>
                if mux_iexpr_eqdec e1 e2 return _ then None
                else Some (e2, true)
              | _, _ => None
            end)
            _ ab1.(eqs) ab2.(eqs)
        in
        let cond_free := mux_iexpr_free_vars cond in
        let free :=
          T.fold (fun free x eb =>
            match eb return _ with
              | (e1, false) =>
                S.fold (fun x' free =>
                  match T.get x' free return _ with
                    | None => free
                    | Some s => T.set x' (S.rem x s) free
                  end)
                  (mux_iexpr_free_vars e1) free
              | (e2, true) =>
                S.fold (fun x' free =>
                  T.set x'
                    (S.add x
                      (match T.get x' free return _ with
                         | None => S.empty
                         | Some s => s
                       end)) free)
                  (mux_iexpr_free_vars_acc e2 cond_free) free
            end) diffs ab1.(eqs_free)
        in
        {| eqs :=
             T.shcombine (fun _ e1 e2 =>
               match e1, e2 return _ with
                 | None, _ | _, None => None
                 | Some e1, Some e2 =>
                   if mux_iexpr_eqdec e1 e2 return _ then Some e1
                   else if swap return _ then Some (Mux cond e1 e2)
                        else Some (Mux cond e2 e1)
               end) _ ab1.(eqs) ab2.(eqs);
           eqs_free := free;
           facts := ab.(facts);
           facts_free := ab.(facts_free)
        |}
      | None =>
        let diffs :=
          T.shcombine_diff (fun _ e1 e2 =>
            match e1, e2 return _ with
              | Some e1, None => Some e1
              | Some e1, Some e2 =>
                if mux_iexpr_eqdec e1 e2 return _ then None
                else Some e1
              | _, _ => None
            end)
            _ ab1.(eqs) ab2.(eqs)
        in
        let free :=
          T.fold (fun free x e1 =>
            S.fold (fun x' free =>
              match T.get x' free return _ with
                | None => free
                | Some s => T.set x' (S.rem x s) free
              end)
              (mux_iexpr_free_vars e1) free)
            diffs ab1.(eqs_free)
        in
        {| eqs :=
             T.shcombine (fun _ e1' e2' =>
               match e1', e2' return _ with
                 | None, _ | _, None => None
                 | Some e1, Some e2 =>
                   if mux_iexpr_eqdec e1 e2 return _ then e1'
                   else None
               end) _ ab1.(eqs) ab2.(eqs);
           eqs_free := free;
           facts := ab.(facts);
           facts_free := ab.(facts_free)
        |}
    end.
  Next Obligation. destruct v; auto. destruct mux_iexpr_eqdec; congruence. Qed.
  Next Obligation. destruct v; auto. destruct mux_iexpr_eqdec; congruence. Qed.
  Next Obligation. destruct v; auto. destruct mux_iexpr_eqdec; congruence. Qed.
  Next Obligation. destruct v; auto. destruct mux_iexpr_eqdec; congruence. Qed.

  Lemma join0_eqs_correct :
    ∀ ab1 ab2 ab w,
      t0_inv ab1 -> t0_inv ab2 -> t0_inv ab ->
      t0_inv (join0_eqs ab1 ab2 ab w) /\
      ∀ ρ, ρ ∈ (gamm0 ab1 ∪ gamm0 ab2) ->
           ρ ∈ gamm0 ab ->
           ρ ∈ gamm0 (join0_eqs ab1 ab2 ab w).
  Proof.
    intros. pose proof (join0_cond_correct ab1 ab2 w).
    unfold join0_eqs. destruct join0_cond as [[cond swap]|].
    - split.
      + split. 2:apply H1. simpl. intros.
        rewrite T.gshcombine in H3.
        destruct (T.get x (eqs ab1)) eqn:EQ1. 2:discriminate.
        destruct (T.get x (eqs ab2)) eqn:EQ2. 2:discriminate.
        rewrite T.fold_spec, <- fold_left_rev_right.
        match goal with
          | |- context [T.elements ?t] =>
            pose proof T.elements_correct (m:=t) (i:=x);
            pose proof (T.elements_complete t x)
        end.
        rewrite T.gshcombine_diff, EQ1, EQ2 in H5, H6.
        setoid_rewrite in_rev in H5. setoid_rewrite in_rev in H6.
        apply or_introl with (B := S.mem y (mux_iexpr_free_vars m)) in H5.
        revert H5 H6.
        induction rev.
        { simpl. intros. eapply H. eauto.
          destruct mux_iexpr_eqdec. congruence. destruct H5. edestruct H5. eauto. auto. }
        { destruct a as (x' & e2 & []); simpl; intros.
          - rewrite S.fold_spec.
            pose proof proj1 (S.elements_spec
                 (mux_iexpr_free_vars_acc e2 (mux_iexpr_free_vars cond)) y).
            match type of IHl with
              | ?P -> _ => eapply or_introl with (B:=P) in H7
            end.
            revert H7. induction S.elements; intros.
            + apply IHl; auto. destruct H5, H7; auto.
              destruct mux_iexpr_eqdec. right. congruence.
              edestruct H5. eauto. 2:left; intros; inv H9; auto.
              inv H8. right.
              rewrite mux_iexpr_free_vars_acc_iff in H7. simpl in H7.
              destruct swap; inv H3; unfold mux_iexpr_free_vars in H4; simpl in H4.
              rewrite mux_iexpr_free_vars_acc_iff in H4.
              rewrite (mux_iexpr_free_vars_acc_iff m) in H4.
              intuition.
              rewrite mux_iexpr_free_vars_acc_iff in H4.
              rewrite (mux_iexpr_free_vars_acc_iff m0) in H4.
              intuition.
            + fold S.t in *. simpl. rewrite T.gsspec.
              destruct (T.elt_eq y a).
              2:destruct IHl0 as (s & -> & Hsx); [simpl in H7; now intuition|eauto].
              subst. eexists. split. eauto. rewrite S.mem_add.
              destruct H5. 2:destruct IHl0 as (s & -> & Hsx); now auto.
              destruct mux_iexpr_eqdec. destruct IHl0 as (s & -> & Hsx); auto. right. left. discriminate.
              edestruct H5. eauto. inv H8. now auto.
              destruct IHl0 as (s & -> & Hsx); auto; right; left.
              intros. congruence.
          - rewrite S.fold_spec. induction S.elements.
            + apply IHl. 2:now auto. destruct H5; auto. left.
              destruct mux_iexpr_eqdec. discriminate.
              intros. edestruct H5. eauto. congruence. auto.
            + simpl. destruct IHl0 as (s & Hs & Hsx).
              match goal with
                  |- context [match ?x with _ => _ end] => destruct x eqn:EQ
              end. 2:now eauto.
              rewrite T.gsspec. destruct T.elt_eq. 2:now eauto.
              subst. eexists. split. eauto. apply S.mem_rem.
              fold S.t in *. assert (s = t1) by congruence. subst. split. auto.
              intro. subst. specialize (H6 _ (or_introl eq_refl)).
              destruct mux_iexpr_eqdec; congruence. }
      + constructor; simpl. 2:apply H4.
        intros. rewrite T.gshcombine in H5.
        destruct (T.get x ab1.(eqs)) eqn:EQ1. 2:discriminate.
        destruct (T.get x ab2.(eqs)) eqn:EQ2. 2:discriminate.
        destruct mux_iexpr_eqdec.
        inv H5. subst. destruct H3; apply H3; auto.
        specialize (H2 ρ _ _ eq_refl).
        destruct H2. destruct swap, H3; inv H5.
        constructor 3. eauto. apply H3; auto.
        constructor 2. eauto. apply H3; auto.
        constructor 2. eauto. apply H3; auto.
        constructor 3. eauto. apply H3; auto.
    - split.
      + split. 2:apply H1. simpl. intros.
        rewrite T.gshcombine in H3.
        destruct (T.get x (eqs ab1)) eqn:EQ1. 2:discriminate.
        destruct (T.get x (eqs ab2)) eqn:EQ2. 2:discriminate.
        rewrite T.fold_spec, <- fold_left_rev_right.
        match goal with
          | |- context [T.elements ?t] =>
            pose proof (T.elements_complete t x)
        end.
        rewrite T.gshcombine_diff, EQ1, EQ2 in H5.
        setoid_rewrite in_rev in H5.
        revert H5.
        induction rev.
        { simpl. intros. eapply H. eauto. destruct mux_iexpr_eqdec; congruence. }
        { destruct a as [x' e1]; simpl; intros.
          rewrite S.fold_spec. induction S.elements. auto.
          simpl. destruct IHl0 as (s & Hs & Hsx).
          match goal with
              |- context [match ?x with _ => _ end] => destruct x eqn:EQ
          end. 2:now eauto.
          rewrite T.gsspec. destruct T.elt_eq. 2:now eauto.
          subst. eexists. split. eauto. apply S.mem_rem.
          fold S.t in *. assert (s = t1) by congruence. subst. split. auto.
          intro. subst. specialize (H5 _ (or_introl eq_refl)).
          destruct mux_iexpr_eqdec; congruence. }
      + constructor; simpl. 2:apply H4.
        intros. rewrite T.gshcombine in H5.
        destruct (T.get x ab1.(eqs)) eqn:EQ1. 2:discriminate.
        destruct (T.get x ab2.(eqs)) eqn:EQ2. 2:discriminate.
        destruct mux_iexpr_eqdec; inv H5. subst. destruct H3; apply H3; auto.
  Qed.

  Program Definition join0_facts (ab1:t0) (ab2:t0) (ab:t0) : t0 :=
    let diffs :=
      PShareTree.shcombine_diff (fun _ e1b1 e2b2 =>
        match e1b1, e2b2 return _ with
          | Some (e1, b1), None => Some e1
          | Some (e1, b1), Some (e2, b2) =>
            if xorb b1 b2 return _ then None
            else
              if mux_iexpr_eqdec e1 e2 return _ then None
              else Some e1
          | _, _ => None
        end)
        _ ab1.(facts) ab2.(facts)
    in
    let free :=
      PShareTree.fold (fun free x e1 =>
        S.fold (fun x' free =>
          match T.get x' free return _ with
            | None => free
            | Some s => T.set x' (PSet.rem x s) free
          end)
          (mux_iexpr_free_vars e1) free)
        diffs ab1.(facts_free)
    in
    {| facts :=
         PShareTree.shcombine (fun _ e1b1 e2b2 =>
           match e1b1, e2b2 return _ with
             | None, _ | _, None => None
             | Some (e1, b1), Some (e2, b2) =>
               if xorb b1 b2 return _ then None
               else
                 if mux_iexpr_eqdec e1 e2 return _ then e1b1
                 else None
           end) _ ab1.(facts) ab2.(facts);
       facts_free := free;
       eqs := ab.(eqs);
       eqs_free := ab.(eqs_free)
    |}.
  Next Obligation. destruct v as [[? []]|]; auto; simpl; destruct mux_iexpr_eqdec; congruence. Qed.
  Next Obligation. destruct v as [[? []]|]; auto; simpl; destruct mux_iexpr_eqdec; congruence. Qed.

  Lemma join0_facts_correct :
    ∀ ab1 ab2 ab,
      t0_inv ab1 -> t0_inv ab2 -> t0_inv ab ->
      t0_inv (join0_facts ab1 ab2 ab) /\
      ∀ ρ, ρ ∈ (gamm0 ab1 ∪ gamm0 ab2) ->
           ρ ∈ gamm0 ab ->
           ρ ∈ gamm0 (join0_facts ab1 ab2 ab).
  Proof.
    intros.
    unfold join0_facts. split.
    - split. apply H1. simpl. intros.
      rewrite PShareTree.gshcombine in H2.
      destruct ((facts ab1)!p) as [[]|] eqn:EQ1. 2:discriminate.
      destruct ((facts ab2)!p) as [[]|] eqn:EQ2. 2:discriminate.
      rewrite PShareTree.fold_spec, <- fold_left_rev_right.
      match goal with
        | |- context [PShareTree.elements ?t] =>
          pose proof (PShareTree.elements_complete t p)
      end.
      rewrite PShareTree.gshcombine_diff, EQ1, EQ2 in H4.
      setoid_rewrite in_rev in H4.
      revert H4.
      induction rev.
      + simpl. intros. eapply H. eauto. destruct mux_iexpr_eqdec, b0, b1; simpl in H2; congruence.
      + destruct a as [x' e1]; simpl; intros.
        rewrite S.fold_spec. induction S.elements; intros. auto.
        simpl. destruct IHl0 as (s & Hs & Hsx).
        match goal with
            |- context [match ?x with _ => _ end] => destruct x eqn:EQ
        end. 2:now eauto.
        rewrite T.gsspec. destruct T.elt_eq. 2:now eauto.
        subst. eexists. split. eauto. apply PSet.mem_rem.
        fold PSet.t in *. assert (s = t1) by congruence. subst. split. auto.
        intro. subst. specialize (H4 _ (or_introl eq_refl)).
        destruct mux_iexpr_eqdec, b0, b1; simpl in H2, H4; congruence.
    - constructor; simpl. apply H3.
      intros. rewrite PShareTree.gshcombine in H4.
      destruct ((facts ab1)!p) as [[]|] eqn:EQ1. 2:discriminate.
      destruct ((facts ab2)!p) as [[]|] eqn:EQ2. 2:discriminate.
      destruct mux_iexpr_eqdec, b0, b1; inv H4; subst.
      destruct H2. apply H2 in EQ1; auto. apply H2 in EQ2; auto.
      destruct H2. apply H2 in EQ1; auto. apply H2 in EQ2; auto.
  Qed.

  Definition join0 (ab1:t0) (ab2:t0) (w:bool) : t0 :=
    let ab :=
        {| eqs := T.empty _;        eqs_free := T.empty _;
           facts := PShareTree.empty _;    facts_free := T.empty _ |}
    in
    let ab := join0_eqs ab1 ab2 ab w in
    let ab := join0_facts ab1 ab2 ab in
    {| eqs := ab.(eqs); eqs_free := ab.(eqs_free);
       facts := ab.(facts); facts_free := ab.(facts_free) |}.

  Lemma join0_correct :
    ∀ ab1 ab2 w,
      t0_inv ab1 -> t0_inv ab2 ->
      t0_inv (join0 ab1 ab2 w) /\
      ∀ ρ, ρ ∈ (gamm0 ab1 ∪ gamm0 ab2) ->
           ρ ∈ gamm0 (join0 ab1 ab2 w).
  Proof.
    intros.
    destruct (join0_eqs_correct ab1 ab2
                {| eqs := T.empty _;        eqs_free := T.empty _;
                   facts := PShareTree.empty _;    facts_free := T.empty _ |} w); auto.
    { constructor; simpl; intros.
      rewrite T.gempty in H1; discriminate.
      rewrite PShareTree.gempty in H1; discriminate. }
    edestruct (join0_facts_correct _ _ _ H H0 H1).
    split; constructor; intros.
    - eapply H3; eauto.
    - eapply H3; eauto.
    - eapply H4; eauto. eapply H2; eauto.
      constructor; simpl; intros.
      rewrite T.gempty in H7; discriminate.
      rewrite PShareTree.gempty in H7; discriminate.
    - eapply H4; eauto. eapply H2; eauto.
      constructor; simpl; intros.
      rewrite T.gempty in H7; discriminate.
      rewrite PShareTree.gempty in H7; discriminate.
  Qed.

  Program Definition join (ab1:t * @in_chan var) (ab2:t * @in_chan var) (chan:in_chan) :
    t * in_chan :=
    extract_inchan (join0 (fst ab1) (fst ab2) false) chan.
  Next Obligation. apply join0_correct; apply proj2_sig. Qed.

  Lemma join_correct :
    ∀ ab1 ab2 chan ρ,
      ρ ∈ γ ab1 \/ ρ ∈ γ ab2 -> ρ ∈ γ chan ->
      ρ ∈ γ (join ab1 ab2 chan).
  Proof.
    intros. apply extract_inchan_correct.
    apply join0_correct; try apply proj2_sig. destruct H; [left|right]; apply H. auto.
  Qed.

  Program Definition widen (ab1:t * @in_chan var) (ab2:t * @in_chan var) (chan:in_chan) :
    t * in_chan :=
    extract_inchan (join0 (fst ab1) (fst ab2) true) chan.
  Next Obligation. apply join0_correct; apply proj2_sig. Qed.

  Instance abVarExpEq_env : ab_ideal_env t :=
    {| id_gamma := gamm;
       id_leb := leb;
       id_top := top;
       id_join := join;
       id_widen := widen;
       assign := assign;
       forget := forget;
       assume := assume |}.
  Proof.
    - apply top_correct.
    - apply leb_correct.
    - apply join_correct.
    - apply assign_correct.
    - apply forget_correct.
    - apply assume_correct.
    - intros. intuition. destruct H0, H1. constructor; auto.
    - intros. intuition. destruct H0, H1. constructor; auto.
    - intros. intuition. destruct H0, H1. constructor; auto.
    - unfold assign. intros.
      destruct match e with
                 | IEvar _ x' => if eq_dec x x' then true else false
                 | _ => false
               end.
      intuition. destruct H0, H1. constructor; auto.
      destruct H. destruct iexpr_is_free.
      destruct H, H0. constructor; auto.
      destruct H, H0. constructor; auto.
    - intros. intuition. destruct H0, H1. constructor; auto.
    - intros. intuition. destruct H0, H1. constructor; auto.
  Qed.


  Global Instance abVarExpEq_env_tostring
    {_:ToString var} {_:ToString (T.t (mux_iexpr var))} : ToString t :=
    fun x =>
      (PShareTree.fold (fun acc _ fact =>
                    acc ++ "[" ++
                    let '(e, b):=fact in
                    to_string e ++ "↦" ++
                    (if b:bool then "true" else "false") ++ "]")
                 (proj1_sig x).(facts)
                 "{" ++ "}"
                 ++
      to_string (proj1_sig x).(eqs))%string.

End T.

End Make.
