(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib Util AdomLib.

Set Implicit Arguments.

Definition block_mon (T:Type) : Type := (T*list (unit -> string))%type.
Definition bm_lift {T T'} (f: T → T') (m: block_mon T) : block_mon T' :=
  (f(fst m), snd m).

Instance block_mon_gamma T T' `(gamma_op T T') : gamma_op (block_mon T) (option T') :=
  λ t, match snd t with
       | nil => fun x => match x with Some x => x ∈ γ(fst t) | None => False end
       | _ => λ _, True
       end.

Definition bm_get {T} (x: block_mon T) : option T :=
  match x with (y, nil) => Some y | _ => None end.

Definition bind_bm {A B:Type} (x:block_mon A) (f:A -> block_mon B) : block_mon B :=
  let '(x, err) := x in
  match err with
    | nil => f x (* Enforce tail call in this most general case *)
    | _ =>
      let '(y, err') := f x in
      (y, err++err')
  end.

Definition ret_bm {A:Type} (x:A) : block_mon A := (x, nil).

Notation err_bm s := (tt, (fun _ => s)::nil).

Notation "'do_bm' X <- A ; B" := (bind_bm A (fun X => B))
 (at level 200, X ident, A at level 100, B at level 200).

Notation "'do_bm' '_' <- A ; B" := (bind_bm A (fun _ => B))
 (at level 200, A at level 100, B at level 200).

Lemma bind_bm_assoc {A B C} (m: block_mon A) (f: A → block_mon B) (g: B → block_mon C) :
  bind_bm (bind_bm m f) g = bind_bm m (λ a, bind_bm (f a) g).
Proof.
  destruct m as (a, q). destruct q; simpl. auto.
  destruct (f a) as (b, q'). simpl. destruct (g b) as (c, q'').
  destruct q'; rewrite <- app_assoc; reflexivity.
Qed.

Lemma bm_bind_case {T T'} (f: block_mon T) (b: T → block_mon T') :
  ∀ P : block_mon T' → Prop,
    (∀ x y z, P (x, y :: z)) →
    (∀ x, f = (x, nil) → P (b x)) →
    P (do_bm a <- f; b a).
Proof.
  intros P A B.
  destruct f as (x & [|]). generalize (B _ eq_refl); simpl; destruct (b x); exact id.
  simpl. destruct (b x). exact (A _ _ _).
Defined.

Notation bm_assert b err :=
  (if b then ret_bm tt else err_bm err).

Lemma bm_assert_spec b err B (k: block_mon B) :
  ∀ P : _ → Prop,
    (∀ x y z, P (x, y :: z)) →
    (b = true → P k) →
    P (do_bm _ <- bm_assert b err; k).
Proof.
  intros P U V.
  apply bm_bind_case. eauto.
  intros (). destruct b. auto. inversion 1.
Qed.

Lemma bm_get_ret A (a: A) : bm_get (ret_bm a) = Some a.
Proof. reflexivity. Qed.

Definition bm_join {A} (join: A → A → A) (a b: block_mon A) : block_mon A :=
  do_bm a' <- a; do_bm b' <- b; ret_bm (join a' b').

Instance bm_wl {A} (WL: weak_lattice A) : weak_lattice (block_mon A) :=
  {| leb a b := match bm_get a, bm_get b with Some a', Some b' => leb a' b'
                                            | _, None => true
                                            | None, Some _ => false
                end
   ; top := do_bm _ <- err_bm "top"%string; ret_bm (⊤)
   ; join := bm_join join
   ; widen := bm_join widen
  |}.

Lemma bm_adom {A} {WL: weak_lattice A} {B} (G: gamma_op A B) (AD: adom A B WL G) :
  adom (block_mon A) (option B) (bm_wl WL) (block_mon_gamma G).
Proof.
  split.
+ (* gamma_monotone *)
  intros (a & [|]) (b & [|]); simpl; try (intros; exact I).
  intros LE [c|]. apply (gamma_monotone AD LE). intros (). intros; eq_some_inv.
+ (* top *) exact (λ _, I).
+ (* join *)
  intros x y c. unfold join. simpl. unfold bm_join.
  apply bm_bind_case. intros; exact I. intros a ->.
  apply bm_bind_case. intros; exact I. intros b ->.
  destruct c as [c|]. apply (join_sound _ _ WL G AD).
  compute. intuition.
Qed.

Definition bm_fold A B (f: A → B → block_mon A) (l: list B) (a: A) : block_mon A :=
  fold_left (λ acc b, do_bm acc' <- acc; f acc' b) l (ret_bm a).

Lemma bm_fold_app A B f  l l' a :
  @bm_fold A B f (l ++ l') a = do_bm a' <- bm_fold f l a; bm_fold f l' a'.
Proof.
  revert l' l a. induction l' as [|x l' IH] using rev_ind.
  intros. rewrite app_nil_r. destruct (bm_fold f l a). simpl. rewrite app_nil_r. destruct l0; reflexivity.
  intros l a. unfold bm_fold at 1. rewrite app_assoc. rewrite fold_left_app.
  fold (bm_fold f (l ++ l') a). rewrite IH. clear IH. simpl. rewrite bind_bm_assoc.
  destruct (bm_fold f l a) as (a', q). simpl.
  unfold bm_fold at 3 4. rewrite fold_left_app. reflexivity.
Qed.

Definition bm_rev_map A B (f: B → block_mon A) (l: list B) : block_mon (list A) :=
  bm_fold (λ acc b, do_bm a <- f b; ret_bm (a :: acc)) l nil.

Inductive bm_map_t (A: Type) : Type :=
| Result (a: A)
| Errorc (a: option A) (msg: unit -> string).
Notation Error a msg := (Errorc a (fun _ => msg)).
Lemma Result_Error {A} (a: A) a' m : Result a = Errorc a' m → ∀ P : Prop, P.
Proof (λ E, match E in _ = t return match t return Prop with Result _ => True | _ => _ end with eq_refl => I end).
Lemma Result_eq_inv {A} (a a': A) : Result a = Result a' → a = a'.
Proof (λ E, match E in _ = t return match t with Result x => a = x | _ => False end with eq_refl => eq_refl end).
Ltac result_eq_inv :=
  repeat match goal with
  | H : Result _ = Errorc _ _ |- _ => exact (Result_Error H _)
  | H : Errorc _ _ = Result _ |- _ => exact (Result_Error (eq_sym H) _)
  | H : Result _ = Result _ |- _ => apply Result_eq_inv in H
  end.

Definition bm_map' A B (f: B → bm_map_t A) (l: list B) : block_mon (list A) :=
  bm_fold (λ acc b, match f b with
                      | Result a => ret_bm (a :: acc)
                      | Errorc a msg =>
                        do_bm _ <- (tt, msg::nil);
                        ret_bm match a with
                                 | None => acc
                                 | Some a => a :: acc
                               end
                    end) l nil.

Lemma bm_rev_map_correct A B (f: B → block_mon A) :
  ∀ l,
    match bm_get (bm_rev_map f l) with
    | Some m =>
      (∀ y, In y m → ∃ x, bm_get (f x) = Some y ∧ In x l)
    ∧ (∀ x, In x l → ∃ y, bm_get (f x) = Some y ∧ In y m)
    | None => ∃ x, In x l ∧ bm_get (f x) = None
    end.
Proof.
  induction l using rev_ind.
+ split; intros ? ().
+ unfold bm_rev_map, bm_fold. rewrite fold_left_app. simpl.
  fold (bm_fold (λ acc b, do_bm a <- f b; ret_bm (a :: acc)) l nil).
  fold (bm_rev_map f l).
  destruct (bm_rev_map f l) as (acc', [|]).
  destruct IHl as (IH & HI).
  destruct (f x) as (a, [|]) eqn: FX.
  simpl. split.
  - intros y [->|IN].
    exists x; split. rewrite FX. reflexivity. rewrite in_app. right. left. reflexivity.
    destruct (IH _ IN) as (? & ? & ?).
    eexists; split; eauto. rewrite in_app. simpl. eauto.
  - intros x' IN. rewrite in_app in IN. destruct IN as [IN|[->|()]].
    destruct (HI x' IN) as (? & ? & H). eauto.
    rewrite FX. eexists. split. reflexivity. auto.
  - exists x. split. rewrite in_app. right. left. reflexivity.
    rewrite FX. reflexivity.
  - destruct IHl as (x' & ? & ?).
    simpl. destruct (do_bm a <- f x; ret_bm (a::acc')).
    exists x'. intuition.
Qed.

Lemma bm_rev_map_inv A B (f: B → block_mon A) :
  ∀ l,
    match bm_get (bm_rev_map f l) with
    | Some m =>
      m = rev (map (fst ∘ f) l)
    | None => True
    end.
Proof.
  induction l using rev_ind.
+ split; intros ? ().
+ unfold bm_rev_map, bm_fold. rewrite fold_left_app. simpl.
  fold (bm_fold (λ acc b, do_bm a <- f b; ret_bm (a :: acc)) l nil).
  fold (bm_rev_map f l).
  destruct (bm_rev_map f l) as (acc', [|]).
  destruct (f x) as (a, [|]) eqn: FX.
  simpl. rewrite map_app, rev_app_distr, IHl. simpl.
  rewrite FX. simpl.
  simpl. reflexivity.
  exact I.
  simpl. destruct (do_bm a <- f x; ret_bm (a::acc')). exact I.
Qed.

Lemma bm_map'_correct A B (f: B → bm_map_t A) :
  ∀ l,
    match bm_get (bm_map' f l) with
    | Some m =>
      (∀ y, In y m → ∃ x, f x = Result y ∧ In x l)
    ∧ (∀ x, In x l → ∃ y, f x = Result y ∧ In y m)
    | None => True (*∃ x, In x l ∧ bm_not_blocked (f x) = false*)
    end.
Proof.
  induction l using rev_ind.
+ split; intros ? ().
+ unfold bm_map', bm_fold. rewrite fold_left_app. simpl.
  apply bm_bind_case. intros; exact I.
  intros al.
  intros AL.
  destruct (f x) eqn: FX. 2: exact I.
  unfold bm_map', bm_fold in IHl. simpl in IHl.
  rewrite AL in IHl.
  destruct IHl as (IH & HI).
  simpl; split.
  - intros y [->|IN].
    exists x; split. rewrite FX. reflexivity. rewrite in_app. right. left. reflexivity.
    destruct (IH _ IN) as (? & ? & ?).
    eexists; split; eauto. rewrite in_app. simpl. eauto.
  - intros x' IN. rewrite in_app in IN. destruct IN as [IN|[->|()]].
    destruct (HI x' IN) as (? & ? & H). eauto.
    rewrite FX. eexists. split. reflexivity. auto.
Qed.

Ltac with_bm_map' :=
  match goal with
  |- appcontext[ bm_map' ?f ?l ] =>
  let map_f := fresh "map_f" in
  remember f as map_f;
  generalize (bm_map'_correct map_f l);
    destruct (bm_map' map_f l) as (? & [|]);
    [subst map_f| easy ]
  end.
