(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Util.
Require Import Maps.
Require Import Integers.
Require Import Values.
Require Import Memory.
Require Import AdomLib.
Require Import Utf8.
Require Import AST.
Require Import Events.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Behaviors.
Require Import Csharpminor.
Require Import CsharpminorLogic.
Require Import CsharpminorIter.
Require Import BlockMon.
Require Import AbMemSignatureCsharpminor.
Require Import AnyDefs.

Section PROG.

Variable (unroll: nat).
Variable (prog: program).
Let ge := Genv.globalenv prog.

Lemma eval_expr_dec:
  ∀ e t m exp,
    (∃ v, eval_expr ge e t m exp v) \/ (∀ v, ~eval_expr ge e t m exp v).
Proof.
  intros.
  edestruct Classical_Prop.classic.
  left. apply H.
  right. repeat intro. eauto.
Qed.

Lemma eval_exprlist_dec:
  ∀ e t m expl,
    (∃ vl, eval_exprlist ge e t m expl vl) \/
    (∀ vl, ~eval_exprlist ge e t m expl vl).
Proof.
  induction expl; intros. repeat econstructor.
  destruct IHexpl as [[vl Hvl]|].
  edestruct eval_expr_dec as [[v Hv]|].
  repeat econstructor; eauto.
  right. repeat intro. inv H0. intuition eauto.
  right. repeat intro. inv H0. intuition eauto.
Qed.

Section ABMEM.

Variable abstate: Type.

Variable abmem : mem_dom abstate.
Existing Instance abmem.

Variable mem_gamma : gamma_op abstate concrete_state.
Existing Instance mem_gamma.

Variable mem_spec : mem_dom_spec ge abmem mem_gamma.

Lemma mem_join_sound:
  ∀ (ab1 ab2:abstate), γ ab1 ∪ γ ab2 ⊆ γ (ab1 ⊔ ab2).
Proof. apply join_sound. eapply mem_adom. eauto. Qed.
Hint Resolve mem_join_sound.

Lemma mem_join_sound_bot:
  ∀ (ab1 ab2:abstate+⊥), γ ab1 ∪ γ ab2 ⊆ γ (ab1 ⊔ ab2).
Proof. apply join_sound. eapply lift_bot. eapply mem_adom. eauto. Qed.
Hint Resolve mem_join_sound_bot.

Lemma ab_builtin_sound:
  ∀ s m ab, (s, m) ∈ γ ab ->
  ∀ ef x args ab', ab_builtin _ _ ef x args ab = (ab', nil) ->
                   (s = nil ->
                    args = nil /\
                    ef_sig ef = mksignature nil (Some AST.Tint) cc_default) ->
    ∃ vargs,
      match s with
        | nil => vargs = nil
        | (_, (t, e))::_ => eval_exprlist ge e t m args vargs
      end /\
      (∃ tr vres m', external_call ef ge vargs m tr vres m') /\
       ∀ tr vres m', external_call ef ge vargs m tr vres m' ->
         (match s with
           | nil => nil
           | (f, (t, e)) :: s' => (f, (Cminor.set_optvar x vres t, e)) :: s'
          end, m') ∈ γ ab' /\
         (s = nil -> ∃ r, vres = Vint r).
Proof.
  intros.
  destruct ef; simpl; try now inv H0.
  - simpl in H0.
    destruct ident_eq; [|destruct ident_eq]; try discriminate;
    destruct signature_eq; try discriminate;
    destruct args; try discriminate.
    + exists nil. split. destruct s. auto. destruct p. destruct p. constructor.
      subst. rewrite any_int64_sem_eq. split.
      econstructor. eexists (Vlong Int64.zero). repeat econstructor.
      intros. destruct H2 as (r & ? & ? & ? & ?). subst.
      split.
      * destruct s. destruct H1; auto. discriminate.
        destruct p. destruct p. destruct x.
        pose proof (assign_any_sound ge _ _ mem_spec i0 TypesDomain.AbTy.VL ab
                     (Some ((i, (Cminor.set_optvar (Some i0) (Vlong r) t, e)) :: s, m))).
        rewrite H0 in H3. apply H3. repeat econstructor. auto. constructor.
        inv H0. auto.
      * intro. subst. destruct H1; auto. discriminate.
    + exists nil. split. destruct s. auto. destruct p. destruct p. constructor.
      subst. rewrite any_double_sem_eq. split.
      econstructor. eexists (Vfloat Floats.Float.zero). repeat econstructor.
      intros. destruct H2 as (r & ? & ? & ? & ?). subst.
      split.
      * destruct s. destruct H1; auto. discriminate.
        destruct p. destruct p. destruct x.
        pose proof (assign_any_sound ge _ _ mem_spec i0 TypesDomain.AbTy.VF ab
                     (Some ((i, (Cminor.set_optvar (Some i0) (Vfloat r) t, e)) :: s, m))).
        rewrite H0 in H3. apply H3. repeat econstructor. auto. constructor.
        inv H0. auto.
      * intro. subst. destruct H1; auto. discriminate.
  - destruct args; try now inv H0.
    destruct args; try now inv H0.
    destruct s as [|[? []]]. destruct (H1 eq_refl). discriminate. clear H1.
    assert (A:=malloc_sound _ _ _ mem_spec e x ab).
    simpl in H0. rewrite H0 in A.
    destruct (eval_expr_dec e0 t m e) as [[varg VARG]|].
    destruct varg;
      try (destruct (A None); econstructor; eauto; intros; inv H2; determ; congruence).
    destruct (Mem.alloc m (-4) (Int.unsigned i0)) eqn:?.
    destruct (Mem.store Mint32 m0 b (-4) (Vint i0)) eqn:?.
    + refine (let A':=A _ _ in _).
      { econstructor. eauto.
        intros. inv H2. determ.
        replace b0 with b by congruence.
        replace m' with m1 by congruence. auto. }
      repeat (econstructor; try discriminate); eauto.
      intros. apply (A (Some _)).
      econstructor. eauto. inv H1.
      intros. determ. inv H2. repeat f_equal; congruence.
    + destruct (A None).
      econstructor; eauto. intros. inv H2. determ. congruence.
    + destruct (A None). econstructor. eauto.
      intros. edestruct H1; eauto.
  - destruct args; try now inv H0.
    destruct args; try now inv H0.
    destruct x; try now inv H0.
    assert (A:=free_sound _ _ _ mem_spec e ab).
    simpl in H0. rewrite H0 in A.
    destruct s as [|[? []]]. destruct (H1 eq_refl). discriminate. clear H1.
    destruct (eval_expr_dec e0 t m e) as [[varg VARG]|].
    destruct varg;
      try (destruct (A None); econstructor; eauto; intros; inv H2; determ; congruence).
    destruct (Mem.load Mint32 m b (Int.unsigned i0 - 4)) eqn:?.
    destruct v;
      try (destruct (A None); econstructor; eauto; intros; inv H2; determ; congruence).
    destruct (Z_gt_dec (Int.unsigned i1) 0).
    destruct (Mem.free m b (Int.unsigned i0 - 4) (Int.unsigned i0 + Int.unsigned i1)) eqn:?.
    + refine (let A':=A _ _ in _).
      { econstructor. eauto.
        intros. inv H2. determ. replace m' with m0 by congruence. auto. }
      repeat (econstructor; try discriminate); eauto.
      intros. apply (A (Some _)).
      econstructor. eauto. inv H1.
      intros. determ. inv H2. repeat f_equal; congruence.
    + destruct (A None).
      econstructor; eauto. intros. inv H2. determ. congruence.
    + destruct (A None).
      econstructor; eauto. intros. inv H2. determ.
      assert (i1 = sz) by congruence. subst. destruct n. auto.
    + destruct (A None).
      econstructor; eauto. intros. inv H2. determ. congruence.
    + destruct (A None). econstructor. eauto.
      intros. edestruct H1; eauto.
  - destruct args; try now inv H0.
    destruct args; try now inv H0.
    destruct args; try now inv H0.
    destruct x; try now inv H0.
    destruct s as [|[? []]]. destruct (H1 eq_refl). discriminate. clear H1.
    assert (A:=memcpy_sound _ _ _ mem_spec sz al e e0 ab).
    simpl in H0. rewrite H0 in A.
    destruct (eval_expr_dec e1 t m e) as [[varg0 VARG0]|].
    destruct (eval_expr_dec e1 t m e0) as [[varg1 VARG1]|].
    destruct varg0;
      try (destruct (A None); econstructor; eauto; intros; inv H3; determ; congruence).
    destruct varg1;
      try (destruct (A None); econstructor; eauto; intros; inv H3; determ; congruence).
    destruct (Mem.loadbytes m b0 (Int.unsigned i1) sz) eqn:?.
    destruct (Mem.storebytes m b (Int.unsigned i0) l) eqn:?.
    assert ((al = 1 ∨ al = 2 ∨ al = 4 ∨ al = 8) \/
            ~(al = 1 ∨ al = 2 ∨ al = 4 ∨ al = 8)) by omega.
    destruct H1.
    assert ((sz >= 0 /\ (al | sz) /\ (sz > 0 -> (al | Int.unsigned i1)) /\ (sz > 0 -> (al | Int.unsigned i0))) \/
            ~(sz >= 0 /\ (al | sz) /\ (sz > 0 -> (al | Int.unsigned i1)) /\ (sz > 0 -> (al | Int.unsigned i0))))
    by tauto.
    destruct H2.
    assert ((b0 ≠ b ∨ Int.unsigned i1 = Int.unsigned i0
            ∨ Int.unsigned i1 + sz <= Int.unsigned i0
            ∨ Int.unsigned i0 + sz <= Int.unsigned i1) \/
           ~(b0 ≠ b ∨ Int.unsigned i1 = Int.unsigned i0
            ∨ Int.unsigned i1 + sz <= Int.unsigned i0
            ∨ Int.unsigned i0 + sz <= Int.unsigned i1)) by tauto.
    destruct H3.
    + refine (let A':=A _ _ in _).
      { econstructor. eauto.
        intros. inv H6. determ.
        replace m' with m0 by congruence. auto. }
      eexists. split. repeat econstructor; eauto.
      split. eexists. eexists. eexists.
      econstructor; eauto; try tauto.
      intros. split.
      apply (A (Some _)). econstructor. eauto. inv H4.
      intros. determ. inv H6. repeat f_equal; congruence.
      discriminate.
    + destruct (A None).
      econstructor; eauto. intros. inv H6. determ. congruence.
    + destruct (A None).
      econstructor; eauto. intros. inv H5. determ.
      inv EQ. inv EQ0. destruct H2. auto.
    + destruct (A None).
      econstructor; eauto. intros. inv H4. destruct H1. auto.
    + destruct (A None).
      econstructor; eauto. intros. inv H3. determ. congruence.
    + destruct (A None).
      econstructor; eauto. intros. inv H3. determ. congruence.
    + destruct (A None). econstructor. eauto.
      intros. edestruct H1; eauto.
    + destruct (A None). econstructor. eauto.
      intros. edestruct H1; eauto.
Qed.

Definition id_function (id:ident) (f:fundef) :=
  match Genv.find_symbol ge id with
    | Some b =>
      match Genv.find_funct ge (Vptr b Int.zero) with
        | Some f' => f' = f
        | _ => False
      end
    | None => False
  end.

Section stmt_iterator_proof.

Variable iter_function:
  forall (sig:signature)
         (name: ident) (function:fundef) (rettemp:option ident)
         (args:list expr) (state:abstate), block_mon (abstate+⊥).
Variable rettemp: option ident.
Variable f : function.
Variable env : env.
Variable stk : stack.

Inductive concrete_state_decomp:
  concrete_state -> temp_env -> mem -> Prop :=
| CSD :
    ∀ idf le cs_stk mem,
      id_function idf (Internal f) ->
      List.Forall2
        (fun cellcs cellstk =>
           let '(idcs, (lecs, ecs)) := cellcs in
           let '(fstk, estk, lestk) := cellstk in
           id_function idcs (Internal fstk) /\
           lecs = lestk /\
           ecs = estk) cs_stk stk ->
      concrete_state_decomp ((idf, (le, env)) :: cs_stk, mem) le mem.

Definition pred_of_ab (ab:abstate+⊥) (le:temp_env) (mem:mem) : Prop :=
  ∃ cs, concrete_state_decomp cs le mem /\ cs ∈ γ ab.

Lemma join_pred_of_ab:
  ∀ ab1 ab2 le mem,
    pred_of_ab ab1 le mem \/ pred_of_ab ab2 le mem ->
    pred_of_ab (ab1 ⊔ ab2) le mem.
Proof.
  unfold pred_of_ab.
  intros ab1 ab2 le mem [(cs & CS & AB)|(cs & CS & AB)]; eauto.
Qed.

Lemma join_exit_pred_of_ab:
  ∀ ab1 ab2 le mem x,
    pred_of_ab (nth x ab1 Bot) le mem \/ pred_of_ab (nth x ab2 Bot) le mem ->
    pred_of_ab (nth x (join_exit _ _ ab1 ab2) Bot) le mem.
Proof.
  intros ab1 ab2 le mem x. revert x ab1 ab2 le mem.
  induction x; destruct ab1, ab2;
    try (apply IHx; now auto);
    try (apply join_pred_of_ab; now auto);
    intros le mem [H|H];
    now auto || now destruct H as (cs & CS & []).
Qed.

Definition ret_pred_of_ab (ab:abstate+⊥) (v:val) (mem:mem) : Prop :=
  ∃ cs_stk,
    List.Forall2
      (fun cellcs cellstk =>
         let '(idcs, (lecs, ecs)) := cellcs in
         let '(fstk, estk, lestk) := cellstk in
         id_function idcs (Internal fstk) /\
         lecs = lestk /\
         ecs = estk) cs_stk stk /\
    match cs_stk with
      | nil => (nil, mem) ∈ γ ab /\
               rettemp = Some 1%positive /\
               ∃ r, v = Vint r
      | (idf, (le', e')) :: cs_stk =>
        ((idf, (Cminor.set_optvar rettemp v le', e')) :: cs_stk, mem) ∈ γ ab
    end.

Lemma join_ret_pred_of_ab:
  ∀ ab1 ab2 v mem,
    ret_pred_of_ab ab1 v mem \/ ret_pred_of_ab ab2 v mem ->
    ret_pred_of_ab (ab1 ⊔ ab2) v mem.
Proof.
  unfold ret_pred_of_ab.
  intros ab1 ab2 v mem [([|[idf [le' e']] cs_stk] & CSSTK & AB)|
                        ([|[idf [le' e']] cs_stk] & CSSTK & AB)];
  eexists; (split; [apply CSSTK|]).
  - destruct AB. auto.
  - cbv iota beta. auto.
  - destruct AB. auto.
  - cbv iota beta. auto.
Qed.

Definition goto_pred_of_ab (ab:PTree.t abstate) (lbl:ident): temp_env -> mem -> Prop :=
  pred_of_ab
    match ab!lbl with
      | Some a => NotBot a
      | None => Bot
    end.

Lemma join_goto_pred_of_ab:
  ∀ ab1 ab2 lbl le mem,
    goto_pred_of_ab ab1 lbl le mem \/ goto_pred_of_ab ab2 lbl le mem ->
    goto_pred_of_ab (join_goto _ _ ab1 ab2) lbl le mem.
Proof.
  unfold join_goto, goto_pred_of_ab. intros. rewrite PTree.gcombine by auto.
  destruct PTree.get, PTree.get.
  apply join_pred_of_ab in H. auto.
  destruct H as [? | (? & ? & [])]. auto.
  destruct H as [(? & ? & []) | ?]. auto.
  destruct H as [(? & ? & []) | (? & ? & [])].
Qed.

Hypothesis iter_function_ok:
  ∀ sig name fundef rettemp args i o,
    id_function name fundef ->
    iter_function sig name fundef rettemp args i = (o, nil) ->
  ∀ le mem, pred_of_ab (NotBot i) le mem ->
    ∃ vargs,
      eval_exprlist ge env le mem args vargs /\
      funsig fundef = sig /\
      hoare_fun prog ((f, env, le)::stk)
                (fun vargs' mem' => vargs = vargs' /\ mem = mem')
                fundef
                (fun ret mem' =>
                   pred_of_ab o (Cminor.set_optvar rettemp ret le) mem').

Section lfp.

Variable igoto:PTree.t abstate.
Variable s:stmt.
Variable it:abstate +⊥ → block_mon (outcome abstate).
Variable low:abstate +⊥.

Hypothesis it_ok:
  ∀ ab o,
    it ab = (o, nil) ->
    hoare_stmt prog f env stk
               (pred_of_ab (low ⊔ ab)) (goto_pred_of_ab igoto)
               s
               (pred_of_ab o.(onormal _)) (ret_pred_of_ab o.(oreturn _))
               (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
               (goto_pred_of_ab o.(ogoto _)).

Definition is_invariant o :=
  ∃ inv,
    (∀ le mem, (inv le mem:Prop) -> pred_of_ab o.(onormal _) le mem) /\
    hoare_stmt prog f env stk
               (fun le mem => pred_of_ab low le mem \/ inv le mem)
               (goto_pred_of_ab igoto)
               s
               inv
               (ret_pred_of_ab o.(oreturn _))
               (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
               (goto_pred_of_ab o.(ogoto _)).

Lemma narrow_ok:
  ∀ n ab o,
    narrow _ _ it n (it ab) = (o, nil) ->

    (∀ o', it ab = (o', nil) -> is_invariant o') ->

    is_invariant o.
Proof.
  induction n; intros. apply H0; auto.
  unfold narrow in H. fold (narrow _ _ it n (it (onormal abstate (fst (it ab))))) in H.
  destruct @is_bot in H. apply H0; auto.
  destruct (it ab) as [o' [|]] eqn:?.
  - simpl in H. destruct (it (onormal abstate o')) as [o'' [|]] eqn:?; simpl in H.
    2:inv H; eauto.
    eapply (IHn o'.(onormal _)). congruence.
    intros. edestruct H0 as (inv & invle & INV). eauto. clear H0.
    rewrite Heqb0 in H1. inv H1.
    eexists. split.
    2:eapply hoare_stmt_post_impl, hoare_stmt_post_and, hoare_stmt_pre_impl, INV; eauto.
    5:eapply hoare_stmt_pre_impl, it_ok, Heqb0.
    simpl; now intuition.  simpl; now intuition.
    simpl; now intuition.  simpl; now intuition.
    destruct 1. auto using join_pred_of_ab.
    destruct H0. auto using join_pred_of_ab.
    now auto. simpl; now intuition.
  - Opaque leb. simpl in H. Transparent leb.
    destruct ((fst (it (onormal abstate o'))).(onormal _) ⊑ o'.(onormal _)) eqn:?.
    2:discriminate.
    eapply IHn. eauto.
    intros. rewrite H1 in Heqb0. eexists. split. eauto.
    eapply hoare_stmt_pre_impl, it_ok, H1; auto.
    intros le mem [|(cs & CS & AB)]. auto using join_pred_of_ab.
    eexists. split. eauto.
    eapply join_sound. eapply lift_bot, mem_adom, mem_spec. right.
    eapply gamma_monotone. eapply lift_bot, mem_adom, mem_spec. eauto. auto.
Qed.

Lemma lfp_ok:
  ∀ o, lfp _ _ it = (o, nil) -> is_invariant o.
Proof.
  unfold lfp. generalize (@Bot abstate).
  induction fuel.
  - intros. eexists. split. eauto.
    eapply hoare_stmt_pre_impl, it_ok, H; auto.
    intros le mem [|(cs & CS & AB)]. auto using join_pred_of_ab.
    eexists. split. eauto.
    eapply join_sound. eapply lift_bot, mem_adom, mem_spec. right.
    apply gamma_top. eapply lift_bot, mem_adom, mem_spec.
  - intros. destruct ((fst (it b)).(onormal _) ⊑ b) eqn:?.
    + eapply narrow_ok; eauto.
      eexists. split. eauto.
      eapply hoare_stmt_pre_impl, it_ok, H0; auto.
      intros le mem [|(cs & CS & AB)]. auto using join_pred_of_ab.
      eexists. split. eauto.
      eapply join_sound. eapply lift_bot, mem_adom, mem_spec. right.
      eapply gamma_monotone; eauto. eapply lift_bot, mem_adom, mem_spec.
      rewrite H0. eauto.
    + eapply IHn; eauto.
Qed.

End lfp.

Section lfp_goto.

Variable inormal:abstate+⊥.
Variable it:PTree.t abstate → block_mon (outcome abstate).

Hypothesis it_ok:
  ∀ ab o,
    it ab = (o, nil) ->
    hoare_stmt prog f env stk
               (pred_of_ab inormal) (goto_pred_of_ab ab)
               f.(fn_body)
               (pred_of_ab o.(onormal _)) (ret_pred_of_ab o.(oreturn _))
               (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
               (goto_pred_of_ab o.(ogoto _)).

Definition is_goto_invariant o :=
  ∃ inv,
    (∀ lbl le mem, (inv lbl le mem:Prop) -> goto_pred_of_ab o.(ogoto _) lbl le mem) /\
    hoare_stmt prog f env stk
               (pred_of_ab inormal) inv
               f.(fn_body)
               (pred_of_ab o.(onormal _)) (ret_pred_of_ab o.(oreturn _))
               (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
               inv.

Definition incl_goto a b :=
  ∀ l,
    match a!l with
      | None => True
      | Some a' =>
        match b!l with
          | None => False
          | Some b' => γ a' ⊆ γ b'
        end
    end.

Lemma leb_goto_incl_goto:
  ∀ a b, leb_goto _ _ a b = true -> incl_goto a b.
Proof.
  intros.
  unfold leb_goto in H. rewrite PTree.bempty_correct in H.
  intro. specialize (H l).
  rewrite PTree.gcombine in H by easy.
  edestruct PTree.get; try easy.
  edestruct PTree.get; try easy.
  destruct (a0 ⊑ a1) eqn:?; try easy.
  eapply gamma_monotone; eauto. eapply mem_adom, mem_spec.
Qed.

Lemma narrow_goto_ok:
  ∀ n ab o,
    narrow_goto _ _ it n (it ab) = (o, nil) ->

    (∀ o', it ab = (o', nil) -> is_goto_invariant o') ->

    is_goto_invariant o.
Proof.
  induction n; intros. apply H0; auto.
  unfold narrow_goto in H. fold (narrow_goto _ _ it n (it (ogoto _ (fst (it ab))))) in H.
  destruct (leb_goto _ _ ((fst (it ab)).(ogoto _)) (bot_goto abstate)) eqn:?.
  apply H0; auto.
  destruct (it ab) as [o' [|]] eqn:?.
  - simpl in H. destruct (it (ogoto abstate o')) as [o'' [|]] eqn:?; simpl in H.
    2:inv H; eauto.
    eapply (IHn o'.(ogoto _)). congruence.
    intros. edestruct H0 as (inv & invle & INV). eauto. clear H0.
    rewrite Heqb1 in H1. inv H1.
    eexists. split.
    2:eapply hoare_stmt_post_impl, hoare_stmt_post_and, hoare_stmt_pre_impl, INV; eauto.
    5:eapply hoare_stmt_pre_impl, it_ok, Heqb1.
    simpl; now intuition.  simpl; now intuition.
    simpl; now intuition.  simpl; now intuition. now auto.
    destruct 1. auto using join_pred_of_ab.
    simpl; now intuition.
  - Opaque leb. simpl in H. Transparent leb.
    destruct (leb_goto _ _ (fst (it (ogoto abstate o'))).(ogoto _) o'.(ogoto _)) eqn:?.
    2:discriminate.
    eapply IHn. eauto.
    intros. rewrite H1 in Heqb1. eexists. split. eauto.
    eapply hoare_stmt_pre_impl, it_ok, H1; auto.
    intros lbl le mem (cs & CS & AB).
    eexists. split. eauto. eapply leb_goto_incl_goto in Heqb1.
    specialize (Heqb1 lbl). simpl in Heqb1.
    destruct PTree.get, PTree.get; try contradiction; auto. apply Heqb1, AB.
Qed.

Lemma lfp_goto_ok:
  ∀ o, lfp_goto _ _ it = (o, nil) -> is_goto_invariant o.
Proof.
  unfold lfp_goto. generalize (bot_goto abstate) at 2.
  induction fuel.
  - destruct (it (bot_goto abstate)); discriminate.
  - intros. destruct (leb_goto _ _ (fst (it t)).(ogoto _) t) eqn:?.
    2:eapply IHn, H.
    eapply narrow_goto_ok. eauto.
    intros. eexists. split. 2:eapply hoare_stmt_pre_impl, it_ok, H0. auto. auto.
    intros. destruct H1 as (cs & CS & AB). exists cs. split. auto.
    eapply leb_goto_incl_goto in Heqb. specialize (Heqb lbl).
    rewrite H0 in Heqb. simpl in Heqb.
    destruct PTree.get, PTree.get; try contradiction. apply Heqb, AB.
Qed.

End lfp_goto.

Lemma iter_ok:
  ∀ s inormal igoto o,
    iter _ unroll _ iter_function rettemp inormal igoto s = (o, nil) ->
    hoare_stmt prog f env stk
               (pred_of_ab inormal) (goto_pred_of_ab igoto)
               s
               (pred_of_ab o.(onormal _)) (ret_pred_of_ab o.(oreturn _))
               (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
               (goto_pred_of_ab o.(ogoto _))
with iter_switch_ok:
  ∀ lbl_s islong inormal idefault iprev igoto e o,
    iter_switch _ unroll _ iter_function rettemp islong inormal
                idefault iprev igoto e lbl_s = (o, nil) ->
    hoare_lbl_stmt prog f env stk
                   (fun n le mem =>
                      eval_expr ge env le mem e
                                (if islong then Vlong (Int64.repr n)
                                 else Vint (Int.repr n)) /\
                      pred_of_ab inormal le mem)
                   (pred_of_ab idefault) (pred_of_ab iprev)
                   (goto_pred_of_ab igoto)
                   lbl_s
                   (pred_of_ab o.(onormal _)) (ret_pred_of_ab o.(oreturn _))
                   (fun x => pred_of_ab (List.nth x o.(oexit _) Bot))
                   (goto_pred_of_ab o.(ogoto _)).
Proof.
  2:destruct lbl_s; intros.
  destruct s; intros.
  - (* Sskip *)
    clear iter_ok iter_switch_ok.
    inv H. apply hoare_Sskip. auto.
  - (* Sset *)
    clear iter_ok iter_switch_ok.
    apply hoare_Sset. intros le mem (cs & CS & AB). destruct CS.
    destruct inormal as [|inormal]. contradiction.
    revert H. unfold iter. simpl. apply bm_bind_case. discriminate.
    intros out OUT eq. inv eq.
    assert (A:=assign_sound ge _ _ mem_spec i e inormal). rewrite OUT in A.
    destruct (eval_expr_dec env le mem e) as [[v Hv]|].
    + eexists. split. eauto. eexists. split. constructor; eauto.
      apply (A (Some _)). econstructor; eauto. intros. determ. auto.
    + destruct (A None). econstructor; eauto. intros. edestruct H. eauto.
  - (* Sstore*)
    clear iter_ok iter_switch_ok.
    apply hoare_Sstore. intros le mem (cs & CS & AB). destruct CS.
    destruct inormal as [|inormal]. contradiction.
    revert H. unfold iter. simpl. apply bm_bind_case. discriminate.
    intros out OUT eq. inv eq.
    assert (S:=store_sound ge _ _ mem_spec m e e0 inormal). rewrite OUT in S.
    destruct (eval_expr_dec env le mem e) as [[v Hv]|].
    destruct (eval_expr_dec env le mem e0) as [[v0 Hv0]|].
    destruct (Mem.storev m mem v v0) eqn:?.
    + eexists. eexists. eexists. split. eauto. split. eauto. split. eauto.
      eexists. split. constructor; eauto.
      apply (S (Some _)). econstructor; eauto. intros. determ. auto.
    + destruct (S None). econstructor; eauto. intros. determ. congruence.
    + destruct (S None). econstructor; eauto. intros. edestruct H. eauto.
    + destruct (S None). econstructor; eauto. intros. edestruct H. eauto.
  - (* Scall*)
    clear iter_ok iter_switch_ok.
    apply hoare_Scall. intros le mem (cs & CS & AB). destruct CS.
    destruct inormal as [|inormal]. contradiction.
    revert H. unfold iter, bind_normal_outcome.
    apply bm_bind_case. discriminate. intros out1 OUT1 eq. inv eq.
    revert OUT1. apply bm_bind_case. discriminate. intros out2 OUT1 OUT2.
    assert (D:=deref_fun_sound ge _ _ mem_spec e inormal). rewrite OUT1 in D.
    destruct (eval_expr_dec env le mem e) as [[vf Hvf]|].
    destruct (Genv.find_funct (Genv.globalenv prog) vf) eqn:?.
    destruct vf; try discriminate.
    destruct (Genv.invert_symbol ge b) eqn:?.
    destruct (Int.eq_dec i Int.zero). subst i.
    + assert (In (i0,f0) out2).
      { apply (D (Some _)). econstructor; eauto.
        intros. determ. inv EQ.
        apply Genv.find_invert_symbol in H3.
        simpl in Heqo0. destruct Int.eq_dec in Heqo0; [|discriminate].
        unfold ge in H2. congruence. }
      rewrite <- fold_left_rev_right in OUT2.
      rewrite in_rev in H. clear OUT1 D.
      revert out1 OUT2. induction (rev out2). contradiction.
      unfold fold_right.
      apply bm_bind_case. discriminate. intros out1 OUT1.
      apply bm_bind_case. discriminate. intros out3 OUT3.
      intros out4 eq. apply pair_eq_inv in eq. destruct eq. subst.
      destruct H.
      * subst. edestruct iter_function_ok as (vargs & VARGS & SIG & HOARE). 2:eauto.
        simpl. unfold id_function.
        apply Genv.invert_find_symbol in Heqo1. unfold ge at 2. rewrite Heqo1, Heqo0. auto.
        eexists. split. econstructor; eauto. eauto.
        eexists. eexists. eexists. split. eauto. split. eauto. split. eauto. split. eauto.
        eapply hoare_fun_post_impl, HOARE. eauto using join_pred_of_ab.
      * edestruct IHl0 as (vf & vargs & fd & EVAL & EVALS & FIND & SIG & HOARE); eauto.
        eexists. eexists. eexists. split. eauto. split. eauto. split. eauto. split. eauto.
        eapply hoare_fun_post_impl, HOARE. eauto using join_pred_of_ab.
    + destruct (D None). econstructor; eauto.
      intros. determ. congruence.
    + destruct (D None). econstructor; eauto.
      intros. apply Genv.find_invert_symbol in H3. determ. congruence.
    + destruct (D None). econstructor; eauto.
      intros. determ. subst. simpl in Heqo0.
      unfold ge in H2. destruct Int.eq_dec; try congruence.
    + destruct (D None). econstructor; eauto.
      intros. edestruct H; eauto.
  - (* Sbuiltin *)
    clear iter_ok iter_switch_ok.
    apply hoare_Sbuiltin.
    intros le mem (cs & CS & AB). destruct CS.
    destruct inormal as [|inormal]. contradiction.
    unfold iter, bind_normal_outcome in H.
    revert H. apply bm_bind_case. discriminate. intros out1 OUT1 eq. inv eq.
    edestruct ab_builtin_sound as (vargs & VARGS & exEXTCALL & EXTCALL); eauto. discriminate.
    eexists. split. apply VARGS. split. auto.
    intros. eexists. split. constructor; eauto.
    eapply EXTCALL. eauto.
  - (* Sseq *)
    assert (IHs1:=iter_ok s1). assert (IHs2:=iter_ok s2).
    clear iter_ok iter_switch_ok.
    revert H. unfold iter.
    apply bm_bind_case. discriminate. intros out1 OUT1.
    apply bm_bind_case. discriminate. intros out2 OUT2.
    intros eq. apply pair_eq_inv in eq. destruct eq. subst.
    eapply hoare_Sseq;
      [eapply hoare_stmt_post_impl, IHs1, OUT1 |
       eapply hoare_stmt_post_impl, IHs2, OUT2]; eauto.
    + eauto using join_ret_pred_of_ab.
    + eauto using join_exit_pred_of_ab.
    + eauto using join_goto_pred_of_ab.
    + eauto using join_ret_pred_of_ab.
    + eauto using join_exit_pred_of_ab.
    + eauto using join_goto_pred_of_ab.
  - (* Sifthenelse *)
    assert (IHs1:=iter_ok s1). assert (IHs2:=iter_ok s2).
    clear iter_ok iter_switch_ok.
    revert H. unfold iter.
    apply bm_bind_case. discriminate. intros [out1t out1f] OUT1.
    apply bm_bind_case. discriminate. intros out2 OUT2.
    apply bm_bind_case. discriminate. intros out3 OUT3.
    intros eq. apply pair_eq_inv in eq. destruct eq. subst.
    eapply hoare_Sifthenelse;
      [|eapply hoare_stmt_post_impl, IHs1, OUT2 |
        eapply hoare_stmt_post_impl, IHs2, OUT3]; eauto.
    + intros le mem (cs & CS & AB). destruct CS.
      destruct inormal as [|inormal]. contradiction.
      assert (A:=assume_sound ge _ _ mem_spec e inormal). rewrite OUT1 in A.
      destruct (eval_expr_dec env le mem e) as [[v Hv]|].
      assert ((exists b, Val.bool_of_val v b) \/ (∀ b, ~Val.bool_of_val v b)).
      { destruct v; try now (right; repeat intro; inv H2). repeat econstructor. repeat econstructor. }
      destruct H2 as [[b Hb]|].
      * eexists. eexists. split. eauto. split. eauto.
        assert (A':=fun cs => A (Some (b, cs))). clear A.
        destruct b; eexists; (split; [constructor; eauto|
                                      apply A'; econstructor; [eauto|
                                                               intros; determ; auto]]).
      * destruct (A None). econstructor; eauto.
        intros. determ. edestruct H2. eauto.
      * destruct (A None). econstructor; eauto.
        intros. edestruct H2. eauto.
    + eauto using join_pred_of_ab.
    + eauto using join_ret_pred_of_ab.
    + eauto using join_exit_pred_of_ab.
    + eauto using join_goto_pred_of_ab.
    + eauto using join_pred_of_ab.
    + eauto using join_ret_pred_of_ab.
    + eauto using join_exit_pred_of_ab.
    + eauto using join_goto_pred_of_ab.
  - (* Sloop *)
    assert (IHs:=iter_ok s).
    clear iter_ok iter_switch_ok iter_function_ok.
    cut (∃ I, (∀ le mem, pred_of_ab inormal le mem -> (I le mem:Prop)) /\
              hoare_stmt prog f env stk I (goto_pred_of_ab igoto) s
                         I (ret_pred_of_ab (oreturn abstate o))
                         (λ x : nat, pred_of_ab (nth x (oexit abstate o) Bot))
                         (goto_pred_of_ab (ogoto abstate o))).
    { intros [I [HI1 HI2]]. eapply hoare_Sloop; eauto. }
    revert H. unfold iter.
    revert inormal igoto o. generalize (unroll_call_lookup s unroll).
    induction n; intros inormal igoto o.
    + apply bm_bind_case; try discriminate.
      intros out OUT eq. inv eq.
      edestruct lfp_ok as (inv & invle & INV).
      * intros. eapply IHs, H.
      * eauto.
      * eexists. split. 2:eapply hoare_stmt_post_impl, INV; eauto.
        simpl. eauto. simpl. eauto.
    + apply bm_bind_case. discriminate. intros out1 OUT1.
      destruct (onormal abstate out1) eqn:E.
      * intros Hr; inv Hr. clear IHn.
        assert (IHss:=IHs _ _ _ OUT1); clear IHs OUT1. simpl.
        eexists (pred_of_ab inormal). split. auto.
        rewrite E in IHss.
        eapply hoare_stmt_post_impl, IHss; auto.
        intros ? ? (? & ? & []).
      * apply bm_bind_case. discriminate. intros out2 OUT2.
        intros Hr; inv Hr.
        assert (IHnn:=IHn _ _ _ OUT2). clear IHn OUT2.
        destruct IHnn as (I & HI1 & HI2).
        assert (IHss:=IHs _ _ _ OUT1); clear IHs OUT1.
        exists (fun le mem => I le mem \/ pred_of_ab inormal le mem). split. auto.
        eapply hoare_stmt_pre_impl, hoare_stmt_pre_or.
        intros. apply H. intros. right. apply H.
        eapply hoare_stmt_post_impl, HI2; eauto.
        intros. apply join_ret_pred_of_ab. auto.
        intros. apply join_exit_pred_of_ab. auto.
        intros. apply join_goto_pred_of_ab. auto.
        eapply hoare_stmt_post_impl, IHss.
        rewrite E. auto.
        intros. apply join_ret_pred_of_ab. auto.
        intros. apply join_exit_pred_of_ab. auto.
        intros. apply join_goto_pred_of_ab. auto.
  - (* Sblock *)
    assert (IHs:=iter_ok s).
    clear iter_ok iter_switch_ok.
    revert H. unfold iter.
    apply bm_bind_case. discriminate. intros out OUT.
    intros eq. apply pair_eq_inv in eq. destruct eq. subst.
    eapply hoare_Sblock, hoare_stmt_post_impl, IHs, OUT.
    + eauto using join_pred_of_ab.
    + auto.
    + intros [|x].
      intros le mem CS. apply join_pred_of_ab. destruct oexit; auto.
      unfold oexit at 2. destruct oexit, x; auto.
    + auto.
  - (* Sexit *)
    clear iter_ok iter_switch_ok.
    apply hoare_Sexit. inv H. simpl.
    clear; induction n; auto.
  - (* Sswitch *)
    assert (IHlbl_s:=iter_switch_ok l b).
    clear iter_ok iter_switch_ok.
    revert H. unfold iter.
    apply bm_bind_case. discriminate. intros out1 OUT1 OUT1'.
    eapply hoare_Sswitch, hoare_lbl_stmt_pre_impl, IHlbl_s, OUT1'; eauto. 2:contradiction.
    intros. destruct inormal. now destruct H as (? & ? & []).
    revert OUT1. apply bm_bind_case. discriminate. intros out2 OUT2 OUT2'.
    assert (N:=nonblock_sound ge _ _ mem_spec (Eunop (if b then Cminor.Onegl else Cminor.Onegint) e) x). rewrite OUT2 in N.
    destruct H as (cs & CS & AB). destruct CS.
    destruct (eval_expr_dec env le mem e) as [[v Hv]|].
    assert ((∃ n, Switch.switch_argument b v n) \/ (∀ n, ~Switch.switch_argument b v n)).
    { destruct b, v; try (left; eexists; constructor); right; intros n Hn; inv Hn. }
    destruct H1 as [[n Hn]|?].
    + eexists. eexists. split. eauto. split. eauto.
      destruct (select_switch_case n l) eqn:?. split.
      destruct Hn. rewrite Int.repr_unsigned; auto. rewrite Int64.repr_unsigned; auto.
      eexists. split. constructor; eauto. eauto.
      eexists. split. constructor; eauto.
      clear - mem_spec OUT2' AB Hv Hn Heqo0.
      revert x out1 OUT2' AB. induction l; intros. inv OUT2'. auto.
      destruct o. 2:inv OUT2'; now eauto.
      simpl in OUT2', Heqo0. destruct zeq. discriminate. specialize (IHl Heqo0).
      destruct zlt. simpl in OUT2'. eauto. destruct zlt. simpl in OUT2'. eauto. simpl in OUT2'.
      revert OUT2'. apply bm_bind_case. discriminate. intros [outt outf] OUT.
      assert (A:=assume_sound ge _ _ mem_spec
        (if b then Ebinop (Cminor.Ocmpl Ceq) e (Econst (Olongconst (Int64.repr z)))
         else Ebinop (Cminor.Ocmp Ceq) e (Econst (Ointconst (Int.repr z)))) x).
      rewrite OUT in A.
      assert (((idf, (le, env)) :: cs_stk, mem) ∈ γ outf).
      { apply (A (Some (false, _))). econstructor; eauto.
        intros. destruct Hn.
        - inv H0. inv H6. inv H1. inv H7. determ.
          unfold Val.cmp, Val.cmp_bool, Int.cmp in H.
          pose proof (Int.eq_spec i (Int.repr z)). destruct Int.eq.
          subst i. rewrite Int.unsigned_repr in n0. congruence. omega.
          inv H. auto.
        - inv H0. inv H6. inv H1. inv H7. determ. inv H1.
          pose proof (Int64.eq_spec i (Int64.repr z)). destruct Int64.eq.
          subst i. rewrite Int64.unsigned_repr in n0. congruence. omega.
          inv H. auto. }
      destruct outf. contradiction. eauto.
    + destruct (N None). econstructor; eauto.
      intros. inv H2. determ.
      destruct b, v; inv H8; try congruence; edestruct H1; constructor.
    + destruct (N None). econstructor; eauto.
      intros. inv H2. edestruct H1. eauto.
  - clear iter_ok iter_switch_ok.
    destruct o; [apply hoare_Sreturn_1|apply hoare_Sreturn_0].
    + (* Sreturn (Some _) *)
      intros le mem (cs & CS & AB). destruct CS.
      destruct inormal as [|inormal]. contradiction.
      revert H. unfold iter. apply bm_bind_case. discriminate. intros out OUT eq. inv eq.
      assert (P:=pop_frame_sound ge _ _ mem_spec (Some e) rettemp inormal). rewrite OUT in P.
      destruct (eval_expr_dec env le mem e) as [[v Hv]|].
      destruct (Mem.free_list mem (blocks_of_env env)) as [mem0|] eqn:MEM.
      * eexists. eexists. split. eauto. split. eauto. eexists. split. eauto.
        destruct cs_stk.
        { split. apply (P (Some _)). econstructor; eauto.
          intros. determ. auto.
          destruct rettemp as [[]|], v; eauto; destruct (P None);
          econstructor; eauto; simpl; intros; determ; discriminate. }
        { destruct p as (idf0 & le' & e'). apply (P (Some _)). econstructor; eauto.
          simpl. intros. determ. auto. }
      * destruct (P None). econstructor; eauto. congruence.
      * destruct (P None). econstructor; eauto. intros. edestruct H; eauto.
    + (* Sreturn None *)
      intros le mem (cs & CS & AB). destruct CS.
      destruct inormal as [|inormal]. contradiction.
      revert H. unfold iter. apply bm_bind_case. discriminate. intros out OUT eq. inv eq.
      assert (P:=pop_frame_sound ge _ _ mem_spec None rettemp inormal). rewrite OUT in P.
      destruct (Mem.free_list mem (blocks_of_env env)) as [mem0|] eqn:MEM.
      * eexists. eexists. eauto. eexists. split. eauto.
        destruct cs_stk.
        { split. apply (P (Some _)). econstructor; eauto.
          simpl. congruence.
          destruct (P None). econstructor; eauto. simpl. congruence. }
        { destruct p as (idf0 & le' & e'). apply (P (Some _)). econstructor; eauto.
          simpl. intros. determ. subst. auto. }
      * destruct (P None). econstructor; eauto. congruence.
  - (* Slabel *)
    assert (IHs:=iter_ok s).
    clear iter_ok iter_switch_ok.
    eapply hoare_Slabel, hoare_stmt_pre_impl, IHs, H.
    + intros le mem HH. apply join_pred_of_ab. destruct HH; auto.
    + auto.
  - (* Sgoto *)
    clear iter_ok iter_switch_ok.
    eapply hoare_Sgoto. inv H. unfold ogoto at 1.
    destruct inormal as [|inormal]. now intros le mem (? & ? & []).
    unfold goto_pred_of_ab. rewrite PTree.gss. auto.
  - (* LSnil *)
    clear iter_ok iter_switch_ok. apply pair_eq_inv in H. destruct H. subst.
    apply hoare_LSnil; eauto using join_pred_of_ab.
  - (* LScons *)
    assert (IHs:=iter_ok s). assert (IHlbl_s:=iter_switch_ok lbl_s islong).
    clear iter_ok iter_switch_ok.
    revert H. unfold iter_switch.
    apply bm_bind_case. discriminate. intros [icase' idefault'] OUT1.
    apply bm_bind_case. discriminate. intros out2 OUT2.
    apply bm_bind_case. discriminate. intros out3 OUT3.
    intros eq. apply pair_eq_inv in eq. destruct eq. subst.
    destruct o; revert OUT1.
    + apply bm_bind_case. discriminate. intros out4 OUT4.
      intros eq. apply triple_eq_inv in eq. destruct eq as [[? ?] _]. subst.
      eapply hoare_LScons_case;
        [eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, IHs, OUT2 |
         eapply hoare_lbl_stmt_post_impl, IHlbl_s, OUT3]; eauto.
      * eauto using join_ret_pred_of_ab.
      * eauto using join_exit_pred_of_ab.
      * intros. apply join_goto_pred_of_ab. auto.
      * intros le mem [[EVAL HH]|HH]; apply join_pred_of_ab; auto.
        right.
        destruct inormal. now destruct HH as (? & ? & []).
        revert OUT4. apply bm_bind_case. discriminate.
        intros [icase idummy] OUT eq. inv eq.
        destruct HH as (cs & CS & AB).
        exists cs. split. auto.
        assert (A:=assume_sound ge _ _ mem_spec
          (if islong then Ebinop (Cminor.Ocmpl Ceq) e (Econst (Olongconst (Int64.repr z)))
           else Ebinop (Cminor.Ocmp Ceq) e (Econst (Ointconst (Int.repr z))))
          x (Some (true, cs))).
        rewrite OUT in A. apply A.
        destruct CS. econstructor; eauto.
        intros. destruct islong.
        inv H3. inv H9. inv H4. inv H10. determ.
        unfold Val.cmpl, Val.cmpl_bool, Int64.cmp in H4. rewrite Int64.eq_true in H4.
        inv H4. inv H2. auto.
        inv H3. inv H9. inv H4. inv H10. determ.
        unfold Val.cmp, Val.cmp_bool, Int.cmp in H2. rewrite Int.eq_true in H2. inv H2. auto.
      * eauto using join_ret_pred_of_ab.
      * eauto using join_exit_pred_of_ab.
      * intros. apply join_goto_pred_of_ab. auto.
    + intros eq. apply triple_eq_inv in eq. destruct eq as [[? ?] _]. subst.
      eapply hoare_LScons_default;
        [eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, IHs, OUT2|
         eapply hoare_lbl_stmt_post_impl, hoare_lbl_stmt_pre_impl, IHlbl_s, OUT3]; eauto.
      * eauto using join_ret_pred_of_ab.
      * eauto using join_exit_pred_of_ab.
      * eauto using join_goto_pred_of_ab.
      * intros. apply join_pred_of_ab. intuition.
      * eauto using join_ret_pred_of_ab.
      * eauto using join_exit_pred_of_ab.
      * eauto using join_goto_pred_of_ab.
      * contradiction.
Qed.

End stmt_iterator_proof.

Lemma labels_stmt_ok:
  ∀ lbl s k,
    (labels_stmt s) ! lbl = Some tt ->
    ∃ sk, find_label lbl s k = Some sk
with labels_sl_ok:
  ∀ lbl sl k,
    (labels_sl sl) ! lbl = Some tt ->
    ∃ sk, find_label_ls lbl sl k = Some sk.
Proof.
{ destruct s; simpl; intros k Hlbl;
  try (rewrite PTree.gempty in Hlbl; discriminate);
  try (rewrite PTree.gcombine in Hlbl by auto).
  - destruct ((labels_stmt s1)!lbl) as [[]|] eqn:?.
    pose proof (labels_stmt_ok lbl s1 (Kseq s2 k)).
    edestruct H as [sk Hsk]; auto. rewrite Hsk. eauto.
    pose proof (labels_stmt_ok lbl s2 k).
    destruct ((labels_stmt s2)!lbl) as [[]|]. 2:discriminate.
    edestruct H as [sk Hsk]; auto. rewrite Hsk.
    destruct (find_label lbl s1 (Kseq s2 k)) as [[]|]; eauto.
  - destruct ((labels_stmt s1)!lbl) as [[]|] eqn:?.
    pose proof (labels_stmt_ok lbl s1 k).
    edestruct H as [sk Hsk]; auto. rewrite Hsk. eauto.
    pose proof (labels_stmt_ok lbl s2 k).
    destruct ((labels_stmt s2)!lbl) as [[]|]. 2:discriminate.
    edestruct H as [sk Hsk]; auto. rewrite Hsk.
    destruct (find_label lbl s1 k) as [[]|]; eauto.
  - eapply (labels_stmt_ok lbl s). auto.
  - eapply (labels_stmt_ok lbl s). auto.
  - eapply (labels_sl_ok lbl l). auto.
  - unfold ident_eq. rewrite PTree.gsspec in Hlbl.
    destruct peq. eauto.
    eapply (labels_stmt_ok lbl s). auto.
}
{ destruct sl; simpl.
  - rewrite PTree.gempty. discriminate.
  - rewrite PTree.gcombine by auto. intros.
    destruct ((labels_stmt s)!lbl) as [[]|] eqn:?.
    pose proof (labels_stmt_ok lbl s (Kseq (seq_of_lbl_stmt sl) k)).
    edestruct H0 as [sk Hsk]; auto. rewrite Hsk. eauto.
    pose proof (labels_sl_ok lbl sl k).
    destruct ((labels_sl sl)!lbl) as [[]|]. 2:discriminate.
    edestruct H0 as [sk Hsk]; auto. rewrite Hsk.
    destruct (find_label lbl s (Kseq (seq_of_lbl_stmt sl) k)) as [[]|]; eauto.
}
Qed.

Lemma iter_function_ok :
  ∀ fuel sig name fundef rettemp args i o,
    id_function name fundef ->
    iter_function _ unroll _ fuel sig name fundef rettemp args i = (o, nil) ->
  ∀ stk mem,
    match stk with
      | nil => (nil, mem) ∈ γ i /\
               sig = (mksignature nil (Some AST.Tint) cc_default) /\
               args = nil
      | (f, env, le)::stk => pred_of_ab f env stk (NotBot i) le mem
    end ->
    ∃ vargs,
      match stk with
        | nil => vargs = nil
        | (_, env, le)::_ => eval_exprlist ge env le mem args vargs
      end /\
      funsig fundef = sig /\
      hoare_fun prog stk
                (fun vargs' mem' => vargs = vargs' /\ mem = mem')
                fundef
                (fun ret mem' =>
                   match stk with
                     | nil =>
                       (nil, mem') ∈ γ o /\ ∃ r, ret = Vint r
                     | (f, env, le)::stk =>
                       pred_of_ab f env stk o (Cminor.set_optvar rettemp ret le) mem'
                   end).
Proof.
  induction fuel. discriminate.
  Opaque join. simpl. Transparent join.
  intros*. destruct signature_eq. 2:discriminate.
  destruct fundef.
  - apply bm_bind_case. discriminate. intros out1 OUT1.
    apply bm_bind_case. discriminate. intros out2 OUT2.
    apply bm_bind_case. discriminate. intros out3 OUT3.
    apply bm_bind_case. discriminate. intros out4 OUT4.
    apply bm_bind_case. discriminate. intros out5 OUT5.
    apply bm_bind_case. discriminate. intros out6 OUT6.
    apply bm_bind_case. discriminate. intros out7 OUT7.
    apply bm_bind_case. discriminate. intros out8 OUT8.
    intros IDFUN eq. apply pair_eq_inv in eq. destruct eq. subst.
    destruct list_norepet_dec; inv OUT1.
    destruct list_norepet_dec; inv OUT2.
    destruct list_disjoint_dec; inv OUT3.
    destruct out5.(oexit _) eqn:?; inv OUT6.
    intros.
    assert (P:=push_frame_sound ge _ _ mem_spec name f args i). rewrite OUT4 in P.
    destruct (Classical_Prop.classic
                (exists vargs,
                   match stk with
                     | nil => vargs = nil
                     | (_, env, le) :: _ => eval_exprlist ge env le mem args vargs
                   end)) as [[vargs VARGS]|].
    destruct (Classical_Prop.classic
                (exists mem1 env le,
                   alloc_variables empty_env mem (fn_vars f) env mem1 /\
                   bind_parameters (fn_params f) vargs (create_undef_temps (fn_temps f)) = Some le)) as [(mem1 & env & le & ALLOC & BIND)|].
    + eexists. split. eauto. split. auto.
      apply hoare_Internal.
      intros ? ? [<- <-]. split. auto. split. auto. split. auto.
      eexists. eexists. eexists. split. eauto. split. eauto.
      edestruct lfp_goto_ok with (rettemp:=rettemp) as (inv & invle & INV); eauto.
      { intros. eapply iter_ok, H1.
        intros. eapply IHfuel with (stk:=(_, _, _)::_); eauto. }
      exists inv. split.
      { pose proof PTree.bempty_correct (PTree.combine
           (λ x y, do_opt _ <- x;
            match y with
              | Some _ => None
              | None => Some tt
            end) (ogoto abstate out5) (labels_stmt (fn_body f))).
        destruct PTree.bempty. 2:discriminate.
        intros. apply proj1 in H1. specialize (H1 eq_refl lbl).
        rewrite PTree.gcombine in H1 by auto.
        destruct (out5.(ogoto _) ! lbl) eqn:?.
        - destruct ((labels_stmt (fn_body f)) ! lbl) eqn:?. destruct u.
          apply labels_stmt_ok with (k:=k) in Heqo0.
          destruct Heqo0 as [sk ->]. congruence. discriminate.
        - unfold goto_pred_of_ab in invle.
          specialize (invle _ _ _ H2). rewrite Heqo in invle.
          destruct invle as (? & ? & []). }
      eapply hoare_stmt_post_impl, hoare_stmt_pre_impl, INV; auto.
      * { intros. destruct out5.(onormal _). now destruct H1 as (? & ? & []).
          assert (P':=pop_frame_sound ge _ _ mem_spec None rettemp x). rewrite OUT8 in P'.
          destruct H1 as (cs & CS & AB). destruct CS.
          destruct (Mem.free_list mem0 (blocks_of_env env)) eqn:?.
          - eexists. split. eauto.
            destruct stk as [|((fstk & estk) & lestk)],
                     cs_stk as [|(idf' & le' & e')]; inv H2.
            + destruct (P' None). econstructor; eauto. simpl. congruence.
            + destruct H6 as (? & -> & ->). eapply join_pred_of_ab. right.
              eexists. split. econstructor; eauto. apply (P' (Some _)).
              econstructor; eauto. simpl. intros. subst. determ. auto.
          - destruct (P' None). econstructor; eauto. congruence. }
      * unfold ret_pred_of_ab.
        intros v mem0 (cs_stk & FA & CSSTK).
        destruct stk as [|((fstk & estk) & lestk)],
                 cs_stk as [|(idf & le' & e')]; inv FA.
        destruct CSSTK as (? & ? & ?). split.
        eapply join_sound. eapply lift_bot, mem_adom, mem_spec. eauto. auto.
        apply join_pred_of_ab. left. econstructor.
        split. 2:eauto. destruct H4 as (? & -> & ->). econstructor; eauto.
      * rewrite Heql2. intros. destruct x, H1 as (? & ? & []).
      * { intros ? ? [<- <-]. destruct stk as [|((f0 & env0) & le)].
          - eexists. split. constructor; eauto.
            destruct H as (? & ? & ?).
            eapply (P (Some _)). econstructor.
            eexists. split. eauto. split. auto.
            intros. subst. determ. auto.
          - destruct H as (cs & CS & AB). destruct CS.
            eexists ((_, (_, _))::(_, (_, _))::_, _). split. constructor; eauto.
            eapply (P (Some _)). econstructor.
            eexists. split. eauto. split. auto. intros. determ. auto. }
    + destruct (P None), stk as [|[[f' env] le]].
      destruct H as (? & ? & ->).
      econstructor. eexists. split. eauto. split. auto.
      intros. subst. destruct H1. eauto.
      destruct H as (cs & CS & AB). destruct CS.
      econstructor. eexists. split. eauto. split. auto.
      intros. determ. destruct H1. eauto.
    + destruct (P None), stk as [|[[f' env] le]].
      destruct H as (? & ? & ->).
      econstructor. eexists. split. eauto. split. auto.
      intros. destruct H1. eauto.
      destruct H as (cs & CS & AB). destruct CS.
      econstructor. eexists. split. eauto. split. auto.
      intros. destruct H1. eauto.
  - intros. subst. destruct stk as [|[[f' env] le]].
    + destruct H1 as (? & ? & ->).
      edestruct (ab_builtin_sound nil) as (vargs & -> & exEXTCALL & EXTCALL); eauto.
      eexists. split. eauto. split. eauto.
      apply hoare_External.
      intros vargs' mem' [<- <-]. split. auto.
      intros. edestruct EXTCALL; eauto.
    + destruct H1 as (cs & CS & AB). destruct CS.
      edestruct ab_builtin_sound as (vargs & VARGS & exEXTCALL & EXTCALL); eauto.
      discriminate.
      eexists. split. apply VARGS. split. auto.
      apply hoare_External.
      intros vargs' mem' [<- <-]. split. auto.
      intros. edestruct EXTCALL; eauto.
      eexists. split. constructor; eauto. auto.
Qed.

Lemma find_label_rel_cont_stack:
  ∀ lbl s k s' k', find_label lbl s k = Some (s', k') ->
  ∀ stk, rel_cont_stack k stk -> rel_cont_stack k' stk
with  find_label_ls_rel_cont_stack:
  ∀ lbl sl k s' k', find_label_ls lbl sl k = Some (s', k') ->
  ∀ stk, rel_cont_stack k stk -> rel_cont_stack k' stk.
Proof.
{ destruct s; intros; try discriminate; simpl in H.
  - destruct (find_label lbl s1 (Kseq s2 k)) eqn:?; inv H.
    eapply (find_label_rel_cont_stack lbl s1); eauto. constructor; auto.
    eapply (find_label_rel_cont_stack lbl s2); eauto.
  - destruct (find_label lbl s1 k) eqn:?; inv H.
    eapply (find_label_rel_cont_stack lbl s1); eauto.
    eapply (find_label_rel_cont_stack lbl s2); eauto.
  - eapply (find_label_rel_cont_stack lbl s); eauto. constructor; auto.
  - eapply (find_label_rel_cont_stack lbl s); eauto. constructor; auto.
  - eapply (find_label_ls_rel_cont_stack lbl l); eauto.
  - destruct ident_eq. inv H. auto.
    eapply (find_label_rel_cont_stack lbl s); eauto.
}
{ destruct sl; intros; try discriminate; simpl in H.
  destruct (find_label lbl s (Kseq (seq_of_lbl_stmt sl) k)) eqn:?; inv H.
  eapply (find_label_rel_cont_stack lbl s); eauto. constructor; auto.
  eapply (find_label_ls_rel_cont_stack lbl sl); eauto.
}
Qed.

Lemma iter_program_sound:
  ∀ tr, program_behaves (semantics prog) (Goes_wrong tr) ->
  iter_program _ unroll prog _ = (tt, nil) ->
  False.
Proof.
  intros tr TR.
  unfold iter_program.
  eapply bm_bind_case. discriminate. intros init INIT.
  pose proof (init_mem_sound ge _ _ mem_spec (prog_defs prog)).
  rewrite INIT in H.
  destruct (Genv.alloc_globals ge Mem.empty (prog_defs prog)) eqn:?.
  - refine (let H' := H _ _ in _).
    exists prog. split. auto. split. auto.
    intro. rewrite Heqo. intro eq. inv eq. auto.
    destruct init as [|init]. contradiction.
    destruct (Genv.find_symbol (Genv.globalenv prog) (prog_main prog)) eqn:?. 2:discriminate.
    destruct (Genv.find_funct_ptr (Genv.globalenv prog) b) eqn:?. 2:discriminate.
    eapply bm_bind_case. discriminate. intros finish FINISH _.
    apply iter_function_ok with (stk:=nil) (mem:=m) in FINISH; eauto.
    destruct FINISH as (? & -> & SIG & HOARE).
    eapply hoare_prog; eauto.
    eapply hoare_fun_pre_impl, hoare_fun_post_impl, HOARE.
    simpl. now intuition. simpl. now intuition.
    unfold id_function, ge, Genv.find_funct. rewrite Heqo0. 
    destruct Int.eq_dec. rewrite Heqo1. auto. congruence.
  - destruct (H None).
    exists prog. split. auto. split. auto. intro. congruence.
Qed.

End ABMEM.

End PROG.
