(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8 String ToString Util.
Require Import AdomLib.
Require Import Fcore_defs.
Require Import Integers Floats.
Require Import Reals.
Require Import Fourier.
Require Import ZIntervals.
Require Import Fappli_IEEE.
Require Import Fappli_IEEE_bits.
Require Import Fappli_IEEE_extra.
Require Import Fcore_FLT.
Require Import Fcore_float_prop.
Require Import Fcore_Raux.
Require Import Fcore_Zaux.
Require Import Fcore_generic_fmt.
Require Import Fcore_rnd_ne.
Require Import Fcore_digits.
Require Import FloatIntervals.
Require Import FloatLib.

Definition gammaR (i:itv) (x:R) : Prop :=
  match i with
    | ITV mx Mx => RF_lt _ _ x Mx /\ RF_lt _ _ (-x) (Float.neg mx)
  end.

Definition unround_itv (x:itv) : itv :=
  match x with
    | ITV mx Mx =>
      ITV (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg mx)))
          (next_ulp 53 1024 eq_refl eq_refl Mx)
  end.

Lemma unround_itv_round :
  forall x_r x,
    (if Rlt_bool
          (Rabs (round radix2 (FLT_exp (3 - 1024 - 53) 53)
                       (round_mode mode_NE) x_r))
          (bpow radix2 1024)
     then B2R _ _ x = round radix2 (FLT_exp (3 - 1024 - 53) 53)
                            (round_mode mode_NE) x_r /\
          is_finite _ _ x = true
     else B2FF _ _ x = binary_overflow 53 1024 mode_NE (Rlt_bool x_r 0)) ->
  forall x_ab, x ∈ γ x_ab ->
  gammaR (unround_itv x_ab) x_r.
Proof.
  intros.
  destruct x_ab; destruct H0; try easy. split.
  - eapply next_ulp_round_overflow; eauto. reflexivity.
  - Transparent Float.neg.
    unfold Float.neg.
    rewrite Bopp_involutive.
    rewrite <- fneg_decr in H0.
    eapply next_ulp_round_overflow with (mode:=mode_NE); eauto. reflexivity.
    rewrite round_NE_opp, Rabs_Ropp.
    unfold Float.neg. rewrite B2R_Bopp, is_finite_Bopp.
    unfold round_mode in H. Rlt_bool_case H.
    now intuition.
    destruct x as [| |?[]|]; inv H.
    destruct (Rlt_bool_spec x_r 0), (Rlt_bool_spec (-x_r) 0); try reflexivity; exfalso.
    fourier.
    destruct (Rcompare_spec x_r 0); try fourier.
    subst. rewrite round_0, Rabs_R0 in H2. pose proof (bpow_gt_0 radix2 1024). fourier.
    apply (valid_rnd_round_mode mode_NE).
    destruct f as [|[]| |]; try reflexivity; try discriminate.
    unfold next_ulp, next_ulp_ff, B2FF, Bopp. rewrite is_nan_FF2B.
    repeat (match goal with |- is_nan_FF (if ?b then _ else _) = false => destruct b end);
      reflexivity.
Qed.

Definition Zoffloat_DN (f:float): option Z :=
  match f with
    | B754_finite s m (Zpos e) _ => Some (cond_Zopp s (Zpos m) * Zpower_pos radix2 e)%Z
    | B754_finite s m 0 _ => Some (cond_Zopp s (Zpos m))
    | B754_finite s m (Zneg e) _ => Some (cond_Zopp s (Zpos m) / Zpower_pos radix2 e)%Z
    | B754_zero _ => Some 0%Z
    | _ => None
  end.

Lemma Zoffloat_DN_correct:
  forall f : float,
    match Zoffloat_DN f with
      | Some n =>
        is_finite 53 1024 f = true /\
        Z2R n = round radix2 (Fcore_FIX.FIX_exp 0) (round_mode mode_DN)
                      (B2R 53 1024 f)
      | None => is_finite 53 1024 f = false
    end.
Proof.
  destruct f; try now intuition.
  - simpl B2R. rewrite round_0. easy. auto.
  - destruct e.
    + split. reflexivity.
      rewrite round_generic. symmetry. now apply Rmult_1_r.
      auto.
      apply Fcore_FIX.generic_format_FIX.
      exists (Float radix2 (cond_Zopp b (Zpos m)) 0). split; reflexivity.
    + split; [reflexivity|].
      rewrite round_generic, Z2R_mult, Z2R_Zpower_pos, <- bpow_powerRZ;
        try easy. apply generic_format_F2R; discriminate.
    + split; [reflexivity|].
      rewrite (Fcalc_round.inbetween_float_DN _ _ _ (cond_Zopp b (Zpos m) / Z.pow_pos radix2 p)
                (Fcalc_bracket.new_location (Zpower_pos radix2 p) (cond_Zopp b (Zpos m) mod Zpower_pos radix2 p) Fcalc_bracket.loc_Exact)).
      * symmetry. apply Rmult_1_r.
      * unfold canonic_exp, Fcore_FIX.FIX_exp; replace Z0 with (Zneg p + Zpos p)%Z by apply Zplus_opp_r.
        apply (Fcalc_bracket.inbetween_float_new_location radix2 _ _ _ _ (Zpos p)); [reflexivity|].
        apply Fcalc_bracket.inbetween_Exact. reflexivity.
Qed.

Definition backward_floatofz (res_ab:itv+⊤) (x_ab:ZIntervals.itv+⊤) : ZIntervals.itv+⊤+⊥ :=
  match res_ab with
    | Just res_ab' =>
      match unround_itv res_ab' with
        | ITV mx Mx =>
          let low :=
              match Zoffloat_DN (Float.neg mx), x_ab with
                | Some m, Just (ZIntervals.ITV m' _) => Some (Z.max (-m) m')
                | Some m, All => Some (Z.opp m)
                | None, Just (ZIntervals.ITV m' _) => Some m'
                | None, All => None
              end
          in
          let high :=
              match Zoffloat_DN Mx, x_ab with
                | Some M, Just (ZIntervals.ITV _ M') => Some (Z.min M M')
                | Some M, All => Some M
                | None, Just (ZIntervals.ITV _ M') => Some M'
                | None, All => None
              end
          in
          match low, high with
            | Some m, Some M =>
              if Z.leb m M then NotBot (Just (ZIntervals.ITV m M))
              else Bot
            | _, _ => NotBot All
          end
      end
    | All => NotBot x_ab
  end.

Lemma backward_floatofz_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall res_ab, Fappli_IEEE_extra.BofZ 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) x ∈ γ res_ab ->
    x ∈ γ (backward_floatofz res_ab x_ab).
Proof.
  intros.
  unfold backward_floatofz.
  destruct res_ab as [|res_ab]. auto.
  assert (gammaR (unround_itv res_ab) (Z2R x)).
  { eapply unround_itv_round; eauto.
    use_float_specs.
    destruct Rlt_bool. intuition.
    destruct (Z.ltb_spec x 0), (Rlt_bool_spec (Z2R x) 0); auto.
    apply (le_Z2R 0 x) in H3. omega.
    apply (lt_Z2R x 0) in H3. omega.
  }
  destruct unround_itv. auto.
  destruct H1. unfold RF_lt, RFF_lt, F2R, Fnum, Fexp, bpow in H1, H2.
  rewrite is_finite_FF_B2FF, FF2R_B2FF in H1, H2.
  assert (forall m, Zoffloat_DN (Float.neg f) = Some m -> (-m <= x)%Z).
  { pose proof (Zoffloat_DN_correct (Float.neg f)).
    destruct (Zoffloat_DN (Float.neg f)). 2:easy.
    destruct H3. rewrite H3 in H2. intros. injection H5. intro. subst z. clear H5.
    apply Zopp_le_cancel, le_Z2R. rewrite Zopp_involutive, H4.
    unfold round, scaled_mantissa, canonic_exp, Fcore_FIX.FIX_exp, F2R, Fexp, Fnum,
           scaled_mantissa, bpow, round_mode, Zopp at 2.
    rewrite !Rmult_1_r.
    apply Z2R_le, Zfloor_lub.
    rewrite Z2R_opp. fourier. }
  assert (forall M, Zoffloat_DN f0 = Some M -> (x <= M)%Z).
  { pose proof (Zoffloat_DN_correct f0).
    destruct (Zoffloat_DN f0). 2:easy.
    destruct H4. rewrite H4 in H1. intros. injection H6. intro. subst z. clear H6.
    apply le_Z2R. rewrite H5.
    unfold round, scaled_mantissa, canonic_exp, Fcore_FIX.FIX_exp, F2R, Fexp, Fnum,
           scaled_mantissa, bpow, round_mode, Zopp.
    rewrite !Rmult_1_r.
    apply Z2R_le, Zfloor_lub. fourier. }
  destruct (Zoffloat_DN (Float.neg f)); [specialize (H3 _ eq_refl)|clear H3];
    (destruct (Zoffloat_DN f0); [specialize (H4 _ eq_refl)|clear H4]);
  destruct x_ab as [|[]]; destruct H; auto; try exact I.
  destruct (Z.leb_spec (-z) z0); [split|]; zify; omega.
  destruct (Z.leb_spec (Z.max (-z) z1) (Z.min z0 z2)); [split|]; zify; omega.
  destruct (Z.leb_spec (Z.max (-z) z0) z1); [split|]; zify; omega.
  destruct (Z.leb_spec z0 (Z.min z z1)); [split|]; zify; omega.
  destruct (Z.leb_spec z z0); [split|]; zify; omega.
Qed.

Definition backward_zoffloat (res_ab:ZIntervals.itv+⊤) (x_ab:itv+⊤) : itv+⊤+⊥ :=
  let new_x_ab :=
      match res_ab with
        | All =>
          ITV (FF2B 53 1024 (binary_overflow 53 1024 mode_UP true) eq_refl)
              (FF2B 53 1024 (binary_overflow 53 1024 mode_DN false) eq_refl)
        | Just (ZIntervals.ITV a b) =>
          ITV (half_backward_zoffloat a 53 1024 eq_refl eq_refl)
              (Float.neg (half_backward_zoffloat (-b) 53 1024 eq_refl eq_refl))
      end
  in
  meet_top x_ab (Just new_x_ab).

Transparent Float.cmp.

Lemma backward_zoffloat_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall res res_ab, res ∈ γ res_ab ->
    ZofB _ _ x = Some res ->
    x ∈ γ (backward_zoffloat res_ab x_ab).
Proof.
  intros. apply meet_top_correct. split. auto.
  destruct res_ab as [|[]].
  - destruct x as [| | |[]]; try discriminate; split; try reflexivity;
      unfold Fleb, Float.cmp, Bcompare, binary_overflow, overflow_to_inf, FF2B, negb;
      pose proof e0; apply andb_true_iff in H2; destruct H2; apply Z.leb_le in H3;
      unfold canonic_mantissa, FLT_exp in H2; apply Zeq_bool_eq in H2;
      assert (Z.pos (digits2_pos m) <= 53)%Z by (zify; omega);
      apply Zpower_le with (r:=radix2) in H4; simpl radix_val in H4;
      rewrite Zpos_digits2_pos in *; destruct (Zdigits_correct radix2 (Z.pos m));
      (destruct (pow2m1_pos 53) as [? []]; [reflexivity|]); rewrite H7;
      simpl Z.abs in *; simpl radix_val in *.
    + destruct (Z.compare_spec (1024-53) e); try reflexivity; try omega.
      fold (Pos.compare x m). destruct (Pos.compare_spec x m); try reflexivity; zify; omega.
    + destruct (Z.compare_spec e (1024-53)); try reflexivity; try omega.
      fold (Pos.compare m x). destruct (Pos.compare_spec m x); try reflexivity; zify; omega.
  - destruct H0; split.
    eapply half_backward_zoffloat_correct; eauto. reflexivity.
    rewrite <- fneg_decr.
    replace (Float.neg (Float.neg (half_backward_zoffloat (- z0) 53 1024 eq_refl eq_refl)))
    with (half_backward_zoffloat (- z0) 53 1024 eq_refl eq_refl).
    apply half_backward_zoffloat_correct with (i:=Z.opp res). reflexivity.
    destruct x; inversion H1; subst; clear H1.
    reflexivity.
    simpl. destruct e, b; inversion H4; subst; clear H4; try reflexivity.
    destruct Z.pow_pos; reflexivity.
    destruct Z.pow_pos; reflexivity.
    rewrite Zopp_involutive; reflexivity.
    omega.
    unfold Float.neg.
    rewrite Bopp_involutive. auto.
    unfold half_backward_zoffloat.
    destruct Z.ltb. apply binary_normalize_is_nan.
    pose proof (binary_normalize_is_nan 53 1024 eq_refl eq_refl (-z0 - 1) 0 false mode_DN).
    unfold next_ulp. rewrite is_nan_FF2B.
    destruct binary_normalize as [|[]| |[]]; try discriminate; try reflexivity; simpl;
    repeat match goal with |- context [if ?b then _ else _] => destruct b eqn:? end; reflexivity.
Qed.

Definition backward_le_l (x y:itv+⊤) : itv+⊤+⊥ :=
  match x, y with
    | Just (ITV mx Mx), Just (ITV _ My) =>
      if Fleb mx My then
        if Fleb Mx My then NotBot (Just (ITV mx Mx))
        else NotBot (Just (ITV mx My))
      else Bot
    | All, Just (ITV _ My) =>
      NotBot (Just (ITV (B754_infinity _ _ true) My))
    | Just (ITV mx Mx), All => NotBot x
    | All, All =>
      NotBot (Just (ITV (B754_infinity _ _ true) (B754_infinity _ _ false)))
  end.

Lemma backward_le_l_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall y y_ab, y ∈ γ y_ab ->
    Fleb x y = true ->
    x ∈ γ (backward_le_l x_ab y_ab).
Proof.
  destruct x_ab as [|[]], y_ab as [|[]]; simpl; intros.
  - destruct x as [|[]| |[]]; split; try reflexivity; discriminate.
  - destruct H0.
    destruct x as [|[]| |[]]; split; try reflexivity; try discriminate;
    eapply Fleb_trans; eauto.
  - apply H.
  - destruct H, H0.
    destruct (Fleb f f2) eqn:?.
    destruct (Fleb f0 f2) eqn:?.
    + split; auto.
    + split. auto. eapply Fleb_trans; eauto.
    + assert (Fleb f f2 = true).
      eapply Fleb_trans. eauto. eapply Fleb_trans; eauto.
      congruence.
Qed.

Definition backward_not_le_l (x y:itv+⊤) : itv+⊤+⊥ :=
  match x, y with
    | Just (ITV _ _), Just (ITV my My) =>
      botlift_fun1 forward_fneg
        (backward_le_l (forward_fneg x)
          (forward_fneg (Just (ITV (next_ulp 53 1024 eq_refl eq_refl my) (next_ulp 53 1024 eq_refl eq_refl My)))))
    | _, _ => NotBot x
  end.

Lemma backward_not_le_l_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall y y_ab, y ∈ γ y_ab ->
    Fleb x y = false ->
    x ∈ γ (backward_not_le_l x_ab y_ab).
Proof.
  destruct x_ab as [|[]], y_ab as [|[]]; intros; try auto.
  eapply botlift_fun1_spec.
  - intros.
    replace x with (Float.neg (Float.neg x)).
    apply forward_fneg_correct. eauto.
    unfold Float.neg. apply Bopp_involutive.
    destruct H, x; try reflexivity; discriminate.
  - eapply backward_le_l_correct.
    apply forward_fneg_correct. auto.
    apply forward_fneg_correct.
    destruct H0; split; apply next_ulp_incr; eauto; reflexivity.
    rewrite fneg_decr.
    destruct H, H0.
    destruct (next_ulp_lub 53 1024 eq_refl eq_refl x y eq_refl); eauto 3. congruence.
Qed.

Definition backward_cmpf_l (c:comparison) (x y:itv+⊤) (res:bool) : itv+⊤+⊥ :=
  match c, res with
    | Ceq, true | Cne, false =>
      botbind2 meet_top
        (backward_le_l x y)
        (botlift_fun1 forward_fneg (backward_le_l (forward_fneg x) (forward_fneg y)))
    | Ceq, false | Cne, true =>
      match y with
        | All => NotBot x
        | Just (ITV my My) =>
          if Fleb My my then
            match x with
              | All => NotBot x
              | Just (ITV mx Mx) =>
                if Float.cmp Ceq my mx then
                  if Fleb Mx mx then Bot
                  else NotBot (Just (ITV (next_ulp 53 1024 eq_refl eq_refl mx) Mx))
                else
                  if Float.cmp Ceq my Mx then
                    NotBot (Just (ITV mx (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg Mx)))))
                  else NotBot x
            end
          else NotBot x
      end
    | Cle, true =>
      backward_le_l x y
    | Cle, false =>
      backward_not_le_l x y
    | Cge, true =>
      botlift_fun1 forward_fneg (backward_le_l (forward_fneg x) (forward_fneg y))
    | Cge, false =>
      botlift_fun1 forward_fneg (backward_not_le_l (forward_fneg x) (forward_fneg y))
    | Clt, true =>
      let (my, My) :=
          match y with
            | All => (B754_infinity _ _ true, B754_infinity _ _ false)
            | Just (ITV my My) => (my, My)
          end
      in
      if Fleb My (B754_infinity _ _ true) then Bot
      else
        backward_le_l x (Just (ITV
          (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg my)))
          (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg My)))))
    | Cgt, true =>
      let (my, My) :=
          match y with
            | All => (B754_infinity _ _ true, B754_infinity _ _ false)
            | Just (ITV my My) => (my, My)
          end
      in
      if Fleb (B754_infinity _ _ false) my then Bot
      else
        botlift_fun1 forward_fneg
                     (backward_le_l (forward_fneg x)
                     (forward_fneg (Just (ITV (next_ulp 53 1024 eq_refl eq_refl my)
                                              (next_ulp 53 1024 eq_refl eq_refl My)))))
    | Cgt, false =>
      match x, y with
        | Just (ITV _ _), Just (ITV _ _) => backward_le_l x y
        | _, _ => NotBot x
      end
    | Clt, false =>
      match x, y with
        | Just (ITV _ _), Just (ITV _ _) =>
          botlift_fun1 forward_fneg (backward_le_l (forward_fneg x) (forward_fneg y))
        | _, _ => NotBot x
      end
  end.

Lemma backward_cmpf_l_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ c res, Float.cmp c x y = res ->
    x ∈ γ (backward_cmpf_l c x_ab y_ab res).
Proof.
  intros.
  assert (forall f1 f2, Float.cmp Ceq f1 f2 = Fleb f1 f2 && Fleb f2 f1).
  { intros. unfold Fleb, Float.cmp.
    rewrite Bcompare_swap. destruct Bcompare as [[]|]; reflexivity. }
  destruct c, res.
  - eapply botbind2_spec.
    2:eapply backward_le_l_correct; eauto 3.
    3:eapply botlift_fun1_spec.
    3:apply forward_fneg_correct.
    3:eapply backward_le_l_correct.
    3:apply forward_fneg_correct; eauto.
    3:apply forward_fneg_correct; eauto.
    intros.
    apply meet_top_correct. split. auto.
    destruct x as [[]|[]|[]|[]]; apply H4.
    unfold Fleb. unfold Float.cmp in H1.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
    rewrite fneg_decr.
    unfold Fleb. rewrite Bcompare_swap. unfold Float.cmp in H1.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
  - destruct x_ab as [|[]], y_ab as [|[]]; auto; unfold backward_cmpf_l; rewrite ?H2 in *; clear H2; unfold andb;
    repeat match goal with |- context [Fleb ?a ?b] => destruct (Fleb a b) eqn:? end; auto;
    destruct H, H0; try split; eauto 1 using Fleb_trans.
    + assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
    + destruct (next_ulp_lub 53 1024 eq_refl eq_refl x f eq_refl); eauto 2.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
    + rewrite <- fneg_decr.
      replace (Float.neg (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0))))
      with (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0)).
      destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (Float.neg f0) eq_refl); auto.
      destruct x; try reflexivity; discriminate.
      destruct f0; try reflexivity; discriminate.
      rewrite fneg_decr in H4.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
      symmetry. apply Bopp_involutive.
      eapply Fleb_nan_r. eapply next_ulp_le. destruct f0; try reflexivity; discriminate.
    + rewrite <- fneg_decr.
      replace (Float.neg (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0))))
      with (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0)).
      destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (Float.neg f0) eq_refl); auto.
      destruct x; try reflexivity; discriminate.
      destruct f0; try reflexivity; discriminate.
      rewrite fneg_decr in H4.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
      symmetry. apply Bopp_involutive.
      eapply Fleb_nan_r. eapply next_ulp_le. destruct f0; try reflexivity; discriminate.
  - rewrite Float.cmp_ne_eq, negb_true_iff in H1.
    destruct x_ab as [|[]], y_ab as [|[]]; auto; unfold backward_cmpf_l; rewrite ?H2 in *; clear H2; unfold andb;
    repeat match goal with |- context [Fleb ?a ?b] => destruct (Fleb a b) eqn:? end; auto;
    destruct H, H0; try split; eauto 1 using Fleb_trans.
    + assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
    + destruct (next_ulp_lub 53 1024 eq_refl eq_refl x f eq_refl); eauto 2.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
    + rewrite <- fneg_decr.
      replace (Float.neg (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0))))
      with (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0)).
      destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (Float.neg f0) eq_refl); auto.
      destruct x; try reflexivity; discriminate.
      destruct f0; try reflexivity; discriminate.
      rewrite fneg_decr in H4.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
      symmetry. apply Bopp_involutive.
      eapply Fleb_nan_r. eapply next_ulp_le. destruct f0; try reflexivity; discriminate.
    + rewrite <- fneg_decr.
      replace (Float.neg (Float.neg (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0))))
      with (next_ulp 53 1024 eq_refl eq_refl (Float.neg f0)).
      destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (Float.neg f0) eq_refl); auto.
      destruct x; try reflexivity; discriminate.
      destruct f0; try reflexivity; discriminate.
      rewrite fneg_decr in H4.
      assert (Fleb x y && Fleb y x = true) by (apply andb_true_intro; split; eauto using Fleb_trans).
      congruence.
      symmetry. apply Bopp_involutive.
      eapply Fleb_nan_r. eapply next_ulp_le. destruct f0; try reflexivity; discriminate.
  - rewrite Float.cmp_ne_eq in H1.
    destruct (Float.cmp Ceq x y) eqn:?; inversion H1; clear H1.
    eapply botbind2_spec.
    2:eapply backward_le_l_correct; eauto 3.
    3:eapply botlift_fun1_spec.
    3:apply forward_fneg_correct.
    3:eapply backward_le_l_correct.
    3:apply forward_fneg_correct; eauto.
    3:apply forward_fneg_correct; eauto.
    intros.
    apply meet_top_correct. split. auto.
    destruct x as [[]|[]|[]|[]]; apply H3.
    unfold Fleb. unfold Float.cmp in Heqb.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
    rewrite fneg_decr.
    unfold Fleb. unfold Float.cmp in Heqb. rewrite Bcompare_swap.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
  - assert (is_nan _ _ x = false).
    { destruct x; try discriminate; reflexivity. }
    assert (is_nan _ _ y = false).
    { destruct x as [|[]| |[]], y; try discriminate; reflexivity. }
    unfold backward_cmpf_l.
    destruct (match y_ab with
                | All => (B754_infinity 53 1024 true, B754_infinity 53 1024 false)
                | Just (ITV my My) => (my, My)
              end) as [my My] eqn:?.
    assert (Fleb my y = true /\ Fleb y My = true).
    { destruct y_ab as [|[]]; inversion Heqp; clear Heqp; subst.
      destruct y as [|[]| |[]]; try (split; reflexivity).
      destruct x as [|[]| |[]]; inversion H1.
      apply H0. }
    clear Heqp. destruct H5.
    destruct (Fleb My (B754_infinity _ _ true)) eqn:?.
    pose proof (Fleb_trans _ _ _ _ _ H6 Heqb).
    destruct x as [|[]| |[]], y as [|[]| |[]]; discriminate.
    eapply backward_le_l_correct. auto.
    apply forward_fneg_correct with (x_ab:=Just (ITV _ _)).
    split; apply next_ulp_incr; try reflexivity; rewrite fneg_decr; eauto.
    replace x with (Float.neg (Float.neg x)).
    rewrite fneg_decr.
    destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (Float.neg y) eq_refl).
    destruct x; try discriminate; reflexivity.
    destruct y; try discriminate; reflexivity.
    rewrite fneg_decr in H7.
    unfold Fleb in H7. unfold Float.cmp in H1. rewrite Bcompare_swap in H7.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
    auto.
    unfold Float.neg. apply Bopp_involutive. auto.
  - destruct x_ab as [|[]], y_ab as [|[]]; try exact I. apply H.
    destruct H, H0.
    eapply botlift_fun1_spec.
    intros.
    replace x with (Float.neg (Float.neg x)). eapply forward_fneg_correct. eauto.
    unfold Float.neg. apply Bopp_involutive. eauto.
    eapply backward_le_l_correct.
    eapply forward_fneg_correct. split; auto.
    eapply forward_fneg_correct. split; eauto.
    rewrite fneg_decr.
    destruct (Fleb_total _ _ y x); eauto 3.
    unfold Fleb in H5 |- *. unfold Float.cmp in H1.
    rewrite Bcompare_swap. destruct Bcompare as [[]|]; try discriminate; try reflexivity.
  - eapply backward_le_l_correct; eauto.
  - eapply backward_not_le_l_correct; eauto.
  - assert (is_nan _ _ x = false).
    { destruct x; try discriminate; reflexivity. }
    assert (is_nan _ _ y = false).
    { destruct x as [|[]| |[]], y; try discriminate; reflexivity. }
    unfold backward_cmpf_l.
    destruct (match y_ab with
                | All => (B754_infinity 53 1024 true, B754_infinity 53 1024 false)
                | Just (ITV my My) => (my, My)
              end) as [my My] eqn:?.
    assert (Fleb my y = true /\ Fleb y My = true).
    { destruct y_ab as [|[]]; inversion Heqp; clear Heqp; subst.
      destruct y as [|[]| |[]]; try (split; reflexivity).
      destruct x as [|[]| |[]]; inversion H1.
      apply H0. }
    clear Heqp. destruct H5.
    destruct (Fleb (B754_infinity _ _ false) my) eqn:?.
    pose proof (Fleb_trans _ _ _ _ _ Heqb H5).
    destruct x as [|[]| |[]], y as [|[]| |[]]; discriminate.
    replace x with (Float.neg (Float.neg x)).
    eapply botlift_fun1_spec.
    apply forward_fneg_correct.
    eapply backward_le_l_correct.
    apply forward_fneg_correct.
    auto.
    apply forward_fneg_correct.
    split; apply next_ulp_incr; eauto; reflexivity.
    rewrite fneg_decr.
    destruct (next_ulp_lub 53 1024 eq_refl eq_refl x y eq_refl).
    destruct x; try discriminate; reflexivity.
    destruct y; try discriminate; reflexivity.
    unfold Fleb in H7. unfold Float.cmp in H1. destruct Bcompare as [[]|]; discriminate.
    auto.
    unfold Float.neg. apply Bopp_involutive. auto.
  - destruct x_ab as [|[]], y_ab as [|[]]; try exact I. apply H.
    eapply backward_le_l_correct; eauto 3.
    destruct H, H0.
    destruct (Fleb_total _ _ x y); eauto 3.
    unfold Fleb in H5 |- *. unfold Float.cmp in H1.
    rewrite Bcompare_swap in H5.
    destruct Bcompare as [[]|]; try discriminate; reflexivity.
  - replace x with (Float.neg (Float.neg x)).
    eapply botlift_fun1_spec.
    apply forward_fneg_correct.
    eapply backward_le_l_correct.
    apply forward_fneg_correct. auto.
    apply forward_fneg_correct. eauto.
    rewrite fneg_decr. rewrite <- Float.cmp_swap in H1. apply H1.
    unfold Float.neg. apply Bopp_involutive.
    destruct x; try discriminate; reflexivity.
  - eapply botlift_fun1_spec.
    instantiate (1:=Float.neg x).
    intros. apply forward_fneg_correct in H3.
    destruct x as [[]|[]|[]|[]]; apply H3.
    eapply backward_not_le_l_correct.
    apply forward_fneg_correct. auto.
    apply forward_fneg_correct. eauto.
    rewrite fneg_decr. rewrite <- Float.cmp_swap in H1. apply H1.
Qed.

Definition backward_absf (res_ab:itv+⊤) (x_ab:itv+⊤) : itv+⊤+⊥ :=
  match res_ab with
    | All => NotBot x_ab
    | Just (ITV _ _) =>
      botbind (fun res_ab' => meet_top x_ab (forward_fneg res_ab' ⊔ res_ab'))
              (meet_top res_ab (Just (ITV (B754_zero _ _ false)
                                          (B754_infinity _ _ false))))
  end.

Lemma backward_absf_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall res_ab, Float.abs x ∈ γ res_ab ->
    x ∈ γ (backward_absf res_ab x_ab).
Proof.
Transparent Float.abs.
  intros.
  destruct res_ab as [|[]]. apply H.
  eapply botbind_spec.
  2:apply meet_top_correct; split; eauto.
  intros.
  apply meet_top_correct.
  split; auto.
  apply join_sound. apply itv_adom.
  destruct x as [|[]| |[]].
  right. destruct a; apply H1.
  left. apply forward_fneg_correct in H1. apply H1.
  right. apply H1.
  destruct H0. discriminate.
  left. apply forward_fneg_correct in H1. apply H1.
  right. apply H1.
  destruct x as [|[]| |[]]; try (split; reflexivity).
  destruct H0. discriminate.
Qed.

Definition backward_addf_l (res x y:itv+⊤) : itv+⊤+⊥ :=
  match y with
    | All => NotBot x
    | Just (ITV my My) =>
      match res with
        | All => NotBot x
        | Just res =>
          match unround_itv res with
            | ITV mres Mres =>
              let m:= Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_UP mres My in
              let M:= Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_DN Mres my in
              if is_nan _ _ m || is_nan _ _ M then NotBot x
              else meet_top x (Just (ITV m M))
          end
      end
  end.

Lemma backward_addf_l_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ res_ab, (Float.add x y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_addf_l res_ab x_ab y_ab).
Proof.
  intros.
  destruct y_ab as [|[]]; auto. destruct H0. simpl.
  destruct res_ab as [|res_ab]; auto.
  Transparent Float.add. unfold Float.add in H1.
  destruct (is_finite _ _ x && is_finite _ _ y) eqn:?.
  - rewrite Bool.andb_true_iff in Heqb. destruct Heqb.
    pose proof (unround_itv_round (B2R 53 1024 x + B2R 53 1024 y)
                 (Bplus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_NE x y)).
    refine (let H5 := H5 _ _ H1 in _).
    { clear H5. use_float_specs.
      destruct Rlt_bool. intuition.
      destruct H5; auto. rewrite H5. f_equal.
      destruct (Rlt_bool_spec (B2R 53 1024 x + B2R 53 1024 y) 0);
      destruct x as [[]| | |[]], y as [[]| | |[]]; try discriminate;
      try reflexivity; simpl in *; try fourier.
      pose proof (F2R_gt_0_compat radix2 {| Fnum := Z.pos m; Fexp := e |} eq_refl);
        pose proof (F2R_gt_0_compat radix2 {| Fnum := Z.pos m0; Fexp := e1 |} eq_refl);
        fourier.
      pose proof (F2R_lt_0_compat radix2 {| Fnum := Z.neg m; Fexp := e |} eq_refl);
        pose proof (F2R_lt_0_compat radix2 {| Fnum := Z.neg m0; Fexp := e1 |} eq_refl);
        fourier. }
    clear H5.
    destruct unround_itv; auto.
    destruct (is_nan _ _ (Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_UP f1 f0)) eqn:?,
             (is_nan _ _ (Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_DN f2 f)) eqn:?;
             auto.
    eapply meet_top_correct; split; auto.
    unfold gammaR, RF_lt, RFF_lt, Float.neg in H6.
    rewrite !is_finite_FF_B2FF, !FF2R_B2FF, is_finite_Bopp, B2R_Bopp in H6.
    split.
    + destruct (is_finite _ _ f1) eqn:?. destruct (is_finite _ _ f0) eqn:?.
      * destruct H6.
        rewrite Fleb_Rle in H2 by auto. Rle_bool_case H2; [|discriminate].
        assert (B2R _ _ f1 - B2R _ _ f0 <= B2R _ _ x) by fourier.
        use_float_specs.
        unfold Z.compare, Pos.compare, Pos.compare_cont in H11.
        Rlt_bool_case H11.
        destruct H11 as [? []]; auto. rewrite Fleb_Rle by auto. apply Rle_bool_true.
        rewrite H11.
        apply round_le_generic; auto. apply generic_format_B2R; auto.
        destruct H11, Bminus as [|[]| |], (Bsign _ _ f1) eqn:?; auto; try discriminate.
        edestruct RIneq.Rle_not_lt. apply H12.
        rewrite Rabs_pos_eq. eapply Rle_lt_trans. apply round_le; eauto using Rlt_le.
        rewrite round_generic by (auto; apply generic_format_B2R).
        eapply Rle_lt_trans. apply Rle_abs. apply abs_B2R_lt_emax.
        eapply round_ge_generic; eauto. apply generic_format_0.
        assert (0 <= B2R _ _ f1).
        { destruct f1 as [| | |[]]; try discriminate. apply Rle_refl.
          apply F2R_ge_0_compat. compute; discriminate. }
        assert (B2R _ _ f0 <= 0).
        { destruct f0 as [| | |[]]; try discriminate. apply Rle_refl.
          apply F2R_le_0_compat. compute; discriminate. }
        fourier.
        inversion H11; subst.
        destruct (next_ulp_lub 53 1024 eq_refl eq_refl x (B754_infinity _ _ true) eq_refl).
        destruct x; try reflexivity; discriminate.
        reflexivity.
        destruct x as [| | |[]]; try discriminate.
        auto.
      * assert (is_nan _ _ f0 = false) by eauto.
        destruct f0; try discriminate.
        destruct b. destruct y as [| | |[]]; try discriminate.
        destruct f1, x; try discriminate; reflexivity.
      * destruct H6. destruct f1 as [|[]|? []|]; inversion H6; subst.
        destruct f0 as [|[]| |], x; try discriminate; try reflexivity.
    + destruct (is_finite _ _ f) eqn:?. destruct (is_finite _ _ f2) eqn:?.
      * destruct H6.
        rewrite Fleb_Rle in H0 by auto. Rle_bool_case H0; [|discriminate].
        assert (B2R _ _ x <= B2R _ _ f2 - B2R _ _ f) by fourier.
        use_float_specs.
        unfold Z.compare, Pos.compare, Pos.compare_cont in H10.
        Rlt_bool_case H10.
        destruct H10 as [? []]; auto. rewrite Fleb_Rle by auto. apply Rle_bool_true.
        rewrite H10.
        apply round_ge_generic; auto. apply generic_format_B2R; auto.
        clear Heqb.
        destruct H10, Bminus as [|[]| |], (Bsign _ _ f2) eqn:?; auto; try discriminate.
        edestruct RIneq.Rle_not_lt. apply H12.
        rewrite Rabs_left1. eapply Rle_lt_trans. apply Ropp_le_contravar.
        apply round_le; eauto.
        rewrite round_generic by (auto; apply generic_format_B2R).
        eapply Rle_lt_trans. apply Rle_abs. rewrite Rabs_Ropp. apply abs_B2R_lt_emax.
        eapply round_le_generic; eauto. apply generic_format_0.
        assert (0 <= B2R _ _ f).
        { destruct f as [| | |[]]; try discriminate. apply Rle_refl.
          apply F2R_ge_0_compat. compute; discriminate. }
        assert (B2R _ _ f2 <= 0).
        { destruct f2 as [| | |[]]; try discriminate. apply Rle_refl.
          apply F2R_le_0_compat. compute; discriminate. }
        fourier.
        inversion H10; subst.
        rewrite <- fneg_decr.
        destruct (next_ulp_lub 53 1024 eq_refl eq_refl (Float.neg x) (B754_infinity _ _ true) eq_refl).
        destruct x; try reflexivity; discriminate.
        reflexivity.
        destruct x as [| | |[]]; try discriminate.
        auto.
      * destruct H6. destruct f2 as [|[]|? []|]; inv H5.
        destruct f, x as [| | |[]]; try discriminate; try reflexivity.
      * assert (is_nan _ _ f = false) by eauto.
        destruct f; try discriminate.
        destruct b.
        destruct f2 as [|[]| |], x as [| | |[]]; try discriminate; reflexivity.
        destruct y as [| | |[]]; try discriminate.
  - destruct (unround_itv res_ab) eqn:?. auto.
    destruct (is_nan 53 1024 (Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_UP f1 f0)) eqn:?,
             (is_nan 53 1024 (Bminus 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) Float.binop_pl mode_DN f2 f)) eqn:?;
             auto.
    apply meet_top_correct; split; auto.
    destruct res_ab; inversion Heqi; clear Heqi; subst.
    destruct H1.
    split.
    + destruct x as [|[]| |];
      try
        (destruct y as [|[]| |]; try discriminate;
          match goal with |- context [Bminus _ _ _ _ _ _ ?f3' _] =>
            destruct f3' as [|[]| |] eqn:?; try reflexivity; try discriminate end;
          destruct f0 as [|[]| |], f3 as [|[]| |[]];
          try reflexivity; discriminate).
      match goal with |- context [Fleb ?m _] => destruct m as [|[]| |[]] end;
        try discriminate; reflexivity.
    + destruct x as [|[]| |[]];
      try
        (destruct y as [|[]| |]; try discriminate;
         match goal with |- context [Bminus _ _ _ _ _ _ ?f4' _] =>
           destruct f4' as [|[]| |] eqn:?; try reflexivity; try discriminate end;
         destruct f as [|[]| |[]], f4 as [|[]| |];
         try reflexivity; discriminate).
      match goal with |- context [Fleb _ ?m] => destruct m as [|[]| |[]] end;
        try discriminate; reflexivity.
Qed.

Definition backward_addf_r (res x y:itv+⊤) : itv+⊤+⊥ := backward_addf_l res y x.

Lemma backward_addf_r_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ res_ab, (Float.add x y)%Z ∈ γ res_ab ->
    y ∈ γ (backward_addf_r res_ab x_ab y_ab).
Proof.
  intros. eapply backward_addf_l_correct; eauto.
  unfold Float.add. rewrite fadd_commut.
  destruct res_ab, x as [|[]| |], y as [|[]| |]; apply H1.
Qed.

Definition backward_subf_l (res x y:itv+⊤) : itv+⊤+⊥ := backward_addf_l res x (forward_fneg y).

Lemma backward_subf_l_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ res_ab, (Float.sub x y)%Z ∈ γ res_ab ->
    x ∈ γ (backward_subf_l res_ab x_ab y_ab).
Proof.
  intros. eapply backward_addf_l_correct; eauto.
  eapply forward_fneg_correct. eauto.
  destruct res_ab, x, y; apply H1.
Qed.

Definition backward_subf_r (res x y:itv+⊤) : itv+⊤+⊥ :=
  botlift_fun1 forward_fneg (backward_addf_l res (forward_fneg y) x).

Lemma backward_subf_r_correct:
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  ∀ res_ab, (Float.sub x y)%Z ∈ γ res_ab ->
    y ∈ γ (backward_subf_r res_ab x_ab y_ab).
Proof.
  intros.
  eapply botlift_fun1_spec.
  2:eapply backward_addf_r_correct; eauto.
  2:eapply forward_fneg_correct; eauto.
  intros. apply forward_fneg_correct in H2.
  destruct forward_fneg, y as [|[]| |[]]; try apply H2.
  destruct res_ab, x, y; apply H1.
Qed.

Definition binary32offloat_DN (f: float) : binary32 :=
  Bconv 53 1024 24 128 eq_refl eq_refl Float.to_single_pl mode_DN f.

Lemma binary32offloat_DN_Fleb :
  ∀  f1 f2,
    Fleb (Float.of_single f1) f2 = true ->
    Fleb f1 (binary32offloat_DN f2) = true.
Proof.
  intros. unfold binary32offloat_DN.
  destruct (floatofsingle_exact f1).
  assert (is_nan _ _ f2 = false) by eauto.
  assert (is_nan _ _ f1 = false).
  { destruct f1; try reflexivity; discriminate. }
  destruct (is_finite _ _ f2) eqn:?.
  - use_float_specs. specialize (H4 Heqb).
    Rlt_bool_case H4.
    + destruct H4 as (? & ? & ?); auto.
      destruct (is_finite _ _ f1) eqn:?.
      * rewrite Fleb_Rle in * by auto. apply Rle_bool_true. rewrite H4.
        Rle_bool_case H. 2:discriminate.
        apply round_ge_generic; auto using generic_format_B2R.
        rewrite <- H0. auto.
      * destruct f1 as [|[]| |]; try discriminate; try reflexivity.
        destruct Bconv; try discriminate; reflexivity.
        destruct f2; discriminate.
    + destruct f2; try discriminate.
      rewrite round_0, Rabs_R0 in H5 by auto. apply RIneq.Rle_not_lt in H5. destruct H5. apply bpow_gt_0.
      destruct b.
      * destruct (is_finite _ _ f1) eqn:?.
        rewrite Rabs_left1 in H5. rewrite Fleb_Rle in H by auto. Rle_bool_case H; [|discriminate].
        rewrite H0 in H6.
        eapply round_ge_generic with (rnd:=round_mode mode_DN) in H6.
        4:apply generic_format_B2R. 2:auto. 2:auto.
        destruct (Rabs_lt_inv _ _ (abs_B2R_lt_emax _ _ f1)).
        fourier.
        erewrite <- round_0. apply round_le. auto. auto. apply F2R_le_0_compat. compute. discriminate. auto.
        destruct f1 as [|[]| |]; try discriminate.
        destruct Bconv as [|[]|?[]|]; try discriminate. reflexivity.
      * destruct Bconv as [| |?[]|]; inv H4.
        destruct f1 as [|[]| |]; try discriminate; try reflexivity.
        unfold Fleb, Bcompare. destruct b. auto. clear H H0 H1 H3.
        unfold bounded in e3. rewrite Bool.andb_true_iff in e3. destruct e3.
        apply Zle_is_le_bool in H0.
        destruct (Z.compare_spec e1 104). 3:omega. 2:auto.
        rewrite Pos.compare_cont_spec. destruct (Pos.compare_spec m0 16777215); try discriminate.
        auto. auto.
        subst e1. apply canonic_canonic_mantissa with (sx:=false) in H.
        unfold canonic, canonic_exp, FLT_exp in H.
        rewrite ln_beta_F2R in H. 2:discriminate. simpl in H.
        assert (bpow radix2 24 <= P2R m0).
        { rewrite <- Z2R_Zpower by omega. apply Z2R_le with (n:=Z.pos m0).
          change (radix2 ^ 24)%Z with 16777216%Z. zify; omega. }
        apply ln_beta_le with (r:=radix2) in H1. 2:apply bpow_gt_0.
        rewrite ln_beta_bpow in H1. zify; omega.
  - destruct f2 as [|[]| |]; try discriminate.
    + destruct (Float.of_single f1) as [|[]| |]eqn:?; try discriminate.
      destruct f1 as [|[]| |]; try discriminate; try reflexivity.
      discriminate (H1 eq_refl).
      destruct b, f1 as [|[]| |]; discriminate.
    + destruct f1 as [|[]| |[]]; try reflexivity. discriminate.
Qed.

Definition unround_itv_single (x:itv) : itv :=
  match x with
    | ITV mx Mx =>
      ITV (Float.neg (Float.of_single
                        (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN (Float.neg mx)))))
          (Float.of_single (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN Mx)))
  end.

Lemma unround_itv_round_single :
  forall x_r x,
    (if Rlt_bool
          (Rabs (round radix2 (FLT_exp (3 - 128 - 24) 24)
                       (round_mode mode_NE) x_r))
          (bpow radix2 128)
     then B2R _ _ x = round radix2 (FLT_exp (3 - 128 - 24) 24)
                            (round_mode mode_NE) x_r /\
          is_finite _ _ x = true
     else B2FF _ _ x = binary_overflow 24 128 mode_NE (Rlt_bool x_r 0)) ->
  forall x_ab, (Float.of_single x) ∈ γ x_ab ->
  gammaR (unround_itv_single x_ab) x_r.
Proof.
  intros.
  destruct x_ab; destruct H0; try easy.
  apply binary32offloat_DN_Fleb in H1.
  rewrite <- fneg_decr in H0.
  Transparent Float.neg Float32.neg Float.of_single.
  assert (Float.neg (Float.of_single x) = Float.of_single (Float32.neg x))
    by (apply Bconv_Bopp; auto).
  rewrite H2 in H0.
  apply binary32offloat_DN_Fleb in H0.
  split.
  - eapply (next_ulp_round_overflow 24 128 eq_refl eq_refl) in H; eauto. 2:reflexivity.
    unfold RF_lt, RFF_lt in *. rewrite FF2R_B2FF, is_finite_FF_B2FF in *.
    assert (is_finite _ _ (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN f0)) =
            is_finite 53 1024
                      (Float.of_single
                         (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN f0)))).
    { destruct next_ulp; try reflexivity. rewrite (proj2 (floatofsingle_exact _)); auto. }
    rewrite <- H3. clear H3.
    destruct is_finite. rewrite (proj1 (floatofsingle_exact _)). auto.
    destruct next_ulp as [| |? []|]; inv H. reflexivity.
  - apply (next_ulp_round_overflow 24 128 eq_refl eq_refl) with (x:=-x_r) (mode:=mode_NE) in H0.
    2:reflexivity.
    unfold RF_lt, RFF_lt in *. rewrite FF2R_B2FF, is_finite_FF_B2FF in *.
    assert (is_finite _ _ (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN (Float.neg f))) =
            is_finite 53 1024
                      (Float.neg (Float.neg
                        (Float.of_single
                           (next_ulp 24 128 eq_refl eq_refl
                                     (binary32offloat_DN (Float.neg f))))))).
    { destruct next_ulp; try reflexivity.
      unfold Float.neg. rewrite !is_finite_Bopp, (proj2 (floatofsingle_exact _)); auto. }
    rewrite <- H3. clear H3.
    clear H. destruct is_finite.
    unfold Float.neg at 1 2. rewrite !B2R_Bopp, (proj1 (floatofsingle_exact _)).
    rewrite Ropp_involutive. auto.
    destruct next_ulp as [| |? []|]; inv H0. reflexivity.
    rewrite round_NE_opp, Rabs_Ropp. unfold round_mode in H.
    Rlt_bool_case H.
    destruct H. unfold Float32.neg. rewrite B2R_Bopp, is_finite_Bopp. split. f_equal. auto. auto.
    assert (x_r <> 0).
    { intro. subst x_r. rewrite round_0, Rabs_R0 in H3. 2:apply (valid_rnd_round_mode mode_NE).
      pose proof (bpow_gt_0 radix2 128). fourier. }
    assert (Rlt_bool (- x_r) 0 = negb (Rlt_bool x_r 0)).
    { destruct (Rcompare_spec x_r 0).
      rewrite Rlt_bool_false, Rlt_bool_true by fourier. auto.
      congruence.
      rewrite Rlt_bool_true, Rlt_bool_false by fourier. auto. }
    rewrite H5.
    destruct x as [| |? []|]; inv H. reflexivity.
Qed.

Definition backward_singleoffloat (res_ab:itv+⊤) (x_ab:itv+⊤) : itv+⊤+⊥ :=
  match res_ab with
    | All => NotBot x_ab
    | Just res_ab => meet_top (Just (unround_itv_single res_ab)) x_ab
  end.

Lemma backward_singleoffloat_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall res_ab, Float.of_single (Float.to_single x) ∈ γ res_ab ->
    x ∈ γ (backward_singleoffloat res_ab x_ab).
 Proof.
  intros. destruct res_ab as [|res_ab]. auto.
  apply meet_top_correct. split. 2:auto.
  destruct (is_finite _ _ x) eqn:?.
  - eapply unround_itv_round_single with (x_r := B2R _ _ x) in H0.
    destruct unround_itv_single. auto. destruct H0.
    split.
    + unfold RF_lt, RFF_lt, Float.neg in H1.
      rewrite is_finite_FF_B2FF, FF2R_B2FF, is_finite_Bopp in H1.
      destruct (is_finite _ _ f) eqn:?.
      rewrite Fleb_Rle by auto. apply Rle_bool_true. rewrite B2R_Bopp in H1. fourier.
      destruct f as [|[]|? []|]; inv H1. destruct x; try discriminate; reflexivity.
    + unfold RF_lt, RFF_lt in H0.
      rewrite is_finite_FF_B2FF, FF2R_B2FF in H0.
      destruct (is_finite _ _ f0) eqn:?.
      rewrite Fleb_Rle by auto. apply Rle_bool_true. fourier.
      destruct f0 as [| |? []|]; inv H0. destruct x as [|[]| |[]]; try discriminate; reflexivity.
    + Transparent Float.to_single. unfold Float.to_single.
      use_float_specs. specialize (H1 Heqb).
      Rlt_bool_case H1. now intuition.
      rewrite <- Bsign_pos0. auto.
      destruct x; try discriminate; reflexivity.
  - destruct res_ab. auto. destruct H0. split.
    + assert (is_nan _ _
                     (Float.of_single
                        (next_ulp 24 128 eq_refl eq_refl
                                  (binary32offloat_DN (Float.neg f)))) = false).
      { apply floatofsingle_nan, next_ulp_is_nan.
        destruct f; try reflexivity; try discriminate.
        apply binary_normalize_is_nan. }
      destruct x as [|[]| |], f as [|[]| |]; try discriminate; try reflexivity.
      destruct b; discriminate.
      clear - H2. destruct Float.of_single as [|[]| |[]]; try reflexivity; discriminate.
    + assert (is_nan _ _
                     (Float.of_single
                        (next_ulp 24 128 eq_refl eq_refl (binary32offloat_DN f0))) = false).
      { apply floatofsingle_nan, next_ulp_is_nan.
        apply Fleb_nan_r in H1.
        destruct f0; try reflexivity. discriminate.
        apply binary_normalize_is_nan. }
      destruct x as [|[]| |], f0 as [|[]| |]; try discriminate; try reflexivity.
      clear - H2. destruct Float.of_single as [|[]| |[]]; try reflexivity; discriminate.
Qed.

Definition backward_singleofz (res_ab:itv+⊤) (x_ab:ZIntervals.itv+⊤) : ZIntervals.itv+⊤+⊥ :=
  match res_ab with
    | All => NotBot x_ab
    | Just res_ab =>
      match unround_itv_single res_ab with
        | ITV mx Mx =>
          let low :=
              match Zoffloat_DN (Float.neg mx), x_ab with
                | Some m, Just (ZIntervals.ITV m' _) => Some (Z.max (-m) m')
                | Some m, All => Some (Z.opp m)
                | None, Just (ZIntervals.ITV m' _) => Some m'
                | None, All => None
              end
          in
          let high :=
              match Zoffloat_DN Mx, x_ab with
                | Some M, Just (ZIntervals.ITV _ M') => Some (Z.min M M')
                | Some M, All => Some M
                | None, Just (ZIntervals.ITV _ M') => Some M'
                | None, All => None
              end
          in
          match low, high with
            | Some m, Some M =>
              if Z.leb m M then NotBot (Just (ZIntervals.ITV m M))
              else Bot
            | _, _ => NotBot All
          end
      end
  end.

Lemma backward_singleofz_correct :
  forall x x_ab, x ∈ γ x_ab ->
  forall res_ab, Float.of_single (BofZ 24 128 (@eq_refl _ Lt) (@eq_refl _ Lt) x) ∈ γ res_ab ->
    x ∈ γ (backward_singleofz res_ab x_ab).
Proof.
  intros.
  unfold backward_singleofz. destruct res_ab as [|res_ab]. auto.
  assert (gammaR (unround_itv_single res_ab) (Z2R x)).
  { eapply unround_itv_round_single; eauto.
    use_float_specs.
    destruct Rlt_bool. intuition.
    destruct (Z.ltb_spec x 0), (Rlt_bool_spec (Z2R x) 0); auto.
    apply (le_Z2R 0 x) in H3. omega.
    apply (lt_Z2R x 0) in H3. omega. }
  destruct unround_itv_single. auto.
  destruct H1. unfold RF_lt, RFF_lt, F2R, Fnum, Fexp, bpow in H1, H2.
  rewrite is_finite_FF_B2FF, FF2R_B2FF in H1, H2.
  assert (forall m, Zoffloat_DN (Float.neg f) = Some m -> (-m <= x)%Z).
  { pose proof (Zoffloat_DN_correct (Float.neg f)).
    destruct (Zoffloat_DN (Float.neg f)). 2:easy.
    destruct H3. rewrite H3 in H2. intros. injection H5. intro. subst z. clear H5.
    apply Zopp_le_cancel, le_Z2R. rewrite Zopp_involutive, H4.
    unfold round, scaled_mantissa, canonic_exp, Fcore_FIX.FIX_exp, F2R, Fexp, Fnum,
           scaled_mantissa, bpow, round_mode, Zopp at 2.
    rewrite !Rmult_1_r.
    apply Z2R_le, Zfloor_lub.
    rewrite Z2R_opp. fourier. }
  assert (forall M, Zoffloat_DN f0 = Some M -> (x <= M)%Z).
  { pose proof (Zoffloat_DN_correct f0).
    destruct (Zoffloat_DN f0). 2:easy.
    destruct H4. rewrite H4 in H1. intros. injection H6. intro. subst z. clear H6.
    apply le_Z2R. rewrite H5.
    unfold round, scaled_mantissa, canonic_exp, Fcore_FIX.FIX_exp, F2R, Fexp, Fnum,
           scaled_mantissa, bpow, round_mode, Zopp.
    rewrite !Rmult_1_r.
    apply Z2R_le, Zfloor_lub. fourier. }
  destruct (Zoffloat_DN (Float.neg f)); [specialize (H3 _ eq_refl)|clear H3];
    (destruct (Zoffloat_DN f0); [specialize (H4 _ eq_refl)|clear H4]);
  destruct x_ab as [|[]]; destruct H; auto; try exact I.
  destruct (Z.leb_spec (-z) z0); [split|]; zify; omega.
  destruct (Z.leb_spec (Z.max (-z) z1) (Z.min z0 z2)); [split|]; zify; omega.
  destruct (Z.leb_spec (Z.max (-z) z0) z1); [split|]; zify; omega.
  destruct (Z.leb_spec z0 (Z.min z z1)); [split|]; zify; omega.
  destruct (Z.leb_spec z z0); [split|]; zify; omega.
Qed.
