(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import ZArith NArith.
Require Import Integers Floats.
Require Import IdealExpr AdomLib.
Require Import Cells AST.

Local Open Scope N_scope.

(* OneAtATime hash *)
Definition MIX (h d:N) : N :=
  let h := h + d in
  let h := N.land 4294967295 (h + N.shiftl h 10) in
  let h := N.lxor h (N.shiftr h 6) in
  h.

Definition final_MIX (h:N) : N :=
  let h := N.land 4294967295 (h + N.shiftl h 3) in
  let h := N.lxor h (N.shiftr h 11) in
  let h := N.land 4294967295 (h + N.shiftl h 15) in
  h.

Function MIX_N_aux (h n:N) {measure N.to_nat n} : N :=
  match n with
    | 0 => MIX h 0
    | _ => MIX_N_aux (MIX h (N.land n 255)) (N.shiftr n 8)
  end.
Proof.
  intros. subst n.
  assert (N.shiftr (N.pos p) 8 < N.pos p). 2:Psatz.lia.
  rewrite N.shiftr_div_pow2.
  change (2 ^ 8)%N with 256%N.
  pose proof N.div_mod' (N.pos p) 256. Psatz.lia.
Qed.

Definition MIX_N (h n:N) : N :=
  MIX_N_aux (MIX h (N.size n)) n.

Definition MIX_Z (h:N) (z:Z) : N :=
  if Z.ltb z 0 then MIX_N (MIX h 0) (Z.to_N (-z))
  else MIX_N (MIX h 1) (Z.to_N z).

Definition MIX_int64 (h:N) (i:int64) : N :=
  let n := Z.to_N (Int64.unsigned i) in
  let h := MIX h (N.land n 255) in
  let n := N.shiftr n 8 in
  let h := MIX h (N.land n 255) in
  let n := N.shiftr n 8 in
  let h := MIX h (N.land n 255) in
  let n := N.shiftr n 8 in
  let h := MIX h n in
  h.

Fixpoint MIX_iexpr {var:Type} (MIX_var:N -> var -> N) (h:N) (e:iexpr var) :=
  match e with
    | IEvar _ x => MIX_var (MIX h 1) x
    | IEZconst (Just (ZIntervals.ITV a b)) =>
      MIX_Z (MIX_Z (MIX h 2) a) b
    | IEFconst (Just (FloatIntervals.ITV a b)) =>
      MIX_int64 (MIX_int64 (MIX h 3) (Float.to_bits a)) (Float.to_bits b)
    | IEunop op e =>
      let key := match op with
                   | IOneg => 4            | IOnot => 5            | IOnegf => 6
                   | IOabsf => 7           | IOsingleoffloat => 8  | IOfloatofz => 9
                   | IOsingleofz => 10     | IOzoffloat => 11
                 end in
      MIX_iexpr MIX_var (MIX h key) e
    | IEbinop op e1 e2 =>
      let key := match op with
                   | IOadd  => 12      | IOsub  => 13      | IOmul  => 14
                   | IOdiv  => 15      | IOmod  => 16      | IOand  => 17
                   | IOor  => 18       | IOxor  => 19      | IOshl  => 20
                   | IOshr  => 21      | IOcmp Ceq => 22   | IOcmp Cne => 23
                   | IOcmp Clt => 24   | IOcmp Cle => 25   | IOcmp Cgt => 26
                   | IOcmp Cge => 27   | IOaddf  => 28     | IOsubf  => 29
                   | IOmulf  => 30     | IOdivf  => 31     | IOcmpf Ceq => 32
                   | IOcmpf Cne => 33  | IOcmpf Clt => 34  | IOcmpf Cle => 35
                   | IOcmpf Cgt => 36  | IOcmpf Cge => 37
                 end in
      MIX_iexpr MIX_var (MIX_iexpr MIX_var (MIX h key) e1) e2
    | IEZconst All => MIX h 38
    | IEFconst All => MIX h 39
  end.

Fixpoint MIX_mux_iexpr {var:Type} (MIX_var:N -> var -> N) (h:N) (e:mux_iexpr var) :=
  match e with
    | Leaf e => MIX_iexpr MIX_var h e
    | Mux e1 e2 e3 =>
      MIX_mux_iexpr MIX_var (MIX_mux_iexpr MIX_var (MIX_mux_iexpr MIX_var (MIX h 40) e1) e2) e3
  end.

Definition MIX_ident (h:N) (i:ident) : N := MIX_N h (N.pos i).

Definition MIX_typed_memory_chunk (h:N) (c:MemChunkTree.typed_memory_chunk) : N :=
  let key := match proj1_sig c with
               | Mint8signed => 0    | Mint8unsigned => 1  | Mint16signed => 2
               | Mint16unsigned => 3 | Mint32 => 4         | Mint64 => 5
               | Mfloat32 => 6       | Mfloat64 => 7
               | Many32 | Many64 => 0 (* dummy *)
             end in
  MIX h key.

Definition MIX_cell (h:N) (c:cell) : N :=
  match c with
    | AClocal x y z t =>
      MIX_typed_memory_chunk (MIX_Z (MIX_ident (MIX_ident (MIX h 0) x) y) z) t
    | ACtemp x y =>
      MIX_ident (MIX_ident (MIX h 1) x) y
    | ACglobal x y z =>
      MIX_typed_memory_chunk (MIX_Z (MIX_N (MIX h 2) (Npos x)) y) z
  end.
