(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib Utf8.
Require Import Maps ShareTree Util.
Require Import AdomLib.
Require Import IdealExpr.
Require Import AbIdealEnv.
Require Import AbIdealNonrel.
Require Import IdealIntervals.

(* TODO : make use of non relationnal info. *)

Module Make(T:SHARETREE).

Module Dom := WPFun T.

Definition var := T.elt.

Section S.

  Context {Val: Type} {Dval: ab_ideal_nonrel Val}.
  Context (Dval_cor: ab_ideal_nonrel_correct Dval).
  Context (Hjoin : ∀ x : Val, pl_join as_pl x x = Just x)
          (Hwiden : ∀ x : Val, pl_widen as_pl x x = Just x)
          (Hleb : ∀ x : Val, pl_leb as_pl x x = true).

  Definition t := T.t Val.

  Instance var_eq : EqDec var := T.elt_eq.
  Instance gamm : gamma_op t (var->ideal_num) := Dom.gamma as_gamma.

  Fixpoint eval_expr (fuel:nat) (chan:in_chan) : iexpr var -> t -> Val+⊤+⊥ :=
    fix eval_expr' (e:iexpr var) (ρ:t) :=
      match e with
        | IEvar _ x =>
          let abx := Dom.get x ρ in
          if forward_try_subst abx then
            match chan.(get_eq_expr) x with
              | Some (Leaf e') =>
                match fuel with
                  | S fuel => botbind (meet abx) (eval_expr fuel chan e' ρ)
                  | O => NotBot abx
                end
              | _ => NotBot abx
            end
          else
            NotBot abx
        | IEZconst (Just itv) => NotBot (z_const itv)
        | IEFconst (Just f) => NotBot (float_const f)
        | IEZconst All | IEFconst All => NotBot All
        | IEunop op e => botbind (forward_unop op) (eval_expr' e ρ)
        | IEbinop op e1 e2 =>
          botbind2 (forward_binop op) (eval_expr' e1 ρ) (eval_expr' e2 ρ)
      end.
  Definition eval_expr_fuel := 5%nat.
  Opaque eval_expr_fuel.

  Lemma eval_expr_correct:
    ∀ fuel chan e ρ ab, ρ ∈ γ ab ->
                        ρ ∈ γ chan ->
                        eval_iexpr ρ e ⊆ γ (eval_expr fuel chan e ab).
  Proof.
    fix 1.
    induction e; intros ρ ab Hρ Hchan x EVAL; inv EVAL.
    - destruct fuel; simpl.
      + destruct forward_try_subst. destruct @get_eq_expr as [[]|]; apply Hρ. apply Hρ.
      + destruct forward_try_subst. destruct (chan.(get_eq_expr) v) as [[]|] eqn:?; try apply Hρ.
        * eapply botbind_spec. intros. apply meet_correct. split. apply Hρ. apply H.
          eapply eval_expr_correct; eauto.
          eapply get_eq_expr_correct in Heqo. inv Heqo. eauto. auto.
        * apply Hρ.
    - destruct t0, fuel; try apply z_const_correct; auto.
    - destruct t0, fuel; try apply float_const_correct; auto.
    - destruct fuel; eapply botbind_spec; eauto using forward_unop_sound.
    - destruct fuel; eapply botbind2_spec; eauto using forward_binop_sound.
  Qed.

  Definition assign (x:var) (e:iexpr var) (ρ:t * in_chan) (chan:@in_chan var) : t+⊥ :=
    botlift_fun1 (Dom.set (fst ρ) x) (eval_expr eval_expr_fuel (snd ρ) e (fst ρ)).

  Lemma assign_correct:
    ∀ x e ρ, ∀ n ab c,
      ρ ∈ γ ab -> (upd ρ x n) ∈ γ c ->
      n ∈ eval_iexpr ρ e ->
      (upd ρ x n) ∈ γ (assign x e ab c).
  Proof.
    intros x e ρ n ab c Hρ Hc Hn.
    eapply (botlift_fun1_spec).
    2: destruct Hρ; eapply eval_expr_correct; eauto.
    intros. intro. rewrite Dom.gsspec.
    unfold upd. unfold eq_dec, var_eq. destruct Hρ, T.elt_eq; auto.
  Qed.

  Fixpoint backward_eval (fuel:nat) (chan:in_chan) :
    iexpr var -> t -> Val+⊤ -> t+⊥ :=
    fix backward_eval' (e:iexpr var) (ρ:t) (v:Val+⊤) : t+⊥ :=
    match e with
      | IEvar _ x =>
        let abx := Dom.get x ρ in
        do_bot abxmeet <- meet v abx;
        let newρ := Dom.set ρ x abxmeet in
        if backward_try_subst v abx then
          match chan.(get_eq_expr) x with
            | Some (Leaf e') =>
              match fuel with
                | S fuel => backward_eval fuel chan e' newρ abxmeet
                | O => NotBot newρ
              end
            | _ => NotBot newρ
          end
        else
          NotBot newρ
      | IEZconst (Just i) =>
        botlift_fun1 (fun _ => ρ) (meet v (z_const i))
      | IEFconst (Just f) =>
        botlift_fun1 (fun _ => ρ) (meet v (float_const f))
      | IEZconst All | IEFconst All => NotBot ρ
      | IEunop op e =>
        do_bot ve <- eval_expr fuel chan e ρ;
        do_bot vres <- botbind (meet v) (forward_unop op ve);
        botbind (backward_eval' e ρ) (backward_unop op vres ve)
      | IEbinop op e1 e2 =>
        do_bot val1 <- eval_expr fuel chan e1 ρ;
        do_bot val2 <- eval_expr fuel chan e2 ρ;
        do_bot vres <- botbind (meet v) (forward_binop op val1 val2);
        do_bot is <- backward_binop op vres val1 val2;
        let '(v1, v2) := is in
        do_bot ρ <- backward_eval' e2 ρ v2;
        backward_eval' e1 ρ v1
    end.
  Definition backward_eval_fuel := 5%nat.

  Lemma backward_eval_correct:
    ∀ fuel e res res_ab, res ∈ γ res_ab ->
    ∀ ρ ρ_ab chan, ρ ∈ γ ρ_ab -> ρ ∈ γ chan ->
              eval_iexpr ρ e res ->
              ρ ∈ γ (backward_eval fuel chan e ρ_ab res_ab).
  Proof.
    fix 1.
    induction e; intros res res_ab Hres ρ ρ_ab chan Hρ Hchan EVAL; inversion EVAL; subst; clear EVAL.
    - destruct fuel; (eapply botbind_spec; [|apply meet_correct; eauto]);
      destruct backward_try_subst; destruct (get_eq_expr chan v) as [[]|] eqn:?; intros;
      try (eapply backward_eval_correct; eauto; [|eapply get_eq_expr_correct in Heqo; inv Heqo; eauto]);
      (intro; rewrite Dom.gsspec; destruct eq_dec; [congruence|eauto]).
    - destruct fuel, t0; auto; (eapply botlift_fun1_spec; [auto|]);
      apply AbIdealNonrel.meet_correct; eauto using z_const_correct.
    - destruct fuel, t0; auto; eapply botlift_fun1_spec; auto;
      apply AbIdealNonrel.meet_correct; eauto using float_const_correct.
    - destruct fuel;
      (repeat (eapply botbind_spec; [intros|]);
        [eapply IHe with (chan:=chan); eauto|
         eapply backward_unop_sound; eauto|
         eapply meet_correct; eauto|
         eapply forward_unop_sound; eauto|
         eapply eval_expr_correct; eauto]).
    - destruct fuel;
      (repeat (eapply botbind_spec; [intros; try destruct a2|]);
       [| |eapply backward_binop_sound; eauto|
        eapply meet_correct; eauto|
        eapply forward_binop_sound; eauto|
        eapply eval_expr_correct; eauto|
        eapply eval_expr_correct; eauto];
       destruct H3; [eapply IHe1 with (chan := chan)|eapply IHe2 with (chan:=chan)];
       eauto).
  Qed.

  Definition assume (e:iexpr var) (b:bool) (ρ:t * in_chan) (c:@in_chan var): t+⊥ :=
    let v := if b then 1 else 0 in
    backward_eval backward_eval_fuel (snd ρ) e (fst ρ) (z_const (ZIntervals.ITV v v)).

  Lemma assume_correct:
    ∀ c ρ ab b chan, ρ ∈ γ ab -> ρ ∈ γ chan ->
                (INz (if b:bool then 1 else 0)) ∈ eval_iexpr ρ c ->
                ρ ∈ γ (assume c b ab chan).
  Proof.
    unfold γ.
    intros. eapply backward_eval_correct; eauto. apply z_const_correct. unfold γ. simpl; omega.
    apply H. apply H.
  Qed.

  Definition forget (x:var) (ρ:t*@in_chan var) (c:@in_chan var): t+⊥ :=
    NotBot (Dom.set (fst ρ) x All).

  Lemma forget_correct:
    ∀ x ρ, ∀ n ab c,
      ρ ∈ γ ab -> (upd ρ x n) ∈ γ c ->
      (upd ρ x n) ∈ γ (forget x ab c).
  Proof.
    intros x ρ n ab c Hρ Hc y. unfold upd.
    rewrite Dom.gsspec. unfold eq_dec, var_eq. destruct T.elt_eq; auto. exact I.
    destruct Hρ. auto.
  Qed.

  Variable extract_inchan : @in_chan var -> t -> t*@in_chan var.
  Hypothesis extract_inchan_correct :
    forall chan ab ρ,
      ρ ∈ γ chan -> ρ ∈ γ ab -> ρ ∈ γ (extract_inchan chan ab).
  Hypothesis extract_inchan_redundant :
    forall chan ab,
      γ chan ∩ γ (fst (extract_inchan chan ab)) ⊆ γ (snd (extract_inchan chan ab)).

  Instance ideal_box_domain : ab_ideal_env (var:=var) t := {
    id_gamma := Dom.gamma _;
    id_top c := extract_inchan c (T.empty _);
    id_leb x y := Dom.leb _ _ (fst x) y;
    id_join x y c := extract_inchan c (Dom.join (pl_join _) _ (fst x) (fst y));
    id_widen x y c := extract_inchan c (Dom.join (pl_widen _) _ (fst x) (fst y));
    assign x e ab c := botlift_fun1 (extract_inchan c) (assign x e ab c);
    forget x ab c := botlift_fun1 (extract_inchan c) (forget x ab c);
    assume e b ab c := botlift_fun1 (extract_inchan c) (assume e b ab c)
  }.
  Proof.
    - intros. apply extract_inchan_correct. auto.
      intro. unfold Dom.get. rewrite T.gempty. exact I.
    - intros. intro. apply Dom.leb_le with (p:=e) (PL:=as_pl) in H.
      destruct H0. specialize (H0 e). unfold flat_top_leb in H.
      destruct Dom.get, Dom.get; try discriminate; try exact I.
      red. red. eapply pl_gamma_monotone; eauto. apply as_pls.
    - intros. apply extract_inchan_correct; auto.
      apply Dom.join_sound. apply as_pls. destruct H as [[]|[]]; auto.
    - intros. eapply botlift_fun1_spec, assign_correct; eauto.
    - intros. eapply botlift_fun1_spec, forget_correct; eauto.
    - intros. eapply botlift_fun1_spec, assume_correct; eauto.
    - intros. apply extract_inchan_redundant; auto.
    - intros. apply extract_inchan_redundant; auto.
    - intros. apply extract_inchan_redundant; auto.
    - unfold botlift_fun1, botbind. intros. destruct assign. now intuition.
      apply extract_inchan_redundant; auto.
    - unfold botlift_fun1, botbind. intros. destruct forget. now intuition.
      apply extract_inchan_redundant; auto.
    - unfold botlift_fun1, botbind. intros. destruct assume. now intuition.
      apply extract_inchan_redundant; auto.
   Grab Existential Variables.
   eauto. eauto. eauto.
  Defined.

End S.

End Make.