(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8 String.
Require Import Coqlib.
Require Import AbIdealEnv IdealExpr.
Require Import AbMachineEnv.
Require Import Integers Floats.
Require Import FloatLib FloatIntervals.
Require Import Reals.
Require Import Fappli_IEEE.
Require Import ZArith.
Require Import Cminor Ctypes.
Require Import AdomLib.
Require Import ShareTree Util.

Open Scope Z_scope.

Ltac destruct_opt :=
  match goal with
    | |- context [match ?opt with | Some _ => _ | None => _ end] =>
      let Heq := fresh "EQOPT" in
      destruct opt eqn:Heq
    | H:context [match ?opt with | Some _ => _ | None => _ end] |- _ =>
      let Heq := fresh "EQOPT" in
      destruct opt eqn:Heq
  end.

Module Make(T:SHARETREE).

Definition var := T.elt.
Instance var_eq : EqDec var := T.elt_eq.

Section Functor.
  Context {abt:Type}  (IdD : ab_ideal_env abt).

  Inductive related : num_val -> ideal_num -> Prop :=
  | rel_int : forall x, related (NVint (Int.repr x)) (INz x)
  | rel_long : forall x, related (NVint64 (Int64.repr x)) (INz x)
  | rel_float : forall x, related (NVfloat x) (INf x)
  | rel_single : forall x, related (NVsingle x) (INf (Float.of_single x)).

  Hint Constructors related.

  Lemma related_unsigned:
    forall x, related (NVint x) (INz (Int.unsigned x)).
  Proof.
    intros. rewrite <- Int.repr_unsigned with (i:=x) at 1. constructor.
  Qed.

  Lemma related_signed:
    forall x, related (NVint x) (INz (Int.signed x)).
  Proof.
    intros. rewrite <- Int.repr_signed with (i:=x) at 1. constructor.
  Qed.

  Lemma related_unsigned_long:
    forall x, related (NVint64 x) (INz (Int64.unsigned x)).
  Proof.
    intros. rewrite <- Int64.repr_unsigned with (i:=x) at 1. constructor.
  Qed.

  Lemma related_signed_long:
    forall x, related (NVint64 x) (INz (Int64.signed x)).
  Proof.
    intros. rewrite <- Int64.repr_signed with (i:=x) at 1. constructor.
  Qed.

  Definition chan_abt := { ab : abt * in_chan | γ (fst ab) ⊆ γ (snd ab) }.
  Definition bot_chan_abt :=
    { ab : (abt * in_chan)+⊥ | γ (botlift_fun1 fst ab) ⊆ γ (botlift_fun1 snd ab) }.
  Definition lift_bot_chan_abt (a:bot_chan_abt) : chan_abt+⊥.
  Proof.
    destruct a as [[|a] AB]. exact Bot.
    apply NotBot. eexists. apply AB.
  Defined.

  Local Instance gamma_id : gamma_op chan_abt (var -> ideal_num) :=
    fun v => γ (proj1_sig v).

  Lemma lift_bot_chan_abt_correct:
    ∀ ab (ρ:var -> ideal_num), γ (proj1_sig ab) ρ -> γ (lift_bot_chan_abt ab) ρ.
  Proof. intros. destruct ab as [[|ab] AB]; auto. Qed.
  Opaque lift_bot_chan_abt.

  Instance gamma_mach : gamma_op chan_abt (var -> num_val) :=
    fun v ρ =>
      ∃ ρ':var -> ideal_num,
        ρ' ∈ γ v ∧ ∀ id, related (ρ id) (ρ' id).

  Definition signednessmap := T.t (option signedness * num_val_ty).

  Instance gamma_sm : gamma_op signednessmap (var -> num_val) :=
    fun ms ρ =>
      ∀ v s, T.get v ms = Some s ->
             match snd s with NVTfloat | NVTsingle => False | _ => True end /\
             ρ v ∈ γ (snd s).

  Definition signed_itv :=
    ZIntervals.ITV Int.min_signed Int.max_signed.

  Lemma signed_itv_correct:
    ∀ x, Int.signed x ∈ γ signed_itv.
  Proof.
    simpl. apply Int.signed_range.
  Qed.

  Definition unsigned_itv :=
    ZIntervals.ITV 0 Int.max_unsigned.

  Lemma unsigned_itv_correct:
    ∀ x, Int.unsigned x ∈ γ unsigned_itv.
  Proof.
    intros. simpl. apply Int.unsigned_range_2.
  Qed.

  Definition signed64_itv :=
    ZIntervals.ITV Int64.min_signed Int64.max_signed.

  Lemma signed64_itv_correct:
    ∀ x, Int64.signed x ∈ γ signed64_itv.
  Proof.
    simpl. apply Int64.signed_range.
  Qed.

  Definition unsigned64_itv :=
    ZIntervals.ITV 0 Int64.max_unsigned.

  Lemma unsigned64_itv_correct:
    ∀ x, Int64.unsigned x ∈ γ unsigned64_itv.
  Proof.
    intros. simpl. apply Int64.unsigned_range_2.
  Qed.

  Fixpoint shiftr_nat_pos (n:positive) (k:nat) (racc:bool) {struct k} : N * bool :=
    match k with
      | O => (Npos n, racc)%N
      | S k' =>
        match n with
          | xH => (0, true)%N
          | xI n' => shiftr_nat_pos n' k' true
          | xO n' => shiftr_nat_pos n' k' racc
        end
    end.

  Lemma shiftr_nat_pos_shiftr:
    ∀ k n racc,
      shiftr_nat_pos n k racc =
      (N.shiftr (Npos n) (N.of_nat k),
       racc || negb (N.eqb (N.shiftl (N.shiftr (Npos n) (N.of_nat k)) (N.of_nat k))
                           (Npos n))).
  Proof.
    induction k; simpl; intros.
    - rewrite Pos.eqb_refl, orb_false_r. auto.
    - destruct n; rewrite ?IHk.
      + rewrite Pos2Nat.inj_iter, SuccNat2Pos.id_succ, nat_iter_succ_r; simpl.
        rewrite Nnat.Nat2N.inj_iter. f_equal.
        replace (N.pos (Pos.of_succ_nat k)) with (N.succ (N.of_nat k)) by (zify; omega).
        rewrite N.shiftl_succ_r.
        destruct N.shiftl, racc; auto.
      + rewrite Pos2Nat.inj_iter, SuccNat2Pos.id_succ, nat_iter_succ_r; simpl.
        rewrite Nnat.Nat2N.inj_iter. f_equal.
        replace (N.pos (Pos.of_succ_nat k)) with (N.succ (N.of_nat k)) by (zify; omega).
        rewrite N.shiftl_succ_r. f_equal. f_equal.
        change (N.shiftr (N.pos n) (N.of_nat k)) with (N.iter (N.of_nat k) N.div2 (N.pos n)).
        destruct N.shiftl; auto.
      + rewrite Pos2Nat.inj_iter, SuccNat2Pos.id_succ, nat_iter_succ_r; simpl.
        replace (nat_iter k N.div2 0%N) with 0%N.
        rewrite N.shiftl_0_l. destruct racc; auto.
        apply nat_iter_invariant. intros; subst; auto. auto.
  Qed.

  Definition shiftr_nat (n:Z) (k:nat) : Z :=
    match n with
      | Z0 => Z0
      | Zpos n' => Z.of_N (fst (shiftr_nat_pos n' k false))
      | Zneg n' => match shiftr_nat_pos n' k false with
                     | (q, false) => - Z.of_N q
                     | (q, true) => - Z.of_N q - 1
                   end
    end.

  Lemma shiftr_nat_shiftr:
    ∀ n k, shiftr_nat n k = Z.shiftr n (Z.of_nat k).
  Proof.
    destruct n; simpl; intro.
    - rewrite Z.shiftr_0_l. auto.
    - rewrite shiftr_nat_pos_shiftr. simpl.
      destruct k. auto.
      simpl. unfold Z.shiftr, Z.shiftl; simpl.
      rewrite !Pos2Nat.inj_iter, SuccNat2Pos.id_succ.
      induction (S k).
      + auto.
      + simpl. rewrite N2Z.inj_div2, IHn. auto.
    - rewrite shiftr_nat_pos_shiftr. simpl.
      rewrite N.shiftl_mul_pow2, N.shiftr_div_pow2, Z.shiftr_div_pow2 by (zify; omega).
      rewrite N2Z.inj_div, N.mul_comm.
      destruct (N.eqb_spec (2 ^ N.of_nat k * (N.pos p / 2 ^ N.of_nat k)) (N.pos p)).
      + symmetry in e. rewrite N.div_exact in e.
        simpl. rewrite N2Z.inj_pow, nat_N_Z, <- Z_div_zero_opp_full. auto.
        change 0 with (Z.of_N 0). rewrite <- e, N2Z.inj_mod, N2Z.inj_pow, nat_N_Z.
        reflexivity.
        apply N.pow_nonzero. discriminate.
        apply N.pow_nonzero. discriminate.
      + apply not_eq_sym in n. rewrite N.div_exact in n.
        simpl. rewrite N2Z.inj_pow, nat_N_Z, <- Z_div_nz_opp_full. auto.
        contradict n. change 0%N with (Z.to_N 0).
        rewrite <- n, Z2N.inj_mod, Z2N.inj_pow; try (zify; omega).
        destruct k; auto. apply Z.pow_pos_nonneg; zify; omega.
        apply N.pow_nonzero. discriminate.
  Qed.

  Definition shiftl_nat (n:Z) (k:nat) : Z :=
    match n with
      | Z0 => Z0
      | Zpos p => Zpos (nat_iter k xO p)
      | Zneg p => Zneg (nat_iter k xO p)
    end.

  Lemma shiftl_nat_shiftl:
    ∀ n k, shiftl_nat n k = Z.shiftl n (Z.of_nat k).
  Proof.
    induction k.
    - destruct n; auto.
    - rewrite Nat2Z.inj_succ. unfold Z.succ.
      rewrite <- Z.shiftl_shiftl by (zify; omega).
      destruct n; simpl in *; rewrite <- IHk; auto.
  Qed.

  Definition itv_normalizer (i:ZIntervals.itv+⊤+⊥) (min max:Z) (logdiv:nat): option Z :=
    match i with
      | NotBot (Just (ZIntervals.ITV a b)) =>
        if Z.eqb (shiftr_nat (b-a) logdiv) 0 then
          let q := shiftr_nat (a-min) logdiv in
          if Zle_bool (b-shiftl_nat q logdiv) max
          then Some q
          else None
        else None
      | _ => None
    end.

  Lemma itv_normalizer_correct:
    ∀ i min max logdiv, ∀ q, itv_normalizer i min max logdiv = Some q ->
      ∀ n, n ∈ γ i -> min <= n - q * two_power_nat logdiv <= max.
  Proof.
    intros [|[|[a b]]] min max logdiv q Hq n Hn; try discriminate Hq; unfold γ in *. simpl in *. unfold γ in *. simpl in *.
    rewrite shiftl_nat_shiftl, !shiftr_nat_shiftr,
      Z.shiftl_mul_pow2, !Z.shiftr_div_pow2,
      <- two_p_equiv, <- two_power_nat_two_p in Hq by (zify; omega).
    destruct (Z.eqb_spec ((b-a) / two_power_nat logdiv) 0). 2:discriminate.
    match goal with
      | H:context [Zle_bool ?x ?y] |- _ =>
        assert (HZle:= Zle_cases x y); destruct (Zle_bool x y)
    end; inv Hq.

    assert (Hlt:=two_power_nat_pos logdiv).
    assert (Hmodeq:=Zmod_eq (a-min) _ Hlt).
    assert (Hmodlt:=Z_mod_lt (a-min) _ Hlt).
    destruct Hn. omega.
  Qed.

  Definition normalize_expr (v:chan_abt+⊥) (e:iexpr var) (min max:Z) (logdiv:nat) : option (iexpr var) :=
    match v with
      | NotBot v =>
        match itv_normalizer (do_bot x <- (snd (proj1_sig v)).(AbIdealEnv.get_itv) e;
                              IdealIntervals.in_z x) min max logdiv with
          | Some q =>
            if Z_eq_dec q 0 then Some e
            else
              let shift := shiftl_nat q logdiv in
              Some (IEbinop IOsub e (IEZconst var (Just (ZIntervals.ITV shift shift))))
          | None => None
        end
      | Bot => None
    end.

  Lemma normalize_expr_correct:
    ∀ v e min max, ∀ logdiv,
    ∀ e', normalize_expr v e min max logdiv = Some e' ->
    ∀ ρ x, ρ ∈ γ v -> eval_iexpr ρ e (INz x) ->
         ∃ x', min <= x' <= max ∧ Int.eqmod (two_power_nat logdiv) x x' ∧
               eval_iexpr ρ e' (INz x').
  Proof.
    unfold normalize_expr.
    intros [|v] e min max logdiv e' Heq ρ x Hγ EVAL.
    destruct Hγ.
    assert (Hnorm:=itv_normalizer_correct (do_bot x <- AbIdealEnv.get_itv (snd (proj1_sig v)) e;
             IdealIntervals.in_z x) min max logdiv).
    destruct itv_normalizer; inv Heq.
    eexists. split.
    apply Hnorm. auto.
    eapply botbind_spec; eauto. eapply AbIdealEnv.get_itv_correct; eauto. eapply (proj2_sig v), Hγ.
    split. exists z. ring.
    destruct (Z_eq_dec z 0); inv H0.
    replace (x-0*two_power_nat logdiv) with x by ring. auto.
    eapply eval_IEbinop; eauto.
    eapply eval_IEZconst; simpl.
    unfold γ. simpl. unfold γ. simpl.
    rewrite shiftl_nat_shiftl, Z.shiftl_mul_pow2, <- two_p_equiv, two_power_nat_two_p by (zify; omega).
    omega.
  Qed.

  Definition normalize_expr_full (v:chan_abt+⊥) (e:iexpr var) (f:signedness) : iexpr var :=
    let '(min, max) :=
      match f with Signed => (Int.min_signed, Int.max_signed) | Unsigned => (0, Int.max_unsigned) end
    in
    match normalize_expr v e min max Int.wordsize with
      | Some e => e
      | None => IEZconst var (Just (ZIntervals.ITV min max))
    end.

  Lemma normalize_expr_full_correct:
    ∀ v e f,
    ∀ ρ x, ρ ∈ γ v -> eval_iexpr ρ e (INz x) ->
    ∃ x', x' ∈ γ (match f with Signed => signed_itv | Unsigned => unsigned_itv end) ∧ 
             Int.eqm x x' ∧ eval_iexpr ρ (normalize_expr_full v e f) (INz x').
  Proof.
    intros v e [] ρ x Hρ EVAL; unfold normalize_expr_full.
    - destruct_opt.
      eapply normalize_expr_correct; eauto.
      eexists. split. apply signed_itv_correct. split.
      eapply Int.eqm_trans, Int.eqm_sym, Int.eqm_signed_unsigned. apply Int.eqm_unsigned_repr.
      econstructor. apply signed_itv_correct.
    - destruct_opt.
      eapply normalize_expr_correct; eauto.
      eexists. split. apply unsigned_itv_correct. split.
      apply Int.eqm_unsigned_repr.
      econstructor. apply unsigned_itv_correct.
  Qed.

  Definition normalize_expr_full64 (v:chan_abt+⊥) (e:iexpr var) (f:signedness) : iexpr var :=
    let '(min, max) :=
      match f with Signed => (Int64.min_signed, Int64.max_signed) | Unsigned => (0, Int64.max_unsigned) end
    in
    match normalize_expr v e min max Int64.wordsize with
      | Some e => e
      | None => IEZconst var (Just (ZIntervals.ITV min max))
    end.

  Lemma normalize_expr_full64_correct:
    ∀ v e f,
    ∀ ρ x, ρ ∈ γ v -> eval_iexpr ρ e (INz x) ->
    ∃ x', x' ∈ γ (match f with Signed => signed64_itv | Unsigned => unsigned64_itv end) ∧ 
             Int64.eqm x x' ∧ eval_iexpr ρ (normalize_expr_full64 v e f) (INz x').
  Proof.
    intros v e [] ρ x Hρ EVAL; unfold normalize_expr_full64.
    - destruct_opt.
      eapply normalize_expr_correct; eauto.
      eexists. split. apply signed64_itv_correct. split.
      eapply Int64.eqm_trans, Int64.eqm_sym, Int64.eqm_signed_unsigned. apply Int64.eqm_unsigned_repr.
      econstructor. apply signed64_itv_correct.
    - destruct_opt.
      eapply normalize_expr_correct; eauto.
      eexists. split. apply unsigned64_itv_correct. split.
      apply Int64.eqm_unsigned_repr.
      econstructor. apply unsigned64_itv_correct.
  Qed.

  Definition smallest_zofb_range prec emax Hprec Hmax min :=
    half_backward_zoffloat min prec emax Hprec Hmax.
  Definition largest_zofb_range prec emax Hprec Hmax max :=
    Bopp _ _ (fun a b => (a,b)) (half_backward_zoffloat (-max) prec emax Hprec Hmax).

  Lemma zofb_range_intv:
    ∀ prec emax Hprec Hmax min max f, 1 < prec ->
      is_finite _ _ (smallest_zofb_range _ _ Hprec Hmax min) = true ->
      Fleb (smallest_zofb_range _ _ Hprec Hmax min) f = true ->
      is_finite _ _ (largest_zofb_range _ _ Hprec Hmax max) = true ->
      Fleb f (largest_zofb_range _ _ Hprec Hmax max) = true ->
      Fappli_IEEE_extra.ZofB_range prec emax f min max = None ->
      False.
  Proof.
    intros.
    assert (is_finite _ _ f = true).
    { destruct f as [|[]| |]; try reflexivity; try discriminate.
      destruct (smallest_zofb_range _ _ Hprec Hmax min) as [| | |[]]; discriminate.
      destruct (largest_zofb_range _ _ Hprec Hmax max); discriminate. }
    unfold Fappli_IEEE_extra.ZofB_range in *.
    destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?.
    2:rewrite Fappli_IEEE_extra.ZofB_correct, H5 in Heqo; discriminate.
    assert (∃ m, Fappli_IEEE_extra.ZofB _ _ (smallest_zofb_range _ _ Hprec Hmax min) = Some m).
    { rewrite Fappli_IEEE_extra.ZofB_correct, H0. eauto. }
    destruct H6 as (? & ?).
    eapply zoffloat_incr in H1; eauto.
    eapply half_backward_zoffloat_complete in H6. 2:auto.
    rewrite Zle_imp_le_bool in H4 by omega.
    assert (∃ m, Fappli_IEEE_extra.ZofB _ _ (largest_zofb_range prec emax Hprec Hmax max) = Some m).
    { rewrite Fappli_IEEE_extra.ZofB_correct, H2. eauto. }
    destruct H7 as (? & ?).
    eapply zoffloat_incr in H3; eauto.
    unfold largest_zofb_range in H7.
    rewrite zofb_neg in H7.
    assert (∀ A B (f:A->B) a b, option_map f a = Some b -> ∃ c, a = Some c /\ b = f c).
    { intros. destruct a; inv H8; eauto. }
    apply H8 in H7. clear H8. destruct H7 as (MA & HMA & ?). subst x0.
    eapply half_backward_zoffloat_complete in HMA. 2:auto.
    rewrite Zle_imp_le_bool in H4 by omega.
    discriminate.
  Qed.

  Transparent Float.to_int.
  Definition smallest_intoffloat: float :=
    smallest_zofb_range 53 1024 eq_refl eq_refl Int.min_signed.
  Definition largest_intoffloat: float :=
    largest_zofb_range 53 1024 eq_refl eq_refl Int.max_signed.

  Lemma intoffloat_intv :
    forall f,
      Fleb smallest_intoffloat f = true ->
      Fleb f largest_intoffloat = true ->
      Float.to_int f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float.to_int in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_intoffloat largest_intoffloat.

  Transparent Float.to_intu.
  Definition smallest_intuoffloat: float :=
    smallest_zofb_range 53 1024 eq_refl eq_refl 0.
  Definition largest_intuoffloat: float :=
    largest_zofb_range 53 1024 eq_refl eq_refl Int.max_unsigned.

  Lemma intuoffloat_intv :
    forall f,
      Fleb smallest_intuoffloat f = true ->
      Fleb f largest_intuoffloat = true ->
      Float.to_intu f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float.to_intu in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_intuoffloat largest_intuoffloat.

  Transparent Float.to_long.
  Definition smallest_longoffloat: float :=
    smallest_zofb_range 53 1024 eq_refl eq_refl Int64.min_signed.
  Definition largest_longoffloat: float :=
    largest_zofb_range 53 1024 eq_refl eq_refl Int64.max_signed.

  Lemma longoffloat_intv :
    forall f,
      Fleb smallest_longoffloat f = true ->
      Fleb f largest_longoffloat = true ->
      Float.to_long f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float.to_long in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_longoffloat largest_longoffloat.

  Transparent Float.to_longu.
  Definition smallest_longuoffloat: float :=
    smallest_zofb_range 53 1024 eq_refl eq_refl 0.
  Definition largest_longuoffloat: float :=
    largest_zofb_range 53 1024 eq_refl eq_refl Int64.max_unsigned.

  Lemma longuoffloat_intv :
    forall f,
      Fleb smallest_longuoffloat f = true ->
      Fleb f largest_longuoffloat = true ->
      Float.to_longu f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float.to_longu in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_longuoffloat largest_longuoffloat.

  Transparent Float32.to_int.
  Definition smallest_intofsingle: float32 :=
    smallest_zofb_range 24 128 eq_refl eq_refl Int.min_signed.
  Definition largest_intofsingle: float32 :=
    largest_zofb_range 24 128 eq_refl eq_refl Int.max_signed.

  Lemma intofsingle_intv :
    forall f,
      Fleb smallest_intofsingle f = true ->
      Fleb f largest_intofsingle = true ->
      Float32.to_int f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float32.to_int in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_intofsingle largest_intofsingle.

  Transparent Float32.to_intu.
  Definition smallest_intuofsingle: float32 :=
    smallest_zofb_range 24 128 eq_refl eq_refl 0.
  Definition largest_intuofsingle: float32 :=
    largest_zofb_range 24 128 eq_refl eq_refl Int.max_unsigned.

  Lemma intuofsingle_intv :
    forall f,
      Fleb smallest_intuofsingle f = true ->
      Fleb f largest_intuofsingle = true ->
      Float32.to_intu f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float32.to_intu in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_intuofsingle largest_intuofsingle.

  Transparent Float32.to_long.
  Definition smallest_longofsingle: float32 :=
    smallest_zofb_range 24 128 eq_refl eq_refl Int64.min_signed.
  Definition largest_longofsingle: float32 :=
    largest_zofb_range 24 128 eq_refl eq_refl Int64.max_signed.

  Lemma longofsingle_intv :
    forall f,
      Fleb smallest_longofsingle f = true ->
      Fleb f largest_longofsingle = true ->
      Float32.to_long f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float32.to_long in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_longofsingle largest_longofsingle.

  Transparent Float32.to_longu.
  Definition smallest_longuofsingle: float32 :=
    smallest_zofb_range 24 128 eq_refl eq_refl 0.
  Definition largest_longuofsingle: float32 :=
    largest_zofb_range 24 128 eq_refl eq_refl Int64.max_unsigned.

  Lemma longuofsingle_intv :
    forall f,
      Fleb smallest_longuofsingle f = true ->
      Fleb f largest_longuofsingle = true ->
      Float32.to_longu f = None ->
      False.
  Proof.
    intros. eapply zofb_range_intv; eauto 2; try reflexivity. unfold Float32.to_longu in *.
    destruct Fappli_IEEE_extra.ZofB_range. discriminate. auto.
  Qed.
  Opaque smallest_longuofsingle largest_longuofsingle.

  Program Definition assume_cmp (c:comparison) (e:iexpr var) (z:Z) (v:chan_abt+⊥) : chan_abt+⊥ :=
    do_bot v <- v;
    let v := proj1_sig v in
    lift_bot_chan_abt
      (AbIdealEnv.assume (IEbinop (IOcmp c) e (IEZconst var (Just (ZIntervals.ITV z z))))
        true v in_chan_top).
  Next Obligation. apply assume_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma assume_cmp_correct:
    ∀ c e x z ρ ab,
      ρ ∈ γ ab ->
      INz x ∈ eval_iexpr ρ e ->
      ZIntervals.Zcmp c x z = true ->
      ρ ∈ γ (assume_cmp c e z ab).
  Proof.
    intros. apply botbind_spec with ρ; auto.
    intros. apply lift_bot_chan_abt_correct.
    eapply AbIdealEnv.assume_correct. auto. apply in_chan_top_correct.
    econstructor. eauto. constructor. split; apply Zle_refl.
    replace 1 with (if ZIntervals.Zcmp c x z then 1 else 0). constructor.
    rewrite H1. auto.
  Qed.

  Definition assume_cmpf (c:comparison) (e:iexpr var) (f:float) (b:bool) (v:chan_abt+⊥) : (abt * in_chan)+⊥ :=
    let itv := Just (FloatIntervals.ITV f f) in
    do_bot v <- v;
    let v := proj1_sig v in
    AbIdealEnv.assume (IEbinop (IOcmpf c) e (IEFconst _ itv)) b v in_chan_top.

  Lemma assume_cmpf_correct:
    ∀ c e x f ρ b ab,
      ρ ∈ γ ab ->
      INf x ∈ eval_iexpr ρ e ->
      is_nan _ _ f = false ->
      Float.cmp c x f = b ->
      ρ ∈ γ (assume_cmpf c e f b ab).
  Proof.
    intros. apply botbind_spec with ρ; auto.
    intros. eapply AbIdealEnv.assume_correct. auto. apply in_chan_top_correct.
    econstructor. eauto. constructor. split; apply Fleb_refl; auto.
    subst. econstructor.
  Qed.

  Program Definition join_chan_abt_bot (ab1 ab2:chan_abt+⊥) : chan_abt+⊥ :=
    match ab1, ab2 with
      | Bot, ab | ab, Bot => ab
      | NotBot ab1, NotBot ab2 =>
        NotBot (id_join (proj1_sig ab1) (proj1_sig ab2) in_chan_top)
    end.
  Next Obligation. apply id_join_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma join_chan_abt_bot_correct :
    ∀ ab1 ab2 (ρ:var -> ideal_num), ρ ∈ γ ab1 \/ ρ ∈ γ ab2 -> ρ ∈ γ (join_chan_abt_bot ab1 ab2).
  Proof.
    intros [|ab1] [|ab2] ρ H; simpl; try now intuition.
    destruct H; apply id_join_correct; auto using in_chan_top_correct.
  Qed.

  Fixpoint convert_expr (v:chan_abt+⊥) (e:nexpr var): (iexpr var * chan_abt+⊥ * bool) :=
    match e with
      | NEvar (NVTint|NVTint64) x => (IEvar INTz x, v, true)
      | NEvar (NVTfloat | NVTsingle) x => (IEvar INTf x, v, true)
      | NEconst (NVIint itv) => (IEZconst var itv, v, true)
      | NEconst (NVIint64 itv) => (IEZconst var itv, v, true)
      | NEconst (NVIfloat itv) => (IEFconst var itv, v, true)
      | NEconst (NVIsingle itv) => (IEFconst var itv, v, true)

      | NEunop op e =>
        let '(e, v, nb) := convert_expr v e in
        match op with
          | Ocast8unsigned =>
            match normalize_expr v e 0 255 8 with
              | Some e => (e, v, nb)
              | None => (IEZconst var (Just (ZIntervals.ITV 0 255)), v, nb)
            end
          | Ocast8signed =>
            match normalize_expr v e (-128) 127 8 with
              | Some e => (e, v, nb)
              | None => (IEZconst var (Just (ZIntervals.ITV (-128) 127)), v, nb)
            end
          | Ocast16unsigned =>
            match normalize_expr v e 0 65535 16 with
              | Some e => (e, v, nb)
              | None => (IEZconst var (Just (ZIntervals.ITV 0 65535)), v, nb)
            end
          | Ocast16signed =>
            match normalize_expr v e (-32768) 32767 16 with
              | Some e => (e, v, nb)
              | None => (IEZconst var (Just (ZIntervals.ITV (-32768) 32767)), v, nb)
            end
          | Olongofint => (normalize_expr_full v e Signed, v, nb)
          | Olongofintu => (normalize_expr_full v e Unsigned, v, nb)
          | Ointoflong => (e, v, nb)

          | Onegint => (IEunop IOneg e, v, nb)
          | Onotint => (IEunop IOnot e, v, nb)

          | Onegl => (IEunop IOneg e, v, nb)
          | Onotl => (IEunop IOnot e, v, nb)

          | Onegf | Onegfs => (IEunop IOnegf e, v, nb)
          | Oabsf | Oabsfs => (IEunop IOabsf e, v, nb)
          | Ofloatofsingle => (e, v, nb)
          | Osingleoffloat => (IEunop IOsingleoffloat e, v, nb)
          | Ointoffloat =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e smallest_intoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_intoffloat false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Ointofsingle =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_intofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_intofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Ointuoffloat =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e smallest_intuoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_intuoffloat false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Ointuofsingle =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_intuofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_intuofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Ofloatofint =>
            (IEunop IOfloatofz (normalize_expr_full v e Signed), v, nb)
          | Osingleofint =>
            (IEunop IOsingleofz (normalize_expr_full v e Signed), v, nb)
          | Ofloatofintu =>
            (IEunop IOfloatofz (normalize_expr_full v e Unsigned), v, nb)
          | Osingleofintu =>
            (IEunop IOsingleofz (normalize_expr_full v e Unsigned), v, nb)
          | Olongoffloat =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e smallest_longoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_longoffloat false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Olongofsingle =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_longofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_longofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Olonguoffloat =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e smallest_longuoffloat false v) &&
              is_bot (assume_cmpf Cle e largest_longuoffloat false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Olonguofsingle =>
            let nb := nb &&
              is_bot (assume_cmpf Cge e (Float.of_single smallest_longuofsingle) false v) &&
              is_bot (assume_cmpf Cle e (Float.of_single largest_longuofsingle) false v)
            in
            (IEunop IOzoffloat e, v, nb)
          | Ofloatoflong =>
            (IEunop IOfloatofz (normalize_expr_full64 v e Signed), v, nb)
          | Ofloatoflongu =>
            (IEunop IOfloatofz (normalize_expr_full64 v e Unsigned), v, nb)
          | Osingleoflong =>
            (IEunop IOsingleofz (normalize_expr_full64 v e Signed), v, nb)
          | Osingleoflongu =>
            (IEunop IOsingleofz (normalize_expr_full64 v e Unsigned), v, nb)
        end

      | NEbinop op e1 e2 =>
        let '(e1, v, nb1) := convert_expr v e1 in
        let '(e2, v, nb2) := convert_expr v e2 in
        match op with
          | Ocmp c =>
            let e1 := normalize_expr_full v e1 Signed in
            let e2 := normalize_expr_full v e2 Signed in
            (IEbinop (IOcmp c) e1 e2, v, (nb1 && nb2)%bool)
          | Ocmpu c =>
            let e1 := normalize_expr_full v e1 Unsigned in
            let e2 := normalize_expr_full v e2 Unsigned in
            (IEbinop (IOcmp c) e1 e2, v, (nb1 && nb2)%bool)
          | Ocmpl c =>
            let e1 := normalize_expr_full64 v e1 Signed in
            let e2 := normalize_expr_full64 v e2 Signed in
            (IEbinop (IOcmp c) e1 e2, v, (nb1 && nb2)%bool)
          | Ocmplu c =>
            let e1 := normalize_expr_full64 v e1 Unsigned in
            let e2 := normalize_expr_full64 v e2 Unsigned in
            (IEbinop (IOcmp c) e1 e2, v, (nb1 && nb2)%bool)

          | Oaddf => (IEbinop IOaddf e1 e2, v, (nb1 && nb2)%bool)
          | Oaddfs => (IEunop IOsingleoffloat (IEbinop IOaddf e1 e2), v, (nb1 && nb2)%bool)
          | Osubf => (IEbinop IOsubf e1 e2, v, (nb1 && nb2)%bool)
          | Osubfs => (IEunop IOsingleoffloat (IEbinop IOsubf e1 e2), v, (nb1 && nb2)%bool)
          | Omulf => (IEbinop IOmulf e1 e2, v, (nb1 && nb2)%bool)
          | Omulfs => (IEunop IOsingleoffloat (IEbinop IOmulf e1 e2), v, (nb1 && nb2)%bool)
          | Odivf => (IEbinop IOdivf e1 e2, v, (nb1 && nb2)%bool)
          | Odivfs => (IEunop IOsingleoffloat (IEbinop IOdivf e1 e2), v, (nb1 && nb2)%bool)
          | Ocmpf c | Ocmpfs c => (IEbinop (IOcmpf c) e1 e2, v, (nb1 && nb2)%bool)
          | Oadd | Oaddl => (IEbinop IOadd e1 e2, v, (nb1 && nb2)%bool)
          | Osub | Osubl => (IEbinop IOsub e1 e2, v, (nb1 && nb2)%bool)
          | Omul =>
            let e1 :=
                match normalize_expr v e1 Int.min_signed Int.max_unsigned Int.wordsize with
                  | Some e => e
                  | None => e1
                end
            in
            let e2 :=
                match normalize_expr v e2 Int.min_signed Int.max_unsigned Int.wordsize with
                  | Some e => e
                  | None => e2
                end
            in
            (IEbinop IOmul e1 e2, v, (nb1 && nb2)%bool)
          | Omull =>
            let e1 :=
                match normalize_expr v e1 Int64.min_signed Int64.max_unsigned Int64.wordsize with
                  | Some e => e
                  | None => e1
                end
            in
            let e2 :=
                match normalize_expr v e2 Int64.min_signed Int64.max_unsigned Int64.wordsize with
                  | Some e => e
                  | None => e2
                end
            in
            (IEbinop IOmul e1 e2, v, (nb1 && nb2)%bool)

          | Odiv =>
            let e1 := normalize_expr_full v e1 Signed in
            let e2 := normalize_expr_full v e2 Signed in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOdiv e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := join_chan_abt_bot
                         (assume_cmp Cgt e1 Int.min_signed v)
                         (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOdiv e1 e2, v, false)
          | Odivl =>
            let e1 := normalize_expr_full64 v e1 Signed in
            let e2 := normalize_expr_full64 v e2 Signed in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int64.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOdiv e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := join_chan_abt_bot
                         (assume_cmp Cgt e1 Int64.min_signed v)
                         (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOdiv e1 e2, v, false)
          | Odivu =>
            let e1 := normalize_expr_full v e1 Unsigned in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOdiv e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOdiv e1 e2, v, false)
          | Odivlu =>
            let e1 := normalize_expr_full64 v e1 Unsigned in
            let e2 := normalize_expr_full64 v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOdiv e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOdiv e1 e2, v, false)
          | Omod =>
            let e1 := normalize_expr_full v e1 Signed in
            let e2 := normalize_expr_full v e2 Signed in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOmod e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := join_chan_abt_bot
                         (assume_cmp Cgt e1 Int.min_signed v)
                         (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOmod e1 e2, v, false)
          | Omodl =>
            let e1 := normalize_expr_full64 v e1 Signed in
            let e2 := normalize_expr_full64 v e2 Signed in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) &&
               is_bot (assume_cmp Cle e1 Int64.min_signed (assume_cmp Ceq e2 (-1) v))
            then (IEbinop IOmod e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              let v := join_chan_abt_bot (assume_cmp Cgt e1 Int64.min_signed v)
                                         (assume_cmp Cne e2 (-1) v) in
              (IEbinop IOmod e1 e2, v, false)
          | Omodu =>
            let e1 := normalize_expr_full v e1 Unsigned in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOmod e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOmod e1 e2, v, false)
          | Omodlu =>
            let e1 := normalize_expr_full64 v e1 Unsigned in
            let e2 := normalize_expr_full64 v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Ceq e2 0 v) then
              (IEbinop IOmod e1 e2, v, true)
            else
              let v := assume_cmp Cne e2 0 v in
              (IEbinop IOmod e1 e2, v, false)


          | Oand | Oandl => (IEbinop IOand e1 e2, v, (nb1 && nb2)%bool)
          | Oor | Oorl => (IEbinop IOor e1 e2, v, (nb1 && nb2)%bool)
          | Oxor | Oxorl => (IEbinop IOxor e1 e2, v, (nb1 && nb2)%bool)
          | Oshl =>
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshl e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshl e1 e2, v, false)
          | Oshll =>
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshl e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshl e1 e2, v, false)
          | Oshr =>
            let e1 := normalize_expr_full v e1 Signed in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshr e1 e2, v, false)
          | Oshrl =>
            let e1 := normalize_expr_full64 v e1 Signed in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshr e1 e2, v, false)
          | Oshru =>
            let e1 := normalize_expr_full v e1 Unsigned in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int.zwordsize v in
              (IEbinop IOshr e1 e2, v, false)
          | Oshrlu =>
            let e1 := normalize_expr_full64 v e1 Unsigned in
            let e2 := normalize_expr_full v e2 Unsigned in
            if nb1 && nb2 && is_bot (assume_cmp Cge e2 Int64.zwordsize v) then
              (IEbinop IOshr e1 e2, v, true)
            else
              let v := assume_cmp Clt e2 Int64.zwordsize v in
              (IEbinop IOshr e1 e2, v, false)
        end
    end.

  Lemma of_bool_itv:
    ∀ ρ b,
    ∃x' : ideal_num,
      eval_iexpr ρ (IEZconst var (Just (ZIntervals.ITV 0 1))) x' ∧ related (of_bool b) x'.
  Proof.
    destruct b.
    exists (INz 1); split. constructor; unfold γ; simpl; unfold γ; simpl; omega. constructor.
    exists (INz 0); split. constructor; unfold γ; simpl; unfold γ; simpl; omega. constructor.
  Qed.

  Lemma convert_expr_ok:
    ∀ e ρ x, eval_nexpr ρ e x ->
    ∀ v ρ', ρ' ∈ γ v -> (∀ id, related (ρ id) (ρ' id)) ->
    ∀ e' v' nb', convert_expr v e = (e', v', nb') ->
      ρ' ∈ γ v' /\
      ∃ x', eval_iexpr ρ' e' x' ∧ related x x'.
  Proof.
  Hint Resolve lift_bot.
    induction 1; simpl convert_expr; intros ab ρ' Hρ' Hρρ' e' v' nb EQ; inv EQ.

    - specialize (Hρρ' id). inv Hρρ'; inv H0; inv H2; try congruence;
      (split; [auto|repeat econstructor; instantiate; rewrite <- H; constructor]).

    - destruct H; inv H1; repeat econstructor; auto.

    - specialize (IHeval_nexpr _ _ Hρ' Hρρ').
      destruct (convert_expr ab a1) as [[e2 ab2] nb2].
      edestruct IHeval_nexpr as [Hρ'0 [x'e [EVAL ?]]]; auto. clear IHeval_nexpr.
      destruct H0; inv H1;
      repeat (match goal with
        | H : context [normalize_expr ?ab ?e ?min ?max ?logmod],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          let NORM := fresh "NORM" in
          assert (NORM := fun e' EQ =>
                    normalize_expr_correct ab e min max logmod e' EQ ρ' x AB EVAL);
          destruct (normalize_expr ab e min max logmod);
          [specialize (NORM _ eq_refl); destruct NORM as (? & ? & ? & ?)|
           clear NORM]
        | H : context [normalize_expr_full ?ab ?e ?sgn],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          destruct (normalize_expr_full_correct ab e sgn ρ' x AB EVAL) as (? & ? & ? & ?);
          clear EVAL
        | H : context [normalize_expr_full64 ?ab ?e ?sgn],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          destruct (normalize_expr_full64_correct ab e sgn ρ' x AB EVAL) as (? & ? & ? & ?);
          clear EVAL
      end);
      inv H2; (split; [now auto | eexists]).

      + split. eauto.
        replace (Int.zero_ext 8 (Int.repr x)) with (Int.repr x0). constructor.
        rewrite <- Int.repr_unsigned with (Int.zero_ext 8 (Int.repr x)). f_equal.
        assert (8 < Int.zwordsize) by reflexivity.
        apply Int.eqmod_small_eq with 256. 2:omega. 2:apply Int.zero_ext_range; omega.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H1.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_zero_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
      + split. 2:apply related_unsigned. constructor.
        destruct (Int.zero_ext_range 8 (Int.repr x)). compute; split; congruence.
        change (two_p 8) with 256 in H1. split; omega.

      + split. eauto.
        replace (Int.sign_ext 8 (Int.repr x)) with (Int.repr x0). constructor.
        rewrite <- Int.repr_signed with (Int.sign_ext 8 (Int.repr x)). f_equal.
        assert (8 < Int.zwordsize) by reflexivity.
        apply Z.add_cancel_r with (p:=128).
        apply Int.eqmod_small_eq with 256. 2:omega.
        apply Int.eqmod_add. 2:apply Int.eqmod_refl.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H1.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_sign_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
        destruct (Int.sign_ext_range 8 (Int.repr x)). omega.
        change (two_p (8-1)) with 128 in H4, H5. omega.
      + split. 2:apply related_signed. constructor.
        destruct (Int.sign_ext_range 8 (Int.repr x)). compute; split; congruence.
        change (two_p (8-1)) with 128 in H0, H1. split; omega.

      + split. eauto.
        replace (Int.zero_ext 16 (Int.repr x)) with (Int.repr x0). constructor.
        rewrite <- Int.repr_unsigned with (Int.zero_ext 16 (Int.repr x)). f_equal.
        assert (16 < Int.zwordsize) by reflexivity.
        apply Int.eqmod_small_eq with 65536. 2:omega. 2:apply Int.zero_ext_range; omega.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H1.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_zero_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
      + split. 2:apply related_unsigned. constructor.
        destruct (Int.zero_ext_range 16 (Int.repr x)). compute; split; congruence.
        change (two_p 16) with 65536 in H1. split; omega.

      + split. eauto.
        replace (Int.sign_ext 16 (Int.repr x)) with (Int.repr x0). constructor.
        rewrite <- Int.repr_signed with (Int.sign_ext 16 (Int.repr x)). f_equal.
        assert (16 < Int.zwordsize) by reflexivity.
        apply Z.add_cancel_r with (p:=32768).
        apply Int.eqmod_small_eq with 65536. 2:omega.
        apply Int.eqmod_add. 2:apply Int.eqmod_refl.
        eapply Int.eqmod_trans. eapply Int.eqmod_sym, H1.
        eapply Int.eqmod_trans, Int.eqmod_sym, Int.eqmod_sign_ext. 2:omega.
        eapply Int.eqmod_divides. eapply Int.eqm_unsigned_repr.
        apply Z.mod_divide. discriminate. reflexivity.
        destruct (Int.sign_ext_range 16 (Int.repr x)). omega.
        change (two_p (16-1)) with 32768 in H4, H5. omega.
      + split. 2:apply related_signed. constructor.
        destruct (Int.sign_ext_range 16 (Int.repr x)). compute; split; congruence.
        change (two_p (16-1)) with 32768 in H0, H1. split; omega.

      + split; eauto. rewrite Int.neg_repr. constructor.

      + split; eauto.
        replace (Int.not (Int.repr x)) with (Int.repr (Z.lnot x)). constructor.
        apply Int.same_bits_eq. intros i Hi.
        rewrite Int.bits_not, !Int.testbit_repr, Z.lnot_spec; intuition.

      + split; econstructor; eauto.

      + split; econstructor; eauto.
        Transparent Float.neg Float32.neg Float.of_single.
        replace (Float.of_single (Float32.neg f))
        with (Float.neg (Float.of_single f))
          by (apply Bconv_Bopp; auto).
        constructor.

      + split; econstructor; eauto.

      + split; econstructor; eauto.
        Transparent Float.abs Float32.abs Float.of_single.
        replace (Float.of_single (Float32.abs f))
        with (Float.abs (Float.of_single f))
          by (apply Bconv_Babs; auto).
        constructor.

      + split; econstructor; eauto.

      + split. eauto. econstructor.

      + instantiate (1:=INz (Int.signed i)).
        split. 2:apply related_signed. econstructor. eauto. constructor.
        unfold Float.to_int, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int.min_signed <=? z) && (z <=? Int.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.signed_repr by auto. auto.

      + instantiate (1:=INz (Int.signed i)).
        split. 2:apply related_signed. econstructor. eauto. constructor.
        unfold Float32.to_int, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int.min_signed <=? z) && (z <=? Int.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.signed_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + instantiate (1:=INz (Int.unsigned i)).
        split. 2:apply related_unsigned. econstructor. eauto. constructor.
        unfold Float.to_intu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.unsigned_repr by auto. auto.

      + instantiate (1:=INz (Int.unsigned i)).
        split. 2:apply related_unsigned. econstructor. eauto. constructor.
        unfold Float32.to_intu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int.unsigned_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float.of_int.
        split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float.of_int.
        rewrite Int.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_int.
        split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float32.of_int.
        rewrite Int.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float.of_intu.
        split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float.of_intu.
        rewrite Int.unsigned_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_intu.
        split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float32.of_intu.
        rewrite Int.unsigned_repr by auto.
        econstructor. eauto. constructor.

      + split; eauto. rewrite Int64.neg_repr. constructor.

      + split; eauto.
        replace (Int64.not (Int64.repr x)) with (Int64.repr (Z.lnot x)). constructor.
        apply Int64.same_bits_eq. intros i Hi.
        rewrite Int64.bits_not, !Int64.testbit_repr, Z.lnot_spec; intuition.

      + split; eauto.
        replace (Int64.loword (Int64.repr x)) with (Int.repr x). constructor.
        eapply Int.eqm_samerepr, Int.eqmod_divides. apply Int64.eqm_unsigned_repr.
        exists Int.modulus. reflexivity.

      + split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        rewrite Int.signed_repr by auto.
        auto.

      + split; [|constructor].
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        rewrite Int.unsigned_repr by auto.
        auto.

      + instantiate (1:=INz (Int64.signed i)).
        split. 2:apply related_signed_long. econstructor. eauto. constructor.
        unfold Float.to_long, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int64.min_signed <=? z) && (z <=? Int64.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.signed_repr by auto. auto.

      + instantiate (1:=INz (Int64.unsigned i)).
        split. 2:apply related_unsigned_long. econstructor. eauto. constructor.
        unfold Float.to_longu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int64.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.unsigned_repr by auto. auto.

      + instantiate (1:=INz (Int64.signed i)).
        split. 2:apply related_signed_long. econstructor. eauto. constructor.
        unfold Float32.to_long, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((Int64.min_signed <=? z) && (z <=? Int64.max_signed))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.signed_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + instantiate (1:=INz (Int64.unsigned i)).
        split. 2:apply related_unsigned_long. econstructor. eauto. constructor.
        unfold Float32.to_longu, Fappli_IEEE_extra.ZofB_range in H0.
        destruct (Fappli_IEEE_extra.ZofB _ _ f) eqn:?; [|discriminate].
        destruct ((0 <=? z) && (z <=? Int64.max_unsigned))%bool eqn:?; inv H0.
        rewrite Bool.andb_true_iff, <- !Zle_is_le_bool in Heqb.
        rewrite Int64.unsigned_repr by auto.
        rewrite Fappli_IEEE_extra.ZofB_correct in *.
        destruct (is_finite _ _ f) eqn:?.
        rewrite (proj2 (floatofsingle_exact _)), (proj1 (floatofsingle_exact _)); auto.
        discriminate.

      + Transparent Float.of_long.
        split; [|constructor].
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float.of_long.
        rewrite Int64.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_long.
        split; [|constructor].
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float32.of_long.
        rewrite Int64.signed_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float.of_longu.
        split; [|constructor].
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float.of_longu.
        rewrite Int64.unsigned_repr by auto.
        econstructor. eauto. constructor.

      + Transparent Float32.of_longu.
        split; [|constructor].
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x0) by trivial.
        unfold Float32.of_longu.
        rewrite Int64.unsigned_repr by auto.
        econstructor. eauto. constructor.

    - specialize (IHeval_nexpr1 _ _ Hρ' Hρρ'). destruct (convert_expr ab a1) as [[e1' ab'] nb1].
      edestruct IHeval_nexpr1 as [Hρ'0 [x'e1 [EVAL ?]]]; eauto. clear IHeval_nexpr1.
      specialize (IHeval_nexpr2 _ _ Hρ'0 Hρρ'). destruct (convert_expr ab' a2) as [[e2' ab''] nb2].
      edestruct IHeval_nexpr2 as [Hρ'1 [x'e2 [EVAL2 ?]]]; eauto; clear IHeval_nexpr2.
      destruct H1; inv H2; inv H4;
      repeat (match goal with
        | H : context [normalize_expr ?ab ?e ?min ?max ?logmod],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          let NORM := fresh "NORM" in
          assert (NORM := fun e' EQ =>
                    normalize_expr_correct ab e min max logmod e' EQ ρ' x AB EVAL);
          destruct (normalize_expr ab e min max logmod);
          [specialize (NORM _ eq_refl); destruct NORM as (? & ? & ? & ?)|
           clear NORM]
        | H : context [normalize_expr_full ?ab ?e ?sgn],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          destruct (normalize_expr_full_correct ab e sgn ρ' x AB EVAL) as (? & ? & ? & ?);
          clear EVAL
        | H : context [normalize_expr_full64 ?ab ?e ?sgn],
          EVAL : eval_iexpr ?ρ' ?e (INz ?x),
          AB : γ ?ab ?ρ' |- _ =>
          destruct (normalize_expr_full64_correct ab e sgn ρ' x AB EVAL) as (? & ? & ? & ?);
          clear EVAL
      end);
      (try (inv H3; split; [now auto | eexists])).

      + split. econstructor; eauto.
        rewrite Int.add_unsigned.
        erewrite Int.eqm_samerepr. constructor. apply Int.eqm_add; eauto with ints.

      + split. econstructor; eauto.
        unfold Int.sub.
        erewrite Int.eqm_samerepr. constructor. apply Int.eqm_sub; eauto with ints.

      + split. econstructor; eauto.
        unfold Int.mul. erewrite Int.eqm_samerepr. constructor. apply Int.eqm_mult; eauto with ints.
      + split. econstructor; eauto.
        unfold Int.mul. erewrite Int.eqm_samerepr. constructor. apply Int.eqm_mult; eauto with ints.
      + split. econstructor; eauto.
        unfold Int.mul. erewrite Int.eqm_samerepr. constructor. apply Int.eqm_mult; eauto with ints.
      + split. econstructor; eauto.
        unfold Int.mul. erewrite Int.eqm_samerepr. constructor. apply Int.eqm_mult; eauto with ints.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int.divs. rewrite !Int.signed_repr; eauto.
        split.
        eapply join_chan_abt_bot_correct.
        unfold γ in *. simpl in *.
        { destruct H5; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto; simpl.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int.eq_spec (Int.repr x) (Int.repr Int.min_signed)).
            rewrite H3 in Heqspec.
            erewrite Int.eqm_samerepr in Heqspec by eauto.
            assert (x2 ≠ Int.min_signed) by congruence.
            destruct (Z.compare_spec x2 Int.min_signed); simpl in *; try congruence. omega.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.mone).
            rewrite H3 in Heqspec.
            erewrite Int.eqm_samerepr in Heqspec by eauto.
            assert (x1 ≠ -1) by (unfold Int.mone in Heqspec; congruence).
            destruct (Z.compare_spec x1 (-1)); simpl in *; congruence. }
        eexists; split; econstructor; eauto.
        erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int.signed_repr by auto.
        constructor. auto.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int.divu. rewrite !Int.unsigned_repr; eauto.
        unfold γ in *.
        simpl in *. rewrite Int.Zquot_Zdiv by omega. destruct zlt. omega. constructor.
        split.
        eapply assume_cmp_correct; eauto.
        simpl. destruct (Z.compare_spec x1 0); try congruence; omega.
        eexists; split; econstructor; eauto.
        erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int.unsigned_repr by auto.
        replace (x2/x1) with (x2÷x1).
        econstructor. auto.
        unfold γ in *.
        simpl in *. rewrite Int.Zquot_Zdiv by omega. destruct zlt; omega.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int.mods. rewrite !Int.signed_repr; eauto.
        split. apply join_chan_abt_bot_correct.
        unfold γ in *; simpl in *.
        { destruct H5; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto; simpl.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int.eq_spec (Int.repr x) (Int.repr Int.min_signed)).
            rewrite H3 in Heqspec.
            erewrite Int.eqm_samerepr in Heqspec by eauto.
            assert (x2 ≠ Int.min_signed) by congruence.
            destruct (Z.compare_spec x2 Int.min_signed); simpl in *; try congruence. omega.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.mone).
            rewrite H3 in Heqspec.
            erewrite Int.eqm_samerepr in Heqspec by eauto.
            assert (x1 ≠ -1) by (unfold Int.mone in Heqspec; congruence).
            destruct (Z.compare_spec x1 (-1)); simpl in *; congruence. }
        eexists; split; econstructor; eauto.
        erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int.signed_repr by auto.
        constructor. auto.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int.eq_spec (Int.repr x0) Int.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int.modu. rewrite !Int.unsigned_repr; eauto.
        unfold γ in *.
        simpl in *. rewrite Z.rem_mod_nonneg by omega. constructor.
        split.
        eapply assume_cmp_correct; eauto.
        simpl. destruct (Z.compare_spec x1 0); try congruence; omega.
        eexists; split; econstructor; eauto.
        erewrite Int.eqm_samerepr with (x:=x), Int.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int.unsigned_repr by auto.
        unfold γ in *.
        simpl in *. rewrite <- Z.rem_mod_nonneg by omega.
        econstructor. auto.

      + split. eauto.
        replace (Int.and (Int.repr x) (Int.repr x0)) with (Int.repr (Z.land x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_and, !Int.testbit_repr, Z.land_spec by trivial.
        reflexivity.

      + split. eauto.
        replace (Int.or (Int.repr x) (Int.repr x0)) with (Int.repr (Z.lor x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_or, !Int.testbit_repr, Z.lor_spec by trivial.
        reflexivity.

      + split. eauto.
        replace (Int.xor (Int.repr x) (Int.repr x0)) with (Int.repr (Z.lxor x x0)).
        constructor.
        apply Int.same_bits_eq. intros.
        rewrite Int.bits_xor, !Int.testbit_repr, Z.lxor_spec by trivial.
        reflexivity.

      + unfold γ in *. simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshl e1' (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint (Int.shl (Int.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          replace (Int.shl (Int.repr x) (Int.repr x1)) with (Int.repr (Z.shiftl x x1)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shl by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          rewrite Z.shiftl_spec by omega.
          apply Z.testbit_neg_r.
          omega.
          apply Z.shiftl_spec. omega. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int.zwordsize); try omega; auto.

      + unfold γ in *; simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshr (normalize_expr_full ab'' e1' Signed)
                                 (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint (Int.shr (Int.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int.shr (Int.repr x2) (Int.repr x1)) with (Int.repr (Z.shiftr x2 x1)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shr by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          rewrite <- Int.signed_repr with x2 by auto.
          rewrite !Int.bits_signed by omega.
          destruct Coqlib.zlt. omega.
          destruct Coqlib.zlt; reflexivity. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int.zwordsize); try omega; auto.

      + unfold γ in *; simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshr (normalize_expr_full ab'' e1' Unsigned)
                                 (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint (Int.shru (Int.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int.shru (Int.repr x2) (Int.repr x1)) with (Int.repr (Z.shiftr x2 x1)).
          constructor.
          apply Int.same_bits_eq. intros.
          rewrite Int.bits_shru by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          apply Z.testbit_false. omega.
          rewrite Z.div_small. reflexivity. split.
          simpl in *; omega.
          eapply Z.lt_le_trans with (2^Int.zwordsize).
          change (0 <= x2 <= 2 ^ Int.zwordsize - 1) in H6. omega.
          apply Z.pow_le_mono_r; omega. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int.iwordsize) with Int.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int.zwordsize); try omega; auto.

      + split; econstructor; eauto.

      + split; econstructor; eauto.

      + split; econstructor; eauto.

      + split; econstructor; eauto.

      + split; econstructor; eauto.
        rewrite plus_double_round. eauto.

      + split; econstructor; eauto.
        rewrite sub_double_round. eauto.

      + split; econstructor; eauto.
        rewrite mult_double_round. eauto.

      + split; econstructor; eauto.
        rewrite div_double_round. eauto.

      + split. econstructor; eauto.
        rewrite Int64.add_unsigned.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_add; eauto with ints.

      + split. econstructor; eauto.
        unfold Int64.sub.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_sub; eauto with ints.

      + split. econstructor; eauto. unfold Int64.mul.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_mult; eauto with ints.
      + split. econstructor; eauto. unfold Int64.mul.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_mult; eauto with ints.
      + split. econstructor; eauto. unfold Int64.mul.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_mult; eauto with ints.
      + split. econstructor; eauto. unfold Int64.mul.
        erewrite Int64.eqm_samerepr. constructor. apply Int64.eqm_mult; eauto with ints.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int64.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int64.divs. rewrite !Int64.signed_repr; eauto.
        split. apply join_chan_abt_bot_correct.
        unfold γ in *. simpl in *.
        { destruct H5; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto; simpl.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int64.eq_spec (Int64.repr x) (Int64.repr Int64.min_signed)).
            rewrite H3 in Heqspec.
            erewrite Int64.eqm_samerepr in Heqspec by eauto.
            assert (x2 ≠ Int64.min_signed) by congruence.
            destruct (Z.compare_spec x2 Int64.min_signed); simpl in *; try congruence. omega.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.mone).
            rewrite H3 in Heqspec.
            erewrite Int64.eqm_samerepr in Heqspec by eauto.
            assert (x1 ≠ -1) by (unfold Int64.mone in Heqspec; congruence).
            destruct (Z.compare_spec x1 (-1)); simpl in *; congruence. }
        eexists; split; econstructor; eauto.
        erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int64.signed_repr by auto.
        constructor. auto.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int64.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int64.divu. rewrite !Int64.unsigned_repr; eauto.
        unfold γ in *.
        simpl in *. rewrite Int64.Zquot_Zdiv by omega. destruct zlt. omega. constructor.
        split.
        eapply assume_cmp_correct; eauto.
        simpl. destruct (Z.compare_spec x1 0); try congruence; omega.
        eexists; split; econstructor; eauto.
        erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int64.unsigned_repr by auto.
        replace (x2/x1) with (x2÷x1).
        econstructor. auto.
        unfold γ in *.
        simpl in *. rewrite Int64.Zquot_Zdiv by omega. destruct zlt; omega.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int64.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int64.mods. rewrite !Int64.signed_repr; eauto.
        split. apply join_chan_abt_bot_correct.
        unfold γ in *; simpl in *.
        { destruct H5; [left|right];
            (eapply assume_cmp_correct; [eapply assume_cmp_correct| |]);
            eauto; simpl.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int64.eq_spec (Int64.repr x) (Int64.repr Int64.min_signed)).
            rewrite H3 in Heqspec.
            erewrite Int64.eqm_samerepr in Heqspec by eauto.
            assert (x2 ≠ Int64.min_signed) by congruence.
            destruct (Z.compare_spec x2 Int64.min_signed); simpl in *; try congruence. omega.
          * destruct (Z.compare_spec x1 0); congruence.
          * assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.mone).
            rewrite H3 in Heqspec.
            erewrite Int64.eqm_samerepr in Heqspec by eauto.
            assert (x1 ≠ -1) by (unfold Int64.mone in Heqspec; congruence).
            destruct (Z.compare_spec x1 (-1)); simpl in *; congruence. }
        eexists; split; econstructor; eauto.
        erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int64.signed_repr by auto.
        constructor. auto.

      + assert (x1 ≠ 0).
        { assert (Heqspec:=Int64.eq_spec (Int64.repr x0) Int64.zero).
          rewrite H1 in Heqspec.
          contradict Heqspec. subst x1. apply Int64.eqm_samerepr; trivial. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3.
        split; [now auto|eexists; split].
        econstructor; eauto.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        unfold Int64.modu. rewrite !Int64.unsigned_repr; eauto.
        unfold γ in *.
        simpl in *. rewrite Z.rem_mod_nonneg by omega. constructor.
        split.
        eapply assume_cmp_correct; eauto.
        simpl. destruct (Z.compare_spec x1 0); try congruence; omega.
        eexists; split; econstructor; eauto.
        erewrite Int64.eqm_samerepr with (x:=x), Int64.eqm_samerepr with (x:=x0); eauto.
        rewrite !Int64.unsigned_repr by auto.
        unfold γ in *.
        simpl in *. rewrite <- Z.rem_mod_nonneg by omega.
        econstructor. auto.

      + split. eauto.
        replace (Int64.and (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.land x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_and, !Int64.testbit_repr, Z.land_spec by trivial.
        reflexivity.

      + split. eauto.
        replace (Int64.or (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.lor x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_or, !Int64.testbit_repr, Z.lor_spec by trivial.
        reflexivity.

      + split. eauto.
        replace (Int64.xor (Int64.repr x) (Int64.repr x0)) with (Int64.repr (Z.lxor x x0)).
        constructor.
        apply Int64.same_bits_eq. intros.
        rewrite Int64.bits_xor, !Int64.testbit_repr, Z.lxor_spec by trivial.
        reflexivity.

      + unfold γ in *. simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshl e1' (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint64 (Int64.shl' (Int64.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          replace (Int64.shl' (Int64.repr x) (Int.repr x1)) with (Int64.repr (Z.shiftl x x1)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shl' by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          rewrite Z.shiftl_spec by omega.
          apply Z.testbit_neg_r. omega.
          apply Z.shiftl_spec. omega. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int64.zwordsize); try omega; auto.

      + unfold γ in *; simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshr (normalize_expr_full64 ab'' e1' Signed)
                                 (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint64 (Int64.shr' (Int64.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int64.shr' (Int64.repr x2) (Int.repr x1)) with (Int64.repr (Z.shiftr x2 x1)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shr' by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          rewrite <- Int64.signed_repr with x2 by auto.
          rewrite !Int64.bits_signed by omega.
          destruct Coqlib.zlt. omega.
          destruct Coqlib.zlt; reflexivity. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int64.zwordsize); try omega; auto.

      + unfold γ in *; simpl in *.
        assert (∃ x' : ideal_num,
          eval_iexpr ρ' (IEbinop IOshr (normalize_expr_full64 ab'' e1' Unsigned)
                                 (normalize_expr_full ab'' e2' Unsigned)) x' ∧
          related (NVint64 (Int64.shru' (Int64.repr x) (Int.repr x0))) x').
        { eexists; split.
          econstructor; eauto.
          rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
          rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
          replace (Int64.shru' (Int64.repr x2) (Int.repr x1)) with (Int64.repr (Z.shiftr x2 x1)).
          constructor.
          apply Int64.same_bits_eq. intros.
          rewrite Int64.bits_shru' by trivial.
          rewrite Int.unsigned_repr by trivial.
          destruct Coqlib.zlt; rewrite !Int64.testbit_repr by omega.
          apply Z.shiftr_spec. omega.
          rewrite Z.shiftr_spec by omega.
          apply Z.testbit_false. omega.
          rewrite Z.div_small. reflexivity. split. omega.
          eapply Z.lt_le_trans with (2^Int64.zwordsize).
          change (0 <= x2 <= 2 ^ Int64.zwordsize - 1) in H6. omega.
          apply Z.pow_le_mono_r; omega. }
        match type of H3 with context [if ?c then _ else _] => destruct c end; inv H3; split; auto.
        eapply assume_cmp_correct; eauto. simpl.
        unfold Int.ltu in H1. erewrite Int.eqm_samerepr in H1; eauto. rewrite Int.unsigned_repr in H1; eauto.
        change (Int.unsigned Int64.iwordsize') with Int64.zwordsize in H1.
        destruct Coqlib.zlt in H1; [|discriminate].
        destruct (Z.compare_spec x1 Int64.zwordsize); try omega; auto.

      + instantiate (1:=INz (if Int.cmp c (Int.repr x) (Int.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int.cmp; constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        replace (Int.cmp c (Int.repr x2) (Int.repr x1)) with (ZIntervals.Zcmp c x2 x1).
        constructor.
        unfold Int.cmp, Int.lt.
        rewrite !Int.signed_repr by eauto.
        pose proof (Int.eq_spec (Int.repr x2) (Int.repr x1)).
        destruct c; simpl.
        * destruct Int.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x1)) by congruence.
          rewrite !Int.signed_repr in H9 by trivial. omega.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x1)) by congruence.
          rewrite !Int.signed_repr in H9 by trivial. omega.
          congruence.
        * destruct Int.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x1)) by congruence.
          rewrite !Int.signed_repr in H9 by trivial. omega.
          assert (Int.signed (Int.repr x2) = Int.signed (Int.repr x1)) by congruence.
          rewrite !Int.signed_repr in H9 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.

      + instantiate (1:=INz (if Int.cmpu c (Int.repr x) (Int.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int.cmpu; constructor.
        rewrite !Int.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        replace (Int.cmpu c (Int.repr x2) (Int.repr x1)) with (ZIntervals.Zcmp c x2 x1).
        constructor.
        unfold Int.cmpu, Int.ltu.
        rewrite !Int.unsigned_repr by auto.
        pose proof (Int.eq_spec (Int.repr x2) (Int.repr x1)).
        destruct c; simpl.
        * destruct Int.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x1)) by congruence.
          rewrite !Int.unsigned_repr in H9 by trivial. omega.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x1)) by congruence.
          rewrite !Int.unsigned_repr in H9 by trivial. omega.
          congruence.
        * destruct Int.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x1)) by congruence.
          rewrite !Int.unsigned_repr in H9 by trivial. omega.
          assert (Int.unsigned (Int.repr x2) = Int.unsigned (Int.repr x1)) by congruence.
          rewrite !Int.unsigned_repr in H9 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.

      + instantiate (1:=INz (if Int64.cmp c (Int64.repr x) (Int64.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int64.cmp; constructor.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        replace (Int64.cmp c (Int64.repr x2) (Int64.repr x1)) with (ZIntervals.Zcmp c x2 x1).
        constructor.
        unfold Int64.cmp, Int64.lt.
        rewrite !Int64.signed_repr by eauto.
        pose proof (Int64.eq_spec (Int64.repr x2) (Int64.repr x1)).
        destruct c; simpl.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x1)) by congruence.
          rewrite !Int64.signed_repr in H9 by trivial. omega.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x1)) by congruence.
          rewrite !Int64.signed_repr in H9 by trivial. omega.
          congruence.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x1)) by congruence.
          rewrite !Int64.signed_repr in H9 by trivial. omega.
          assert (Int64.signed (Int64.repr x2) = Int64.signed (Int64.repr x1)) by congruence.
          rewrite !Int64.signed_repr in H9 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.

      + instantiate (1:=INz (if Int64.cmpu c (Int64.repr x) (Int64.repr x0) then 1 else 0)).
        split. econstructor; eauto. 2:destruct Int64.cmpu; constructor.
        rewrite !Int64.eqm_samerepr with (x:=x) (y:=x2) by trivial.
        rewrite !Int64.eqm_samerepr with (x:=x0) (y:=x1) by trivial.
        replace (Int64.cmpu c (Int64.repr x2) (Int64.repr x1)) with (ZIntervals.Zcmp c x2 x1).
        constructor.
        unfold Int64.cmpu, Int64.ltu.
        rewrite !Int64.unsigned_repr by auto.
        pose proof (Int64.eq_spec (Int64.repr x2) (Int64.repr x1)).
        destruct c; simpl.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x1)) by congruence.
          rewrite !Int64.unsigned_repr in H9 by trivial. omega.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x1)) by congruence.
          rewrite !Int64.unsigned_repr in H9 by trivial. omega.
          congruence.
        * destruct Int64.eq; destruct (Z.compare_spec x2 x1); try constructor; exfalso.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x1)) by congruence.
          rewrite !Int64.unsigned_repr in H9 by trivial. omega.
          assert (Int64.unsigned (Int64.repr x2) = Int64.unsigned (Int64.repr x1)) by congruence.
          rewrite !Int64.unsigned_repr in H9 by trivial. omega.
          congruence.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.
        * destruct Coqlib.zlt; destruct (Z.compare_spec x2 x1); try constructor; exfalso; omega.

      + split. econstructor; eauto.
        destruct Float.cmp; constructor.

      + split. econstructor; eauto.
        rewrite Float32.cmp_double. unfold Float32.to_double.
        destruct Float.cmp; constructor.
  Qed.

  Lemma convert_expr_nonblock_id:
    ∀ e e' v v', convert_expr v e = (e', v', true) ->
      v = v'.
  Proof.
    induction e; intros; simpl in H; try congruence.
    - destruct n; congruence.
    - destruct n; congruence.
    - destruct (convert_expr v e) as [[] []] eqn:EQ.
      apply IHe in EQ. subst.
      destruct u; try congruence;
      destruct normalize_expr; congruence.
      destruct u; try discriminate;
      destruct normalize_expr; congruence.
    - destruct (convert_expr v e1) as [[] []] eqn:EQ.
      apply IHe1 in EQ. subst.
      destruct (convert_expr b0 e2) as [[] []] eqn:EQ.
      apply IHe2 in EQ. subst.
      destruct b; try congruence;
      try (destruct normalize_expr_full; congruence);
      try (destruct assume_cmp, assume_cmp; simpl in H; congruence).
      destruct b; discriminate.
      destruct (convert_expr b0 e2) as [[] ?] eqn:EQ'.
      destruct b; discriminate.
  Qed.

  Lemma convert_expr_nonblock:
    ∀ ρ ρ', (∀ id, related (ρ id) (ρ' id)) ->
    ∀ e v, ρ' ∈ γ v ->
    ∀ e' v', convert_expr v e = (e', v', true) ->
      block_nexpr ρ e ->
      False.
  Proof.
    intros ρ ρ' Hρρ'. induction e; intros ab Hρ' e' ab' EQCVT BLOCK; inv BLOCK.
    - simpl in EQCVT.
      destruct (convert_expr ab e) as [[e0 ab0] nb0] eqn:EQCVT'.
      eapply convert_expr_ok in EQCVT'; eauto.
      destruct EQCVT' as [Hρ'0 [x' [EVAL REL]]].
      destruct H2; inv REL; inv EQCVT;
      rewrite !Bool.andb_true_iff in H4; destruct H4 as [[? ?] ?]; subst.
      + eapply intoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_intoffloat f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_intoffloat) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply intuoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_intuoffloat f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_intuoffloat) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply longoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_longoffloat f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_longoffloat) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply longuoffloat_intv; [| |now eauto].
        * destruct (Fleb smallest_longuoffloat f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. apply EQ.
        * destruct (Fleb f largest_longuoffloat) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto.
      + eapply intofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_intofsingle f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_intofsingle) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto. rewrite <- Float32.cmp_double. apply EQ.
      + eapply intuofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_intuofsingle f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_intuofsingle) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto. rewrite <- Float32.cmp_double. apply EQ.
      + eapply longofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_longofsingle f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_longofsingle) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto. rewrite <- Float32.cmp_double. apply EQ.
      + eapply longuofsingle_intv; [| |now eauto].
        * destruct (Fleb smallest_longuofsingle f) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H2. assumption.
          eapply assume_cmpf_correct; eauto.
          rewrite <- Float.cmp_swap. rewrite <- Float32.cmp_double. apply EQ.
        * destruct (Fleb f largest_longuofsingle) eqn:EQ. reflexivity.
          erewrite (is_bot_spec (AD:= WProd.gamma _ _)) in H3. assumption.
          eapply assume_cmpf_correct; eauto. rewrite <- Float32.cmp_double. apply EQ.
    - simpl convert_expr in EQCVT.
      destruct (convert_expr ab e) as [[e0 ab0] nb0] eqn:EQCVT'.
      assert (nb0 = true).
      { destruct u; try congruence;
        try (destruct normalize_expr; congruence);
        injection EQCVT; destruct nb0; simpl; congruence. }
      subst. eapply IHe; eauto.
    - simpl convert_expr in EQCVT.
      destruct (convert_expr ab e1) as [[e01 ab01] nb01] eqn:EQCVT'1.
      eapply convert_expr_ok in EQCVT'1; eauto.
      destruct EQCVT'1 as [Hρ'01 [x'1 [EVAL1 REL1]]].
      destruct (convert_expr ab01 e2) as [[e02 ab02] nb02] eqn:EQCVT'2.
      eapply convert_expr_ok in EQCVT'2; eauto.
      destruct EQCVT'2 as [Hρ'02 [x' [EVAL2 REL2]]].
      destruct H4; inv REL1; inv REL2;
      match type of EQCVT with context [if ?c then _ else _] => destruct c eqn:EQCOND end;
        inv EQCVT; rewrite !Bool.andb_true_iff in EQCOND;
      repeat match type of EQCOND with | _ /\ _ => destruct EQCOND as [EQCOND ?] end; subst.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        edestruct (normalize_expr_full_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite <- Int.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int.eqm_samerepr, Int.signed_repr in H0 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H1. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x) = Int.unsigned (Int.repr (Int.min_signed))) in H0.
        change (Int.unsigned (Int.repr x0) = Int.unsigned (Int.repr (-1))) in H1.
        edestruct (normalize_expr_full_correct ab' e01 Signed) as [x'1 [Hx'1bounded [Hxx'1 EVAL'1]]]; simpl; eauto.
        edestruct (normalize_expr_full_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H0.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H1.
        rewrite !Int.repr_unsigned in H0, H1.
        erewrite Int.eqm_samerepr, Int.signed_repr in H0, H1 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        erewrite Int.eqm_samerepr, Int.unsigned_repr in H0 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        edestruct (normalize_expr_full_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite <- Int.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int.eqm_samerepr, Int.signed_repr in H0 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H1. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x) = Int.unsigned (Int.repr (Int.min_signed))) in H0.
        change (Int.unsigned (Int.repr x0) = Int.unsigned (Int.repr (-1))) in H1.
        edestruct (normalize_expr_full_correct ab' e01 Signed) as [x'1 [Hx'1bounded [Hxx'1 EVAL'1]]]; simpl; eauto.
        edestruct (normalize_expr_full_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H0.
        eapply f_equal with (f:=Int.signed ∘ Int.repr) in H1.
        rewrite !Int.repr_unsigned in H0, H1.
        erewrite Int.eqm_samerepr, Int.signed_repr in H0, H1 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int.unsigned (Int.repr x0) = 0) in H0.
        edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        erewrite Int.eqm_samerepr, Int.unsigned_repr in H0 by eauto.
        subst.
        erewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int.zwordsize); try omega; auto.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int.zwordsize); try omega; auto.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int.zwordsize); try omega; auto.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        edestruct (normalize_expr_full64_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite <- Int64.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H1. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x) = Int64.unsigned (Int64.repr (Int64.min_signed))) in H0.
        change (Int64.unsigned (Int64.repr x0) = Int64.unsigned (Int64.repr (-1))) in H1.
        edestruct (normalize_expr_full64_correct ab' e01 Signed) as [x'1 [Hx'1bounded [Hxx'1 EVAL'1]]]; simpl; eauto.
        edestruct (normalize_expr_full64_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H0.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H1.
        rewrite !Int64.repr_unsigned in H0, H1.
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0, H1 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        edestruct (normalize_expr_full64_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        erewrite Int64.eqm_samerepr, Int64.unsigned_repr in H0 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        edestruct (normalize_expr_full64_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite <- Int64.signed_eq_unsigned in H0 by (rewrite H0; compute; congruence).
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H1. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x) = Int64.unsigned (Int64.repr (Int64.min_signed))) in H0.
        change (Int64.unsigned (Int64.repr x0) = Int64.unsigned (Int64.repr (-1))) in H1.
        edestruct (normalize_expr_full64_correct ab' e01 Signed) as [x'1 [Hx'1bounded [Hxx'1 EVAL'1]]]; simpl; eauto.
        edestruct (normalize_expr_full64_correct ab' e02 Signed) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H0.
        eapply f_equal with (f:=Int64.signed ∘ Int64.repr) in H1.
        rewrite !Int64.repr_unsigned in H0, H1.
        erewrite Int64.eqm_samerepr, Int64.signed_repr in H0, H1 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + change (Int64.unsigned (Int64.repr x0) = 0) in H0.
        edestruct (normalize_expr_full64_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        erewrite Int64.eqm_samerepr, Int64.unsigned_repr in H0 by eauto.
        subst.
        rewrite (is_bot_spec ρ') in H. discriminate.
        eauto using assume_cmp_correct.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int64.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int64.zwordsize); try omega; auto.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int64.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int64.zwordsize); try omega; auto.
      + edestruct (normalize_expr_full_correct ab' e02 Unsigned) as [x'2 [Hx'2bounded [Hxx'2 EVAL'2]]]; simpl; eauto.
        rewrite (is_bot_spec ρ') in H0. discriminate.
        eapply assume_cmp_correct; eauto.
        erewrite Int.eqm_samerepr in H by eauto.
        unfold Int.ltu in H. rewrite Int.unsigned_repr in H by auto. destruct zlt. discriminate.
        change (x'2 >= Int64.zwordsize) in g.
        simpl. destruct (Z.compare_spec x'2 Int64.zwordsize); try omega; auto.

    - simpl convert_expr in EQCVT.
      destruct (convert_expr ab e1) as [[e01 ab01] nb01] eqn:EQCVT'1.
      destruct (convert_expr ab01 e2) as [[e02 ab02] nb02] eqn:EQCVT'2.
      assert (nb01 = true).
      { destruct b, nb01, nb02; simpl in EQCVT; congruence. }
      subst. eapply IHe1; eauto.
    - simpl convert_expr in EQCVT.
      destruct (convert_expr ab e1) as [[e01 ab01] nb01] eqn:EQCVT'1.
      destruct (convert_expr ab01 e2) as [[e02 ab02] nb02] eqn:EQCVT'2.
      assert (nb01 = true).
      { destruct b, nb01, nb02; simpl in EQCVT; congruence. }
      assert (nb02 = true).
      { destruct b, nb01, nb02; simpl in EQCVT; congruence. }
      subst.
      eapply convert_expr_nonblock_id in EQCVT'1.
      subst.
      subst. eapply IHe2; eauto.
  Qed.

  Definition get_itv (e:nexpr var) (v:chan_abt * signednessmap) : num_val_itv+⊥ :=
    let '(ie, v, _) := convert_expr (NotBot (fst v)) e in
    do_bot v <- v;
    do_bot itv <- (snd (proj1_sig v)).(AbIdealEnv.get_itv) ie;
      match itv, nexpr_ty e with
        | Just (IdealIntervals.Az ((ZIntervals.ITV a b) as itv')), NVTint =>
          NotBot (NVIint
            (match itv_normalizer (NotBot (Just itv')) Int.min_signed Int.max_unsigned Int.wordsize with
               | Some q => Just (ZIntervals.ITV (a-Int.modulus*q) (b-Int.modulus*q))
               | None => All
             end))
        | All, NVTint => NotBot (NVIint All)
        | Just (IdealIntervals.Az ((ZIntervals.ITV a b) as itv')), NVTint64 =>
          NotBot (NVIint64
            (match itv_normalizer (NotBot (Just itv')) Int64.min_signed Int64.max_unsigned Int64.wordsize with
               | Some q => Just (ZIntervals.ITV (a-Int64.modulus*q) (b-Int64.modulus*q))
               | None => All
             end))
        | All, NVTint64 => NotBot (NVIint64 All)
        | Just (IdealIntervals.Af itv), NVTfloat => NotBot (NVIfloat (Just itv))
        | Just (IdealIntervals.Af itv), NVTsingle => NotBot (NVIsingle (Just itv))
        | All, NVTfloat => NotBot (NVIfloat All)
        | All, NVTsingle => NotBot (NVIsingle All)
        | _, _ => Bot
      end.

  Lemma get_itv_correct:
    ∀ e ρ ab, ρ ∈ γ ab -> eval_nexpr ρ e ⊆ γ (get_itv e ab).
  Proof.
    intros e ρ [ab sm] [[ρ' [Hγρ' Hρρ']] Hsm] x EVAL. unfold get_itv.
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot ab) _ Hγρ' Hρρ'). unfold fst.
    destruct convert_expr as [[e' ab'] nb']. edestruct H as [Hab' [x' [EVAL' REL]]]; eauto. clear H.
    assert (TY:=nexpr_ty_correct EVAL).
    revert TY. generalize (nexpr_ty e) as ty. intros.
    clear Hγρ' EVAL e ab sm Hsm.
    eapply botbind_spec with ρ'; eauto. intros.
    eapply botbind_spec with x';
      [|eapply AbIdealEnv.get_itv_correct]; eauto. 2:apply (proj2_sig a), H.
    clear EVAL' e' ρ ρ' Hρρ' Hab' H.
    intros. destruct a0 as [|[]], x'; try contradiction; destruct TY; inv REL;
    try (constructor; auto).
    - destruct i.
      destruct itv_normalizer; [|now repeat constructor].
      rewrite Int.eqm_samerepr with (y:=z-Int.modulus * z2). destruct H; repeat constructor; omega.
      exists z2. ring.
    - destruct i.
      destruct itv_normalizer; [|now repeat constructor].
      rewrite Int64.eqm_samerepr with (y:=z-Int64.modulus * z2). destruct H; repeat constructor; omega.
      exists z2. ring.
  Qed.

  Program Definition forget (x:var) (v:chan_abt * signednessmap) : (chan_abt * signednessmap)+⊥ :=
    do_bot fstres <- lift_bot_chan_abt (AbIdealEnv.forget x (proj1_sig (fst v)) in_chan_top);
    NotBot (fstres, T.remove x (snd v)).
  Next Obligation. apply forget_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma forget_correct:
    ∀ x ρ, ∀ n ab,
      ρ ∈ γ ab ->
      (upd ρ x n) ∈ γ (forget x ab).
  Proof.
    intros x ρ n ab Hρ.
    destruct Hρ as [[ρ' [Hρ' Hρρ']] Hsm].
    eapply botbind_spec.
    - intros. split. eauto.
      simpl. intros y s Hx. erewrite T.grspec in Hx; eauto.
      unfold upd. destruct T.elt_eq. discriminate. destruct eq_dec. easy. auto.
    - eapply (AbIdealEnv.forget_correct (x:=x) (chan:=in_chan_top)) in Hρ'. 2:apply in_chan_top_correct.
      generalize (forget_obligation_1 x ab).
      destruct AbIdealEnv.forget. contradiction.
      intros. eexists. split. apply Hρ'.
      intro. unfold upd. destruct eq_dec. 2:auto. subst.
      instantiate (1:=match n with NVint _ => _ | _ => _ end).
      destruct n. apply related_signed. apply related_signed_long. constructor. constructor.
  Qed.

  Fixpoint infer_var_signedness (e:nexpr var) (ctx:option signedness) (sm:signednessmap)
  : signednessmap :=
    match e with
      | NEvar nvt x =>
        match ctx with
          | None =>
            match T.get x sm with
              | None =>
                match nvt with
                  | NVTfloat | NVTsingle => sm
                  | _ => T.set x (None, nvt) sm
                end
              | Some _ => sm
            end
          | Some _ => T.set x (ctx, nvt) sm
        end
      | NEconst _ => sm
      | NEunop op e =>
        let ctx :=
            match op with
              | (Ofloatofint | Ofloatoflong | Olongofint | Osingleofint | Osingleoflong |
                 Ocast16signed | Ocast8signed) => Some Signed
              | (Ofloatofintu | Ofloatoflongu | Olongofintu | Osingleofintu | Osingleoflongu |
                 Ocast16unsigned | Ocast8unsigned) => Some Unsigned
              | _ => None
            end
        in
        infer_var_signedness e ctx sm
      | NEbinop op e1 e2 =>
        let ctx1 :=
          match op with
            | (Odiv | Omod | Odivl | Omodl | Ocmp _ | Ocmpl _ | Oshr | Oshrl) =>
              Some Signed
            | (Odivu | Omodu | Odivlu | Omodlu | Ocmpu _ | Ocmplu _ | Oshru | Oshrlu) =>
              Some Unsigned
            | _ => None
          end
        in
        let ctx2 :=
          match op with
            | (Odiv | Omod | Odivl | Omodl | Ocmp _ | Ocmpl _) => Some Signed
            | (Odivu | Omodu | Odivlu | Omodlu | Ocmpu _ | Ocmplu _ |
               Oshl | Oshr | Oshru | Oshll | Oshrl | Oshrlu) => Some Unsigned
            | _ => None
          end
        in
        (infer_var_signedness e2 ctx2 (infer_var_signedness e1 ctx1 sm))
    end.

  Lemma infer_var_signedness_consistent :
    ∀ ρ e nv ctx sm, eval_nexpr ρ e nv ->
                     ρ ∈ γ sm ->
                     (ctx <> None -> match nv with NVfloat _ | NVsingle _ => False | _ => True end) ->
                     ρ ∈ γ (infer_var_signedness e ctx sm).
  Proof.
    induction e; simpl; intros nv ctx sm H Hsm Hfl v' [s ty] Hmap; inv H; simpl.
    - destruct ctx.
      + erewrite T.gsspec in Hmap; eauto.
        destruct T.elt_eq.
        inv Hmap. split; auto. inv H4; try easy.
        rewrite <- H1 in Hfl. destruct Hfl. congruence.
        rewrite <- H1 in Hfl. destruct Hfl. congruence.
        apply Hsm in Hmap. auto.
      + destruct (T.get v sm) eqn:?. apply Hsm in Hmap. auto.
        destruct n.
        * erewrite T.gsspec in Hmap; eauto.
          destruct T.elt_eq.
          inv Hmap. auto. apply Hsm in Hmap. auto.
        * erewrite T.gsspec in Hmap; eauto.
          destruct T.elt_eq.
          inv Hmap. auto. apply Hsm in Hmap. auto.
        * apply Hsm in Hmap. auto.
        * apply Hsm in Hmap. auto.
    - apply Hsm in Hmap. auto.
    - match goal with Hmap : context [infer_var_signedness e ?ctx' sm] |- _ => set ctx' in Hmap end.
      edestruct (IHe _ o sm H2); eauto. destruct H4; auto.
    - match goal with Hmap : context [infer_var_signedness e1 ?ctx' sm] |- _ => set ctx' in Hmap end.
      match goal with Hmap : context [infer_var_signedness e2 ?ctx' _] |- _ => set ctx' in Hmap end.
      edestruct (fun sm => IHe2 _ o0 sm H5). 3:eauto. 2:destruct H6; auto. 2:auto.
      intros v [s' ty'] Hmap'.
      edestruct (IHe1 _ o sm H3); eauto. destruct H6; auto.
  Qed.

  (* If e is a variable, and we cannot get a better approximation for this variable than 
     top then we assign it to the interval corresponding to this sign, before assume.

     This is usefull in the following case (for example): if we have an assume of the form x < e,
     where e is well approximated but where we do not know anything about x, we need to force x
     to be in [-min_int; +max_int] before forwarding the assume. *)
  Program Definition assign_vars_to_top (e:nexpr var) (v:chan_abt * signednessmap) :
    (chan_abt * signednessmap)+⊥ :=
    let signedness := infer_var_signedness e None (T.empty _) in
    T.fold (fun ab v sty =>
                do_bot ab <- ab;
                let '(ab, sm) := ab return _ in
                match sty return _ with
                  | (Some s, ty) =>
                  let sm := T.set v sty sm in
                  match get_itv (NEvar ty v) (ab, sm) return _ with
                    | NotBot (NVIint All) | NotBot (NVIint64 All) =>
                      do_bot ab <-
                        lift_bot_chan_abt
                          (AbIdealEnv.assign
                             v (IEZconst var
                                 (Just match s, ty return _ with
                                    | Signed, NVTint => signed_itv
                                    | Unsigned, NVTint => unsigned_itv
                                    | Signed, NVTint64 => signed64_itv
                                    | Unsigned, NVTint64 => unsigned64_itv
                                    | _, _ => signed_itv (* Dummy *)
                                  end)) (proj1_sig ab) in_chan_top);
                      NotBot (ab, sm)
                    | _ => NotBot (ab, sm)
                  end
                  | (None, _) =>
                    let sm :=
                        match T.get v sm return _ with
                          | None => T.set v sty sm
                          | Some _ => sm
                        end
                    in
                    NotBot (ab, sm)
                end)
             signedness (NotBot v).
  Next Obligation. eapply assign_redundant. split. apply in_chan_top_correct. auto. Qed.
  Next Obligation. eapply assign_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma assign_vars_to_top_correct:
    forall ab e ρ v,
      eval_nexpr ρ e v ->
      γ ab ρ -> γ (assign_vars_to_top e ab) ρ.
  Proof.
    unfold assign_vars_to_top. intros.
    lapply (fun HH => infer_var_signedness_consistent ρ e _ None (T.empty _) H
            HH (fun N => match N eq_refl with end)).
    2:intro; rewrite T.gempty; discriminate.
    change (γ (NotBot ab) ρ) in H0. revert H0.
    generalize (NotBot ab).
    generalize (assign_vars_to_top_obligation_1 e ab).
    generalize (assign_vars_to_top_obligation_2 e ab).
    clear ab.
    generalize (infer_var_signedness e None). clear e v H.
    intros. revert ρ H0 H.
    assert (∀ v s0, T.get v (s (T.empty _)) = Some s0 <->
                    List.In (v, s0) (T.elements (s (T.empty _)))).
    { split. apply T.elements_correct. apply T.elements_complete. }
    unfold γ at 2, gamma_sm. setoid_rewrite H. clear H.
    rewrite T.fold_spec, <- fold_left_rev_right. setoid_rewrite in_rev.
    induction (rev (T.elements (s (T.empty _)))). auto. simpl. intros.
    eapply botbind_spec with ρ; [|apply IHl]; eauto.
    intros. destruct a. specialize (H0 _ _ (or_introl eq_refl)). destruct H0, p.
    destruct a0, H1 as [[? []] ?].
    destruct o.
    2:simpl; destruct (T.get e t); (split; [eexists; split; now eauto|]); auto;
      simpl; intro; rewrite T.gsspec; destruct T.elt_eq; [|now apply H4];
      intros s0 Hs0; inv Hs0; auto.
    match goal with
      | |- appcontext [fold_right ?f] => set (F:=f)
    end.
    assert (eval_nexpr ρ (NEvar n e) (ρ e)) by (constructor; auto).
    assert (γ (T.set e (Some s0, n) t) ρ).
    { intro. simpl. rewrite T.gsspec. destruct T.elt_eq.
      - intros. inv H6. auto.
      - apply H4. }
    unfold snd in H2. inv H2; try discriminate.
    + assert (ASS:=AbIdealEnv.assign_correct (x:=e)
                (e:=IEZconst var (Just match s0 with
                       | Signed => signed_itv
                       | Unsigned => unsigned_itv
                     end))
                (n:=INz (match s0 with
                           | Signed => Int.signed i
                           | Unsigned => Int.unsigned i
                         end)) (ab:=proj1_sig c) H1 (in_chan_top_correct _)).
      lapply ASS.
      2:constructor; destruct s0;
      [apply signed_itv_correct|apply unsigned_itv_correct].
      clear ASS; intro. simpl.
      generalize (γ0 (fold_right F b l) e (Some s0, NVTint)
                     (c, t) c t (Some s0) NVTint s0).
      generalize (γ1 (fold_right F b l) e (Some s0, NVTint)
                     (c, t) c t (Some s0) NVTint s0).
      simpl. change T.elt with var in *.
      destruct AbIdealEnv.assign. contradiction. intros.
      assert (∀ Hx0, γ (do_bot ab <- lift_bot_chan_abt (exist _ (NotBot x0) Hx0);
                        NotBot(ab, T.set e (Some s0, NVTint) t)) ρ).
      { intros. eapply botbind_spec. split; now eauto.
        eexists. split. apply H2.
        intro. unfold upd. destruct eq_dec; auto. subst. rewrite <- H9.
        destruct s0. apply related_signed. apply related_unsigned. }
      destruct get_itv as [|[[]|[]| |]]; (try now (econstructor; eauto)); eauto;
      (split; [eexists; eauto |auto]).
    + assert (ASS:=AbIdealEnv.assign_correct (x:=e)
                (e:=IEZconst var (Just match s0 with
                            | Signed => signed64_itv
                            | Unsigned => unsigned64_itv
                          end))
                (n:=INz (match s0 with
                           | Signed => Int64.signed i
                           | Unsigned => Int64.unsigned i
                         end)) (ab:=proj1_sig c) H1 (in_chan_top_correct _)).
        lapply ASS.
        2:constructor; destruct s0;
          [apply signed64_itv_correct|apply unsigned64_itv_correct].
        clear ASS; intro. simpl.
      generalize (γ0 (fold_right F b l) e (Some s0, NVTint64)
                     (c, t) c t (Some s0) NVTint64 s0).
      generalize (γ1 (fold_right F b l) e (Some s0, NVTint64)
                     (c, t) c t (Some s0) NVTint64 s0).
      simpl. change T.elt with var in *.
      destruct AbIdealEnv.assign. contradiction. intros.
      assert (∀ Hx0, γ (do_bot ab <- lift_bot_chan_abt (exist _ (NotBot x0) Hx0);
                        NotBot(ab, T.set e (Some s0, NVTint64) t)) ρ).
      { intros. eapply botbind_spec. split; now eauto.
        eexists. split. apply H2.
        intro. unfold upd. destruct eq_dec; auto. subst. rewrite <- H9.
        destruct s0. apply related_signed_long. apply related_unsigned_long. }
      destruct get_itv as [|[[]|[]| |]]; (try now (econstructor; eauto)); eauto;
      (split; [eexists; eauto |auto]).
    + easy.
    + easy.
  Qed.

  Program Definition assign (x:var) (e:nexpr var) (v:chan_abt*signednessmap) :
    (chan_abt*signednessmap)+⊥ :=
    let v := assign_vars_to_top e v in
    do_bot v <- v;
    let '(v, sm) := v return _ in
    let '(e', v, _) := convert_expr (NotBot v) e return _ in
    do_bot v <- v;
    do_bot v <- lift_bot_chan_abt (AbIdealEnv.assign x
      (match nexpr_ty e return _ with
         | NVTint =>
           match normalize_expr (NotBot v) e'
                 Int.min_signed Int.max_unsigned Int.wordsize return _ with
             | Some e => e
             | None => IEZconst var All
           end
         | NVTint64 =>
           match normalize_expr (NotBot v) e'
                 Int64.min_signed Int64.max_unsigned Int64.wordsize return _ with
             | Some e => e
             | None => IEZconst var All
           end
         | NVTfloat | NVTsingle => e'
       end) (proj1_sig v) in_chan_top);
  let sm := match nexpr_ty e return _ with
              | NVTfloat | NVTsingle => T.remove x sm
              | ty => T.set x (None, ty) sm
            end in
  NotBot (v, sm).
  Next Obligation. eapply assign_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma assign_correct:
    ∀ x e ρ, ∀ n ab,
      ρ ∈ γ ab ->
      n ∈ eval_nexpr ρ e ->
      (upd ρ x n) ∈ γ (assign x e ab).
  Proof.
    intros x e ρ n ab0 Hρ EVAL.
    unfold assign. apply botbind_spec with ρ.
    2:eauto using assign_vars_to_top_correct; eauto. clear Hρ.
    intros [ab sm] [[ρ' [Hρ' Hρρ']] Hsm].
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot ab) _ Hρ' Hρρ').
    destruct convert_expr as [[] ?].
    edestruct H as [? [? []]]; auto. clear H.
    apply botbind_spec with ρ'. 2:eauto. intros.
    eapply botbind_spec.
    { split.
      - simpl. eauto.
      - intros y smy. simpl.
        pose proof (nexpr_ty_correct EVAL). destruct (nexpr_ty e).
        + rewrite T.gsspec. unfold upd, eq_dec, var_eq. destruct T.elt_eq.
          intros eq. inv eq. split; easy. auto.
        + rewrite T.gsspec. unfold upd, eq_dec, var_eq. destruct T.elt_eq.
          intros eq. inv eq. split; easy. auto.
        + rewrite T.grspec. unfold upd, eq_dec, var_eq. destruct T.elt_eq. easy. eauto.
        + rewrite T.grspec. unfold upd, eq_dec, var_eq. destruct T.elt_eq. easy. eauto. }
    assert (∀ (ab:chan_abt+⊥) x0, related n x0 -> γ ab (upd ρ' x x0) -> γ ab (upd ρ x n)).
    { unfold upd. intros [|ab']. auto.
      eexists. split. apply H4. intro. destruct eq_dec; eauto. }
    pose proof (nexpr_ty_correct EVAL).
    generalize (assign_obligation_1 x e ab0 (ab, sm) ab sm (i, b) b0 i b a).
    destruct H4; intro.
    - inv H2. destruct_opt.
      edestruct normalize_expr_correct as [x'0 [Hx'0bounded [Hx'x'0 EVALx'0]]]; eauto.
      eapply H3. erewrite Int.eqm_samerepr. constructor. eauto.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
      eapply H3. constructor.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
      constructor. constructor.
    - inv H2. destruct_opt.
      edestruct normalize_expr_correct with (logdiv:=Int64.wordsize)
        as [x'0 [Hx'0bounded [Hx'x'0 EVALx'0]]]; eauto.
      eapply H3. erewrite Int64.eqm_samerepr. constructor. eauto.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
      eapply H3. constructor.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
      constructor. constructor.
    - eapply H3. eauto.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
    - eapply H3. eauto.
      eapply lift_bot_chan_abt_correct, AbIdealEnv.assign_correct; eauto using in_chan_top_correct.
  Qed.

  Fixpoint simpl_assume (e:iexpr var) (b:bool) (ab:chan_abt) : (iexpr var * bool) :=
    match e with
      | IEbinop (IOcmp Ceq) e' (IEZconst (Just (ZIntervals.ITV 0 0)))
      | IEbinop (IOcmp Ceq) (IEZconst (Just (ZIntervals.ITV 0 0))) e' =>
        match AbIdealEnv.get_itv (snd (proj1_sig ab)) e' with
          | NotBot (Just (IdealIntervals.Az (ZIntervals.ITV 0 1))) =>
            simpl_assume e' (negb b) ab
          | _ => (e, b)
        end
      | IEbinop (IOcmp Cne) e' (IEZconst (Just (ZIntervals.ITV 0 0)))
      | IEbinop (IOcmp Cne) (IEZconst (Just (ZIntervals.ITV 0 0))) e' =>
        match AbIdealEnv.get_itv (snd (proj1_sig ab)) e' with
          | NotBot (Just (IdealIntervals.Az (ZIntervals.ITV 0 1))) =>
            simpl_assume e' b ab
          | _ => (e, b)
        end
      | _ => (e, b)
    end.

  Lemma simpl_assume_correct:
    ∀ e ρ ab b,
      ρ ∈ γ ab ->
      eval_iexpr ρ e (INz (if b:bool then 1 else 0))->
    ∀ e' b', simpl_assume e b ab = (e', b') ->
      eval_iexpr ρ e' (INz (if b':bool then 1 else 0)).
  Proof.
    fix 1.
    intros.
    destruct e; try (injection H1; intros; subst; exact H0).
    destruct i; try (injection H1; intros; subst; exact H0).
    pose proof H0. inv H0. inv H9.
    assert (∀ x, eval_iexpr ρ (IEZconst var (Just (ZIntervals.ITV 0 0))) (INz x) -> x = 0).
    { intros. inv H0. destruct H5. omega. }
    pose proof (AbIdealEnv.get_itv_correct (proj2 H) H8).
    pose proof (AbIdealEnv.get_itv_correct (proj2 H) H6).
    simpl in H1.
    pose proof (simpl_assume_correct e1).
    pose proof (simpl_assume_correct e2).
    clear simpl_assume_correct.
    destruct c; try (injection H1; intros; subst; exact H2);
    (repeat match goal with
              | H1 : match ?x with _ => _ end = (e', b') |- _ => destruct x
              | H1 : (_, _) = (e', b') |- _ =>
                injection H1; intros; subst e'; subst b'; now auto
              | H : simpl_assume _ _ _ = (e', b') |- _ =>
                eapply H5; [now eauto| |exact H]; clear H
              | H : simpl_assume _ _ _ = (e', b') |- _ =>
                eapply H9; [now eauto| |exact H]; clear H
              | H : eval_iexpr ρ ?e (INz ?v) |- eval_iexpr ρ ?e (INz ?v') =>
                replace v' with v; [exact H|]
              | H : eval_iexpr ρ (IEZconst var (Just (ZIntervals.ITV 0 0))) (INz ?x) |- _ =>
                apply H0 in H; subst x
              | H : γ (NotBot (Just (IdealIntervals.Az (ZIntervals.ITV 0 1)))) (INz ?i) |- _ =>
                destruct H, b, i as [|[]|]; try reflexivity; try discriminate; zify; omega
            end).
    destruct b; try discriminate; reflexivity.
    destruct b; try discriminate; reflexivity.
  Qed.

  (* TODO : This might be a good idea to do this as an output channel. *)
  Program Fixpoint assume_mux_iexpr (fuel:nat) (c:mux_iexpr var) (b:bool) (ab:chan_abt) : chan_abt+⊥ :=
    match fuel return _ with
      | O => NotBot ab
      | S fuel =>
        match c return _ with
          | Mux c0 ct cf =>
            let abt :=
                do_bot ab <- assume_mux_iexpr fuel c0 true ab;
                  assume_mux_iexpr fuel ct b ab
            in
            let abf :=
                do_bot ab <- assume_mux_iexpr fuel c0 false ab;
                  assume_mux_iexpr fuel cf b ab
            in
            match abt, abf return _ with
              | Bot, _ => abf
              | _, Bot => abt
              | NotBot abt, NotBot abf => NotBot (id_join abt abf in_chan_top)
            end
          | Leaf c =>
            let '(c, b) := simpl_assume c b ab return _ in
            do_bot ab <- lift_bot_chan_abt (AbIdealEnv.assume c b ab in_chan_top);
            match c return _ with
              | IEvar _ x =>
                match AbIdealEnv.get_eq_expr (snd (ab:chan_abt)) x return _ with
                  | None => NotBot ab
                  | Some e => assume_mux_iexpr fuel e b ab
                end
              | _ => NotBot ab
            end
        end
    end.
  Next Obligation. apply AbIdealEnv.assume_redundant. split. apply in_chan_top_correct. auto. Qed.
  Next Obligation. apply AbIdealEnv.id_join_redundant. split. apply in_chan_top_correct. auto. Defined.
  Definition assume_mux_iexpr_fuel := 10%nat.

  Lemma assume_mux_iexpr_correct:
    forall fuel c ρ ab b,
      ρ ∈ γ ab ->
      (INz (if b:bool then 1 else 0)) ∈ eval_mux_iexpr ρ c ->
      ρ ∈ γ (assume_mux_iexpr fuel c b ab).
  Proof.
    induction fuel. auto. intros. simpl. inv H0.
    - pose proof (simpl_assume_correct e ρ ab b H H1). clear H1.
      destruct (simpl_assume e b ab) as [e' b']. specialize (H0 _ _ eq_refl).
      eapply botbind_spec with ρ.
      + destruct e'; auto. intros.
        pose proof (get_eq_expr_correct (proj2 H1) v).
        destruct (get_eq_expr (snd (proj1_sig a)) v); auto.
        apply IHfuel. auto. inv H0. rewrite <- H5. auto.
      + apply lift_bot_chan_abt_correct.
        apply AbIdealEnv.assume_correct. apply H. apply in_chan_top_correct. auto.
    - assert (ρ ∈ γ (do_bot ab0 <- assume_mux_iexpr fuel c0 false ab; assume_mux_iexpr fuel ef b ab0)).
      { eapply botbind_spec with ρ; eauto. }
      destruct (do_bot ab0 <- assume_mux_iexpr fuel c0 false ab; assume_mux_iexpr fuel ef b ab0).
      contradiction.
      destruct (do_bot ab0 <- assume_mux_iexpr fuel c0 true ab; assume_mux_iexpr fuel et b ab0).
      auto. apply id_join_correct. auto. apply in_chan_top_correct.
    - assert (ρ ∈ γ (do_bot ab0 <- assume_mux_iexpr fuel c0 true ab; assume_mux_iexpr fuel et b ab0)).
      { eapply botbind_spec with ρ; eauto. }
      destruct (do_bot ab0 <- assume_mux_iexpr fuel c0 true ab; assume_mux_iexpr fuel et b ab0).
      contradiction.
      destruct (do_bot ab0 <- assume_mux_iexpr fuel c0 false ab; assume_mux_iexpr fuel ef b ab0).
      auto. apply id_join_correct. auto. apply in_chan_top_correct.
  Qed.
  Opaque assume_mux_iexpr.

  Definition assume (e:nexpr var) (b:bool) (v:chan_abt * signednessmap) :
    (chan_abt*signednessmap)+⊥ :=
    let v := assign_vars_to_top e v in
    do_bot v <- v;
    let '(v, sm) := v return _ in
    let '(e, v, _) := convert_expr (NotBot v) e return _ in
    do_bot v <- v;
    match normalize_expr (NotBot v) e Int.min_signed Int.max_unsigned Int.wordsize return _ with
      | Some e =>
        do_bot v <- assume_mux_iexpr assume_mux_iexpr_fuel (Leaf e) b v;
        NotBot (v, sm)
      | None => NotBot (v, sm)
    end.

  Lemma assume_correct:
    ∀ e ρ ab b,
      ρ ∈ γ ab ->
      of_bool b ∈ eval_nexpr ρ e ->
      ρ ∈ γ (assume e b ab).
  Proof.
    intros e ρ ab b Hρ EVAL. unfold assume.
    eapply botbind_spec with ρ. 2:eauto using assign_vars_to_top_correct.
    intros. destruct a, H as [[ρ' [Hρ' Hρρ']] Hsm].
    pose proof (convert_expr_ok _ _ _ EVAL (NotBot c) _ Hρ' Hρρ').
    destruct convert_expr as [[] ?]. edestruct H as [? [x' [EVAL2 REL]]]; auto. clear H.
    eapply botbind_spec with ρ'; eauto.
    assert (∃ z, x' = INz z) by (destruct b; inv REL; eauto).
    destruct H; subst.
    intros.
    destruct_opt.
    - edestruct normalize_expr_correct as [? [? []]]; eauto.
      eapply botbind_spec with ρ'.
      + split. eexists. eauto. auto.
      + apply assume_mux_iexpr_correct; eauto.
        constructor. replace (if b then 1 else 0) with x0. auto.
        inv REL. 2:destruct b; discriminate.
        apply Int.eqm_samerepr in H2. rewrite H2 in H5.
        assert (Int.repr x0 = if b then Int.one else Int.zero) by
            (destruct b; compute in H5; unfold Int.one, Int.zero; congruence).
        destruct (Coqlib.zlt x0 0).
        * exfalso. apply (f_equal Int.signed) in H4.
          rewrite Int.signed_repr in H4. subst x0. destruct b; discriminate.
          split. now intuition. assert (0 < Int.max_signed) by reflexivity. omega.
        * apply (f_equal Int.unsigned) in H4.
          rewrite Int.unsigned_repr in H4 by intuition. destruct b; apply H4.
    - split. eexists; eauto. auto.
  Qed.

  Definition nonblock (e:nexpr var) (v:chan_abt * signednessmap) : bool :=
    let '(_, _, nb) := convert_expr (NotBot (fst v)) e in
    nb.

  Lemma nonblock_correct:
    ∀ e ρ ab,
      ρ ∈ γ ab ->
      nonblock e ab = true ->
      block_nexpr ρ e ->
      False.
  Proof.
    unfold nonblock. intros.
    destruct H as [[? [? ?]] ?].
    destruct (convert_expr (NotBot (fst ab)) e) as [[] ?] eqn:EQ. subst.
    eapply convert_expr_nonblock with (v:=NotBot (fst ab)); eauto.
  Qed.

  Function zitv_concretize_int_aux (from: Z) (nb: N)
           (acc: IntSet.int_set) {measure N.to_nat nb} : TreeAl.ZShareTree.t unit :=
    match nb with
    | N0 => acc
    | Npos p =>
      let nb' := Pos.pred_N p in
      zitv_concretize_int_aux (Z.succ from) nb' (TreeAl.ZShareTree.set from tt acc)
    end.
  Proof.
    intros from nb acc p ->. rewrite N.pos_pred_spec, Nnat.N2Nat.inj_pred. zify. intuition.
  Defined.

  Definition zitv_concretize_int (max: N) (i: ZIntervals.itv) : TreeAl.ZShareTree.t unit+⊤+⊥ :=
    match i with
    | ZIntervals.ITV m M =>
      if zlt M m then Bot
      else
        let n := Z.to_N (M-m) in
        if N.leb n max
        then NotBot (Just (zitv_concretize_int_aux m (N.succ n) (TreeAl.ZShareTree.empty _)))
        else NotBot All
    end.

  Lemma zitv_concretize_int_correct:
    ∀ itv max z, γ itv z ->
      match zitv_concretize_int max itv with
        | NotBot (Just ab) => TreeAl.ZShareTree.get z ab = Some tt
        | NotBot All => True
        | Bot => False
      end.
  Proof.
    intros [a b] max z ITV.
    unfold zitv_concretize_int.
    destruct ITV as [A B].
    destruct zlt. omega.
    destruct N.leb. 2:exact I.
    remember (TreeAl.ZShareTree.empty unit) as init.
    assert (TreeAl.ZShareTree.get z init = Some tt \/ a <= z) by auto.
    clear Heqinit max A.
    remember (N.succ (Z.to_N (b-a))).
    assert (b-a+1= Z.of_N n).
    { subst n. rewrite N2Z.inj_succ, Z2N.id. Psatz.lia. omega. }
    clear Heqn g.
    functional induction (zitv_concretize_int_aux a n init).
    - destruct H. apply H. zify; omega.
    - eapply IHt.
      rewrite TreeAl.ZShareTree.gsspec. destruct (TreeAl.ZShareTree.elt_eq z from). auto.
      destruct H. auto. right. unfold TreeAl.ZShareTree.elt, TreeAl.Z_EQ_POS.s in *. omega.
      rewrite N.pos_pred_spec, N2Z.inj_pred_max. zify; omega.
  Qed.

  Definition concretize_int (max: N) (e: nexpr var) (ab: chan_abt*signednessmap) : IntSet.int_set+⊤+⊥ :=
    let ab := fst ab in
    let '(e, _, _) := convert_expr (NotBot ab) e in
    do_bot cong <- (snd (proj1_sig ab)).(AbIdealEnv.get_congr) e;
    let '{| ZCongruences_defs.zc_rem := rem; ZCongruences_defs.zc_mod := modn |} := cong in
    let modz := Z.of_N modn in
    match modn with
      | N0 => NotBot (Just (IntSet.singleton (Int.repr rem)))
      | Npos modn =>
        do_bot itv <- (snd (proj1_sig ab)).(AbIdealEnv.get_itv) e;
        do_bot itv <- IdealIntervals.in_z itv;
        match itv with
          | All => NotBot All
          | Just (ZIntervals.ITV a b) =>
            let itvdiv := ZIntervals.ITV (-((rem-a)/(Zpos modn))) ((b-rem)/(Zpos modn)) in
            do_bot set <- zitv_concretize_int max itvdiv;
            NotBot (toplift_lift (IntSet.map (fun x => IntSet.proj (Int.repr (x*Zpos modn+rem)))) set)
        end
    end.

  Lemma concretize_int_correct :
    ∀ max e ρ ab, ρ ∈ γ ab ->
      (eval_nexpr ρ e ∘ NVint) ⊆ γ (concretize_int max e ab).
  Proof.
    intros. unfold concretize_int.
    simpl in H0. destruct H as [[ρ' [Hρ' REL]] _].
    destruct (convert_expr (NotBot (fst ab)) e)  as [[e' ab'] ?]  eqn:EQ.
    edestruct convert_expr_ok with (v:=NotBot (fst ab)) as (_ & z & EVAL & RELz); eauto.
    inv RELz.
    eapply botbind_spec. 2:eapply get_congr_correct; eauto; apply Hρ'.
    intros [] (q & Hq). simpl in Hq.
    destruct zc_mod.
    - simpl. unfold IntSet.singleton. do 6 red.
      rewrite TreeAl.ZShareTree.gsspec. destruct TreeAl.ZShareTree.elt_eq. auto.
      destruct n. f_equal. f_equal. subst zc_rem. ring.
    - eapply botbind_spec.
      2:eapply AbIdealEnv.get_itv_correct; [apply Hρ'|eauto].
      intros. eapply botbind_spec; [|apply IdealIntervals.in_z_correct; eauto].
      intros. destruct a0 as [|[]]. exact I.
      assert ((-q) ∈ γ (ZIntervals.ITV (- ((zc_rem - z) / Z.pos p))
                                       ((z0 - zc_rem) / Z.pos p))).
      { destruct H1. split; [apply -> Z.opp_le_mono|]; apply Z.div_le_lower_bound; Psatz.lia. }
      pose proof (zitv_concretize_int_correct _ max _ H2).
      destruct zitv_concretize_int as [|[|set]]. auto. exact I. simpl.
      unfold botbind. red. red. red. red.
      replace x with ((-q) * Z.pos p + zc_rem).
      apply (IntSet.map_in (fun _ =>  _)). auto.
      rewrite Hq. simpl. ring.
  Qed.

  Program Definition transfer_signedness (v2:chan_abt) (sm1 sm2:signednessmap) : chan_abt :=
    let sm := T.shcombine_diff (fun _ s1 s2 =>
      match s1, s2 return _ with
        | Some (Some s, NVTint), Some (None, NVTint) =>
          Some (Some s, NVTint)
        | Some (Some s, NVTint64), Some (None, NVTint64) =>
          Some (Some s, NVTint64)
        | _, _ => None
      end) _ sm1 sm2
    in
    T.fold (fun ab x sty =>
              match sty return _ with
                | (Some s, ty) =>
                  match get_itv (NEvar ty x) (ab, sm2) return _ with
                    | (NotBot (NVIint All)) | NotBot (NVIint64 All) =>
                      match lift_bot_chan_abt
                              (AbIdealEnv.assign
                                 x (IEZconst var
                                      (Just match s, ty return _ with
                                         | Signed, NVTint => signed_itv
                                         | Unsigned, NVTint => unsigned_itv
                                         | Signed, NVTint64 => signed64_itv
                                         | Unsigned, NVTint64 => unsigned64_itv
                                         | _, _ => signed_itv (* Dummy *)
                                       end)) (proj1_sig ab) in_chan_top)
                            return _
                      with Bot => ab (* I see no reason why this would happend *)
                        | NotBot ab => ab
                      end
                    | _ => ab
                  end
                | _ => ab
              end)
           sm v2.
  Next Obligation. destruct v as [[[] []]|]; auto. Qed.
  Next Obligation. eapply assign_redundant. split. apply in_chan_top_correct. auto. Qed.
  Next Obligation. eapply assign_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma transfer_signedness_correct:
    ∀ sm1 sm2 v2 (ρ:var -> num_val),
      ρ ∈ γ (v2, sm2) ->
      ρ ∈ γ (transfer_signedness v2 sm1 sm2).
  Proof.
    intros. unfold transfer_signedness.
    match goal with |- context [T.shcombine_diff ?f' ?Hf' sm1 sm2] => set (f:=f'); set (Hf:=Hf') end.
    assert (ρ ∈ γ (T.shcombine_diff f Hf sm1 sm2)).
    { intros x s. rewrite T.gshcombine_diff by auto.
      destruct H. red in H0.
      destruct (T.get x sm1) as [[[] []]|] eqn:?, (T.get x sm2) as [[[] []]|] eqn:?; simpl;
        intro eq; inv eq; apply (H0 _ (None, _)); auto. }
    revert H0.
    assert (∀ v s0, T.get v (T.shcombine_diff f Hf sm1 sm2) = Some s0 <->
                    List.In (v, s0) (T.elements (T.shcombine_diff f Hf sm1 sm2))).
    { split. apply T.elements_correct. apply T.elements_complete. }
    unfold γ at 1, gamma_sm. setoid_rewrite H0. clear H0.
    rewrite T.fold_spec, <- fold_left_rev_right. setoid_rewrite in_rev.
    match goal with
      | |- appcontext [fold_right ?f] => set (F:=f)
    end.
    induction (rev (T.elements (T.shcombine_diff f Hf sm1 sm2))); intros. apply H.
    lapply IHl. 2:intros; apply H0; simpl; auto.
    clear IHl. intro IHl.
    destruct a as [? [[] ?]]. 2:now auto. simpl. unfold F at 1. simpl.
    edestruct H0. left. eauto.
    pose proof (get_itv_correct (NEvar n e) ρ (_, sm2) (conj IHl (proj2 H)) (ρ e)
                                (eval_NEvar _ _ eq_refl H2)).
    generalize (transfer_signedness_obligation_2 v2 sm1 sm2
                  (fold_right F v2 l) e (Some s, n) (Some s) n s).
    generalize (transfer_signedness_obligation_3 v2 sm1 sm2
                  (fold_right F v2 l) e (Some s, n) (Some s) n s).
    Transparent lift_bot_chan_abt.
    pose proof IHl. destruct IHl as (ρ' & Hρ' & Hρρ').
    change T.elt with var in *.
    destruct (get_itv (NEvar n e) (fold_right F v2 l, sm2)) as [|[[]|[]| |]]; auto;
    simpl in H2; inv H2; try contradiction; simpl;
    match goal with |- context [AbIdealEnv.assign e ?e' ?ab in_chan_top] =>
      refine (let ASS' :=
        let val := match s with Signed => _ | _ => _ end in
        @AbIdealEnv.assign_correct _ _ _ _ e e' ρ' val ab in_chan_top Hρ'
                                   (in_chan_top_correct _) _
              in (fun ASS => _) ASS'); [
                        (destruct s; constructor;
                         [apply signed_itv_correct|apply unsigned_itv_correct]) ||
                        (destruct s; constructor;
                         [apply signed64_itv_correct|apply unsigned64_itv_correct])
                       |clear ASS'; intros;
                        destruct (AbIdealEnv.assign e e' ab in_chan_top)]
    end; auto;
    eexists; (split; [apply ASS|]);
    intro y; unfold upd; (destruct eq_dec; [subst y|now auto]); destruct s;
    rewrite <- H7;
    apply related_signed || apply related_unsigned ||
          apply related_signed_long || apply related_unsigned_long.
    Opaque lift_bot_chan_abt.
  Qed.

  Program Definition join_sm (sm1 sm2:signednessmap) : signednessmap :=
    T.shcombine (fun _ s1 s2 =>
      match s1, s2 return _ with
        | Some (Some Signed, NVTint),     Some (Some Signed | None,   NVTint)
        | Some (Some Signed, NVTint64),   Some (Some Signed | None,   NVTint64)
        | Some (Some Unsigned, NVTint),   Some (Some Unsigned | None, NVTint)
        | Some (Some Unsigned, NVTint64), Some (Some Unsigned | None, NVTint64)
          => s1
        | Some (None, NVTint),   Some (Some Signed,   NVTint)
        | Some (None, NVTint64), Some (Some Signed,   NVTint64)
        | Some (None, NVTint),   Some (Some Unsigned, NVTint)
        | Some (None, NVTint64), Some (Some Unsigned, NVTint64)
          => s2
        | Some (Some _, NVTfloat), Some (Some _, NVTfloat)
        | Some (Some _, NVTsingle), Some (Some _, NVTsingle)
        | Some (None, NVTint), Some (None, NVTint)
        | Some (None, NVTint64), Some (None, NVTint64)
        | Some (None, NVTfloat), Some (None, NVTfloat)
        | Some (None, NVTsingle), Some (None, NVTsingle)
          => s1

        | _, _ => None
      end) _ sm1 sm2.
  Next Obligation. destruct v as [[[[]|] []]|]; auto. Qed.

  Program Instance wl : weak_lattice (chan_abt * signednessmap) := {
    top := (id_top in_chan_top, T.empty _);
    leb ab1 ab2 :=
      id_leb (proj1_sig (fst ab1)) (fst (proj1_sig (fst ab2))) &&
      T.shforall2 (fun _ s1 s2 =>
                       match s1, s2 return _ with
                         | _, None
                         | Some (_, NVTint), Some (_, NVTint)
                         | Some (_, NVTint64), Some (_, NVTint64)
                         | Some (_, NVTsingle), Some (_, NVTsingle)
                         | Some (_, NVTfloat), Some (_, NVTfloat) => true
                         | _, _ => false
                       end) _
        (snd ab1) (snd ab2);
    join ab1 ab2 :=
      (let ab1' := transfer_signedness (fst ab1) (snd ab2) (snd ab1) in
       let ab2' := transfer_signedness (fst ab2) (snd ab1) (snd ab2) in
       id_join (proj1_sig ab1') (proj1_sig ab2') in_chan_top,
       join_sm (snd ab1) (snd ab2));
    widen ab1 ab2 := (id_widen (proj1_sig (fst ab1)) (proj1_sig (fst ab2)) in_chan_top,
                      join_sm (snd ab1) (snd ab2))
  }.
  Next Obligation. destruct v as [[? []]|]; auto. Qed.
  Next Obligation. apply id_top_redundant. split. apply in_chan_top_correct. auto. Qed.
  Next Obligation. apply id_join_redundant. split. apply in_chan_top_correct. auto. Qed.
  Next Obligation. apply id_widen_redundant. split. apply in_chan_top_correct. auto. Qed.

  Lemma idealize_adom : Adom.adom (chan_abt * signednessmap) (var->num_val) wl _.
  Proof.
    Hint Resolve in_chan_top_correct.
    unfold wl. split.
    - intros [a1 sm1] [a2 sm2]. simpl. rewrite andb_true_iff.
      intros [Hleb Hlebsm] ρ [[ρ' [Hρ'γ Hρρ']] Hρsm]. split; simpl.
      + eapply id_leb_correct in Hleb; eauto.
        eexists. intuition eauto. split. auto. apply (proj2_sig a2 ). auto.
      + intros x [? s] Hs. simpl.
        rewrite T.shforall2_correct in Hlebsm. specialize (Hlebsm x). rewrite Hs in Hlebsm.
        specialize (Hρsm x). simpl in Hρsm.
        destruct (T.get x sm1) as [[? []]|], s; try discriminate;
        specialize (Hρsm _ eq_refl); auto.
    - intro ρ. split.
      exists (fun id => match ρ id with NVint i => INz (Int.unsigned i)
                                     | NVint64 i => INz (Int64.unsigned i)
                                     | NVfloat f => INf f
                                     | NVsingle f => INf (Float.of_single f) end).
      split. apply id_top_correct. apply in_chan_top_correct.
      intro. destruct (ρ id). apply related_unsigned. apply related_unsigned_long. constructor. constructor.
      intros x s. simpl. erewrite T.gempty. discriminate.
    - intros [[x Hx] smx] [[y Hy] smy] ρ [Hρ|Hρ]; split; simpl; unfold join_sm.
      + eapply transfer_signedness_correct in Hρ.
        destruct Hρ as (ρ' & Hρ' & Hρρ').
        eexists; split; eauto. apply id_join_correct; eauto.
      + destruct Hρ.
        intros v s. simpl. erewrite T.gshcombine; eauto.
        destruct (T.get v smx) as [[[[]|] []]|] eqn:?, (T.get v smy) as [[[[]|] []]|] eqn:?;
          try discriminate; intro H'; inv H'; eauto;
        eapply H0 in Heqo; eapply Heqo.
      + eapply transfer_signedness_correct in Hρ.
        destruct Hρ as (ρ' & Hρ' & Hρρ').
        eexists; split; eauto. apply id_join_correct; eauto.
      + destruct Hρ.
        intros v s. simpl. erewrite T.gshcombine; eauto.
        destruct (T.get v smx) as [[[[]|] []]|] eqn:?,
                 (T.get v smy) as [[[[]|] []]|] eqn:?;
          try discriminate; intro H'; inv H'; eauto;
        eapply H0 in Heqo0; eapply Heqo0.
  Qed.

  Instance idealenv_to_machineenv : ab_machine_env (chan_abt * signednessmap) :=
    { concretize_int := concretize_int
    ; assign := assign
    ; assume := assume
    ; nonblock := nonblock
    ; forget := forget
    }.
  Proof.
  - exact idealize_adom.
  - apply concretize_int_correct.
  - apply assign_correct.
  - apply forget_correct.
  - apply assume_correct.
  - apply nonblock_correct.
  Defined.

  Instance t_mach_ToString {T:ToString.ToString abt} : ToString.ToString (chan_abt * signednessmap) :=
    fun x => T (fst (proj1_sig (fst x))).

End Functor.

Existing Instance t_mach_ToString.

End Make.