(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Coqlib String ToString DebugCminor Util.
Require Import Integers Floats.
Require Import AdomLib.
Require ZIntervals FloatIntervals.

Set Implicit Arguments.
Section iexpr.
  Variable var : Type.

  Inductive ideal_num :=
  | INf : float -> ideal_num
  | INz : Z -> ideal_num.

  Inductive ideal_num_ty : Type :=
  | INTz | INTf.

  Inductive ideal_num_ty_gamma : gamma_op ideal_num_ty ideal_num :=
  | INTSz : forall i, ideal_num_ty_gamma INTz (INz i)
  | INTSf : forall f, ideal_num_ty_gamma INTf (INf f).
  Global Existing Instance ideal_num_ty_gamma.

  Inductive i_binary_operation : Type :=
  | IOadd | IOsub | IOmul | IOdiv | IOmod
  | IOand | IOor | IOxor | IOshl | IOshr
  | IOcmp : comparison -> i_binary_operation
  | IOaddf | IOsubf | IOmulf | IOdivf
  | IOcmpf : comparison -> i_binary_operation.


  Inductive i_unary_operation :=
  | IOneg | IOnot
  | IOnegf | IOabsf
  | IOsingleoffloat | IOfloatofz | IOsingleofz | IOzoffloat.

  Inductive iexpr : Type :=
  | IEvar : ideal_num_ty -> var -> iexpr
  | IEZconst : ZIntervals.itv+⊤ -> iexpr
  | IEFconst : FloatIntervals.itv+⊤ -> iexpr
  | IEunop : i_unary_operation -> iexpr -> iexpr
  | IEbinop : i_binary_operation -> iexpr -> iexpr -> iexpr.

  Inductive mux_iexpr : Type :=
  | Leaf : iexpr -> mux_iexpr
  | Mux : mux_iexpr -> mux_iexpr -> mux_iexpr -> mux_iexpr.

  Section eval_iexpr.
    Variable ρ: var -> ideal_num.

    Inductive eval_ibinop : i_binary_operation -> ideal_num -> ideal_num -> ideal_num -> Prop :=
    | eval_ibinop_IOadd : forall i1 i2,
      eval_ibinop IOadd  (INz i1)  (INz i2)  (INz (i1+i2))
    | eval_ibinop_IOsub : forall i1 i2,
      eval_ibinop IOsub  (INz i1)  (INz i2)  (INz (i1-i2))
    | eval_ibinop_IOmul : forall i1 i2,
      eval_ibinop IOmul  (INz i1)  (INz i2)  (INz (i1*i2))
    | eval_ibinop_IOdiv : forall i1 i2,
      i2 <> 0 ->
      eval_ibinop IOdiv  (INz i1)  (INz i2)  (INz (Z.quot i1 i2))
    | eval_ibinop_IOmod : forall i1 i2,
      i2 <> 0 ->
      eval_ibinop IOmod  (INz i1)  (INz i2)  (INz (Z.rem i1 i2))
    | eval_ibinop_IOand : forall i1 i2,
      eval_ibinop IOand  (INz i1)  (INz i2)  (INz (Z.land i1 i2))
    | eval_ibinop_IOor : forall i1 i2,
      eval_ibinop IOor  (INz i1)  (INz i2)  (INz (Z.lor i1 i2))
    | eval_ibinop_IOxor : forall i1 i2,
      eval_ibinop IOxor  (INz i1)  (INz i2)  (INz (Z.lxor i1 i2))
    | eval_ibinop_IOshl : forall i1 i2,
      eval_ibinop IOshl  (INz i1)  (INz i2)  (INz (Z.shiftl i1 i2))
    | eval_ibinop_IOshr : forall i1 i2,
      eval_ibinop IOshr  (INz i1)  (INz i2)  (INz (Z.shiftr i1 i2))
    | eval_ibinop_IOcmp : forall c i1 i2,
      eval_ibinop (IOcmp c)  (INz i1)  (INz i2)
                  (INz (if (ZIntervals.Zcmp c i1 i2) then 1 else 0))
    | eval_ibinop_IOaddf : forall f1 f2,
      eval_ibinop IOaddf  (INf f1)  (INf f2)  (INf (Float.add f1 f2))
    | eval_ibinop_IOsubf : forall f1 f2,
      eval_ibinop IOsubf (INf f1)  (INf f2)  (INf (Float.sub f1 f2))
    | eval_ibinop_IOmulf : forall f1 f2,
      eval_ibinop IOmulf  (INf f1)  (INf f2)  (INf (Float.mul f1 f2))
    | eval_ibinop_IOdivf : forall f1 f2,
      eval_ibinop IOdivf  (INf f1)  (INf f2)  (INf (Float.div f1 f2))
    | eval_ibinop_IOcmpf : forall c f1 f2,
      eval_ibinop (IOcmpf c)  (INf f1)  (INf f2) (INz (if (Float.cmp c f1 f2) then 1 else 0)).

    Inductive eval_iunop : i_unary_operation -> ideal_num -> ideal_num -> Prop :=
    | eval_iunop_IOneg : forall i,
      eval_iunop IOneg  (INz i)  (INz (-i))
    | eval_iunop_IOnot : forall i,
      eval_iunop IOnot  (INz i)  (INz (Z.lnot i))
    | eval_iunop_IOnegf : forall f,
      eval_iunop IOnegf  (INf f)  (INf (Float.neg f))
    | eval_iunop_IOabsf : forall f,
      eval_iunop IOabsf  (INf f)  (INf (Float.abs f))
    | eval_iunop_IOzoffloat : forall f i,
      Fappli_IEEE_extra.ZofB _ _ f = Some i ->
      eval_iunop IOzoffloat  (INf f)  (INz i)
    | eval_iunop_IOfloatofz : forall i,
      eval_iunop IOfloatofz (INz i)
            (INf (Fappli_IEEE_extra.BofZ 53 1024 (@eq_refl _ Lt) (@eq_refl _ Lt) i))
    | eval_iunop_IOsingleofz : forall i,
      eval_iunop IOsingleofz  (INz i)
                 (INf (Float.of_single (Fappli_IEEE_extra.BofZ 24 128 (@eq_refl _ Lt) (@eq_refl _ Lt) i)))
    | eval_iunop_IOsingleoffloat : forall f,
      eval_iunop IOsingleoffloat  (INf f)  (INf (Float.of_single (Float.to_single f))).

    Inductive eval_iexpr: iexpr -> ideal_num -> Prop :=
    | eval_IEvar: forall id i ty,
      ρ id = i -> γ ty i ->
      eval_iexpr (IEvar ty id) i
    | eval_IEZconst: forall x itv,
      x ∈ γ itv ->
      eval_iexpr (IEZconst itv) (INz x)
    | eval_IEFconst: forall x itv,
      x ∈ γ itv ->
      eval_iexpr (IEFconst itv) (INf x)
    | eval_IEunop: forall op a n r,
      eval_iexpr a n ->
      eval_iunop op n r->
      eval_iexpr (IEunop op a) r
    | eval_IEbinop: forall op a1 a2 n1 n2 n,
      eval_iexpr a1 n1 ->
      eval_iexpr a2 n2 ->
      eval_ibinop op n1 n2 n ->
      eval_iexpr (IEbinop op a1 a2) n.

    Inductive eval_mux_iexpr: mux_iexpr -> ideal_num -> Prop :=
    | eval_Leaf : forall e v,
      eval_iexpr e v ->
      eval_mux_iexpr (Leaf e) v
    | eval_Mux_0 : forall c et ef v,
      eval_mux_iexpr c (INz 0) ->
      eval_mux_iexpr ef v ->
      eval_mux_iexpr (Mux c et ef) v
    | eval_Mux_1 : forall c et ef v,
      eval_mux_iexpr c (INz 1) ->
      eval_mux_iexpr et v ->
      eval_mux_iexpr (Mux c et ef) v.

  End eval_iexpr.

  Definition iexpr_ty e : ideal_num_ty :=
    match e with
      | IEvar ty _ => ty
      | IEZconst _ => INTz
      | IEFconst _ => INTf
      | IEunop (IOneg | IOnot | IOzoffloat) _ => INTz
      | IEunop (IOnegf | IOabsf | IOsingleoffloat | IOfloatofz | IOsingleofz) _ => INTf
      | IEbinop (IOadd | IOsub | IOmul | IOdiv | IOmod | IOand | IOor | IOxor |
                 IOshl | IOshr | IOcmp _ | IOcmpf _) _ _ => INTz
      | IEbinop (IOaddf | IOsubf | IOmulf | IOdivf) _ _ => INTf
    end.

  Lemma iexpr_ty_correct:
    forall n e ρ, eval_iexpr ρ e n -> γ (iexpr_ty e) n.
  Proof.
    destruct 1.
    - destruct H0; constructor.
    - constructor.
    - constructor.
    - destruct H0; constructor.
    - destruct H1; constructor.
  Qed.

  Section ToString.

  Local Open Scope string.

  Context {VS: ToString var}.

  Definition string_of_ideal_num_ty (ty:ideal_num_ty) : string :=
    match ty with
      | INTz => "%z"
      | INTf => "%f"
    end.

  Definition string_of_unop (op:i_unary_operation) :=
    match op with
      | IOneg => "-"
      | IOnot => "~"
      | IOnegf => "-."
      | IOabsf => "absf"
      | IOsingleoffloat => "singleoffloat"
      | IOfloatofz => "floatofz"
      | IOsingleofz => "singleofz"
      | IOzoffloat => "zoffloat"
    end.

  Definition string_of_binop (op:i_binary_operation) :=
    match op with
      | IOadd => "+"
      | IOsub => "-"
      | IOmul => "*"
      | IOdiv => "/"
      | IOmod => "%"
      | IOand => "&"
      | IOor => "|"
      | IOxor => "^"
      | IOshl => "<<"
      | IOshr => ">>"
      | IOcmp c => string_of_comparison c
      | IOaddf => "+."
      | IOsubf => "-."
      | IOmulf => "*."
      | IOdivf => "/."
      | IOcmpf c => string_of_comparison c
    end.

  Fixpoint string_of_iexpr (e: iexpr) {struct e} : string :=
    match e with
      | IEvar ty v => to_string v ++ string_of_ideal_num_ty ty
      | IEZconst itv => to_string itv
      | IEFconst itv => to_string itv
      | IEunop op e' => string_of_unop op ++ "(" ++ string_of_iexpr e' ++ ")"
      | IEbinop op x y =>
        "(" ++ string_of_iexpr x ++ ")" ++ string_of_binop op ++
            "(" ++ string_of_iexpr y ++ ")"
    end.

  Global Instance IExprToString : ToString iexpr := string_of_iexpr.

  Fixpoint string_of_muxiexpr (e: mux_iexpr) {struct e} : string :=
    match e with
      | Leaf e => to_string e
      | Mux c et ef => "(" ++ string_of_muxiexpr c ++ ") ? (" ++
                           string_of_muxiexpr et ++ ") : (" ++
                           string_of_muxiexpr ef ++ ")"
    end.

  Global Instance MuxIExprToString : ToString mux_iexpr := string_of_muxiexpr.

  End ToString.

End iexpr.

Hint Constructors eval_ibinop.
Hint Constructors eval_iunop.
Hint Constructors eval_iexpr.
Hint Constructors ideal_num_ty_gamma.