(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import ShareTree.
Require Import IdealExpr.
Require Import ZArith.
Require Import Coqlib ToString.
Require Import ZIntervals FloatIntervals.
Require Import AdomLib.

Inductive abs :=
| Az : ZIntervals.itv -> abs
| Af : FloatIntervals.itv -> abs.

Program Definition itv_num_plat : pre_lattice abs :=
  {| pl_leb x y :=
      match x, y return _ with
        | Az x, Az y => pl_leb ZIntervals.itv_pl x y
        | Af x, Af y => pl_leb FloatIntervals.itv_pl x y
        | Af _, Az _ | Az _, Af _ => false
      end;
    pl_join x y :=
      match x, y with
        | Az x', Az y' =>
          match pl_join ZIntervals.itv_pl x' y' return _ with
            | All => All
            | Just res =>
              Just (physEq x' res (fun _ => x) (fun _ =>
                      physEq y' res (fun _ => y) (fun _ => Az res) _) _)
          end
        | Af x', Af y' =>
          match pl_join FloatIntervals.itv_pl x' y' return _ with
            | All => All
            | Just res =>
              Just (physEq x' res (fun _ => x) (fun _ =>
                      physEq y' res (fun _ => y) (fun _ => Af res) _) _)
          end
        | Af _, Az _ | Az _, Af _ => All
      end;
    pl_widen x y :=
      match x, y with
        | Az x', Az y' =>
          match pl_widen ZIntervals.itv_pl x' y' return _ with
            | All => All
            | Just res =>
              Just (physEq x' res (fun _ => x) (fun _ =>
                      physEq y' res (fun _ => y) (fun _ => Az res) _) _)
          end
        | Af x', Af y' =>
          match pl_widen FloatIntervals.itv_pl x' y' return _ with
            | All => All
            | Just res =>
              Just (physEq x' res (fun _ => x) (fun _ =>
                      physEq y' res (fun _ => y) (fun _ => Af res) _) _)
          end
        | Af _, Az _ | Az _, Af _ => All
      end |}.

Instance itv_num_gamma : gamma_op abs ideal_num := fun v x =>
   match v, x with
     | Az v, INz x => γ v x
     | Af v, INf x => γ v x
     | _, _ => False
   end.

Lemma itv_num_spec : pre_lattice_spec itv_num_plat itv_num_gamma.
Proof.
  split.
  - intros. destruct x, y, b; try easy; unfold γ in *; simpl in *.
    eapply pl_gamma_monotone; eauto using ZIntervals.itv_pls.
    eapply pl_gamma_monotone; eauto using FloatIntervals.itv_pls.
  - destruct x, y, a; try now (simpl; intuition).
    intros.
    red; red; red; red. unfold physEq.
    pose proof pl_join_sound ZIntervals.itv_pls i i0 H.
    destruct pl_join. auto. auto.
    intros.
    red; red; red; red. unfold physEq.
    pose proof pl_join_sound FloatIntervals.itv_pls i i0 H.
    destruct pl_join. auto. auto.
Qed.

Definition in_z v :=
  match v with
    | Just (Az v) => NotBot (Just v)
    | Just (Af v) => Bot
    | All => NotBot All
  end.

Lemma in_z_correct:
  ∀ ab x, γ ab (INz x) -> γ (in_z ab) x.
Proof.
  destruct ab as [|[]]; simpl; auto.
Qed.

Hint Resolve in_z_correct.

Definition in_f v :=
  match v with
    | Just (Af v) => NotBot (Just v)
    | Just (Az v) => Bot
    | All => NotBot All
  end.

Lemma in_f_correct:
  ∀ ab x, γ ab (INf x) -> γ (in_f ab) x.
Proof.
  destruct ab as [|[]]; simpl; auto.
Qed.

Hint Resolve in_f_correct.

Instance AbsToString : ToString abs := λ i,
  match i with
    | Az j => to_string j
    | Af j => to_string j
  end.
