(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import Utf8.
Require Import IdealExpr.
Require Import AbIdealNonrel.
Require Import ZArith.
Require Import Coqlib.
Require Import ZIntervals FloatIntervals FloatIntervalsBackward IdealIntervals.
Require Import Integers.
Require Import AdomLib.
Require Import AbIdealEnv IdealBoxDomain ShareTree Util.

Definition forward_binop (op:i_binary_operation) (x y:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOadd =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match x, y with
        | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV my My) =>
          NotBot (Just (Az (ZIntervals.ITV (mx+my) (Mx+My))))
        | _, _ => NotBot All
      end
    | IOsub =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match x, y with
        | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV my My) =>
          NotBot (Just (Az (ZIntervals.ITV (mx-My) (Mx-my))))
        | _, _ => NotBot All
      end
    | IOmul => botlift_fun1 (toplift_lift Az) (botbind2 forward_mult (in_z x) (in_z y))
    | IOdiv => botlift_fun1 (toplift_lift Az) (botbind2 forward_div (in_z x) (in_z y))
    | IOmod => botlift_fun1 (toplift_lift Az) (botbind2 forward_mod (in_z x) (in_z y))

    | IOand => botlift_fun1 (toplift_lift Az) (botbind2 forward_land (in_z x) (in_z y))
    | IOor => botlift_fun1 (toplift_lift Az) (botbind2 forward_lor (in_z x) (in_z y))
    | IOxor => botlift_fun1 (toplift_lift Az) (botbind2 forward_lxor (in_z x) (in_z y))
    | IOshl => botlift_fun1 (toplift_lift Az) (botbind2 forward_shl (in_z x) (in_z y))
    | IOshr => botlift_fun1 (toplift_lift Az) (botbind2 forward_shr (in_z x) (in_z y))

    | IOcmp c =>
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      let hasOne := is_bot (backward_cmp_l c x y) in
      let hasZero := is_bot (backward_cmp_l (negate_comparison c) x y) in
      match hasOne, hasZero with
        | false, false => NotBot (Just (Az (ZIntervals.ITV 0 1)))
        | true, false => NotBot (Just (Az (ZIntervals.ITV 0 0)))
        | false, true => NotBot (Just (Az (ZIntervals.ITV 1 1)))
        | true, true => Bot
      end
    | IOcmpf c =>
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      let hasOne := is_bot (backward_cmpf_l c x y true) in
      let hasZero := is_bot (backward_cmpf_l c x y false) in
      match hasOne, hasZero with
        | false, false => NotBot (Just (Az (ZIntervals.ITV 0 1)))
        | true, false => NotBot (Just (Az (ZIntervals.ITV 0 0)))
        | false, true => NotBot (Just (Az (ZIntervals.ITV 1 1)))
        | true, true => Bot
      end

    | IOaddf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fadd (in_f x) (in_f y))
    | IOsubf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fsub (in_f x) (in_f y))
    | IOmulf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fmult (in_f x) (in_f y))
    | IOdivf => botlift_fun1 (toplift_lift Af) (botbind2 forward_fdiv (in_f x) (in_f y))
  end.

Lemma forward_binop_correct:
  ∀ op,
  ∀ x x_ab, x ∈ γ x_ab ->
  ∀ y y_ab, y ∈ γ y_ab ->
  eval_ibinop op x y ⊆ γ (forward_binop op x_ab y_ab).
Proof.
  intros [] x x_ab Hx y y_ab Hy z EVAL; try reflexivity; inversion EVAL; clear EVAL; subst;
    try eapply botlift_fun1_spec; eauto; intros;
    try eapply toplift_lift_spec; eauto;
    try eapply botbind2_spec; eauto; intros.

  - destruct x_ab0 as [|[]], y_ab0 as [|[]]; try easy.
    destruct H, H0; split; omega.
  - destruct x_ab0 as [|[]], y_ab0 as [|[]]; try easy.
    destruct H, H0; split; omega.
  - auto using forward_mult_correct.
  - auto using forward_div_correct.
  - auto using forward_mod_correct.
  - auto using forward_land_correct.
  - auto using forward_lor_correct.
  - auto using forward_lxor_correct.
  - auto using forward_shl_correct.
  - auto using forward_shr_correct.
  - destruct (Zcmp c i1 i2) eqn:?.
    erewrite (is_bot_spec _ (backward_cmp_l c x_ab0 y_ab0)).
    destruct @is_bot; split; omega.
    eapply backward_cmp_l_correct; eauto.
    erewrite (is_bot_spec _ (backward_cmp_l (negate_comparison c) x_ab0 y_ab0)).
    destruct @is_bot; split; omega.
    eapply backward_cmp_l_correct; eauto.
    destruct c; simpl in Heqb |- *; destruct Z.compare; congruence.
  - auto using forward_fadd_correct.
  - auto using forward_fsub_correct.
  - auto using forward_fmult_correct.
  - auto using forward_fdiv_correct.
  - destruct (Floats.Float.cmp c f1 f2) eqn:?.
    erewrite (is_bot_spec _ (backward_cmpf_l c x_ab0 y_ab0 true)).
    destruct @is_bot; split; omega.
    eapply backward_cmpf_l_correct; eauto.
    erewrite (is_bot_spec _ (backward_cmpf_l c x_ab0 y_ab0 false)).
    destruct @is_bot; split; omega.
    eapply backward_cmpf_l_correct; eauto.
Qed.

Definition forward_unop (op:i_unary_operation) (x:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOneg => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_neg (in_z x))
    | IOnot => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_lnot (in_z x))
    | IOnegf => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_fneg (in_f x))
    | IOabsf => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_fabs (in_f x))
    | IOsingleoffloat => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_singleoffloat (in_f x))
    | IOfloatofz => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_floatofz (in_z x))
    | IOsingleofz => botlift_fun1 (toplift_lift Af) (botlift_fun1 forward_singleofz (in_z x))
    | IOzoffloat => botlift_fun1 (toplift_lift Az) (botlift_fun1 forward_zoffloat (in_f x))
  end.

Lemma forward_unop_correct:
  ∀ op,
  ∀ x x_ab, x ∈ γ x_ab ->
  eval_iunop op x ⊆ γ (forward_unop op x_ab).
Proof.
  intros [] x x_ab Hx z EVAL; inversion EVAL; clear EVAL; subst;
    eapply botlift_fun1_spec; eauto; intros;
  try eapply toplift_lift_spec; eauto; intros;
  eapply botlift_fun1_spec; eauto; intros.
  - apply forward_neg_correct; auto.
  - apply forward_lnot_correct; auto.
  - apply forward_fneg_correct; auto.
  - apply forward_fabs_correct; auto.
  - apply forward_singleoffloat_correct; auto.
  - apply forward_floatofz_correct; auto.
  - apply forward_singleofz_correct; auto.
  - intros. eapply forward_zoffloat_correct; eauto.
Qed.

Definition backward_binop (op:i_binary_operation) (res x y:abs+⊤) : (abs+⊤*abs+⊤)+⊥ :=
  match op with
    | IOadd =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <-
        match y, res with
          | Just (ZIntervals.ITV my My), Just (ZIntervals.ITV mres Mres) =>
            ZIntervals.meet_top x (Just (ZIntervals.ITV (mres - My) (Mres - my)))
          | _, _ => NotBot All
        end;
      do_bot y' <-
        match x, res with
          | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV mres Mres) =>
            ZIntervals.meet_top y (Just (ZIntervals.ITV (mres - Mx) (Mres - mx)))
          | _, _ => NotBot All
        end;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOsub =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <-
        match y, res with
          | Just (ZIntervals.ITV my My), Just (ZIntervals.ITV mres Mres) =>
            ZIntervals.meet_top x (Just (ZIntervals.ITV (mres + my) (Mres + My)))
          | _, _ => NotBot All
        end;
      do_bot y' <-
        match x, res with
          | Just (ZIntervals.ITV mx Mx), Just (ZIntervals.ITV mres Mres) =>
            ZIntervals.meet_top y (Just (ZIntervals.ITV (mx - Mres) (Mx - mres)))
          | _, _ => NotBot All
        end;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOmul =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      do_bot x' <- backward_mul_l res x y;
      do_bot y' <- backward_mul_l res y x;
      NotBot (toplift_lift Az x', toplift_lift Az y')
    | IOcmp c =>
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      match res with
        | Just (ZIntervals.ITV m M) =>
          match (Zlt_bool 1 m || Zlt_bool M 1)%bool,
                (Zlt_bool 0 m || Zlt_bool M 0)%bool with
            | false, false => NotBot (toplift_lift Az x, toplift_lift Az y)
            | false, true =>
              do_bot x' <- backward_cmp_l c x y;
              do_bot y' <- backward_cmp_l (swap_comparison c) y x;
              NotBot (toplift_lift Az x', toplift_lift Az y')
            | true, false =>
              do_bot x' <- backward_cmp_l (negate_comparison c) x y;
              do_bot y' <- backward_cmp_l (swap_comparison (negate_comparison c)) y x;
              NotBot (toplift_lift Az x', toplift_lift Az y')
            | true, true => Bot
          end
        | All =>
          NotBot (toplift_lift Az x, toplift_lift Az y)
      end
    | IOdiv | IOmod | IOand | IOor | IOxor | IOshl | IOshr =>
      (* TODO *)
      do_bot res <- in_z res;
      do_bot x <- in_z x;
      do_bot y <- in_z y;
      NotBot (toplift_lift Az x, toplift_lift Az y)
    | IOaddf =>
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      do_bot x' <- backward_addf_l res x y;
      do_bot y' <- backward_addf_r res x y;
      NotBot (toplift_lift Af x', toplift_lift Af y')
    | IOsubf =>
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      do_bot x' <- backward_subf_l res x y;
      do_bot y' <- backward_subf_r res x y;
      NotBot (toplift_lift Af x', toplift_lift Af y')
    | IOmulf | IOdivf =>
      (* TODO *)
      do_bot res <- in_f res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      NotBot (toplift_lift Af x, toplift_lift Af y)
    | IOcmpf c =>
      do_bot res <- in_z res;
      do_bot x <- in_f x;
      do_bot y <- in_f y;
      match res with
        | Just (ZIntervals.ITV m M) =>
          match (Zlt_bool 1 m || Zlt_bool M 1)%bool,
                (Zlt_bool 0 m || Zlt_bool M 0)%bool with
            | false, false => NotBot (toplift_lift Af x, toplift_lift Af y)
            | false, true =>
              do_bot x' <- backward_cmpf_l c x y true;
              do_bot y' <- backward_cmpf_l (swap_comparison c) y x true;
              NotBot (toplift_lift Af x', toplift_lift Af y')
            | true, false =>
              do_bot x' <- backward_cmpf_l c x y false;
              do_bot y' <- backward_cmpf_l (swap_comparison c) y x false;
              NotBot (toplift_lift Af x', toplift_lift Af y')
            | true, true => Bot
          end
        | All =>
          NotBot (toplift_lift Af x, toplift_lift Af y)
      end
  end.

Local Arguments γ {A} {B} {gamma_op} a b /.
Lemma backward_binop_correct:
  ∀ op,
    ∀ x x_ab, x ∈ γ x_ab ->
    ∀ y y_ab, y ∈ γ y_ab ->
    ∀ res res_ab, res ∈ γ res_ab ->
      eval_ibinop op x y res ->
      (x, y) ∈ γ (backward_binop op res_ab x_ab y_ab).
Proof.
Opaque backward_subf_r.
  intros op x x_ab Hx y y_ab Hy res res_ab Hres EVAL.
  destruct op;
    inv EVAL;
    try (eapply botbind_spec; eauto; intros);
    try (eapply botbind2_spec; eauto; intros);
    try (eapply botbind2_spec; eauto; intros);
    try (split; eapply toplift_lift_spec; now eauto).
  - destruct y_ab0 as [|[]], a as [|[]]; try easy.
    eapply ZIntervals.meet_top_correct.
    split. auto. simpl in *. omega.
  - destruct x_ab0 as [|[]], a as [|[]]; try easy.
    eapply ZIntervals.meet_top_correct.
    split. auto. simpl in *. omega.
  - destruct y_ab0 as [|[]], a as [|[]]; try easy.
    eapply ZIntervals.meet_top_correct.
    split. auto. simpl in *. omega.
  - destruct x_ab0 as [|[]], a as [|[]]; try easy.
    eapply ZIntervals.meet_top_correct.
    split. auto. simpl in *. omega.
  - eapply backward_mul_l_correct; eauto.
  - eapply backward_mul_l_correct; eauto.
    rewrite Zmult_comm. auto.
  - destruct a as [|[]]. split; eapply toplift_lift_spec; eauto.
    assert (Zlt_bool 1 z || Zlt_bool z0 1 = true -> Zcmp c i1 i2 = false)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Zcmp. omega. auto. }
    assert (Zlt_bool 0 z || Zlt_bool z0 0 = true -> Zcmp c i1 i2 = true)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Zcmp. auto. omega. }
    destruct (Zlt_bool 1 z || Zlt_bool z0 1)%bool eqn:Hres1,
             (Zlt_bool 0 z || Zlt_bool z0 0)%bool eqn:Hres0;
      (specialize (H2 eq_refl) || clear H2); (specialize (H3 eq_refl) || clear H3).
    + congruence.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; eauto.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (i1 ?= i2); congruence.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (Zcompare_spec i1 i2), (Zcompare_spec i2 i1);
        try congruence; exfalso; omega.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; eauto.
      eapply backward_cmp_l_correct; eauto.
      eapply backward_cmp_l_correct; eauto.
      destruct c; simpl in *; destruct (Zcompare_spec i1 i2), (Zcompare_spec i2 i1);
        try congruence; exfalso; omega.
    + split; eapply toplift_lift_spec; eauto.
  - eapply backward_addf_l_correct; eauto.
  - eapply backward_addf_r_correct; eauto.
  - eapply backward_subf_l_correct; eauto.
  - eapply backward_subf_r_correct; eauto.
  - destruct a as [|[]]. split; eapply toplift_lift_spec; eauto.
    assert (Zlt_bool 1 z || Zlt_bool z0 1 = true -> Floats.Float.cmp c f1 f2 = false)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Floats.Float.cmp. omega. auto. }
    assert (Zlt_bool 0 z || Zlt_bool z0 0 = true -> Floats.Float.cmp c f1 f2 = true)%bool.
    { rewrite Bool.orb_true_iff, <- !Zlt_is_lt_bool. simpl in H. destruct Floats.Float.cmp. auto. omega. }
    destruct (Zlt_bool 1 z || Zlt_bool z0 1)%bool eqn:Hres1,
             (Zlt_bool 0 z || Zlt_bool z0 0)%bool eqn:Hres0;
      (specialize (H2 eq_refl) || clear H2); (specialize (H3 eq_refl) || clear H3).
    + congruence.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; eauto.
      eapply backward_cmpf_l_correct; eauto.
      eapply backward_cmpf_l_correct; eauto.
      rewrite Floats.Float.cmp_swap. auto.
    + eapply botbind2_spec. split; eapply toplift_lift_spec; eauto.
      eapply backward_cmpf_l_correct; eauto.
      eapply backward_cmpf_l_correct; eauto.
      rewrite Floats.Float.cmp_swap. auto.
    + split; eapply toplift_lift_spec; eauto.
Qed.

Definition backward_unop (op:i_unary_operation) (res x:abs+⊤) : abs+⊤+⊥ :=
  match op with
    | IOneg =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Az) (ZIntervals.meet_top x (forward_neg res)))
        (in_z x) (in_z res)
    | IOnot =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Az) (ZIntervals.meet_top x (forward_lnot res)))
        (in_z x) (in_z res)
    | IOnegf =>
      botbind2 (fun x res =>
        botlift_fun1 (toplift_lift Af) (FloatIntervals.meet_top x (forward_fneg res)))
        (in_f x) (in_f res)

    | IOabsf =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_absf (in_f res) (in_f x))
    | IOsingleoffloat =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_singleoffloat (in_f res) (in_f x))
    | IOzoffloat =>
      botlift_fun1 (toplift_lift Af) (botbind2 backward_zoffloat (in_z res) (in_f x))
    | IOfloatofz =>
      botlift_fun1 (toplift_lift Az) (botbind2 backward_floatofz (in_f res) (in_z x))
    | IOsingleofz =>
      botlift_fun1 (toplift_lift Az) (botbind2 backward_singleofz (in_f res) (in_z x))
  end.

Lemma backward_unop_correct:
  ∀ op,
    ∀ x x_ab, x ∈ γ x_ab ->
    ∀ res res_ab, res ∈ γ res_ab ->
      eval_iunop op x res ->
      x ∈ γ (backward_unop op res_ab x_ab).
Proof.
  intros op x x_ab Hx res res_ab Hres EVAL.
  destruct op; inv EVAL;
    try (eapply botbind2_spec; eauto; intros);
    try (eapply botlift_fun1_spec; eauto; intros);
    try (eapply toplift_lift_spec; eauto).
  - eapply ZIntervals.meet_top_correct. split. auto.
    rewrite <- Z.opp_involutive. apply forward_neg_correct; auto.
  - eapply ZIntervals.meet_top_correct. split. auto.
    rewrite <- Z.lnot_involutive. apply forward_lnot_correct; auto.
  - eapply FloatIntervals.meet_top_correct. split. auto.
    apply forward_fneg_correct in H0.
    destruct f as [[]|[]|[]|[]]; easy.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_absf_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_singleoffloat_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_floatofz_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_singleofz_correct; auto.
  - eapply botbind2_spec; eauto.
    intros.
    eapply backward_zoffloat_correct; eauto.
Qed.

Definition pl_spec : pre_lattice_spec itv_num_plat itv_num_gamma.
Proof.
  constructor.
  - destruct x, y; try discriminate. intros. simpl in H.
    + destruct b. destruct H0.
      simpl in H0. eapply pl_gamma_monotone in H0. eauto.
      apply ZIntervals.itv_pls. auto.
    + intros. destruct b. 2:destruct H0.
      simpl in H0. eapply pl_gamma_monotone in H0. eauto.
      apply itv_pls. auto.
  - intros.
    destruct x, y, a, H; (try now elim H); try exact I;
    (apply (pl_join_sound ZIntervals.itv_pls) ||
     apply (pl_join_sound itv_pls)); auto.
Qed.

Instance itv_dom : ab_ideal_nonrel abs :=
  { as_pl := itv_num_plat
  ; meet x y :=
      match x, y with
        | Just (Az x), Just (Az y) => botlift_fun1 (@Just _ ∘ Az) (ZIntervals.meet x y)
        | Just (Af x), Just (Af y) => botlift_fun1 (@Just _ ∘ Af) (FloatIntervals.meet x y)
        | Just (Af _), Just (Az _) | Just (Az _), Just (Af _) => Bot
        | All, x | x, All => NotBot x
      end
  ; z_const := @Just _ ∘ Az
  ; float_const := @Just _ ∘ Af
  ; forward_unop := forward_unop
  ; forward_binop := forward_binop
  ; backward_unop := backward_unop
  ; backward_binop := backward_binop

  ; forward_try_subst x :=
      match x with
        | All => true
        | Just _ => false
      end
  ; backward_try_subst res x :=
      (* TODO : should we do something else here ? *)
      false
 }.

Instance itv_dom_correct : ab_ideal_nonrel_correct itv_dom := {}.
Proof.
  - apply pl_spec.
  - destruct x as [|[]], y as [|[]], a; intuition (try easy); eapply botlift_fun1_spec;
    try (simpl; now eauto).
    apply ZIntervals.meet_correct; auto.
    apply FloatIntervals.meet_correct; auto.
  - auto.
  - auto.
  - apply forward_unop_correct.
  - apply forward_binop_correct.
  - apply backward_unop_correct.
  - apply backward_binop_correct.
Qed.

Module Make(T:SHARETREE).

  Module IBD := IdealBoxDomain.Make(T).

  Definition var := IBD.var.
  Definition t := @IBD.t abs.
  Existing Instance IBD.var_eq.

  Definition extract_inchan (chan:in_chan) (ab:IBD.t) : IBD.t*in_chan :=
    (ab,
     {| get_itv expr :=
          do_bot ab <- IBD.eval_expr IBD.eval_expr_fuel chan expr ab;
          do_bot ab' <- chan.(get_itv) expr;
          match ab return (abs+⊤)+⊥ with
            | All => NotBot ab'
            | Just ab =>
              match ab with
                | Az i1 =>
                  do_bot i2 <- in_z ab';
                  botlift_fun1 (toplift_lift Az) (ZIntervals.meet_top (Just i1) i2)
                | Af i1 =>
                  do_bot i2 <- in_f ab';
                  botlift_fun1 (toplift_lift Af) (FloatIntervals.meet_top (Just i1) i2)
              end
          end;
        get_congr := chan.(get_congr);
        get_eq_expr := chan.(get_eq_expr) |}).

  Instance gamm : gamma_op t (var -> ideal_num) := IBD.gamm.

  Lemma extract_inchan_correct :
    forall chan ab (ρ:var -> ideal_num),
      ρ ∈ γ chan -> ρ ∈ γ  ab -> ρ ∈ γ (extract_inchan chan ab).
  Proof.
    Opaque ZIntervals.meet_top FloatIntervals.meet_top IBD.eval_expr_fuel.
    split. auto. simpl. constructor.
    - intros.
      apply botbind_spec with (x:=a). intros.
      eapply botbind_spec. intros. 2:eapply get_itv_correct; eauto.
      destruct a0 as [|[]], a; eauto; try contradiction.
      eapply botbind_spec. intros. 2:apply in_z_correct, H3.
      eapply botlift_fun1_spec. intros. eapply toplift_lift_spec; eauto.
      eapply ZIntervals.meet_top_correct. auto.
      eapply botbind_spec. intros. 2:apply in_f_correct, H3.
      eapply botlift_fun1_spec. intros. eapply toplift_lift_spec; eauto.
      eapply FloatIntervals.meet_top_correct. auto.
      eapply IBD.eval_expr_correct in H1; eauto using itv_dom_correct.
    - intros. apply H. auto.
    - intros. apply H. auto.
  Qed.

  Lemma extract_inchan_redundant :
    forall chan ab,
      γ chan ∩ γ (fst (extract_inchan chan ab)) ⊆ γ (snd (extract_inchan chan ab)).
  Proof. intros. destruct H. edestruct extract_inchan_correct; eauto. Qed.

  Global Instance intervals_ideal_env : ab_ideal_env (var:=var) t :=
    IBD.ideal_box_domain itv_dom_correct _ _ _ _ extract_inchan_correct extract_inchan_redundant.
  Proof.
    abstract (destruct x; simpl; unfold physEq; f_equal; f_equal;
              (apply ZIntervals.join_eq || apply FloatIntervals.join_eq)).
    abstract (destruct x; simpl; unfold physEq;
              rewrite ?ZIntervals.widen_eq, ?FloatIntervals.widen_eq; auto).
    abstract (destruct x; simpl; rewrite ?ZIntervals.order_eq, ?FloatIntervals.order_eq; auto).
  Defined.

End Make.
