(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

(** Abstract domain for simple type reconstruction. *)

Require Import Coqlib Utf8 String.
Require Import Globalenvs.
Require Import Integers.
Require Import Floats.
Require Import Values.
Require Import AST.
Require Import ToString.
Require Import Util.
Require Import AdomLib.
Require Import Cminor.
Require Import PExpr.

(* trivial types analysis *)
Module AbTy.

(* sup-semi-lattice

         ( _TOP_ )
       /   |   |  \
     VIP  VL  VF  VS
    /   \
   VI    VP
*)

Inductive t := VI | VP | VIP | VL | VF | VS.

Instance toString : ToString t :=
  (λ x, match x with VI => "int" | VP => "ptr" | VIP => "intorptr"
                  | VL => "long" | VF => "float" | VS => "single" end)%string.

Definition leb (x y: t) : bool :=
  match x, y with
  | VI, VI
  | VP, VP
  | VIP, VIP
  | VL, VL
  | VF, VF
  | VS, VS
  | (VI | VP), VIP => true
  | _, _ => false
  end.

Definition join (x y: t) : t+⊤ :=
  match x, y with
    | VI, VI => Just VI
    | VP, VP => Just VP
    | (VIP|VP), (VIP|VI)
    | (VIP|VI), VP | VI, VIP => Just VIP
    | VL, VL => Just VL
    | VF, VF => Just VF
    | VS, VS => Just VS
    | _, _ => All
  end.


Definition plat : pre_lattice t :=
  {| pl_leb := leb
   ; pl_join := join
   ; pl_widen := join |}.

Instance gamma : gamma_op t val :=
  fun ab v =>
    match ab, v with
    | VI, Vint _
    | VP, Vptr _ _
    | VIP, (Vint _ | Vptr _ _)
    | VL, Vlong _
    | VF, Vfloat _
    | VS, Vsingle _ => True
    | _, _ => False
    end.

Lemma plat_correct : pre_lattice_spec plat gamma.
Proof. split. destruct x, y, b; intuition.
       destruct x, y, a; intuition.
Qed.

  Definition dom := adom_of_pre_lattice plat_correct.

  Definition eval_constant (cst: constant) : t+⊤ :=
    match cst with
      | Ointconst n => Just VI
      | Ofloatconst n => Just VF
      | Osingleconst n => Just VS
      | Olongconst n => Just VL
      | Oaddrsymbol s ofs => Just VP
  end.

  Definition eval_local (_: ident) : t+⊤ := Just VP.

  Definition eval_unop (op: unary_operation) (arg: t+⊤) : t+⊤ :=
    match op with
      | Ocast8unsigned | Ocast8signed | Ocast16unsigned | Ocast16signed
      | Onegint | Onotint | Ointoffloat | Ointuoffloat | Ointofsingle
      | Ointuofsingle | Ointoflong => Just VI
      | Onegf | Oabsf | Ofloatofsingle | Ofloatofint | Ofloatofintu
      | Ofloatoflong | Ofloatoflongu => Just VF
      | Onegl | Onotl | Olongofint | Olongofintu | Olongoffloat
      | Olonguoffloat | Olongofsingle | Olonguofsingle => Just VL
      | Onegfs | Oabsfs | Osingleoffloat | Osingleofint | Osingleofintu
      | Osingleoflong | Osingleoflongu => Just VS
  end.

  Definition eval_binop (op: binary_operation) (arg1 arg2: t+⊤): t+⊤ :=
    match op with
    | Oadd =>
      match arg1, arg2 with
      | Just VP, Just VI
      | Just VI, Just VP => Just VP
      | Just VI, Just VI => Just VI
      | _, _ => Just VIP
      end
    | Osub =>
      match arg1, arg2 with
      | Just VP, Just VI => Just VP
      | Just VP, Just VP
      | Just VI, Just VI => Just VI
      | _, _ => Just VIP
      end
    | Omul | Odiv | Odivu | Omod | Omodu | Oand | Oor | Oxor | Oshl
    | Oshr | Oshru | Ocmp _ | Ocmpu _ | Ocmpf _ | Ocmpfs _ | Ocmpl _ | Ocmplu _ => Just VI
    | Oaddf | Osubf | Omulf | Odivf => Just VF
    | Oaddfs | Osubfs | Omulfs | Odivfs => Just VS
    | Oaddl | Osubl | Omull | Odivl | Odivlu | Omodl | Omodlu | Oandl
    | Oorl | Oxorl | Oshll | Oshrl | Oshrlu => Just VL
    end.

  Lemma eval_constant_correct F V : forall (ge: Genv.t F V) c v,
    pexpr_eval_constant ge c = v ->
    v ≠ Vundef →
    γ (eval_constant c) v.
  Proof.
    unfold γ.
    destruct c; simpl; intros v Hv; inv Hv; try easy.
    destruct Genv.find_symbol. intro; exact I.
    intros K; elim (K eq_refl).
  Qed.

  Lemma eval_unop_correct : forall op arg x v,
    Cminor.eval_unop op arg = Some v ->
    v ≠ Vundef →
    γ x arg ->
    γ (eval_unop op x) v.
  Proof.
    unfold γ.
    destruct op; destruct arg; simpl; intros x v Hv U;
    try (destruct (Float.to_int f));
    try (destruct (Float.to_intu f));
    try (destruct (Float.to_long f));
    try (destruct (Float.to_longu f));
    try (destruct (Float32.to_int f));
    try (destruct (Float32.to_intu f));
    try (destruct (Float32.to_long f));
    try (destruct (Float32.to_longu f));
    simpl in Hv;
    eq_some_inv; subst;
    (try (intros _; exact I));
    congruence.
  Qed.

  Lemma gamma_Vint_inv {x : t+⊤} {i} : γ x (Vint i) →
    x = Just VI ∨ x = All ∨ x = Just VIP.
  Proof. destruct x as [|()]; intuition. Qed.

  Lemma gamma_VI_inv {v} :
    γ (Just VI) v → ∃ i, v = Vint i.
  Proof. destruct v; intros (). vauto. Qed.

  Lemma gamma_Vptr_inv {x : t+⊤} {b i} : γ x (Vptr b i) →
    x = Just VP ∨ x = All ∨ x = Just VIP.
  Proof. destruct x as [|()]; intuition. Qed.

  Lemma gamma_VP_inv {v} :
    γ (Just VP) v → ∃ b i, v = Vptr b i.
  Proof. destruct v; intros (); vauto. Qed.

  Lemma gamma_VIP_inv {v} :
    γ (Just VIP) v → (∃ b i, v = Vptr b i) ∨ (∃ i, v = Vint i).
  Proof. destruct v; intros (); vauto. Qed.

  Lemma gamma_Vfloat_inv {x : t+⊤} {f} : γ x (Vfloat f) → x = Just VF ∨ x = All.
  Proof. destruct x as [|()]; intuition. Qed.

  Lemma gamma_Vsingle_inv {x : t+⊤} {f} : γ x (Vsingle f) → x = Just VS ∨ x = All.
  Proof. destruct x as [|()]; intuition. Qed.

  Lemma gamma_VF_inv {v} :
    v ∈ γ(Just VF) → ∃ f, v = Vfloat f.
  Proof. destruct v; intros (); vauto. Qed.

  Lemma gamma_VS_inv {v} :
    v ∈ γ(Just VS) → ∃ f, v = Vsingle f.
  Proof. destruct v; intros (); vauto. Qed.

  Lemma gamma_Vlong_inv {x : t+⊤} {l} : γ x (Vlong l) → x = Just VL ∨ x = All.
  Proof. destruct x as [|()]; intuition. Qed.

  Lemma gamma_VL_inv {v} :
    v ∈ γ(Just VL) → ∃ l, v = Vlong l.
  Proof. destruct v; intros (); vauto. Qed.

  Lemma gamma_Vundef_inv {x: t+⊤} : γ x Vundef → x = All.
  Proof. destruct x as [|()]; intuition. Qed.

  Ltac gamma_inv :=
    repeat
      match goal with
      | H : γ _ (Vint _) |- _ => apply gamma_Vint_inv in H
      | H : γ (Just VI) _ |- _ => apply gamma_VI_inv in H
      | H : γ _ (Vptr _ _) |- _ => apply gamma_Vptr_inv in H
      | H : γ (Just VP) _ |- _ => apply gamma_VP_inv in H
      | H : γ (Just VIP) _ |- _ => apply gamma_VIP_inv in H
      | H : γ _ (Vfloat _) |- _ => apply gamma_Vfloat_inv in H
      | H : γ (Just VF) _ |- _ => apply gamma_VF_inv in H
      | H : γ _ (Vsingle _) |- _ => apply gamma_Vsingle_inv in H
      | H : γ (Just VS) _ |- _ => apply gamma_VS_inv in H
      | H : γ _ (Vlong _) |- _ => apply gamma_Vlong_inv in H
      | H : γ (Just VL) _ |- _ => apply gamma_VL_inv in H
      | H : γ _ Vundef |- _ => apply gamma_Vundef_inv in H
      end.

  Lemma  of_optbool_lemma: forall v,
   match Val.of_optbool v with
   | Vundef => True
   | Vint _ => True
   | Vlong _ => False
   | Vfloat _ => False
   | Vsingle _ => False
   | Vptr _ _ => False
   end.
  Proof.
    destruct v; simpl; auto.
    destruct b; simpl; auto.
  Qed.

  Lemma  of_bool_lemma: forall v,
   match Val.of_bool v with
   | Vundef => True
   | Vint _ => True
   | Vlong _ => False
   | Vfloat _ => False
   | Vsingle _ => False
   | Vptr _ _ => False
   end.
  Proof.
    destruct v; simpl; auto.
  Qed.

  Lemma opt_map_of_bool_inv b v :
    option_map Val.of_bool b = Some v → v = Vtrue ∨ v = Vfalse.
  Proof.
    destruct b as [()|]; simpl; intuition congruence.
  Qed.

  Lemma if_opt {A: Type} (b: bool) (a a': A) :
    (if b then Some a else None) = Some a' →
    a = a' ∧ b = true.
  Proof. bif; intros; eq_some_inv; vauto. Qed.

  Lemma eval_binop_correct : forall op arg1 arg2 x1 x2 m v,
    pexpr_eval_binop m op arg1 arg2 = Some v ->
    v ≠ Vundef →
    γ x1 arg1 ->
    γ x2 arg2 ->
    γ (eval_binop op x1 x2) v.
  Proof.
    intros [] [] [] x1 x2 m v Hv Hvdef Hx1 Hx2; simpl in Hv;
    symmetry in Hv;
    repeat match goal with
      | c: comparison |- _ =>
        destruct c;
        unfold Val.cmp, Val.cmpu, Val.cmplu, Val.cmpf, Val.cmpfs, Val.cmpl, Val.of_bool in Hv;
        simpl in Hv
      | Hv : Some v = Some _ |- _ => eq_some_inv
      | Hv : Some v = None |- _ => eq_some_inv
      | Hv : v = Vundef |- _ => now destruct (Hvdef Hv)
      | Hv : _ = if ?c then _ else _ |- _ =>
        destruct c eqn:?
    end;
    try (subst v; destruct x1 as [|[]], Hx1, x2 as [|[]], Hx2; exact I).
  Qed.

  Instance wlat : weak_lattice (t+⊤) := weak_lattice_of_pre_lattice plat.
  Definition eval_cond (b: t+⊤) (l r: (t+⊤)+⊥) : (t+⊤)+⊥ :=
    match b with
      | Just VI => l ⊔ r
      | Just VP => l
      | All | Just VIP => l ⊔ r
      | Just (VF|VS|VL) => Bot
    end.

  Hint Resolve lift_bot dom.
  Definition adom : Cell2Val.adom t.
    refine (
    Cell2Val.Build_adom
    eval_constant
    eval_local
    eval_unop
    eval_binop
    eval_cond

    plat_correct
    eval_constant_correct
    _
    eval_unop_correct
    eval_binop_correct
    _) .
  Proof.
  + easy.
  + unfold eval_cond.
    intros [|[| | | | |]] l r B bv H H0 K H1.
    apply join_sound; auto. destruct bv; auto.
    apply join_sound; auto. destruct bv; auto.
    destruct B; inv H; inv H0. easy.
    apply join_sound; auto. destruct bv; auto.
    inv H0; destruct H.
    inv H0; destruct H.
    inv H0; destruct H.
  Defined.

End AbTy.
