(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib Integers ArithLib AdomLib ToString.

Require Psatz.

Require Import
  ZIntervals
  AbIdealNonrel
  IdealExpr
  ZCongruences_defs
  AbIdealEnv IdealBoxDomain
  Util.

Local Coercion Z_of_N : N >-> Z.

(** * Abstract Operators. *)
Definition zc_const (z: Z) : zcongr := ZC z 0.

Definition zc_true : zcongr := zc_const 1.
Definition zc_false: zcongr := zc_const 0.
Definition zc_bool (b: bool) : zcongr := if b then zc_true else zc_false.

Definition zc_itv (lo hi: Z) : zcongr :=
  if Z.eqb lo hi
  then zc_const lo
  else zc_top.

Definition zc_add (plus: Z → Z → Z) (x y: zcongr) : zcongr :=
  let m := N.gcd x.(zc_mod) y.(zc_mod) in
  let r := plus x.(zc_rem) y.(zc_rem) in
  ZC (if m then r else r mod m) m.

Definition zc_mul_r (x: zcongr) (y: Z) : zcongr :=
  match y with
  | Z0 => zc_const 0
  | Zpos p => ZC (Zpos p * x.(zc_rem)) (Npos p * x.(zc_mod))
  | Zneg p =>
    if x.(zc_mod)
    then ZC (Zneg p * x.(zc_rem)) 0
    else let m : N := (Npos p * x.(zc_mod))%N in
         ZC ((Zneg p * x.(zc_rem)) mod m) m
  end.

Definition zc_mul (x y: zcongr) : zcongr :=
  if x.(zc_mod)
  then zc_mul_r y x.(zc_rem)
  else if y.(zc_mod)
       then zc_mul_r x y.(zc_rem)
       else zc_top.

Definition zc_quot_pos (x: zcongr) (y: positive) : zcongr :=
  if Zdivides_dec (Zpos y) x.(zc_rem) && Ndivides_dec (Npos y) x.(zc_mod)
  then {| zc_rem := x.(zc_rem) / Zpos y ; zc_mod := x.(zc_mod) / Npos y |}
  else zc_top.

Definition zc_div_pos (x: zcongr) (y: positive) : zcongr :=
  if Ndivides_dec (Npos y) x.(zc_mod)
  then {| zc_rem := x.(zc_rem) / Zpos y; zc_mod := x.(zc_mod) / Npos y |}
  else zc_top.

Definition zc_quot (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | Zpos y' => zc_quot_pos x y'
       | _ => zc_top
       end
  else zc_top.

Definition zc_shl (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | 0 => x
       | Zpos y' => zc_mul_r x (2 ^ Zpos y')
       | Zneg y' => zc_div_pos x (2 ^ y')
       end
  else zc_top.

Definition zc_shr (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | 0 => x
       | Zpos y' => zc_div_pos x (2 ^ y')
       | Zneg y' => zc_mul_r x (2 ^ Zpos y')
       end
  else zc_top.

Definition zc_remainder_pos (x: zcongr) (y: positive) : zcongr :=
  if x.(zc_mod)
  then zc_const (Z.rem x.(zc_rem) (Zpos y))
  else
    if Zdivides_dec (Zpos y) x.(zc_rem)
    && Ndivides_dec (Npos y) x.(zc_mod)
    then {| zc_rem := 0; zc_mod := N0 |}
    else zc_top.

Definition zc_remainder (x y: zcongr) : zcongr :=
  if y.(zc_mod)
  then match y.(zc_rem) with
       | Zpos y' => zc_remainder_pos x y'
       | _ => zc_top
       end
  else zc_top.

Definition zc_cmp_eq (b: bool) (x y: zcongr) : zcongr :=
  match x.(zc_mod), y.(zc_mod) with
  | N0, N0 => zc_bool (eqb b (Z.eqb x.(zc_rem) y.(zc_rem)))
  | N0, Npos m
  | Npos m, N0 => if congr_dec (Zpos m) x.(zc_rem) y.(zc_rem)
                  then zc_top
                  else zc_bool (negb b)
  | Npos x', Npos y' =>
    (* TODO *) zc_top
  end.

Definition zc_cmp (cmp: comparison) (x y: zcongr) : zcongr :=
  match cmp with
  | Ceq => zc_cmp_eq true x y
  | Cne => zc_cmp_eq false x y
  | _ => (* TODO *)
    zc_top
  end.

Definition zc_forward_unop op x :=
  match op with
  | IOneg => NotBot (Just (ZC (if x.(zc_mod)
                               then - x.(zc_rem)
                               else (- x.(zc_rem)) mod x.(zc_mod))
                              x.(zc_mod)))
  | IOnot => NotBot (Just (if x.(zc_mod)
                           then ZC (Z.lnot x.(zc_rem)) N0
                           else zc_top))
  | IOsingleoffloat
  | IOzoffloat => NotBot (Just zc_top)
  | IOnegf
  | IOabsf => Bot
  | IOfloatofz
  | IOsingleofz => NotBot All
  end.

Definition zc_forward_binop op x y :=
  match op with
  | IOadd => NotBot (Just (zc_add Z.add x y))
  | IOsub => NotBot (Just (zc_add Z.sub x y))
  | IOmul => NotBot (Just (zc_mul x y))
  | IOdiv => NotBot (Just (zc_quot x y))
  | IOshl => NotBot (Just (zc_shl x y))
  | IOshr => NotBot (Just (zc_shr x y))
  | IOmod => NotBot (Just (zc_remainder x y))
  | IOcmp cmp => NotBot (Just (zc_cmp cmp x y))
  | IOand   (* TODO *)
  | IOor    (* TODO *)
  | IOxor   (* TODO *)
    =>
    NotBot (Just zc_top)
  | IOcmpf _ => NotBot (Just zc_top) (* exact! *)
  | IOaddf
  | IOsubf
  | IOmulf
  | IOdivf => Bot
  end.

Definition zc_bw_cmp (cmp: comparison) (res: zcongr) (x y: zcongr+⊤) : (zcongr+⊤ * zcongr+⊤)+⊥ :=
  if res.(zc_mod)
  then let cmp := if Z_zerop res.(zc_rem) then negate_comparison cmp else cmp in
       match cmp with
       | Ceq =>
         match x, y with
         | Just x', Just y' => do_bot z <- zc_meet x' y'; NotBot (Just z, Just z)
         | Just x', All => NotBot (Just x', Just x')
         | All, Just y' => NotBot (Just y', Just y')
         | All, All => NotBot (All, All)
         end
       | _ => NotBot (x, y)
       end
  else NotBot (x, y).

(** * Soundness. *)

Lemma zc_bool_sound b : (if b then 1 else 0) ∈ γ(zc_bool b).
Proof. destruct b; exists 0; reflexivity. Qed.

Lemma zc_add_sound (plus: Z → Z → Z) (x y: zcongr) (i j: Z) :
  (∀ n u u' v v', congr n u u' → congr n v v' → congr n (plus u v) (plus u' v')) →
  γ x i → γ y j → γ (zc_add plus x y) (plus i j).
Proof.
  intros COMPAT.
  unfold zc_add.
  destruct x as [xr xm]. destruct y as [yr ym].
  intros [p ->] [q ->].
  unfold γ, zc_gamma. simpl.
  case_eq (N.gcd xm ym). rewrite N.gcd_eq_0. intros [-> ->].
  repeat rewrite Z.mul_0_r, Z.add_0_r. auto with congr.
  intros gcd G.
  eapply congr_trans. 2: apply congr_mod_compat. 2: reflexivity.
  rewrite <- G. clear G gcd.
  rewrite Ngcd_is_Zgcd.
  apply COMPAT.
  rewrite <- (Z.add_0_r i) at 1. apply congr_plus_compat.
  auto with congr. eapply congr_divide. 2: apply Zgcd_divides_l.
  exists p; ring.
  rewrite <- (Z.add_0_r j) at 1. apply congr_plus_compat.
  auto with congr. eapply congr_divide. 2: apply Zgcd_divides_r.
  exists q; ring.
Qed.

Lemma zc_mul_r_sound x y i :
  γ x i → γ (zc_mul_r x y) (i * y).
Proof.
  unfold zc_mul_r.
  destruct x as [xr xm]. intros [q H]. simpl in H. subst.
  destruct y as [|y|y].
  exists 0; simpl; ring.
  exists q. unfold zc_rem, zc_mod. zify. ring.
  unfold zc_rem, zc_mod. destruct xm.
  exists q. unfold zc_rem, zc_mod. ring.
  unfold γ, zc_gamma, zc_rem, zc_mod. zify.
  eapply congr_trans. 2: apply congr_mod_compat; intuition.
  exists (-q). ring.
Qed.

Lemma zc_mul_sound x y i j :
  γ x i → γ y j → γ (zc_mul x y) (i * j).
Proof.
  unfold zc_mul.
  destruct x as [xr xm].
  simpl. destruct xm.
  intros H. apply congr_0_eq in H. subst. simpl. rewrite (Z.mul_comm xr j).
  apply zc_mul_r_sound.
  destruct y as [yr ym].
  simpl. destruct ym.
  intros K H. apply congr_0_eq in H. subst. simpl.
  apply zc_mul_r_sound; auto.
  exists (- i * j). simpl. ring.
Qed.

Local Hint Resolve (gamma_top _ _ _ _ zc_adom).

Lemma zc_quot_pos_sound x y i :
  γ x i → γ (zc_quot_pos x y) (i ÷ Zpos y).
Proof.
  unfold zc_quot_pos.
  destruct x as [xr xm]. intros [q H]. simpl in H. subst.
  simpl. destruct Zdivides_dec as [[x Hx]|]; auto.
  destruct Ndivides_dec as [[d H]|]; auto.
  simpl.
  exists q. simpl.
  assert (Z.pos y ≠ 0) as YNZ by discriminate.
  rewrite Hx.
  rewrite N2Z.inj_div. rewrite H in *. simpl in *.
  rewrite !Z_div_mult_full; auto.
  assert (i = (x - q * d) * Z.pos y). Psatz.lia.
  subst. rewrite Z.quot_mul; auto. ring.
Qed.

Lemma zc_div_pos_sound x y i :
  γ x i → γ (zc_div_pos x y) (i / Zpos y).
Proof.
  unfold zc_div_pos.
  destruct x as [xr xm]. intros [q H]. simpl in *. subst.
  destruct Ndivides_dec as [[d H]|]; auto.
  exists q. simpl.
  rewrite N2Z.inj_div, H. simpl.
  assert (Z.pos y ≠ 0) as YNZ by discriminate.
  rewrite Z_div_mult_full; auto.
  replace (q * (d * Zpos y)) with ((q * d) * Zpos y) by ring.
  rewrite Z.div_add; auto.
Qed.

Lemma zc_quot_sound x y i j :
  γ x i → γ y j → γ (zc_quot x y) (i ÷ j).
Proof.
  unfold zc_quot.
  destruct y as [yr ym].
  simpl. destruct ym; auto.
  destruct yr as [|y'|]; auto.
  intros K H. apply congr_0_eq in H. subst. simpl.
  apply zc_quot_pos_sound; auto.
Qed.

Lemma zc_shl_sound x y i j :
  γ x i → γ y j → γ (zc_shl x y) (Z.shiftl i j).
Proof.
  unfold zc_shl.
  destruct y as [yr ym]. destruct ym; auto; simpl.
  intros K H. apply congr_0_eq in H. subst. simpl.
  destruct yr as [|y'|y'].
  exact K.
  rewrite Z.shiftl_mul_pow2. apply zc_mul_r_sound; auto.
  discriminate.
  rewrite Z.shiftl_div_pow2.
  change (- Z.neg y') with (Z.pos y').
  rewrite <- Pos2Z.inj_pow.
  apply zc_div_pos_sound; auto.
  discriminate.
Qed.

Lemma zc_shr_sound x y i j :
  γ x i → γ y j → γ (zc_shr x y) (Z.shiftr i j).
Proof.
  unfold zc_shr.
  destruct y as [yr ym]. destruct ym; auto.
  intros K H. apply congr_0_eq in H. subst.
  unfold zc_mod, zc_rem.
  destruct yr as [|y'|y'].
  exact K.
  rewrite Z.shiftr_div_pow2.
  rewrite <- Pos2Z.inj_pow. apply zc_div_pos_sound; auto.
  discriminate.
  rewrite Z.shiftr_mul_pow2.
  change (- Z.neg y') with (Z.pos y').
  rewrite <- Pos2Z.inj_pow.
  apply zc_mul_r_sound; auto.
  discriminate.
Qed.

Lemma zc_remainder_pos_sound x y i :
  γ x i → γ (zc_remainder_pos x y) (Z.rem i (Zpos y)).
Proof.
  unfold zc_remainder_pos.
  destruct x as [xr xm]. intros [q H]. simpl in H. subst.
  simpl.
  destruct xm as [|xm].
- exists 0. simpl. replace (i + q * 0) with i by ring. ring.
- destruct Zdivides_dec as [[d H]|]; auto.
  destruct Ndivides_dec as [[d' H']|]; auto.
  simpl in *.
  rewrite H' in H.
  replace i with ((d - q * d') * Z.pos y) by Psatz.lia.
  rewrite Z.rem_mul. exists 0; reflexivity. discriminate.
Qed.

Lemma zc_remainder_sound x y i j :
  γ x i → γ y j → γ (zc_remainder x y) (Z.rem i j).
Proof.
  unfold zc_remainder.
  destruct y as [yr ym].
  simpl. destruct ym; auto.
  destruct yr as [|y'|]; auto.
  intros K H. apply congr_0_eq in H. subst. simpl.
  apply zc_remainder_pos_sound; auto.
Qed.

Lemma zcmp_eq_spec i j : Zcmp Ceq i j = eqb true (i =? j).
Proof.
  simpl. rewrite Z.eqb_compare.
  generalize (Zcompare_elim (i = j) (i < j) (i > j) i j id id id).
  destruct (i ?= j); intuition.
Qed.

Lemma zcmp_ne_spec i j : Zcmp Cne i j = eqb false (i =? j).
Proof.
  simpl. rewrite Z.eqb_compare.
  generalize (Zcompare_elim (i = j) (i < j) (i > j) i j id id id).
  destruct (i ?= j); intuition.
Qed.

Lemma zc_cmp_eq_sound b x y i j :
  i ∈ γ(x) → j ∈ γ(y) → (if eqb b (i =? j) then 1 else 0) ∈ γ(zc_cmp_eq b x y).
Proof.
  unfold zc_cmp_eq.
  intros I J.
  destruct x as [xr [|x']];
  destruct y as [yr [|y']];
  try apply congr_0_eq in I;
  try apply congr_0_eq in J;
  simpl in *; subst.
- generalize (Z.eqb_spec xr yr). inversion 1; apply zc_bool_sound.
- destruct congr_dec as [|K]. auto.
  generalize (Z.eqb_spec xr j); inversion 1.
  elim K. destruct J as [q J]. simpl in J.
  exists q. congruence.
  destruct b; exists 0; reflexivity.
- destruct congr_dec as [|K]. auto.
  generalize (Z.eqb_spec i yr); inversion 1.
  elim K. destruct I as [q I]. simpl in I.
  exists (-q). subst. ring.
  destruct b; exists 0; reflexivity.
- auto.
Qed.

Lemma zc_cmp_sound c x y i j :
  γ x i → γ y j → γ (zc_cmp c x y) (if Zcmp c i j then 1 else 0).
Proof.
  intros I J.
  destruct c; auto.
  rewrite zcmp_eq_spec. apply zc_cmp_eq_sound; auto.
  rewrite zcmp_ne_spec. apply zc_cmp_eq_sound; auto.
Qed.

Lemma zc_meet_sound x y a :
  γ x a ∧ γ y a → γ (zc_meet x y) a.
Proof.
  unfold zc_meet.
  destruct x as [xr xm]; destruct y as [yr ym].
  intros [X Y]. unfold γ, zc_gamma, zc_rem, zc_mod in X, Y.
  simpl. destruct congr_dec as [C|NC].
  destruct rem_eqn_particular as (p & P).
  destruct xm as [|xm]. easy.
  destruct ym as [|ym]. easy.
  case_eq (N.gcd (Npos xm) (Npos ym)). easy.
  intros gcd GCD. rewrite <- GCD.
  eapply congr_trans. 2: eapply congr_mod_compat.
  generalize (rem_eqn_all_solutions xr yr (Npos xm) (Npos ym)).
  rewrite <- Nlcm_is_Zlcm. unfold N.lcm.
  intros H. eapply H; try now zify.
  split; auto with congr.
  fold (N.lcm (Npos xm) (Npos ym)).
  generalize (N.lcm_eq_0 (Npos xm) (Npos ym)). destruct N.lcm; Psatz.lia.
  apply NC; clear NC. rewrite Ngcd_is_Zgcd.
  apply congr_trans with a; eapply congr_divide; eauto.
  apply congr_sym; eauto.
  apply Z.gcd_divide_l.
  apply Z.gcd_divide_r.
Qed.

Lemma zc_forward_unop_sound op i a :
  γ a i → ∀ n, eval_iunop op (INz i) (INz n) → γ (zc_forward_unop op a) n.
Proof.
  destruct op; intros H n K; unfold γ; simpl; unfold γ; simpl; auto; inv K.
- (* neg *)
  destruct H as [q Q].
  rewrite Q. unfold γ, zc_gamma. simpl zc_rem; simpl zc_mod.
  destruct (a.(zc_mod)). ring_simplify. auto with congr.
  eapply congr_trans. 2: apply congr_mod_compat.
  apply congr_neg_compat.
  rewrite <- (Z.add_0_r i) at 1.
  apply congr_plus_compat. auto with congr.
  exists q. ring.
  reflexivity.
- (* not *)
  destruct a as [r m]. destruct m; auto.
  apply congr_0_eq in H. subst.
  exists 1. simpl. ring.
Qed.

Lemma zc_forward_binop_sound op :
  ∀ i a, γ a i →
  ∀ j b, γ b j →
  ∀ n, eval_ibinop op (INz i) (INz j) (INz n) → γ (zc_forward_binop op a b) n.
Proof.
  destruct op; intros i x X j y Y a H; unfold γ; simpl; auto;
  try (red; simpl; auto);
  inv H.
  apply zc_add_sound; auto. apply congr_plus_compat.
  apply zc_add_sound; auto. apply congr_minus_compat.
  apply zc_mul_sound; auto.
  apply zc_quot_sound; auto.
  apply zc_remainder_sound; auto.
  apply zc_shl_sound; auto.
  apply zc_shr_sound; auto.
  apply zc_cmp_sound; auto.
Qed.

Instance zc_ideal_non_rel : ab_ideal_nonrel zcongr :=
  { as_pl := {| pl_leb x y := zc_LE_dec x y
              ; pl_join x y := Just (zc_join x y)
              ; pl_widen x y := Just (zc_join x y) |}
  ; as_gamma ab x :=
      match x with INz z => z ∈ γ(ab) | INf _ => False end
  ; meet x y :=
      match x, y with
      | Just x', Just y' => botlift_fun1 (@Just _) (zc_meet x' y')
      | Just z, All | All, Just z => NotBot (Just z)
      | All, All => NotBot All
      end
  ; z_const cst :=
      match cst with
      | ITV lo hi => Just (zc_itv lo hi)
      end
  ; float_const cst := All
  ; forward_unop op ab :=
      match ab with
      | Just ab' => zc_forward_unop op ab'
      | All => NotBot All
      end
  ; forward_binop op l r :=
      match l, r with
      | Just l', Just r' => zc_forward_binop op l' r'
      | _, _ => NotBot All
      end
  ; backward_unop op res ab := NotBot ab
  ; backward_binop op res l r :=
      match op with
      | IOcmp cmp => match res with Just res => zc_bw_cmp cmp res l r | All => NotBot (l, r) end
      | _ => NotBot (l, r)
      end

  ; forward_try_subst x :=
      match x with
        | All => true
        | Just _ => false
      end
  ; backward_try_subst x y := false
  }.

Instance zc_ideal_non_rel_correct : ab_ideal_nonrel_correct zc_ideal_non_rel.
Proof.
  split.
- split.
  + simpl. intros x y LE [f|z] G. exact G.
    exact (gamma_monotone zc_adom LE z G).
  + simpl. intros x y [f|z]. intros [G|G]; exact G.
    red. red. apply (join_sound _ _ _ _ zc_adom).
- intros [|x] [|y]; try (intuition; fail).
  intros [f|z]. intros [H _]; exfalso; exact H.
  intros H. generalize (zc_meet_sound _ _ _ H).
  simpl. destruct zc_meet; auto.
- intros [hi lo] a H.
  simpl. unfold zc_itv.
  generalize (Z.eqb_eq hi lo). destruct (hi =? lo).
  intuition.
  assert (a = lo). hnf in H. intuition. subst.
  exists 0. simpl. ring.
  intros _. apply (gamma_top _ _ _ _ zc_adom).
- easy.
- intros op [f|z] [|ab]; try easy.
  simpl. intros G [f|n] EV.
  inv EV; exact I.
  generalize (zc_forward_unop_sound op z ab G n EV).
  destruct zc_forward_unop as [|[|]]; auto.
- intros op [f1|z1] [|ab1]; auto. inversion 1.
  intros G1.
  intros [f2|z2] [|ab2]; auto. inversion 1.
  intros G2.
  intros [f|z] EV. inv EV.
  simpl.
  generalize (zc_forward_binop_sound op z1 ab1 G1 z2 ab2 G2 z EV).
  destruct zc_forward_binop as [|[|]]; auto.
- easy.
- destruct op; try (split; easy).
  intros x x_ab H y y_ab H0 res [|(rl, rr)]. split; easy.
  destruct rr. 2: split; easy.
  simpl. unfold zc_bw_cmp. simpl.
  intros HR EV. inv EV. 
  apply congr_0_eq in HR.
  destruct Z_zerop as [Hz|Hnz]; destruct c; try (split; easy).
  simpl in *.
  assert (i2 = i1) as EQ.
  {
    generalize (Z.compare_spec i1 i2).
    destruct (i1 ?= i2); try congruence; inversion 1; intuition.
  }
  clear HR. subst.
  destruct x_ab as [|(xl,xr)]; destruct y_ab as [|(yl,yr)]; simpl.
  easy.
  split; auto.
  split; auto.
  generalize (zc_meet_sound (ZC xl xr) (ZC yl yr) i1 (conj H H0)).
  destruct zc_meet. exact id. split; easy.
  simpl in *.
  assert (i2 = i1) as EQ.
  {
    generalize (Z.compare_spec i1 i2).
    destruct (i1 ?= i2); try congruence; inversion 1; intuition.
  }
  clear HR. subst.
  destruct x_ab as [|(xl,xr)]; destruct y_ab as [|(yl,yr)]; simpl.
  easy.
  split; auto.
  split; auto.
  generalize (zc_meet_sound (ZC xl xr) (ZC yl yr) i1 (conj H H0)).
  destruct zc_meet. exact id. split; easy.
Qed.

Require Import ShareTree.

Module Make(T:SHARETREE).

  Module IBD := IdealBoxDomain.Make(T).

  Definition var := IBD.var.
  Definition t := @IBD.t zcongr.
  Existing Instance IBD.var_eq.

  Definition extract_inchan (chan:in_chan) (ab:IBD.t) : t*in_chan :=
    (ab,
     {| get_itv := chan.(get_itv);
        get_congr expr :=
          do_bot ab <- IBD.eval_expr IBD.eval_expr_fuel chan expr ab;
          do_bot ab' <- chan.(get_congr) expr;
          match ab with
            | All => NotBot ab'
            | Just ab => zc_meet ab ab'
          end;
        get_eq_expr := chan.(get_eq_expr) |}).

  Instance gamm : gamma_op t (var -> ideal_num) := IBD.gamm.

  Opaque IBD.eval_expr_fuel.

  Lemma extract_inchan_correct :
    forall chan ab (ρ:var -> ideal_num),
      ρ ∈ γ chan -> ρ ∈ γ  ab -> ρ ∈ γ (extract_inchan chan ab).
  Proof.
    split. auto. simpl. constructor.
    - intros. apply H. auto.
    - intros. simpl.
      apply botbind_spec with (INz a). intros.
      2:eapply (IBD.eval_expr_correct (Dval:=zc_ideal_non_rel)); eauto using zc_ideal_non_rel_correct.
      eapply botbind_spec. intros.
      2:eapply get_congr_correct; eauto.
      destruct a0. auto. apply zc_meet_sound. auto.
    - intros. apply H. auto.
  Qed.

  Lemma extract_inchan_redundant :
    forall chan ab,
      γ chan ∩ γ (fst (extract_inchan chan ab)) ⊆ γ (snd (extract_inchan chan ab)).
  Proof. intros. destruct H. edestruct extract_inchan_correct; eauto. Qed.

  Global Instance zcongr_ideal_env : ab_ideal_env (var:=var) t :=
    IBD.ideal_box_domain zc_ideal_non_rel_correct _ _ _ _ extract_inchan_correct extract_inchan_redundant.
  Proof.
    - simpl. intro. rewrite zc_join_eq. auto.
    - simpl. intro. rewrite zc_join_eq. auto.
    - simpl. intro. unfold zc_LE_dec.
      destruct Ndivides_dec. destruct congr_dec. auto.
      destruct n. apply congr_refl.
      destruct n. apply Z.divide_refl.
  Defined.

End Make.