(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib ShareTree Integers ArithLib AdomLib ToString.

Require Psatz.

Local Unset Elimination Schemes.
Local Coercion Z_of_N : N >-> Z.

(** * Data type. *)
Record zcongr := ZC { zc_rem : Z; zc_mod : N }.

Instance zc_to_string : ToString zcongr := λ zc, 
  ("≡ " ++ to_string zc.(zc_rem) ++
   " [" ++ to_string zc.(zc_mod) ++ "]")%string.

(** Expected invariant. *)
Definition zc_reduced (z: zcongr) : Prop :=
  z.(zc_mod) = N0 ∨ 0 <= z.(zc_rem) < z.(zc_mod).

(** * Partial order. *)
Definition zc_LE (x y: zcongr) : Prop :=
    ( y.(zc_mod) | x.(zc_mod) )
  ∧ ( x.(zc_rem) ≡ y.(zc_rem) [y.(zc_mod)] ).

Definition zc_LE_dec (x y: zcongr) : { zc_LE x y } + { ¬ zc_LE x y } :=
  match Ndivides_dec y.(zc_mod) x.(zc_mod) with
  | left D => match congr_dec y.(zc_mod) x.(zc_rem) y.(zc_rem) with
              | left C => left (conj D C)
              | right C => right (fun X => C (proj2 X))
              end
  | right D => right (fun X => D (proj1 X))
  end.

Definition zc_top : zcongr :=
  {| zc_rem := 0; zc_mod := 1 |}.

Definition zc_share a a' :=
  if Z.eqb a.(zc_rem) a'.(zc_rem) &&
     N.eqb a.(zc_mod) a'.(zc_mod) then a' else a.

Lemma zc_share_eq:
  forall a a', zc_share a a' = a.
Proof.
  unfold zc_share. intros.
  destruct (Z.eqb_spec a.(zc_rem) a'.(zc_rem)),
           (N.eqb_spec a.(zc_mod) a'.(zc_mod)); auto.
  destruct a, a'; simpl in e, e0; simpl; congruence.
Qed.

(** * Lattice Structure. *)
Definition zc_join (x y: zcongr) : zcongr :=
  if Z.eqb x.(zc_rem) y.(zc_rem) && N.eqb x.(zc_mod) y.(zc_mod) then x
  else
    let m : N := N.gcd (N.gcd x.(zc_mod) y.(zc_mod))
                       (Nabs_diff x.(zc_rem) y.(zc_rem)) in
    let res :=
        {| zc_rem :=
             match m with
               | N0 => x.(zc_rem)
               | _ => x.(zc_rem) mod m
             end
           ; zc_mod := m
        |}
    in
    zc_share (zc_share res x) y.

Lemma zc_join_eq:
  forall x, zc_join x x = x.
Proof.
  intros. unfold zc_join. rewrite Z.eqb_refl, N.eqb_refl. auto.
Qed.

Definition zc_meet (x y: zcongr) : zcongr+⊥ :=
  let gcd : N := N.gcd x.(zc_mod) y.(zc_mod) in
  match congr_dec gcd x.(zc_rem) y.(zc_rem) with
  | left H =>
    if x.(zc_mod) then NotBot x else
    if y.(zc_mod) then NotBot y else
    let m : N := (x.(zc_mod) * (y.(zc_mod) / gcd))%N in
    let (n, K) := rem_eqn_particular x.(zc_rem) y.(zc_rem) 
                                     x.(zc_mod) y.(zc_mod)
                                     (N2Z.is_nonneg _)
                                     (N2Z.is_nonneg _)
                                     (eq_ind _ (fun n : Z => _ ≡ _ [n]) H _ (Ngcd_is_Zgcd _ _))
    in NotBot {| zc_rem := n mod m; zc_mod := m |}
  | right _ => Bot
  end.

Instance zc_wlat : weak_lattice zcongr :=
  {| leb x y := zc_LE_dec x y
   ; top := zc_top
   ; join := zc_join
   ; widen := zc_join (* there is no infinite increasing chain *)
  |}.

Instance zc_gamma : gamma_op zcongr Z :=
  fun x z => z ≡ x.(zc_rem) [x.(zc_mod)].

Lemma zc_adom : adom zcongr Z zc_wlat zc_gamma.
Proof.
  split.
  - intros x y. simpl. destruct zc_LE_dec as [((a & A),(b & B))|].
    intros _ i.
    intros (q & Q). exists (q * a + b).
    rewrite B, Q, A. ring.
    discriminate.
  - intros i. exists (-i). simpl. ring.
  - intros [xr xm] [yr ym] i.
    simpl. unfold zc_join. simpl.
    destruct ((xr =? yr) && (xm =? ym)%N) eqn:?.
    rewrite Bool.andb_true_iff, Z.eqb_eq, N.eqb_eq in Heqb. destruct Heqb as [-> ->].
    now intuition.
    rewrite !zc_share_eq.
    unfold γ, zc_gamma. simpl.
    case_eq (N.gcd (N.gcd xm ym) (Nabs_diff xr yr)).
    repeat rewrite N.gcd_eq_0.
    intros ((-> & ->) & XY).
    assert (xr = yr).
    destruct (Nabs_diff_case xr yr) as [(A,B)|(A,B)];
      rewrite B in XY; clear B.
    destruct (Z_dec (yr - xr) 0) as [[|]|];
      try Psatz.lia. destruct (yr - xr); simpl in *; discriminate.
    destruct (Z_dec (xr - yr) 0) as [[|]|];
      try Psatz.lia. destruct (xr - yr); simpl in *; discriminate.
    subst. intuition.
    intros p H; rewrite <- H.
    intros [X|Y].
    eapply congr_trans.
    2: eapply congr_mod_compat.
    eapply congr_divide. eassumption.
    repeat rewrite Ngcd_is_Zgcd.
    eapply Zdivides_trans; apply Zgcd_divides_l.
    rewrite H; reflexivity.
    eapply congr_trans.
    2: eapply congr_mod_compat.
    2: rewrite H; reflexivity.
    repeat rewrite Ngcd_is_Zgcd.
    apply congr_trans with (xr + (yr - xr)).
    replace (xr + (yr - xr)) with yr by ring.
    eapply congr_divide. eassumption.
    eapply Zdivides_trans. apply Zgcd_divides_l. apply Zgcd_divides_r.
    eapply congr_divide.
    2: apply Zgcd_divides_r.
    destruct (Nabs_diff_case xr yr) as
        [(A,B)|(A,B)]; rewrite B; rewrite Z2N.id.
    exists (-1). ring. Psatz.lia.
    exists (1). ring. Psatz.lia.
Qed.