(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Bool ToString
  List Maps ShareTree Coqlib Integers AST Util.

(** * Definition of an abstract domain *)
Module Adom. (* Spurious module to work around an extraction bug. *)
Unset Elimination Schemes.
Class weak_lattice (A: Type) : Type := {
  leb: A -> A -> bool;
  top: A;
  join: A -> A -> A;
  widen: A -> A -> A
}.

Class gamma_op (A B: Type) : Type := γ : A -> ℘ B.

Record adom (A B:Type) (WL: weak_lattice A) (Gamma: gamma_op A B) : Prop := {
  gamma_monotone: forall {a1} {a2},
    leb a1 a2 = true -> γ a1 ⊆ γ a2;
  gamma_top: forall x, x ∈ γ top;
  join_sound: ∀ x y, γ x ∪ γ y ⊆ γ (join x y)
}.
Arguments gamma_monotone [A] [B] [WL] [Gamma] _ {a1} {a2} _ _ _.
End Adom.
Export Adom.

Notation "x ⊑ y" := (leb x y) (at level 39).
Notation "x ⊔ y" := (join x y) (at level 40).
Notation "x ∇ y" := (widen x y) (at level 40).
Notation "⊤" := top (at level 40).

(** * Union of a list of abstract values *)
Section union_list.
Context {A B C:Type}.
Variable f : C -> A.
Variable bot : A.
Variable union : A -> A -> A.
Variable gamma : A -> B -> Prop.
Variable union_correct : forall ab1 ab2 m,
  gamma ab1 m \/ gamma ab2 m -> gamma (union ab1 ab2) m.

Definition union_list (l: list A) : A :=
  match l with
  | nil => bot
  | a::q =>
    List.fold_left union q a
  end.

Lemma union_list_correct : forall l ab m,
  In ab l ->
  m ∈ gamma ab ->
  m ∈ gamma (union_list l).
Proof.
  unfold union_list.
  destruct l as [|a q]. intuition.
  revert q.
  refine (rev_ind _ _ _).
  intros ab m [->|()]. easy.
  intros b q HI.
  intros ab m.
  rewrite app_comm_cons, in_app, fold_left_app. simpl.
  intros [[->|H]|[->|()]] AB.
  apply union_correct. left. eapply HI; eauto. now left.
  apply union_correct. left. eapply HI; eauto. now right.
  apply union_correct. now right.
Qed.

Definition union_list_map (l: list C) : A :=
  match l with
  | nil => bot
  | a :: q => List.fold_left (λ u x, union u (f x)) q (f a)
  end.

Lemma union_list_map_correct l :
  union_list_map l = union_list (map f l).
Proof.
  unfold union_list_map, union_list.
  destruct l as [|a q]. reflexivity.
  induction q as [|b l IH] using rev_ind.
  reflexivity.
  simpl. rewrite fold_left_app, map_app, fold_left_app. simpl in *.
  now rewrite IH.
Qed.

End union_list.

Set Implicit Arguments.

(** * Adding a bottom element to an abstract domain *)
Inductive botlift (A:Type) : Type := Bot | NotBot (x:A).
Implicit Arguments Bot [A].
Notation "t +⊥" := (botlift t) (at level 39).

Definition bot {A: Type} (l: weak_lattice (A +⊥)) : A+⊥ := Bot.

Instance lift_bot_wl {A:Type} (l:weak_lattice A) : weak_lattice (A+⊥) :=
{ leb :=
    (fun x y =>
      match x, y with
        | Bot, _ => true
        | NotBot x, NotBot y => leb x y
        | _, _ => false
      end);
  top := (NotBot top);
  join :=
    (fun x y =>
      match x, y with
        | Bot, _ => y
        |_ , Bot => x
        | NotBot x, NotBot y => NotBot (join x y)
      end);
  widen :=
    (fun x y =>
      match x, y with
        | Bot, _ => y
        |_ , Bot => x
        | NotBot x, NotBot y => NotBot (widen x y)
      end)
}.

Instance lift_gamma (A B: Type) (G: gamma_op A B) : gamma_op (A+⊥) B :=
    (fun x =>
      match x with
      | Bot => fun _ => False
      | NotBot x => γ x
      end).

Definition lift_bot {A B:Type} WL G (l:adom A B WL G) : adom (A+⊥) B _ _.
Proof.
  split.
  (* gamma_monotone *)
  destruct a1; destruct a2; simpl; intuition; try congruence.
  apply (gamma_monotone l H); auto.
  (* gamma_top *)
  intros; simpl; apply gamma_top; auto.
  (* join sound *)
  intros [|x] [|y]; simpl; intuition; apply join_sound; auto.
Qed.

Definition botbind {A B} (f: A → B+⊥) (a: A+⊥) : B+⊥ :=
  match a with
    | Bot => Bot
    | NotBot x => f x
  end.

Notation "'do_bot' X <- A ; B" := (botbind (fun X => B) A)
 (at level 200, X ident, A at level 100, B at level 200).

Lemma botbind_spec {A A' B B':Type} {GA: gamma_op A A'} {GB: gamma_op B B'} :
  ∀ f:A→B+⊥, ∀ x y, (∀ a, x ∈ γ a -> y ∈ γ (f a)) ->
                     (∀ a, x ∈ γ a -> y ∈ γ (botbind f a)).
Proof (λ f x y H a,
       match a with
       | Bot => λ K, K
       | NotBot a' => λ K, H _ K
       end).

Definition botbind2 {A B C} (f: A → B → C+⊥) (a: A+⊥) (b: B+⊥) : C+⊥ :=
  botbind (fun x => botbind (f x) b) a.

Lemma botbind2_spec {A A' B B' C C':Type} {AD:gamma_op A A'} {BD:gamma_op B B'} {CD:gamma_op C C'}:
  ∀ f:A→B→C+⊥, ∀ x y z, (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (f x_ab y_ab)) ->
                        (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (botbind2 f x_ab y_ab)).
Proof.
  unfold botbind2. repeat (intros; eapply botbind_spec; eauto).
Qed.

Definition botlift_fun1 {A B:Type} (f:A->B) (x:A+⊥) : B+⊥ :=
  botbind (@NotBot _ ∘ f) x.

Lemma botlift_fun1_spec {A A' B B':Type} {AD:gamma_op A A'} {BD:gamma_op B B'}:
  ∀ f:A→B, ∀ x y, (∀ x_ab, x ∈ γ x_ab -> y ∈ γ (f x_ab)) ->
                  (∀ x_ab, x ∈ γ x_ab -> y ∈ γ (botlift_fun1 f x_ab)).
Proof.
  unfold botlift_fun1. intros. eapply botbind_spec, H0. apply H.
Qed.

Definition is_bot {A:Type} (v:A+⊥) : bool :=
  match v with
    | Bot => true
    | NotBot _ => false
  end.

Lemma is_bot_spec {A A':Type} {AD:gamma_op A A'} :
  forall x x_ab, x ∈ γ x_ab -> is_bot x_ab = false.
Proof. destruct x_ab. destruct 1. auto. Qed.

Instance bot_lift_toString A `{ToString A} : ToString (A+⊥) :=
  { to_string x := match x with
                   | Bot => "⊥"%string
                   | NotBot x' => to_string x'
                   end
  }.

Lemma union_list_bot A B WL G {AD: adom A B WL G} (z: A) (l: list (A+⊥)) :
  union_list (NotBot z) join l = Bot →
  l ≠ nil ∧
  ∀ x, In x l → x = Bot.
Proof.
  unfold union_list.
  destruct l as [|a l]. discriminate.
  induction l as [|b l IH] using rev_ind.
  simpl. intros ->. split. discriminate. intros [|x]; intuition discriminate.
  rewrite fold_left_app.
  intros H. split. discriminate.
  assert (∀ a b: A+⊥, a ⊔ b = Bot → a = Bot ∧ b = Bot) as K.
  intros [|] [|]; simpl; auto; discriminate.
  destruct (K _ _ H) as (U & ->).
  specialize (IH U).
  intros x [->|L]. now apply IH; left.
  rewrite in_app_iff in L. destruct L as [L|[->|()]].
  now apply IH; right.
  easy.
Qed.

(** * Adding a top element to a domain. *)
Inductive toplift (A: Type) := All | Just : A → toplift A.
Notation "x +⊤" := (toplift x) (at level 39).
Arguments All [A].
Arguments Just [A] v.

Lemma All_not_Just A : ∀ a : A, All = Just a → ∀ P, P.
Proof.
  exact (λ a H, match H in _ = x return if x then True else _ with eq_refl => I end).
Qed.

Instance toplift_gamma t B (GAMMA: gamma_op t B) : gamma_op (t+⊤) B :=
  λ x, match x with All => λ _, True | Just y => γ y end.

Definition toplift_bind {A B: Type} (f: A → B+⊤) (x: A+⊤) : B+⊤ :=
  match x with
  | All => All
  | Just x' => f x'
  end.

Notation "'do_top' X <- A ; B" := (toplift_bind (fun X => B) A)
 (at level 200, X ident, A at level 100, B at level 200).

Lemma toplift_bind_spec {A A' B B':Type} {GA: gamma_op A A'} {GB: gamma_op B B'} :
  ∀ f:A→B+⊤, ∀ x y, (∀ a, x ∈ γ a -> y ∈ γ (f a)) ->
                    (∀ a, x ∈ γ a -> y ∈ γ (toplift_bind f a)).
Proof. destruct a; eauto. Qed.

Ltac destruct_noitpo_in_goal :=
  match goal with [ |- context[match ?x with Just _ => _ | All => _ end]] => destruct x end.

Ltac destruct_noitpo_in_goal_eq :=
  match goal with [ |- context[match ?x with Just _ => _ | All => _ end]] => case_eq x end.

Definition toplift_lift {A B: Type} (f: A → B) : A+⊤ → B+⊤ :=
  toplift_bind (fun x => Just (f x)).

Lemma toplift_lift_spec {A A' B B':Type} {GA: gamma_op A A'} {GB: gamma_op B B'} :
  ∀ f:A→B, ∀ x y, (∀ a, x ∈ γ a -> y ∈ γ (f a)) ->
                  (∀ a, x ∈ γ a -> y ∈ γ (toplift_lift f a)).
Proof. intros. eapply toplift_bind_spec; eauto. Qed.

Definition toplift_bind2 {A B C: Type} (f: A → B → C+⊤) (x: A+⊤) (y: B+⊤) : C+⊤ :=
  match x, y with
  | All, _ | _, All => All
  | Just x', Just y' => f x' y'
  end.

Lemma toplift_bind2_spec {A A' B B' C C':Type} {AD:gamma_op A A'} {BD:gamma_op B B'} {CD:gamma_op C C'}:
  ∀ f:A→B→C+⊤, ∀ x y z, (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (f x_ab y_ab)) ->
                        (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (toplift_bind2 f x_ab y_ab)).
Proof. destruct x_ab, y_ab; eauto. Qed.

Definition toplift_lift2 {A B C: Type} (f: A → B → C) : A+⊤ → B+⊤ → C+⊤ :=
  toplift_bind2 (fun x y => Just (f x y)).

Lemma toplift_lift2_spec {A A' B B' C C':Type} {AD:gamma_op A A'} {BD:gamma_op B B'} {CD:gamma_op C C'}:
  ∀ f:A→B→C, ∀ x y z, (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (f x_ab y_ab)) ->
                      (∀ x_ab y_ab, x ∈ γ x_ab -> y ∈ γ y_ab -> z ∈ γ (toplift_lift2 f x_ab y_ab)).
Proof. intros. eapply toplift_bind2_spec; eauto. Qed.

Definition flat_top_leb {A} (beq: A → A → bool) (x y: A+⊤) : bool :=
  match x, y with
  | _, All => true
  | All, _ => false
  | Just x', Just y' => beq x' y'
  end.

Instance toplift_toString A `{ToString A} : ToString (A+⊤) :=
  { to_string x := match x with
                   | All => "⊤"%string
                   | Just y => to_string y
                   end
  }.

(** * Partial order with partial least upper bound. *)

Local Unset Elimination Schemes.
Record pre_lattice t :=
  { pl_leb  : t → t → bool
  ; pl_join : t → t → t+⊤
  ; pl_widen: t → t → t+⊤
  }.

Record pre_lattice_spec t B (PL: pre_lattice t) (GAMMA: gamma_op t B) : Prop :=
  { pl_gamma_monotone: ∀ x y : t, pl_leb PL x y = true →
                                  ∀ b : B, γ x b → γ y b
  ; pl_join_sound: ∀ x y: t, (γ x ∪ γ y) ⊆ γ (pl_join PL x y)
  }.

Instance weak_lattice_of_pre_lattice t (PL: pre_lattice t) : weak_lattice (t+⊤) :=
  let join f x y :=
      match x, y with
      | All, _
      | _, All => All
      | Just x', Just y' => f x' y'
      end in
  {| leb := flat_top_leb (pl_leb PL)
   ; top := All
   ; join := join (pl_join PL)
   ; widen:= join (pl_widen PL)
  |}.

Lemma adom_of_pre_lattice t B (PL: pre_lattice t) G (PLP: pre_lattice_spec PL G)
  : adom (t+⊤) B (weak_lattice_of_pre_lattice PL) (toplift_gamma G).
Proof.
  split.
+ intros [|a] [|b]; simpl; auto. discriminate. intros _ ? _. easy.
  apply (pl_gamma_monotone PLP).
+ easy.
+ intros [|a] [|b]; intros; try easy.
  apply pl_join_sound; auto.
Qed.

Definition flat_pre_lattice T (beq: T → T → bool ) : pre_lattice T :=
  {| pl_leb := beq
   ; pl_join x y := if beq x y then Just x else All
   ; pl_widen x y := if beq x y then Just x else All
  |}.

Lemma flat_pre_lattice_correct T B beq (BEQ: ∀ x y, beq x y = true → x = y)
      (G: gamma_op T B) : pre_lattice_spec (flat_pre_lattice beq) G.
Proof.
  split.
- intros x y H. rewrite (BEQ _ _ H). easy.
- intros x y a H. simpl. generalize (BEQ x y). destruct (beq x y). intuition. subst. easy. easy.
Qed.

(** * Direct product of two abstract domains *)
Module WProd.
Section gamm.
  Context {t1 t2 B: Type}.
  Context (G1:gamma_op t1 B) (G2:gamma_op t2 B).

  Definition A: Type := (t1*t2)%type.

  Instance gamma: gamma_op A B :=
    fun x =>
      γ (fst x) ∩ γ (snd x).

  Section wlat.
  Context WL1 (lat1 : adom t1 B WL1 G1).
  Context WL2 (lat2 : adom t2 B WL2 G2).

  Definition leb: A -> A -> bool :=
    fun x y =>
      let (x1,x2) := x in
      let (y1,y2) := y in
        (leb x1 y1
      &&
        leb x2 y2)%bool.

  Definition top: A :=
    (top, top).

  Definition join: A -> A -> A :=
    fun x y =>
      let (x1,x2) := x in
      let (y1,y2) := y in
        (join x1 y1,
          join x2 y2).

  Definition widen: A -> A -> A :=
    fun x y =>
      let (x1,x2) := x in
      let (y1,y2) := y in
        (widen x1 y1,
          widen x2 y2).

  Lemma  gamma_monotone: forall a1 a2,
    leb a1 a2 = true -> gamma a1 ⊆ gamma a2.
  Proof.
    destruct a1; destruct a2; simpl.
    rewrite Bool.andb_true_iff in *.
    destruct 1.
    intros b [T1 T2].
    split; eapply gamma_monotone; eauto.
  Qed.

  Lemma gamma_top: forall x, gamma top x.
  Proof.
    simpl; split; apply gamma_top; auto.
  Qed.

  Lemma join_sound: forall x y, gamma x ∪ gamma y ⊆ gamma (join x y).
  Proof.
    intros (a,b)(c,d) q.
    simpl. intros [[]|[]]; split; simpl; apply join_sound; auto.
  Qed.

  Instance make_wl : weak_lattice A :=
   {leb := leb;
    top := top;
    join := join;
    widen := widen
   }.

  Lemma make : adom A B _ gamma.
  Proof.
    split.
    exact gamma_monotone.
    exact gamma_top.
    exact join_sound.
  Qed.

End wlat.

End gamm.
End WProd.

Existing Instance WProd.make_wl.
Existing Instance WProd.gamma.

(** * Abstract domain of pairs of concrete elements. *)
Module WTensor.
Section gamm.
  Context {t1 t2 B1 B2: Type}.
  Context (G1:gamma_op t1 B1) (G2:gamma_op t2 B2).

  Definition A: Type := (t1*t2)%type.
  Definition B: Type := (B1*B2)%type.

  Instance gamma : gamma_op A B :=
    fun x y =>
      let (x1,x2) := x in
      let (y1,y2) := y in
      y1 ∈ γ x1 /\ y2 ∈ γ x2.

  Section wlat.

  Context WL1 (AD1 : adom t1 B1 WL1 G1) WL2 (AD2 : adom t2 B2 WL2 G2).

  Lemma gamma_monotone:
    forall a1 a2, leb a1 a2 = true -> gamma a1 ⊆ gamma a2.
  Proof.
    destruct a1; destruct a2; simpl.
    rewrite Bool.andb_true_iff in *.
    destruct 1.
    intros [b1 b2] [T1 T2].
    split; eapply gamma_monotone; eauto.
  Qed.

  Lemma gamma_top:
    forall x, gamma top x.
  Proof.
    intros [b1 b2]; split; apply gamma_top; auto.
  Qed.

  Lemma join_sound: forall x y, gamma x ∪ gamma y ⊆ gamma (join x y).
  Proof.
    intros (a,b)(c,d) (b1,b2). simpl.
    intuition; apply join_sound; intuition.
  Qed.

  Lemma make : adom A B _ gamma.
  Proof.
    split.
    exact gamma_monotone.
    exact gamma_top.
    exact join_sound.
  Qed.

  End wlat.

  Section plat.

  Context PL1 (HPL1:pre_lattice_spec PL1 G1) PL2 (HPL2:pre_lattice_spec PL2 G2).

  Definition make_pre_lattice : pre_lattice (t1 * t2) :=
    {| pl_leb x x' :=
         let '(a,b) := x in let '(a',b') := x' in
         pl_leb PL1 a a' && pl_leb PL2 b b'
       ; pl_join x x'  :=
         let '(a,b) := x in let '(a',b') := x' in
         do_top α <- pl_join PL1 a a';
         do_top β <- pl_join PL2 b b';
         Just (α, β)
       ; pl_widen x x' :=
         let '(a,b) := x in let '(a',b') := x' in
         do_top α <- pl_widen PL1 a a';
         do_top β <- pl_widen PL2 b b';
         Just (α, β)
    |}.

  Lemma make_pre_lattice_correct : pre_lattice_spec make_pre_lattice gamma.
  Proof.
    split.
    - intros (a,b) (a',b'). simpl. rewrite andb_true_iff.
      intros [H H'] (u,v) [K K'].
      split; eapply pl_gamma_monotone; eauto.
    - intros (a,b) (a',b') (α, β).
      generalize (@pl_join_sound _ _ _ _ HPL2 b b' β).
      generalize (@pl_join_sound _ _ _ _ HPL1 a a' α).
      simpl.
      destruct (pl_join PL1 a a'). easy.
      destruct (pl_join PL2 b b'). easy.
      intros ? ? [(?,?)|(?,?)]; split; intuition.
  Qed.

End plat.

End gamm.
End WTensor.

Existing Instance WTensor.gamma.

(** * Exponentiation of an abstract domain to an abstract domain of total environments *)
Module WPFun (T:SHARETREE).
Module P := Tree_Properties T.
Module SP := ShareTree_Properties T.
Section elt.

  Context (elt B: Type)
          (PL: pre_lattice elt)
          (elt_gamma: gamma_op elt B)
          (PLP: pre_lattice_spec PL elt_gamma)
          (H_join_eq: ∀ x, PL.(pl_join) x x = Just x)
          (H_widen_eq: ∀ x, PL.(pl_widen) x x = Just x)
          (H_leb_eq: ∀ x, PL.(pl_leb) x x = true).

  Definition t : Type := T.t elt.

  Definition get (p: T.elt) (x: t) : elt+⊤ :=
    match T.get p x with
    | Some r => Just r
    | None => All
    end.

  Definition set (x:t) (p: T.elt) (v: elt+⊤) : t :=
    match v with
    | All => T.remove p x
    | Just v' => T.set p v' x
    end.

  Lemma gsspec x p v p' :
    get p' (set x p v) = if T.elt_eq p' p then v else get p' x.
  Proof.
    unfold get.
    destruct v as [|v]; simpl.
    now rewrite T.grspec; destruct T.elt_eq.
    now rewrite T.gsspec; destruct T.elt_eq.
  Qed.

  Lemma gss x p v :
    get p (set x p v) = v.
  Proof. rewrite gsspec. destruct T.elt_eq; congruence. Qed.

  Lemma gso x p v p' :
    p ≠ p' →
    get p' (set x p v) = get p' x.
  Proof. rewrite gsspec. destruct T.elt_eq; congruence. Qed.

  Program Definition leb : t → t → bool :=
    λ x y,
    T.shforall2 (λ _ a b, match a, b return _ with
                            | _, None => true
                            | None, Some _ => false
                            | Some a', Some b' => pl_leb PL a' b'
                          end)
                _ x y.
  Next Obligation. destruct v; auto. Qed.

  Lemma leb_le (x y: t) :
    leb x y = true <->
    ∀ p : T.elt, flat_top_leb (pl_leb PL) (get p x) (get p y) = true.
  Proof.
    unfold leb, get, flat_top_leb.
    rewrite T.shforall2_correct.
    split; intros H p; specialize (H p);
    destruct (T.get p x);
    destruct (T.get p y); eauto;
    congruence.
  Qed.

  Program Definition join f (Hf:∀ x, f x x = Just x): t → t → t :=
    T.shcombine
      (λ _ u v, match u, v return _ with
                  | None, _
                  | _, None => None
                  | Some u', Some v' =>
                    match f u' v' return _ with
                      | Just r => Some r
                      | All => None
                    end
                end)
      _.
  Next Obligation. destruct v. rewrite Hf. auto. auto. Qed.

  Lemma get_join f Hf a b x :
    get x (join f Hf a b) = toplift_bind2 f (get x a) (get x b).
  Proof.
    unfold get, join.
    rewrite T.gshcombine.
    destruct (T.get x a) as [u|].
    destruct (T.get x b) as [v|].
    simpl. destruct (f u v); reflexivity.
    reflexivity.
    reflexivity.
  Qed.

  Lemma join_monotone f Hf x a b :
    (∀ u v, pl_leb PL u v = true →
            (∀ w v', f v w = Just v' → pl_leb PL u v' = true)
          ∧ (∀ w v', f w v = Just v' → pl_leb PL u v' = true)) →
    leb x a = true ∨ leb x b = true →
    leb x (join f Hf a b) = true.
  Proof.
    intros F [H|H]; rewrite leb_le in *.
    intros p. specialize (H p).
    unfold join, get in *.
    rewrite T.gshcombine.
    destruct (T.get p x) as [u|].
    destruct (T.get p a) as [v|]; auto.
    destruct (T.get p b) as [w|]; auto.
    specialize (F _ _ H). destruct F as [F _]. specialize (F w).
    destruct (f v w). auto. eauto.
    destruct (T.get p a) as [v|]; auto.
    discriminate.
    intros p. specialize (H p).
    unfold join, get in *.
    rewrite T.gshcombine.
    destruct (T.get p x) as [u|].
    destruct (T.get p a) as [v|];
    destruct (T.get p b) as [w|]; auto.
    specialize (F _ _ H). destruct F as [_ F]. specialize (F v).
    destruct (f v w). auto. eauto.
    destruct (T.get p a) as [v|];
    destruct (T.get p b) as [w|]; auto.
    discriminate.
  Qed.

  Instance wlat : weak_lattice t :=
    {| leb := leb
     ; top := T.empty _
     ; join := join (pl_join PL) H_join_eq
     ; widen := join (pl_widen PL) H_widen_eq
    |}.

  Lemma get_top x : get x (T.empty _) = All.
  Proof. unfold get. rewrite T.gempty. reflexivity. Qed.

  Instance gamma : gamma_op t (T.elt → B) :=
    λ x y, ∀ e, y e ∈ γ (get e x).

  Lemma gamma_set x a b p v :
    γ x a →
    γ v (b p) →
    (∀ q, q ≠ p → a q = b q) →
    γ (set x p v) b.
  Proof.
    intros A V H c.
    rewrite gsspec. destruct T.elt_eq.
    subst; auto.
    rewrite <- H; auto.
  Qed.

  Lemma gamma_forget x a b p :
    γ x a →
    (∀ q, q ≠ p → a q = b q) →
    γ (set x p All) b.
  Proof.
    intros A H c. specialize (H c).
    rewrite gsspec. destruct T.elt_eq as [EQ|NE].
    easy.
    rewrite <- (H NE). exact (A c).
  Qed.

  Lemma join_sound :
   ∀ (x y : t) (a : T.elt → B), γ x a ∨ γ y a → γ (join (pl_join PL) H_join_eq x y) a.
  Proof.
    intros x y a H c. simpl. unfold get, join.
    rewrite T.gshcombine.
    destruct H as [H|H]; specialize (H c); unfold get in H.
    - destruct (T.get c x) as [u|].
      * destruct (T.get c y) as [v|].
        generalize (pl_join_sound PLP u v (or_introl H)).
        destruct pl_join; eauto.
        easy.
      * easy.
    - destruct (T.get c x) as [u|].
      * destruct (T.get c y) as [v|].
        generalize (pl_join_sound PLP u v (or_intror H)).
        destruct pl_join. eauto.
        eauto.
        easy.
      * easy.
  Qed.

  Lemma adom : adom t (T.elt → B) wlat gamma.
  Proof.
    split.
    + intros. rewrite (leb_le a1 a2) in H.
      intro. eapply gamma_monotone, H0. apply adom_of_pre_lattice. eauto.
      apply H.
    + intros x y. unfold get. rewrite T.gempty. easy.
    + exact join_sound.
  Qed.

  Instance toString `{ToString T.elt} `{ToString elt} : ToString (T.t elt) :=
    (λ x, T.fold (λ s k v, s ++ "["++ to_string k ++ "↦" ++ to_string v ++"]" ) x "{" ++ "}")%string.

End elt.

Opaque get. (* TODO *)
Opaque set. (* TODO *)

End WPFun.
