(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Utf8 String Coqlib Util
  Maps Sets TreeAl AssocList
  AST Values
  Integers Floats FloatLib.

Require Import
  Globalenvs Memory
  Registers
  Cminor Csharpminor
  Ctypes.

Require Import
  ToString PExpr
  AdomLib ListAdom
  Cells TypesDomain Nconvert
  AbMachineEnv
  AbMemSignatureCsharpminor
  BlockMon
  NonBlock
  OnMem
  MemChunkTree.

Require Import
  DebugCshm
  DebugAbMachineEnv
  Debug.

Open Scope string.

Set Implicit Arguments.
Unset Elimination Schemes.

Definition injective_map {A B} (m: A → option B) : Prop :=
  ∀ x x' a,
    m(x) = Some a → m(x') = Some a →
    x = x'.

Definition stacks_tmp (stk stk': list (ident * (temp_env * env))) : Prop :=
  map (λ ite, (fst ite, snd (snd ite))) stk =
  map (λ ite, (fst ite, snd (snd ite))) stk'.

Lemma stacks_tmp_in : ∀ stk stk',
  stacks_tmp stk stk' →
  ∀ f t e,
    In (f, (t, e)) stk' →
    ∃ t', In (f, (t', e)) stk.
Proof.
  unfold stacks_tmp.
  induction stk as [|(i, (t, e)) stk IH]; intros stk' H.
  generalize (map_nil_inv _ _ (eq_sym H)). intros -> ? ? ? ().
  destruct stk' as [|(i', (t', e')) stk'].
  apply map_nil_inv in H. eq_some_inv.
  simpl in H. eq_some_inv.
  subst.
  intros f t0 e [?|H]. eq_some_inv. subst. vauto.
  edestruct IH. eassumption. eassumption.
  vauto.
Qed.

Remark block_of_ablock_tmp ge : ∀ stk stk',
  stacks_tmp stk stk' →
  ∀ ab, block_of_ablock ge stk ab = block_of_ablock ge stk' ab.
Proof.
  unfold stacks_tmp.
  induction stk as [|(i, (t, e)) stk IH]; intros stk' H.
  generalize (map_nil_inv _ _ (eq_sym H)). clear H. intros ->. reflexivity.
  destruct stk' as [|(i', (t', e')) stk'].
  apply map_nil_inv in H. eq_some_inv.
  simpl in H. eq_some_inv.
  subst.
  intros [f x|b]; simpl; auto.
  destruct @eq_dec. reflexivity.
  refine (IH stk' _ (ABlocal f x)).
  assumption.
Qed.

Lemma norec_tmp : ∀ (stk stk': list (ident * (temp_env * env))),
  stacks_tmp stk stk' →
  map fst stk = map fst stk'.
Proof.
  unfold stacks_tmp.
  induction stk as [|(i, (t, e)) stk IH]; intros stk' H.
  generalize (map_nil_inv _ _ (eq_sym H)). congruence.
  destruct stk' as [|(i', (t', e')) stk'].
  apply map_nil_inv in H. eq_some_inv.
  simpl in H. eq_some_inv.
  subst. simpl. f_equal. eauto.
Qed.

Definition read_cell (ge: genv) (m: concrete_state) (c: cell) : option val :=
  match c with
  | AClocal f l ofs κ =>
    do_opt t_e <- assoc f (fst m);
    let (_, e) := t_e in
    do_opt b_sz <- PTree.get l e;
    let (b, sz) := b_sz in
    Mem.load (proj1_sig κ) (snd m) b ofs
  | ACtemp f r =>
    do_opt t_e <- assoc f (fst m);
    let (t, _) := t_e in
    PTree.get r t
  | ACglobal b ofs κ =>
    do_opt s <- Genv.invert_symbol ge b;
    Mem.load (proj1_sig κ) (snd m) b ofs
  end.

Definition extend {A} (f: A → option val) : A → val :=
  λ a, match f a with Some v => v | None => Vundef end.
Arguments extend [A] f _ /.

Definition mem_as_fun ge (m: concrete_state) : cell → val :=
  extend (read_cell ge m).
Arguments mem_as_fun ge m _ /.

Module N := Nconvert ACTree.
Module ACTreeDom := N.TDom.
Module ABTreeDom := WPFun ABTree.
Existing Instances ACTreeDom.toString ABTreeDom.toString.

Definition PTSet : Type := BlockSet.t +⊤.

Instance ptset_gamma ge stk : gamma_op BlockSet.t val :=
  λ s i,
  match i with
         | Vptr b _ =>
           ∃ ab, BlockSet.mem ab s ∧ block_of_ablock ge stk ab = Some b
         | _ => True
         end.

Lemma ptset_plat_correct ge stk : pre_lattice_spec BSO.plat (ptset_gamma ge stk).
Proof.
  split.
- intros x y H [| | | | |b i] X; auto.
  destruct X as (ab & IN & AB). exists ab. intuition.
  eapply BlockSet.le_spec; eauto.
- intros x y [|?|?|?|?|b ?] H; try now intuition.
  destruct H as [(ab & IN & AB)|(ab & IN & AB)]; exists ab; intuition;
  apply BlockSet.mem_union; auto.
Qed.

Definition points_to : Type := ACTreeDom.t BlockSet.t.
Definition types : Type := ACTreeDom.t AbTy.t.

Definition lvarset : Type := PSet.t.
Definition stack : Type := list (fname * lvarset).
Instance stack_wlat : weak_lattice (stack+⊤) :=
  weak_lattice_of_pre_lattice (list_plat (WTensor.make_pre_lattice (flat_pre_lattice Pos.eqb) (flat_pre_lattice PSet.beq))).

Definition points_to_get (pt: points_to) (c: cell) : PTSet :=
  ACTreeDom.get c pt.
Local Coercion points_to_get : points_to >-> Funclass.

Instance points_to_wlat : weak_lattice points_to := ACTreeDom.wlat BSO.plat _ _ _.
Proof.
  abstract (simpl; unfold BlockSet.union; intro; rewrite ABTree.shcombine_eq; auto).
  abstract (simpl; unfold BlockSet.union; intro; rewrite ABTree.shcombine_eq; auto).
  abstract (simpl; unfold BlockSet.le; intros; rewrite ABTree.shforall2_correct;
            intros; destruct ABTree.get; auto).
Defined.

Instance points_to_gamma ge stk : gamma_op points_to (cell → val) := ACTreeDom.gamma (ptset_gamma ge stk).
Lemma points_to_adom ge stk : adom points_to (cell → val) points_to_wlat (points_to_gamma ge stk).
Proof (ACTreeDom.adom (ptset_plat_correct ge stk) _ _ _).
Hint Resolve points_to_adom.

Existing Instance N.tree_wl.
Existing Instance N.tree_gamma.
Hint Resolve N.tree_adom.

Definition env_as_fun : env → ident → option block :=
  λ e i, match e ! i with Some (b, _) => Some b | _ => None end.
Coercion env_as_fun : env >-> Funclass.

Section PT_EVAL.

Context (ge: genv) (μ: fname) (pt: points_to).
Fixpoint pt_eval_pexpr (pe: pexpr cell) : PTSet :=
  match pe with
  | PEvar x => ACTreeDom.get x pt
  | PElocal s => Just (BSO.singleton (ABlocal μ s))
  | PEconst None => All
  | PEconst (Some (Oaddrsymbol s _)) =>
    match Genv.find_symbol ge s with
    | Some b => Just (BSO.singleton (ABglobal b))
    | None => Just BlockSet.empty
    end
  | PEconst (Some _) => Just BlockSet.empty
  | PEunop  _ pe' => pt_eval_pexpr pe'
  | PEbinop _ x y =>
    do_top p <- pt_eval_pexpr x;
    do_top q <- pt_eval_pexpr y;
    Just (BlockSet.union p q)
  end.

Lemma pt_eval_pexpr_correct m stk ρ ε tmp
      (PT: ρ ∈ points_to_gamma ge ((μ, (tmp, ε)):: stk) pt) :
  ∀ pe,
  eval_pexpr ge m ρ ε pe ⊆ toplift_gamma (ptset_gamma ge ((μ, (tmp, ε)) :: stk)) (pt_eval_pexpr pe).
Proof.
  induction pe; intros w W; eval_pexpr_inv.
  generalize (PT v). rewrite W. exact id.
  subst. eexists. split. apply BSO.mem_singleton. eauto. simpl. rewrite dec_eq_true. exact W.
  destruct o as [()|]; eval_pexpr_inv; subst; try exact I.
  simpl. destruct (Genv.find_symbol ge i) eqn:?. 2: exact I.
  eexists. split. apply BSO.mem_singleton. eauto. simpl. erewrite Genv.find_invert_symbol; eauto.
  simpl.
  destruct w; try (destruct (pt_eval_pexpr pe); exact I).
  eelim (eval_pexpr_unop_ptr). vauto.
  repeat match goal with H : (∀ _ _, _), K : _ |- _ => specialize (H _ K) end.
  simpl. destruct (pt_eval_pexpr pe1) as [|p]. exact I.
  simpl. destruct (pt_eval_pexpr pe2) as [|q]. exact I.
  simpl in *.
  destruct b; simpl in *; eq_some_inv; subst;
  destruct v₁; simpl in *; eq_some_inv; try exact I; destruct v₂; simpl in *; eq_some_inv; try exact I;
  subst;
  try (unfold Val.cmp, Val.cmpu, Val.cmpf, Val.cmpfs; simpl; bif; exact I);
  try (revert H0; bif; intro; eq_some_inv; subst; exact I);
  try (
  unfold Val.cmpl, Val.cmplu, Val.cmpl_bool, Val.cmplu_bool in *;
      try (apply AbTy.if_opt in H0; hsplit; subst);
  simpl in *; eq_some_inv; subst;
  match goal with |- appcontext[ Val.of_bool ?b ] => destruct b; exact I end
    );
  try (unfold Val.cmpu; repeat (simpl; bif; try exact I); destruct c; exact I);
  hnf in *; hsplit; vauto.
  eexists. split. apply BlockSet.mem_union. eauto. auto.
  eexists. split. apply BlockSet.mem_union. eauto. auto.
  eexists. split. apply BlockSet.mem_union. eauto. auto.
Qed.

End PT_EVAL.

(* Size of each block. *)
Definition sizes : Type := ABTreeDom.t (Z * bool). (* boolean denotes write permission *)
Definition sizes_gamma ge stk : gamma_op sizes mem :=
  λ sz m,
  ∀ b hi rw b',
    ABTreeDom.get b sz = Just (hi, rw) →
    block_of_ablock ge stk b = Some b' →
    Mem.range_perm m b' 0 hi Cur (match b with ABglobal _ => if rw then Writable else Readable | _ => Freeable end).

Instance BoolEqDec : EqDec bool := eq_dec_of_beq _ Bool.eqb_true_iff.

Instance sizes_wlat : weak_lattice sizes := ABTreeDom.wlat (flat_pre_lattice (λ a b, eq_dec a b)) _ _ _.
Proof.
  abstract (simpl; intro; destruct (eq_dec x x); intuition).
  abstract (simpl; intro; destruct (eq_dec x x); intuition).
  abstract (simpl; intro; destruct (eq_dec x x); intuition).
Defined.

Lemma sizes_adom ge stk : adom sizes mem sizes_wlat (sizes_gamma ge stk).
Proof.
  constructor.
- (* gamma monotone *)
  intros a1 a2 H m G b hi rw b' B B' ofs Hofs.
  specialize (G b hi rw b').
  unfold leb in H. simpl in H.
  rewrite ABTreeDom.leb_le in H. simpl in H.
  specialize (H b). rewrite B in H.
  destruct (ABTreeDom.get b a1); simpl in H. eq_some_inv.
  InvBooleans.
  subst. exact (G eq_refl B' ofs Hofs).
- (* gamma top *)
  intros x b hi b'. unfold sizes_wlat. rewrite ABTreeDom.get_top. inversion 1.
- (* join monotone *)
  intros x y a H b hi rw b' G B' ofs Hofs.
  unfold join in G. simpl in G. rewrite ABTreeDom.get_join in G.
  destruct (ABTreeDom.get b x) eqn: Bx; simpl in G. discriminate.
  destruct (ABTreeDom.get b y) eqn: By; simpl in G. discriminate.
  destruct (@eq_dec _ _); subst; simpl in *; inv G.
  destruct H as [H|H]; specialize (H b hi rw b');
  first [ specialize (H Bx) | specialize (H By) ];
  exact (H B' ofs Hofs).
Qed.
Hint Resolve sizes_adom.

Lemma ptset_gamma_tmp ge : ∀ stk stk',
  stacks_tmp stk stk' →
  ∀ bs, ptset_gamma ge stk bs ⊆ ptset_gamma ge stk' bs.
Proof.
  intros stk stk' H bs v.
  destruct v; try exact id.
  intros (b' & Hb' & K).
  exists b'. split. exact Hb'.
  erewrite <- block_of_ablock_tmp; eassumption.
Qed.

Lemma points_to_gamma_tmp ge : ∀ stk stk',
  stacks_tmp stk stk' →
  ∀ ptr, points_to_gamma ge stk ptr ⊆ points_to_gamma ge stk' ptr.
Proof.
  intros stk stk' H ptr ρ Ρ c.
  generalize (Ρ c).
  destruct (ACTreeDom.get c ptr). exact id.
  apply ptset_gamma_tmp. exact H.
Qed.

Local Coercion constant_of_constant (c: Csharpminor.constant) : constant :=
  match c with
  | Csharpminor.Ointconst i => Ointconst i
  | Csharpminor.Olongconst i => Olongconst i
  | Csharpminor.Ofloatconst f => Ofloatconst f
  | Csharpminor.Osingleconst f => Osingleconst f
  end.

Lemma sizes_gamma_tmp ge stk stk' :
  stacks_tmp stk stk' →
  ∀ sz, sizes_gamma ge stk sz ⊆ sizes_gamma ge stk' sz.
Proof.
  intros H sz a A b hi b' B B' ofs Hofs.
  apply (A b hi b'); auto.
  rewrite (block_of_ablock_tmp ge H).
  assumption.
Qed.

Section num.

  Context
    {num : Type}
    `{numToString: ToString num}
    (NumDom: ab_machine_env (var:=cell) num).

  Lemma gamma_num_ext :
   ∀ ρ ρ' : cell → num_val,
   (∀ x : cell, ρ x = ρ' x) → ∀ m : num +⊥, γ m ρ → γ m ρ'.
  Proof.
    intros ρ ρ' H [|m]. intros (). replace ρ' with ρ. exact id.
    apply Axioms.extensionality. exact H.
  Qed.

  Definition compat := (λ τ ρ ρn, N.compat _ τ (λ x, x) ρ ρn).

  Variable ge : genv.

  Record invariant (cs: concrete_state) : Prop :=
    { localFreeable :
        (* Local variables have Freeable permission *)
        ∀ f t e, In (f, (t, e)) (fst cs) →
          ∀ b hi, In (b, 0, hi) (blocks_of_env e) →
                  Mem.range_perm (snd cs) b 0 hi Cur Freeable
    ; localNotGlobal :
    (* Globals and locals are in different blocks *)
        ∀ f t e, In (f, (t, e)) (fst cs) →
             ∀ x b y b' z,
               Genv.find_symbol ge x = Some b →
               e ! y = Some (b', z) →
               b ≠ b'
    ; localInj :
    (* Different local variables refer to different blocks *)
   injective_map (λ x_v,
                  do_opt t_e <- assoc (fst x_v) (fst cs);
                  option_lift fst ((snd t_e) ! (snd x_v)))
    ; noRec:
    (* All functions are different (i.e., no recursion) *)
   list_norepet (map fst (fst cs))
   ; globalValid:
   (* Globals are allocated *)
   ∀ x b, Genv.find_symbol ge x = Some b →
          b ∈ Mem.valid_block (snd cs)
   ; localValid:
   (* Locals are allocated *)
       ∀ f t e, In (f, (t, e)) (fst cs) →
                ∀ x b z,
               e ! x = Some (b, z) →
               b ∈ Mem.valid_block (snd cs)
   }.

  Lemma invariant_tmp : ∀ stk stk',
    stacks_tmp stk stk' →
    ∀ m, invariant (stk, m) → invariant (stk', m).
  Proof.
    intros stk stk' H m INV.
    split.
  - intros f t e H0.
    apply (stacks_tmp_in H) in H0. hsplit.
    generalize (INV.(localFreeable) f t' e H0). auto.
  - intros f t e H0.
    apply (stacks_tmp_in H) in H0. hsplit.
    generalize (INV.(localNotGlobal) f t' e H0). auto.
  - intros (f, x) (f', x') b.
    generalize (INV.(noRec)). intros NR.
    generalize (INV.(@localInj _) (f, x) (f', x') b). clear -H NR.
    simpl.
    generalize (assoc_in f stk'). destruct (assoc f stk') as [(t,e)|]. 2: intros; eq_some_inv.
    intros X. specialize (X _ eq_refl). apply (stacks_tmp_in H) in X. hsplit.
    rewrite (In_norepet_assoc _ _ _ NR X). simpl. clear X.
    generalize (assoc_in f' stk'). destruct (assoc f' stk') as [(q,g)|]. 2: intros; eq_some_inv.
    intros X. specialize (X _ eq_refl). apply (stacks_tmp_in H) in X. hsplit.
    rewrite (In_norepet_assoc _ _ _ NR X). simpl. clear X.
    exact id.
  - simpl. rewrite <- (norec_tmp H). exact INV.(noRec).
  - exact INV.(globalValid).
  - intros f t e H0.
    apply (stacks_tmp_in H) in H0. hsplit.
    generalize (INV.(localValid) _ _ _ H0). auto.
  Qed.

  Lemma invariant_block_of_ablock_inj stk m :
    (stk, m) ∈ invariant →
    ∀ ab₁ b₁ ab₂ b₂,
      block_of_ablock ge stk ab₁ = Some b₁ →
      block_of_ablock ge stk ab₂ = Some b₂ →
      ab₁ ≠ ab₂ →
      b₁ ≠ b₂.
  Proof.
    intros INV [f x|b] b₁ [g y|b'] b₂; simpl.
    generalize (INV.(localInj) (f, x) (g, y)). simpl.
    destruct (assoc f stk) as [(t, e)|]. 2: intros; eq_some_inv.
    destruct (assoc g stk) as [(t', e')|]. 2: intros; eq_some_inv. simpl.
    destruct (e!x) as [(b, ?)|]. 2: intros; eq_some_inv.
    destruct (e'!y) as [(b', ?)|]. 2: intros; eq_some_inv.
    simpl. intros H H0 H1 H2 ?. eq_some_inv. subst. apply H2. specialize (H _ eq_refl eq_refl). congruence.
    destruct (assoc f stk) as [(t, e)|] eqn: A. 2: intros; eq_some_inv.
    destruct (e!x) as [(b, ?)|] eqn: Hx. 2: intros; eq_some_inv.
    destruct (Genv.invert_symbol ge b')  as [q|] eqn: Q; intros; eq_some_inv.
    subst. apply not_eq_sym.
    exact (INV.(localNotGlobal) f t e (assoc_in _ _ A) _ _ (Genv.invert_find_symbol _ _ Q) Hx).
    destruct (Genv.invert_symbol ge b)  as [q|] eqn: Q; intro; eq_some_inv.
    destruct (assoc g stk) as [(t', e')|] eqn: A. 2: intros; eq_some_inv.
    destruct (e'!y) as [(b', ?)|] eqn: Hy; intros; eq_some_inv. subst.
    exact (INV.(localNotGlobal) g t' e' (assoc_in _ _ A) _ _ (Genv.invert_find_symbol _ _ Q) Hy).
    destruct (Genv.invert_symbol ge b)  as [q|] eqn: Q; intro; eq_some_inv.
    destruct (Genv.invert_symbol ge b')  as [q'|] eqn: Q'; intros; eq_some_inv.
    congruence.
  Qed.

  Lemma assign_const c z (nm : num) :
    ∀ ρ, ρ ∈ γ(nm) →
    match AbMachineEnv.assign c (ne_const_i _ z) nm with
    | Bot => False
    | NotBot nm' => (upd ρ c (NVint (Int.repr z))) ∈ γ(nm')
    end.
  Proof.
    intros ρ H.
    assert (eval_nexpr ρ (ne_const_i _ z) (NVint (Int.repr z))) as V
    by (repeat econstructor; reflexivity).
    generalize (AbMachineEnv.assign_correct c nm H V).
    destruct AbMachineEnv.assign; exact id.
  Qed.

  Definition tnum : Type := (types * num)%type.

  Instance tnum_gamma : gamma_op tnum (cell → val) :=
    λ τ_ν ρ,
    let '(τ, ν) := τ_ν in
    ρ ∈ γ(τ) ∧
    ∃ ρ',
      ρ' ∈ (compat τ ρ ∩ γ ν).

  Existing Instance WProd.make_wl.
  Hint Resolve id_adom.
  Lemma tnum_adom : adom tnum (cell → val) _ tnum_gamma.
  Proof.
    split.
    - intros (τ,ν) (τ',ν'); simpl. rewrite andb_true_iff.
      intros [H K] ρ (TP & ρ' & NC & NM).
      split. eapply gamma_monotone; eauto.
      exists ρ'. split. eapply N.compat_monotone; eauto.
      eapply gamma_monotone; eauto.
    - intros ρ. split.
      apply gamma_top; eauto.
      exists (compat_fun ∘ ρ). split.
      intros c. split. apply N.ncompat_compat_fun.
      simpl. unfold N.Eval.TDom.get.
      rewrite ACTree.gempty. apply N.wtype_num_top.
      apply gamma_top; eauto.
    - intros (τ,ν) (τ',ν') ρ [(TP & ρ' & H & K)|(TP & ρ' & H & K)];
      (split;[|exists ρ'; split]);
      first [ apply join_sound | apply N.compat_join]; intuition.
  Qed.

  Ltac ssubst :=
    repeat
    match goal with
    | H : ?a = ?b |- _ => first [subst a | subst b]
    end.

  Variable max_concretize : N.

  Section CONVERT.

  (** Current function. *)
  Variable μ : fname.

  (** Set of local variables. *)
  Variable ε : lvarset.

  Definition is_local (v: ident) : bool := PSet.mem v ε.

  Definition addr_of (v: ident) : BlockSet.t :=
    if is_local v
    then BSO.singleton (ABlocal μ v)
    else match Genv.find_symbol ge v with
         | Some b => BSO.singleton (ABglobal b)
         | None => BlockSet.empty
         end.

  Variable sz : sizes.
  Variable pt : points_to.
  Variable tp : types.
  Variable nm : num.

  Definition ptype (pe: pexpr cell) : AbTy.t+⊤ := N.Eval.eval_expr AbTy.adom tp pe.

  Definition ne_of_pe (pe: pexpr cell) : bm_map_t (nexpr cell) :=
    match N.nconvert _ tp (λ x, x) pe with
    | Some ne =>
      (*print ("ne_of_pe: " ++ to_string pe ++ " → " ++ to_string ne) *)
      if AbMachineEnv.nonblock ne nm
      then Result ne
      else Error (Some ne)
                 ("ne_of_pe(" ++ to_string pe ++ "): blocked")
    | None => Error None
                    ("ne_of_pe(" ++ to_string pe ++ "): conversion failed")
    end.

  Definition as_pexpr_list (l: list (pexpr cell)+⊤) : block_mon _ :=
    match l with
    | Just l' => ret_bm l'
    | All => do_bm _ <- err_bm "convert failed"; ret_bm nil
    end.

  Lemma const_sound z :
    NVint (Int.repr z) ∈ γ(NVIint (Just (ZIntervals.ITV z z))).
  Proof. constructor; split; reflexivity. Qed.
  Local Hint Resolve const_sound.

  Definition is_one (z: Z) : bool :=
    match z with 1 => true | _ => false end.

  Lemma is_one_elim z (P: Z → bool → Prop) :
    P 1 true →
    (∀ z, z ≠ 1 → P z false) →
    P z (is_one z).
  Proof.
    intros H1 H.
    destruct z as [|[z|z|]|z];
      first [ exact H1 | apply H; discriminate ].
  Qed.

  Definition check_align (sz: Z) (ne: nexpr cell) : bool :=
    is_one sz ||
           is_bot (AbMachineEnv.assume
              (NEbinop (Ocmp Cne)
                 (NEbinop Omodu ne (ne_const_i _ sz))
                 (ne_const_i _ 0))
              true nm).

  Definition realize κ (lpe:list (pexpr cell)) (ptr: BlockSet.t) : block_mon (CellSet.t+⊤) :=
    let al := align_chunk (proj1_sig κ) in
    bm_fold
      (λ acc pe,
       do_bm offs <-
           match N.Eval.eval_expr AbTy.adom tp pe with
           | Just AbTy.VP => match ne_of_pe pe with
           | Result ne =>
             do_bm _ <- bm_assert (check_align al ne)
                   ("bad align ("++ to_string (proj1_sig κ) ++"): " ++ to_string ne);
             ret_bm (concretize_int max_concretize ne nm)
           | Errorc ne e =>
             do_bm _ <- err_bm (e tt);
             ret_bm match ne with
                      | Some ne => concretize_int max_concretize ne nm
                      | None => NotBot All
                    end
           end
           | ty =>
             do_bm _ <- err_bm ("realize: maybe not a pointer (" ++ to_string ty ++ ")");
             ret_bm (NotBot All)
           end;
       ret_bm match acc with All => All
       | Just acc => toplift_lift (CellSet.union acc) (set_product ptr κ offs)
       end)
      lpe
      (Just CellSet.empty).

  Lemma realize_inv κ lpe ptr :
    match bm_get (realize κ lpe ptr) with
    | Some cells =>
      let al := align_chunk (proj1_sig κ) in
    ∀ pe, In pe lpe →
          N.Eval.eval_expr AbTy.adom tp pe = Just AbTy.VP ∧
          ∃ ne,
            ne_of_pe pe = Result ne ∧
            check_align al ne = true ∧
            let ofs := concretize_int max_concretize ne nm in
            match cells with All => True | Just cells =>
            ofs ≠ NotBot All ∧
            ∀ b i,
              BlockSet.mem b ptr →
              i ∈ γ(ofs) →
              CellSet.mem (cell_from b (IntSet.proj i, κ)) cells
            end
    | None => True
    end.
  Proof.
    unfold realize.
    set (al := align_chunk (proj1_sig κ)).
    match goal with |- appcontext[ bm_fold ?f ] => set ( F := f ) end.
    induction lpe as [|pe lpe IH] using rev_ind.
    intros _ (); fail.
    unfold ACTree.elt in IH.
    rewrite bm_fold_app.
    apply bm_bind_case. intros; exact I.
    intros acc REC. rewrite REC in IH. simpl in IH. clear REC.
    subst F. unfold bm_fold. simpl.
    destruct (N.Eval.eval_expr AbTy.adom tp pe) as [|()] eqn:TY;
      try exact I.
    destruct (ne_of_pe pe) as [ne|] eqn: NE. 2: exact I.
    destruct (check_align al ne) eqn: ALIGN. 2: exact I.
    intros pe' IN. rewrite in_app in IN.
    destruct IN as [IN| [ <- | () ]].
    specialize (IH _ IN). destruct IH as (Ty & ne' & IH); hsplit.
    split. assumption.
    exists ne'. split. assumption. split. assumption.
    destruct acc as [|acc]. exact I. hsplit.
    destruct (set_product) as [|cells]. exact I.
    split. assumption.
    intros; apply CellSet.mem_union; auto; fail.
    split. assumption.
    exists ne. split. assumption. split. assumption.
    destruct acc as [|acc]. exact I.
    destruct (concretize_int _ ne nm) as [|[|offs]] eqn: Hoffs.
    + split. discriminate. intros b i H ().
    + exact I.
    + split. discriminate.
      intros b i H H0.
      generalize (set_product_correct _ κ (@NotBot _ (Just offs)) _ _ H H0).
      intros. apply CellSet.mem_union. auto.
  Qed.

  Definition is_null (ne: nexpr cell) : bool :=
    is_bot (AbMachineEnv.assume (NEbinop (Ocmpu Ceq) ne (ne_const_i _ 0)) false nm).

  Inductive cmp_blockset_t := Equal (b: ablock) | Disjoint | DontKnow | Contradiction.
  Definition cmp_blockset (x y: BlockSet.t) : cmp_blockset_t :=
    match BSO.classify x, BSO.classify y with
    | BSO.Empty, _ | _, BSO.Empty => Contradiction
    | BSO.Singleton v, BSO.Singleton w => if eq_dec v w then Equal v else Disjoint
    | _, _ => if BlockSet.le (BlockSet.inter x y) BlockSet.empty then Disjoint else DontKnow
    end.

  Lemma cmp_blockset_correct x y v w :
    BlockSet.mem v x → BlockSet.mem w y →
    match cmp_blockset x y with
    | Equal b => v = b ∧ w = b
    | Disjoint => v ≠ w
    | DontKnow => True
    | Contradiction => False
    end.
  Proof.
    intros V W.
    unfold cmp_blockset.
    pose proof (BSO.classify_correct x) as X.
    pose proof (BSO.classify_correct y) as Y.
    destruct (BSO.classify x) as [|v'|]; try (now intuition eauto);
    destruct (BSO.classify y) as [|w'|]; try (now intuition eauto);
    repeat
    match goal with |- appcontext[ BlockSet.le ?a ?b ] =>
      generalize (proj2 (BlockSet.le_spec a b));
      destruct BlockSet.le; auto;
      intros HLE ->;
      refine (BlockSet.mem_empty _ (HLE eq_refl _ _));
      rewrite BlockSet.mem_inter; split; eauto
    end.
    assert (v = v'). intuition eauto.
    assert (w = w'). intuition eauto.
    subst v' w'. destruct (eq_dec v w); easy.
  Qed.

  Inductive min_size_t := MinSize | Size (z: Z)  | MaxSize.
  Definition size_of_option (a: option (Z * bool)) := match a with Some (a, _) => Size a | None => MinSize end.

  Definition omin (x: option (Z * bool)) (y: min_size_t) : min_size_t :=
    match x, y with
    | None, _ => MinSize
    | _, MinSize => y
    | Some (x, _), Size y => Size (Z.min x y)
    | Some (x, _), MaxSize => Size x
    end.

  Definition ozle (x y: min_size_t) : Prop :=
    match x, y with
    | MinSize, _ | _, MaxSize => True
    | MaxSize, _ | _, MinSize => False
    | Size x, Size y => Z.le x y
    end.

  Remark ozle_refl : ∀ a, ozle a a.
  Proof. intros [|a|]; simpl; Psatz.lia. Qed.

  Remark ozle_min_2 : ∀ a b c,
    ozle (size_of_option a) c ∨ ozle b c →
    ozle (omin a b) c.
  Proof. intros [(a, ?)|] [|b|] [|c|]; simpl; Psatz.lia. Qed.

  Definition min_size := BlockSet.fold (λ b, omin (ABTree.get b sz)).

  Lemma min_size_min bs z:
    let m := min_size bs z in
    (∀ b, BlockSet.mem b bs → ozle m (size_of_option (ABTree.get b sz))) ∧ ozle m z.
  Proof.
    unfold min_size.
    apply BSO.foldspec.
    2: refine (λ _ H, conj (λ _ K, False_ind _ (H _ K)) (ozle_refl _)).
    intros s s' z' acc x H [H0 H1] [IH Hacc].
    split. 2: apply ozle_min_2; auto.
    intros c Hc. apply ozle_min_2.
    destruct (H0 _ Hc); [ left | right ].
    subst; apply ozle_refl. eauto.
  Qed.

  Definition is_valid_ptr (weak: bool) (bs: BlockSet.t) (off: nexpr cell) : bool :=
    match min_size bs MaxSize with
    | MaxSize => false (* all blocks in bs are infinite: cannot happen *)
    | Size (Zpos min) =>
      Pos.ltb min (Z.to_pos Int.max_unsigned) &&
      is_bot (AbMachineEnv.assume (NEbinop (Ocmpu Clt) off (ne_const_i _ (Zpos (if weak then Pos.succ min else min)))) false nm)
    | Size _
    | MinSize => false
    end.

  Definition convert_err (e:expr) msg :=
    ("convert(" ++ to_string e ++ "): " ++ msg ++ new_line ++
     "Function: " ++ to_string μ ++ new_line ++
     "Locals: " ++ to_string ε ++ new_line ++
     "Points-to: " ++ to_string pt ++ new_line ++
     "Types: " ++ to_string tp ++ new_line ++
     "Num: " ++ to_string nm ++ new_line ++
     "------------------------------------------").
  Opaque convert_err.

  Fixpoint convert (e: expr) : block_mon (list (pexpr cell)) :=
    let err := convert_err e in
    let res :=
    match e with
    | Evar v =>
      let c := ACtemp μ v in
      match ACTreeDom.get c tp with
      | Just _ => ret_bm (PEvar c :: nil)
      | All => do_bm _ <- err_bm (err ("bad temp in " ++ to_string μ)); ret_bm nil
      end
    | Eaddrof i =>
      if is_local i
      then ret_bm (PElocal _ i :: nil)
      else match Genv.find_symbol ge i with
           | Some _ => ret_bm (PEconst _ (Some (Oaddrsymbol i Int.zero)) :: nil)
           | None => do_bm _ <- err_bm (err "bad global variable");
                     ret_bm nil
           end
    | Econst cst => ret_bm (PEconst _ (Some (constant_of_constant cst)) :: nil)
    | Eunop op e1 =>
      do_bm e1' <- convert e1;
      ret_bm (rev_map (PEunop op) e1')
    | Ebinop op e1 e2 =>
      do_bm e1' <- convert e1;
      do_bm e2' <- convert e2;
      match op with
      | Ocmpu c =>
        bm_map'
          (λ x_y,
           let '(x, y) := x_y in
           match ptype x, ptype y with
           | Just AbTy.VI, Just AbTy.VI => Result (PEbinop op x y)
           | Just AbTy.VP, Just AbTy.VI =>
             match ne_of_pe x with
             | Result _ => (* checks not blocked *)
             match ne_of_pe y with
             | Result ny =>
               if is_null ny
               then
                 match c with
                 | Ceq => Result (PEconst _ (Some (Ointconst Int.zero)))
                 | Cne => Result (PEconst _ (Some (Ointconst Int.one)))
                 | _ => Error None ("Bad comparison with NULL pointer: " ++ to_string c)
                 end
               else Error None ("Cannot prove " ++ to_string y ++ " definitely null")
             | Errorc _ msg => Errorc None msg
             end
             | Errorc _ msg => Errorc None msg
             end
           | Just AbTy.VP, Just AbTy.VP =>
             match ne_of_pe x with
             | Result nx =>
             match ne_of_pe y with
             | Result ny =>
             match pt_eval_pexpr ge μ pt x, pt_eval_pexpr ge μ pt y with
             | Just xb, Just yb =>
               match cmp_blockset xb yb with
               | Equal b =>
                 let res := PEbinop (Ocmpu c) x y in
                 if is_valid_ptr true xb nx && is_valid_ptr true yb ny
                 then Result res
                 else Error (Some res)  (err "cannot prove pointer weakly valid")
               | Disjoint =>
                 if is_valid_ptr false xb nx && is_valid_ptr false yb ny
                 then match c with
                        | Ceq | Cne => Result (PEbinop (Ocmpu c) (PEconst _ (Some (Ointconst Int.zero))) (PEconst _ (Some (Ointconst Int.one))))
                        | _ => Error None (err ("bad comparison of pointers to different blocks " ++ to_string c))
                      end
                 else Error None (err "cannot prove pointer valid")
               | DontKnow => Error None (err ("cannot compare points-to sets: " ++ to_string xb ++ "; " ++ to_string yb))
               | Contradiction => Result (PEconst _ (Some (Ointconst Int.one)))
               end
             | _, _ => Error None (err "too large points-to set")
             end
             | Errorc _ msg => Errorc None msg
             end
             | Errorc _ msg => Errorc None msg
             end
           | tx, ty => Error None (err ("bad cmpu(" ++ to_string c ++ "): " ++ to_string tx ++ ", " ++ to_string ty))
           end)
          (map2 pair e1' e2')
      | _ => ret_bm (map2 (PEbinop op) e1' e2')
      end
    | Eload κ e1 =>
      do_bm e1' <- convert e1;
      do_bm ptr <- eval_expr_ptr e1;
      do_bm κ <-
        match κ return block_mon (option typed_memory_chunk) with
          | Many32 | Many64 =>
            do_bm _ <- err_bm (err "Many32 or Many64"); ret_bm None
          | κ => ret_bm (Some (exist _ κ I))
        end;
      do_bm cells <-
            match ptr with
            | All => do_bm _ <- err_bm (err "eval_expr_ptr failed"); ret_bm All
            | Just ptr' => match κ with
                             | Some κ => realize κ e1' ptr'
                             | None => ret_bm All
                           end
            end;
      match cells with
      | All => do_bm _ <- err_bm (err "too much cells"); ret_bm nil
      | Just cells =>
      bm_map' (λ c, match ACTreeDom.get c tp with
                    | All => Error None (err ("bad cell: " ++ to_string c))
                    | Just _ => Result (PEvar c)
                    end)
              (CellSet.elements cells)
      end
    end
      in
    (* print ("convert " ++ to_string e ++ " → " ++ to_string (fst res)) *)
          res
  with eval_expr_ptr (e: expr) : block_mon PTSet :=
         let res :=
    match e with
    | Evar v => ret_bm (pt (ACtemp μ v))
    | Eaddrof v => ret_bm (Just (addr_of v))
    | Econst _
    | Eunop _ _ => ret_bm (Just BlockSet.empty)
    | Ebinop _ e1 e2 =>
      do_bm e1' <- eval_expr_ptr e1;
      do_bm e2' <- eval_expr_ptr e2;
      ret_bm (
      do_top e1' <- e1';
      do_top e2' <- e2';
      Just (BlockSet.union e1' e2')
        )
    | Eload κ e =>
      do_bm e' <- convert e;
      do_bm ptr <- eval_expr_ptr e;
      do_bm κ <-
        match κ return block_mon (option typed_memory_chunk) with
          | Many32 | Many64 =>
            do_bm _ <- err_bm (convert_err e "Many32 or Many64"); ret_bm None
          | κ => ret_bm (Some (exist _ κ I))
        end;
      match ptr, κ with
      | All, _ | _, None => ret_bm All
      | Just ptr', Some κ =>
        do_bm cells <- realize κ e' ptr';
        ret_bm (
        do_top cells <- cells;
        BSO.union_list_noitpo
          (rev_map (fun c => pt c) (CellSet.elements cells))
        )
      end
    end
      in
    (* print ("eval_expr_ptr " ++ to_string e ++ " → " ++ to_string (fst res)) *)
          res.

  Definition ne_of_pe_list := bm_map' ne_of_pe.

  Definition nconvert (e:expr) : block_mon (list (nexpr cell)) :=
    bind_bm (convert e) ne_of_pe_list.

  Definition realize_cells κ (e: expr) : block_mon (list cell) :=
    do_bm lpe <- convert e;
    do_bm ptr <- eval_expr_ptr e;
    match ptr with
    | All => do_bm _ <- err_bm ("eval_expr_ptr(" ++ to_string e ++ "): All");
            ret_bm nil
    | Just ptr' =>
      do_bm cells <- realize κ lpe ptr';
      match cells with
      | All => do_bm _ <- err_bm ("realize_cells(" ++ to_string e ++ "): All");
              ret_bm nil
      | Just cell_set => ret_bm (CellSet.elements cell_set)
      end
    end.

    Section CONVERT_CORRECT.
    Context (e: env) (tmp: temp_env) (m: mem).
    Context (stk: list (ident * (temp_env * env))).
    Let ρ := mem_as_fun ge ((μ, (tmp, e)) :: stk, m).
    Local Instance pt_gamma : gamma_op points_to _ := points_to_gamma ge ((μ, (tmp, e)) :: stk).
    Hypothesis E: ∀ i, is_local i = true ↔ e ! i ≠ None.
    Hypothesis INV : invariant ((μ, (tmp, e)) :: stk, m).
    Hypothesis SZ: m ∈ sizes_gamma ge ((μ, (tmp, e)) :: stk) sz.
    Hypothesis TP: ρ ∈ γ(tp).
    Hypothesis PT: ρ ∈ γ(pt).
    Context (ρn : cell → num_val).
    Hypothesis NM: ρn ∈ γ(nm).
    Hypothesis Nc: N.compat _ tp (λ x, x) ρ ρn.

    Lemma Tmp: ∀ i v, tmp ! i = Some v → ρ (ACtemp μ i) = v.
    Proof.
      intros i v H.
      subst ρ.
      simpl. rewrite dec_eq_true, H. reflexivity.
    Qed.
    Hint Resolve Tmp.

  Lemma is_null_correct ne :
    is_null ne = true →
    ∀ i, NVint i ∈ eval_nexpr ρn ne → i = Int.zero.
  Proof.
    unfold is_null.
    intros N i Hi. generalize (assume_bot NumDom false nm N NM). clear N.
    simpl. intros K.
    destruct (Int.cmpu Ceq i Int.zero) eqn: Hz.
    simpl in Hz. generalize (Int.eq_spec i Int.zero). rewrite Hz. exact id.
    elim K. econstructor; eauto. vauto.
    change Nfalse with (of_bool false). rewrite <- Hz. vauto.
  Qed.
  Arguments is_null_correct: clear implicits.

  Ltac with_is_null q :=
    repeat match goal with
    | H : appcontext[ is_null ?ne ] |- _ =>
      generalize (is_null_correct ne);
      let K := fresh in
      destruct (is_null ne);[intros K; specialize (K eq_refl)|intros _]; q
    end.

    Lemma ne_of_pe_correct pe :
      match ne_of_pe pe with
      | Result ne => N.nconvert _ tp (λ x, x) pe = Some ne ∧ ¬ block_nexpr ρn ne
      | Errorc _ _ => True
      end.
    Proof.
      unfold ne_of_pe.
      destruct N.nconvert as [ne'|]. 2: exact I.
      destruct (AbMachineEnv.nonblock ne' nm) eqn: NB. 2: exact I.
      split. reflexivity.
      exact (nonblock_correct _ NM NB).
    Qed.

    Ltac with_ne_of_pe q :=
      repeat match goal with
      | H : appcontext[ ne_of_pe ?pe ] |- _ =>
        generalize (ne_of_pe_correct pe);
        destruct (ne_of_pe pe); q
      end.

    Lemma ne_of_pe_list_nonblock lpe :
      match bm_get (ne_of_pe_list lpe) with
      | Some lne => (∀ ne, In ne lne → ¬ block_nexpr ρn ne)
                  ∧ (∀ pe, In pe lpe → ∃ ne, In ne lne ∧ N.nconvert _ tp (λ x, x) pe = Some ne)
      | None => True
      end.
    Proof.
      unfold ne_of_pe_list.
      with_bm_map'.
      intros (H & H'). split.
      intros ne NE. destruct (H ne NE) as (pe & PE & K).
      generalize (ne_of_pe_correct pe). rewrite PE. intuition.
      intros pe PE. destruct (H' pe PE) as (ne & NE & K).
      generalize (ne_of_pe_correct pe). rewrite NE.
      intros [? _]. vauto.
    Qed.

    Lemma ptype_correct (pe: pexpr cell) v :
      eval_pexpr ge m ρ e pe v → v ≠ Vundef → v ∈ γ(ptype pe).
    Proof.
      intros V D. eapply N.Eval.eval_expr_correct; eauto.
    Qed.

    Lemma check_align_correct al ne :
      al = 1 ∨ al = 2 ∨ al = 4 ∨ al = 8 →
      check_align al ne = true →
      ∀ n, eval_nexpr ρn ne (NVint n) →
           Int.unsigned n = al * Int.unsigned (Int.divu n (Int.repr al)).
    Proof.
      unfold check_align.
      apply is_one_elim. intros _ _ n _. rewrite Int.divu_one. Psatz.lia.
      clear al.
      intros al Hz Hsz AL n EV. rewrite orb_false_l in AL.
      clear -AL EV NM Hsz.
      pose proof (AbMachineEnv.assume_bot NumDom _ _ AL NM) as K. simpl in K.
      assert (Int.eq (Int.repr al) Int.zero = false)
      by intuition (subst al; reflexivity).
      destruct (Int.cmp Cne (Int.modu n (Int.repr al)) Int.zero) eqn: L.
      elim K. econstructor; vauto.
      replace Ntrue with (of_bool (Int.cmp Cne (Int.modu n (Int.repr al)) Int.zero)).
      vauto. rewrite L. reflexivity.
      simpl in L. apply negb_false_iff in L.
      generalize (Int.eq_spec (Int.modu n (Int.repr al)) Int.zero). rewrite L. clear L.
      intro.
      assert (0 < al <= Int.max_unsigned).
      { destruct Hsz as [|[|[|]]]; subst; compute; split; auto; discriminate. }
      assert (Int.repr al ≠ Int.zero) as Hal.
      clear -Hsz. repeat destruct Hsz as [ Hsz | Hsz ]; rewrite Hsz; discriminate.
      rewrite (Int.modu_divu_Euclid n (Int.repr al)) at 1; auto. rewrite H0, Int.add_zero.
      unfold Int.mul. rewrite Int.unsigned_repr, Z.mul_comm, Int.unsigned_repr. auto.
      Psatz.lia.
      unfold Int.divu. rewrite !Int.unsigned_repr; try Psatz.lia.
      split.
      assert (0 <= Int.unsigned n / al).
      apply Z.div_pos. pose proof (Int.unsigned_range n). omega. omega.
      Psatz.nia.
      assert (Int.unsigned n / al * al = Int.unsigned n - Int.unsigned n mod al).
      { rewrite Zmod_eq; omega. }
      rewrite H2. pose proof (Int.unsigned_range n). unfold Int.max_unsigned.
      destruct (Z_mod_lt (Int.unsigned n) al). omega. omega.
      split.
      apply Z.div_pos. pose proof (Int.unsigned_range n). Psatz.lia.
      destruct Hsz as [Hsz|[Hsz|[Hsz|Hsz]]]; subst al; reflexivity.
      apply Z.div_le_upper_bound. rewrite Int.unsigned_repr; Psatz.lia.
      pose proof (Int.unsigned_range n).
      rewrite Int.unsigned_repr by Psatz.lia.
      unfold Int.max_unsigned. Psatz.nia.
    Qed.

    Lemma is_valid_ptr_valid weak bs off :
      is_valid_ptr weak bs off = true →
      ∀ ab, BlockSet.mem ab bs →
      ∀ b, block_of_ablock ge ((μ, (tmp, e)) :: stk) ab = Some b →
      ∀ i, eval_nexpr ρn off (NVint i) →
           if weak
           then Mem.weak_valid_pointer m b (Int.unsigned i)
           else Mem.valid_pointer m b (Int.unsigned i).
    Proof.
      unfold is_valid_ptr.
      generalize (min_size_min bs MaxSize). simpl.
      destruct (min_size bs MaxSize) as [|z|]; intros [H _] LT; eq_some_inv.
      destruct z as [|min|]; eq_some_inv.
      rewrite !andb_true_iff in LT. destruct LT as [Hz LT].
      rewrite Pos2Z.inj_ltb, Z.ltb_lt in Hz.
      pose proof (assume_bot NumDom _ _ LT NM) as K. clear LT.
      intros ab Hab b Hb i EV.
      specialize (H _ Hab). simpl in H.
      generalize (λ hi rw EQ, SZ ab (hi:=hi) (rw:=rw) EQ Hb).
      unfold ABTreeDom.get. destruct (ABTree.get ab sz) as [(hi, rw)|]. 2: exact (False_ind _ H).
      simpl in H.
      intros Hperm. specialize (Hperm _ _ eq_refl).
      unfold Mem.weak_valid_pointer, Mem.valid_pointer.
      destruct weak.
      *
      generalize (Int.unsigned_range i). intros [Hi Hi'].
      assert (0 < Int.unsigned i ∨ 0 = Int.unsigned i) as [Hi''|Hi''].  Psatz.lia.
      destruct (Mem.perm_dec _ _ (_ - 1) _ _) as [|Q]. rewrite orb_true_r. reflexivity.
      destruct (Int.cmpu Clt i (Int.repr (Zpos (Pos.succ min)))) eqn: Hmin.
      2: elim K; rewrite <- Hmin; vauto.
      elim Q; clear Q.
      simpl in Hmin. unfold Int.ltu in Hmin. rewrite Int.unsigned_repr in Hmin.
      2: change Int.max_unsigned with 4294967295; Psatz.lia.
      destruct (zlt (Int.unsigned i) _); eq_some_inv. clear Hmin.
      specialize (Hperm (Int.unsigned i - 1)).
      eapply Mem.perm_implies. eapply Hperm.  Psatz.lia. destruct ab; vauto.
      destruct (Mem.perm_dec _ _ (Int.unsigned i) _ _) as [|Q]. reflexivity.
      destruct (Int.cmpu Clt i (Int.repr (Zpos (Pos.succ min)))) eqn: Hmin.
      2: elim K; rewrite <- Hmin; vauto.
      elim Q; clear Q.
      simpl in Hmin. unfold Int.ltu in Hmin. rewrite Int.unsigned_repr in Hmin.
      2: change Int.max_unsigned with 4294967295; Psatz.lia.
      destruct (zlt (Int.unsigned i) _); eq_some_inv. clear Hmin.
      specialize (Hperm (Int.unsigned i)).
      eapply Mem.perm_implies. eapply Hperm.  Psatz.lia. destruct ab; vauto.
       *
      destruct (Mem.perm_dec _ _ _ _ _) as [|Q]. reflexivity.
      elim Q; clear Q.
      destruct (Int.cmpu Clt i (Int.repr (Zpos min))) eqn: Hi.
      2: elim K; rewrite <- Hi; vauto. clear K.
      simpl in Hi. unfold Int.ltu in Hi. rewrite Int.unsigned_repr in Hi.
      2: change Int.max_unsigned with 4294967295; Psatz.lia.
      destruct (zlt (Int.unsigned i)); eq_some_inv. clear Hi.
      specialize (Hperm (Int.unsigned i)).
      eapply Mem.perm_implies. eapply Hperm.
      hsplit. split. apply Int.unsigned_range. Psatz.lia. destruct ab; vauto.
    Qed.

    Lemma convert_wf (q: expr) :
      match bm_get (convert q) with
      | Some lp => ∀ pe, In pe lp → N.nconvert_def_pre ge e m ρ pe
      | None => True
      end.
    Proof.
      induction q.
    - simpl.
      set (c := ACtemp μ i).
      generalize (TP c).
      destruct (ACTreeDom.get c tp) as [ty|]; simpl. exact id.
      simpl. intros TY.
      intros pe [ <- | () ]. simpl. intros X. rewrite X in TY.
      AbTy.gamma_inv. discriminate.
    - simpl.
      generalize (E i). destruct is_local.
      intros (A & _) pe [ <- | () ]. simpl. unfold env_as_fun.
      destruct (e ! i) as [()|]. intro; eq_some_inv.
      elim (A eq_refl eq_refl).
      intros _.
      destruct (Genv.find_symbol ge i) eqn: FS.
      intros pe [ <- | () ]. simpl. rewrite FS. intro; eq_some_inv.
      exact I.
    - intros pe [ <- | () ].
      destruct c; exact I.
    - simpl.
      apply bm_bind_case. intros; exact I.
      intros lp LP pe PE.
      rewrite rev_map_correct, <- in_rev, in_map_iff in PE.
      destruct PE as (pe' & <- & PE).
      rewrite LP in IHq. simpl. apply IHq; auto.
    - simpl.
      apply bm_bind_case. intros; exact I.
      intros lp1 LP1.
      apply bm_bind_case. intros; exact I.
      intros lp2 LP2.
      rewrite LP1 in IHq1; rewrite LP2 in IHq2.
      destruct b;
      try (
        intros pe PE;
        rewrite map2_correct in PE; unfold map2_spec in PE;
        rewrite <- in_rev, in_flat_map in PE;
        destruct PE as (pe1 & PE1 & PE);
        rewrite in_map_iff in PE;
        destruct PE as (pe2 & <- & PE);
        vauto).
      with_bm_map'. intros [X _] pe PE.
      destruct (X _ PE) as ((pe1 & pe2) & Hpe & Hpe'). clear X.
      destruct (ptype pe1) as [|()] eqn: Hpe1; result_eq_inv;
      destruct (ptype pe2) as [|()] eqn: Hpe2; result_eq_inv.
      * (* int, int *)
      subst pe. simpl in *. apply map2_in in Hpe'. hsplit. eq_some_inv. subst.
      split. auto. split. auto.
      intros b1 i1 b2 i2 H0 H1. exfalso.
      generalize (ptype_correct H0). rewrite Hpe1. intros X. refine (X _). discriminate.
      * (* ptr, int *)
      with_ne_of_pe result_eq_inv.
      intros (Hpe2ne & NB2) (Hpe1ne & NB1).
      with_is_null result_eq_inv.
      destruct c; result_eq_inv;
      subst pe; vauto.
      * (* ptr, ptr *)
      with_ne_of_pe result_eq_inv.
      intros (Hpe2ne & NB2) (Hpe1ne & NB1).
      destruct (pt_eval_pexpr ge μ pt pe1) as [|xb]eqn: Hpt1; result_eq_inv.
      destruct (pt_eval_pexpr ge μ pt pe2) as [|yb]eqn: Hpt2; result_eq_inv.
      rewrite map2_correct in Hpe'; unfold map2_spec in Hpe';
      rewrite <- in_rev, in_flat_map in Hpe';
      destruct Hpe' as (pe1' & PE1' & Hpe').
      rewrite in_map_iff in Hpe';
      destruct Hpe' as (pe2' & ? & PE2');
      eq_some_inv; subst pe1' pe2'.
      specialize (IHq1 _ PE1').
      specialize (IHq2 _ PE2').
      generalize (@cmp_blockset_correct xb yb). destruct (cmp_blockset xb yb); result_eq_inv.
      intros EQ.
      revert Hpe. bif'. intros X; result_eq_inv. subst. InvBooleans.
      split. assumption. split. assumption.
      intros b1 i1 b2 i2 H1 H2.
      generalize (pt_eval_pexpr_correct PT H1). rewrite Hpt1. intros (ab1 & Habx & Hab1).
      generalize (pt_eval_pexpr_correct PT H2). rewrite Hpt2. intros (ab2 & Haby & Hab2).
      specialize (EQ _ _ Habx Haby). hsplit. subst.
      assert (b1 = b2) by congruence. subst.
      split. reflexivity.
      generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ H1 _ Hpe1ne).
      intros (n1 & Hn1 & Evn1 & Ht1). discriminate. inv Hn1.
      generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ H2 _ Hpe2ne).
      intros (n2 & Hn2 & Evn2 & Ht2). discriminate. inv Hn2.
      rewrite (is_valid_ptr_valid _ _ H _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ H0 _ Haby Hab2 Evn2). reflexivity.
      intros NE.
      revert Hpe. bif'. intros ?; result_eq_inv. destruct c; result_eq_inv; subst.
      split. easy. split. easy. intros. eval_pexpr_inv. discriminate.
      split. easy. split. easy. intros. eval_pexpr_inv. discriminate.
      subst. vauto.
    - rename m0 into κ.
      simpl.
      apply bm_bind_case. intros; exact I.
      intros lp LP.
      apply bm_bind_case. intros; exact I.
      intros ptr' PTR.
      apply bm_bind_case. intros; exact I.
      intros κ' K. assert (Hκ:∃ Hκ, κ' = Some (exist _ κ Hκ)) by (destruct κ; inv K; eauto).
      destruct Hκ as [Hκ ->]. clear K.
      destruct ptr' as [|ptr']. exact I.
      apply bm_bind_case. intros; exact I.
      intros [|cells] Hcells. exact I.
      generalize (realize_inv (exist _ κ Hκ) lp ptr').
      rewrite Hcells.
      intros K.
      with_bm_map'.
      simpl.
      intros (X & Y) pe PE.
      destruct (X pe PE) as (c & C & C').
      destruct (ACTreeDom.get c tp) as [|ty] eqn: TY; inv C.
      unfold N.nconvert_def_pre.
      rewrite LP in IHq. simpl in IHq.
      generalize (TP c). rewrite TY. intros A L. rewrite L in A.
      AbTy.gamma_inv. discriminate.
    Qed.

    Remark align_chunk_4_cases (κ: memory_chunk) :
      align_chunk κ = 1 ∨ align_chunk κ = 2 ∨ align_chunk κ = 4 ∨ align_chunk κ = 8.
    Proof. destruct κ; vauto. Qed.

    Lemma ty_Just_def (t: AbTy.t) v :
      v ∈ γ(Just t) → v ≠ Vundef.
    Proof. intros; destruct v; AbTy.gamma_inv; intuition congruence. Qed.
    Hint Resolve ty_Just_def.

    Lemma convert_correct (q: expr) :
      match bm_get (convert q) with
      | Some lp =>
        ∀ v, eval_expr ge e tmp m q v →
             ∃ a, In a lp ∧ eval_pexpr ge m ρ e a v
      | None => True
      end
    ∧
      match bm_get (eval_expr_ptr q) with
      | Some (Just ptr) =>
         ∀ b i, eval_expr ge e tmp m q (Vptr b i) →
                ∃ ab, BlockSet.mem ab ptr
                    ∧ block_of_ablock ge ((μ, (tmp, e)) :: stk) ab = Some b
      | Some All
      | None => True
      end.
    Proof.
      Ltac ok :=
        let K := fresh in
        intros ? K; eexists; split; [left; reflexivity| eval_expr_inv; try econstructor; eauto].
      induction q; simpl;
      try (split;[ok|auto]).
    - set (c := ACtemp μ i).
      split.
      destruct (ACTreeDom.get c tp). exact I.
      subst c. intros ? K.
      eval_expr_inv. vauto2.
      generalize (PT c). fold (points_to_get pt c).
      destruct (pt c) as [|C]. exact (λ _, I).
      unfold γ. simpl. intros HC b j K. eval_expr_inv.
      subst c. rewrite (Tmp _ K) in HC. eauto.
    - split.
      bif'. ok. unfold env_as_fun. apply E in Heqb. subst. constructor.
      inv H. rewrite H0. auto. congruence.
      bif. simpl.
      ok. inv H. exfalso. generalize (proj2 (E i)). intuition congruence.
      simpl. rewrite H1. reflexivity.
      exact I.
      intros b i0 H. inv H. unfold addr_of. inv H2.
      rewrite (proj2 (E i)). 2: congruence.
      eexists. split. apply BSO.mem_singleton. eauto. simpl. rewrite dec_eq_true, H. reflexivity.
      generalize (proj1 (E i)). rewrite H. bif. intuition. intros _.
      eexists. rewrite H0. split. apply BSO.mem_singleton. eauto. simpl.
      rewrite (Genv.find_invert_symbol _ _ H0). reflexivity.
    - destruct c; simpl in *; eq_some_inv; auto.
    - inversion 1. subst. destruct c; simpl in *; discriminate.
    - split.
      apply bm_bind_case. intros; exact I.
      intros lp Hlp. rewrite Hlp in IHq.
      intros v K. inv K. destruct (proj1 IHq _ H1) as (a & A & B). clear IHq.
      exists (PEunop u a). split.
      rewrite rev_map_correct, <- in_rev. exact (in_map _ _ _ A).
      econstructor; eauto.
      intros b i H. exfalso.
      apply (eval_expr_unop_ptr H).
    - destruct (convert q1) as [lp1 LP1] eqn: Hq1.
      destruct (convert q2) as [lp2 LP2] eqn: Hq2.
      destruct (eval_expr_ptr q1) as [e1' E1'].
      destruct (eval_expr_ptr q2) as [e2' E2'].
      simpl in *.
      destruct IHq1 as [IH1 IH1'].
      destruct IHq2 as [IH2 IH2'].
      split.
      + destruct LP1, LP2;
        try (simpl; destruct b; try exact I; destruct E1', E2', bm_map'; exact I).
        destruct b;
          try (
          intros v V; eval_expr_inv;
          destruct (IH1 _ H) as (a1 & A1 & A1');
          destruct (IH2 _ H0) as (a2 & A2 & A2');
          try (eexists; split; [eapply in_map2; eassumption | econstructor; eauto])
            ).
        with_bm_map'. intros [_ Y].
        intros v V; eval_expr_inv;
        destruct (IH1 _ H) as (a1 & A1 & A1');
        destruct (IH2 _ H0) as (a2 & A2 & A2');
        destruct (Y (a1, a2)) as (a & Ha & Ha');
          [apply in_map2; eassumption|].
        clear Y.
        exists a; split; [ exact Ha' | ].
        generalize (ptype_correct A2').
        generalize (ptype_correct A1').
        destruct (ptype a1) as [|()] eqn: Ha1; result_eq_inv;
        destruct (ptype a2) as [|()] eqn: Ha2; result_eq_inv;
        intros Ha1' Ha2'.
        * (* int, int *)
          subst. econstructor (eauto).
          destruct v1; try assumption.
          destruct v2; try assumption.
          elim (Ha2'). discriminate.
          elim (Ha1'). discriminate.
        * (* ptr, int *)
          with_ne_of_pe result_eq_inv.
          intros (Ha2n2 & NB2) (Ha1n1 & NB1).
          with_is_null result_eq_inv.
          generalize (convert_wf q1). rewrite Hq1. intros K1. specialize (K1 _ A1).
          generalize (convert_wf q2). rewrite Hq2. intros K2. specialize (K2 _ A2).
          pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K2 _ Ha2n2 NB2) as Q2.
          pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K1 _ Ha1n1 NB1) as Q1.
          pose proof (block_pexpr_def Q1 A1') as Def1.
          pose proof (block_pexpr_def Q2 A2') as Def2.
          specialize (Ha1' Def1). specialize (Ha2' Def2).
          AbTy.gamma_inv. hsplit. subst.
          simpl in V. eq_some_inv. unfold Val.cmpu in V. simpl in V.
          generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ A2' _ Ha2n2 Def2).
          intros (n2 & Hn2 & Evn2 & Ht2). inv Hn2. specialize (H1 _ Evn2). subst.
          rewrite Int.eq_true.
          destruct c; result_eq_inv; subst; vauto.
        * with_ne_of_pe result_eq_inv.
          intros (Ha2n2 & NB2) (Ha1n1 & NB1).
          destruct (pt_eval_pexpr ge μ pt a1) as [|xb]eqn: Hpt1; result_eq_inv.
          destruct (pt_eval_pexpr ge μ pt a2) as [|yb]eqn: Hpt2; result_eq_inv.
          generalize (convert_wf q1). rewrite Hq1. intros K1. specialize (K1 _ A1).
          generalize (convert_wf q2). rewrite Hq2. intros K2. specialize (K2 _ A2).
          pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K2 _ Ha2n2 NB2) as Q2.
          pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K1 _ Ha1n1 NB1) as Q1.
          pose proof (block_pexpr_def Q1 A1') as Def1.
          pose proof (block_pexpr_def Q2 A2') as Def2.
          specialize (Ha1' Def1). specialize (Ha2' Def2).
          AbTy.gamma_inv. hsplit. subst.
          generalize (pt_eval_pexpr_correct PT A1'). rewrite Hpt1. intros (ab1 & Habx & Hab1).
          generalize (pt_eval_pexpr_correct PT A2'). rewrite Hpt2. intros (ab2 & Haby & Hab2).
          generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ A1' _ Ha1n1 Def1).
          intros (n1 & Hn1 & Evn1 & _). inv Hn1.
          generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ A2' _ Ha2n2 Def2).
          intros (n2 & Hn2 & Evn2 & _). inv Hn2.
          generalize (@cmp_blockset_correct xb yb). destruct (cmp_blockset xb yb); result_eq_inv.
          destruct (is_valid_ptr  _ xb _) eqn: Xvalid; simpl in Ha; result_eq_inv.
          destruct (is_valid_ptr  _ yb _) eqn: Yvalid; simpl in Ha; result_eq_inv.
          subst. intros K. specialize (K _ _ Habx Haby). hsplit. subst.
          assert (b0 = b) by congruence. subst.
          simpl in V. eq_some_inv. subst.
          econstructor (eauto). simpl. unfold Val.cmpu. simpl.
          destruct eq_block as [K|K]. 2: elim (K eq_refl). simpl. clear K.
          destruct eq_block as [K|K]. 2: elim (K eq_refl). simpl. clear K.
          fold (Mem.weak_valid_pointer m b (Int.unsigned i)).
          fold (Mem.weak_valid_pointer m b (Int.unsigned i0)).
          rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2).
          reflexivity.
          destruct (is_valid_ptr  _ xb _) eqn: Xvalid; simpl in Ha; result_eq_inv.
          destruct (is_valid_ptr  _ yb _) eqn: Yvalid; simpl in Ha; result_eq_inv.
          destruct c; result_eq_inv; simpl in V; eq_some_inv; subst; intros K.
          assert (ab1 ≠ ab2) by intuition eauto. clear K.
          assert (b ≠ b0) by eauto using invariant_block_of_ablock_inj.
          econstructor (vauto). simpl. f_equal.
          unfold Val.cmpu, Val.cmpu_bool. simpl. rewrite Int.eq_false. 2: clear; discriminate.
          destruct eq_block. easy.
          idtac.
          rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2).
          reflexivity.
          assert (ab1 ≠ ab2) by intuition eauto. clear K.
          assert (b ≠ b0) by eauto using invariant_block_of_ablock_inj.
          econstructor (vauto). simpl. f_equal.
          unfold Val.cmpu, Val.cmpu_bool. simpl. rewrite Int.eq_false. 2: clear; discriminate.
          destruct eq_block. easy.
          rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2).
          reflexivity.
          intros K. eelim K; eassumption.
      + destruct E1', E2'; try exact I.
        simpl.
        destruct e1' as [|e1']. exact I.
        destruct e2' as [|e2']. exact I.
        simpl. intros b' i B'.
        destruct (eval_expr_binop_ptr' B') as (j & [B|B]);
          [destruct (IH1' _ _ B) as (ab & IN & BA)
          |destruct (IH2' _ _ B) as (ab & IN & BA)];
        eexists; split; eauto; apply BlockSet.mem_union; auto.
  - split.
    + rename m0 into κ.
      destruct IHq as [IHq IHq'].
      apply bm_bind_case. intros; exact I.
      intros lp Hlp. rewrite Hlp in IHq.
      apply bm_bind_case. intros; exact I.
      intros ptr' Hptr. rewrite Hptr in IHq'.
      apply bm_bind_case. intros; exact I.
      intros κ' K. assert (Hκ:∃ Hκ, κ' = Some (exist _ κ Hκ)) by (destruct κ; inv K; eauto).
      destruct Hκ as [Hκ ->]. clear K.
      destruct ptr' as [|ptr']. exact I.
      apply bm_bind_case. intros; exact I.
      intros [|cells] Hcells. exact I.
      generalize (realize_inv (exist _ κ Hκ) lp ptr').
      rewrite Hcells.
      intros Q.
      with_bm_map'.
      intros (X & Y).
      intros v H; inv H.
      destruct v1; try discriminate.
      specialize (IHq _ H2).
      destruct IHq as (a & A & B).
      specialize (IHq' _ _ H2). destruct IHq' as (ab & IN & BA).
      specialize (Q _ A). destruct Q as (Q & ne & NE & CA & CI & CS).
      generalize (ne_of_pe_correct a). rewrite NE. intros (NC & NB).
      exploit @N.nconvert_correct; try eassumption.
      discriminate.
      intros (n & Z & EV & ?); inv Z; try eassumption.
      set (c := cell_from ab (IntSet.proj i, exist _ κ Hκ)).
      edestruct (Y c) as (pe & PE).
      apply CellSet.elements_spec. subst c. apply CS. apply IN.
      eapply concretize_int_correct.
      eassumption. assumption.
      destruct PE as [PE PE'].
      destruct (ACTreeDom.get c tp); inversion PE; clear PE. subst pe.
      exists (PEvar c). split. exact PE'.
      subst c.
      constructor.
      simpl in H4.
      destruct ab as [f loc|b'].
      * simpl in BA.
        unfold ρ. simpl.
        destruct (eq_dec μ f).
        destruct (e ! loc) as [(b' & _)|]; eq_some_inv.
        subst b'. unfold IntSet.proj. rewrite H4. reflexivity.
        match goal with |- context[match ?a with Some _ => _ | _ => _ end] =>
                        destruct a as [(_, e')|]
        end.
        destruct (e' ! loc) as [(? & _)|]; eq_some_inv. subst.
        unfold IntSet.proj; rewrite H4; reflexivity.
        eq_some_inv.
      * unfold ρ. simpl in *. destruct (Genv.invert_symbol); eq_some_inv. subst. unfold IntSet.proj. rewrite H4.
        reflexivity.
    + rename m0 into κ.
      apply bm_bind_case. intros (); intros; exact I. intros e' E'.
      apply bm_bind_case. intros (); intros; exact I. intros ptr Hptr.
      apply bm_bind_case. intros (); intros; exact I.
      intros κ' K. assert (Hκ:∃ Hκ, κ' = Some (exist _ κ Hκ)) by (destruct κ; inv K; eauto).
      destruct Hκ as [Hκ ->]. clear K.
      destruct ptr as [|ptr]. exact I.
      apply bm_bind_case. intros (); intros; exact I. intros [|cells] Hcells. exact I.
      rewrite bm_get_ret.
      simpl.
      generalize (realize_inv (exist _ κ Hκ) e' ptr). rewrite Hcells.
      simpl.
      intros Q.
      simpl. destruct_noitpo_in_goal_eq. intros _; exact I.
      intros blk Hblk'.
      pose proof (BSO.union_list_noitpo_Just_inv _ Hblk') as Hblk; clear Hblk'.
      intros b i H1. inv H1.
      destruct v1; try discriminate.
      destruct IHq as [IH HI]. rewrite Hptr in HI. rewrite E' in IH.
      destruct (IH _ H2) as (a & IN & EV).
      destruct (HI _ _ H2) as (ab & IN' & AB). clear H2 IH HI.
      specialize (Q _ IN).
      destruct Q as (Q & ne & NE & CA & CI & CS).
      generalize (ne_of_pe_correct a). rewrite NE. intros (NC & NB).
      exploit @N.nconvert_correct; try eassumption.
      discriminate.
      intros (n & Y & EV' & ?); inv Y; try eassumption.
      set (c := cell_from ab (IntSet.proj i0, exist _ κ Hκ)).
      edestruct Hblk as (bs & BS & Hbs).
      { rewrite rev_map_correct, <- in_rev, in_map_iff.
        exists c. split. reflexivity.
        rewrite <- CellSet.elements_spec.
        subst c.
        apply CS; auto.
        eapply concretize_int_correct. eassumption.
        assumption.
      }
      assert (mem_as_fun ge ((μ, (tmp, e)) :: stk, m) c = Vptr b i) as HC.
      {
        unfold Mem.loadv in *. unfold IntSet.proj in *.
        subst c.
        destruct ab; simpl in *; try (destruct (eq_dec μ f));
        repeat match goal with |- appcontext[match ?a with Some _ => _ | _ => _ end] =>
                               destruct a as [(? & ?)|]
               end; eq_some_inv; ssubst; auto;
        try (rewrite H4; reflexivity).
        destruct Genv.invert_symbol; eq_some_inv. subst.
        rewrite H4; reflexivity.
      }
      generalize (PT c). fold (points_to_get pt c). unfold ρ.
      rewrite HC, BS.
      intros (ab' & Hab' & AB').
      exists ab'.
      split. 2: exact AB'.
      eapply Hbs, Hab'.
    Qed.

    Lemma convert_nonblock (q: expr) :
      match bm_get (convert q) with
      | Some lp => (∀ pe, In pe lp → ¬ block_pexpr ge m ρ e pe) → ¬ block_expr ge e tmp m q
      | None => True
      end.
    Proof.
      unfold ρ.
      induction q.
    - simpl. set (c := ACtemp μ i). destruct (ACTreeDom.get c tp) eqn: TY.
      exact I. intros NB. specialize (NB _ (or_introl eq_refl)).
      intros B. subst c.
      inv B; apply NB.
      econstructor. simpl. rewrite dec_eq_true, H0. reflexivity.
      econstructor. simpl. rewrite dec_eq_true, H0. reflexivity.
    - generalize (E i). simpl. destruct is_local.
      intros [K _]. specialize (K eq_refl).
      intros _ B; inv B. intuition.
      intros [_ K]. destruct (Genv.find_symbol ge i) eqn: FS.
      intros _ B; inv B. congruence.
      exact I.
    - intros _ B. inversion B.
    - simpl. apply bm_bind_case. intros; exact I.
      intros lp LP. rewrite LP in IHq. simpl in IHq.
      simpl. intros NB.
      assert ( ∀ pe, In pe lp → ¬ block_pexpr ge m ρ e pe) as NB'.
      {
        intros pe PE B. eapply NB.
        rewrite rev_map_correct, <- in_rev, in_map_iff. eexists. split. reflexivity. exact PE.
        vauto.
      }
      specialize (IHq NB').
      intros B. apply block_expr_Unop_inv in B.
      destruct B as [(v & V & B) | B]. 2: auto.
      generalize (proj1 (convert_correct q)). rewrite LP. simpl.
      intros X. specialize (X _ V). destruct X as (pe & PE & EV).
      eapply NB.
      rewrite rev_map_correct, <- in_rev, in_map_iff. eexists. split. reflexivity. exact PE.
      vauto.

    - simpl. apply bm_bind_case. intros; exact I.
      intros lp1 LP1. rewrite LP1 in IHq1. simpl in IHq1.
      apply bm_bind_case. intros; exact I.
      intros lp2 LP2. rewrite LP2 in IHq2. simpl in IHq2.
      generalize (convert_correct q2). rewrite LP2. intros [H2 _].
      destruct lp2 as [|pe2 lp2].
      destruct (nonblock_eval_expr (IHq2 (λ _ K, False_ind _ K))) as (v2 & Hv2).
      destruct (H2 v2 Hv2) as (? & () & _).
      generalize (convert_correct q1). rewrite LP1. intros [H1 _].
      destruct lp1 as [|pe1 lp1].
      destruct (nonblock_eval_expr (IHq1 (λ _ K, False_ind _ K))) as (v1 & Hv1).
      destruct (H1 v1 Hv1) as (? & () & _).
      unfold bm_get in H1, H2.
      destruct b;
        try (with_bm_map'; intros [_ Y]);
        intros NB;
        try (
      assert ( ∀ pe, In pe (pe1::lp1) → ¬ block_pexpr ge m ρ e pe) as NB1
      by (intros pe PE B; eapply NB; [ apply in_map2; first [ eassumption | left; reflexivity] | vauto]);
      assert ( ∀ pe, In pe (pe2::lp2) → ¬ block_pexpr ge m ρ e pe) as NB2
      by (intros pe PE B; eapply NB; [ apply in_map2; first [ eassumption | left; reflexivity] | vauto]);
      specialize (IHq1 NB1); specialize (IHq2 NB2)
          );
      intros B; apply block_expr_Binop_inv in B;
      destruct B as [(v & w & V & W & B) | [B | B] ]; auto;
      try (
      specialize (H1 _ V); destruct H1 as (pe1' & PE1 & EV1);
      specialize (H2 _ W); destruct H2 as (pe2' & PE2 & EV2)
        );
      try (
      (eapply NB; [apply in_map2; eauto | vauto])
          ).
      + destruct (Y (pe1', pe2')). apply in_map2; assumption.
      clear Y. hsplit.
      generalize (ptype_correct EV2).
      generalize (ptype_correct EV1).
      destruct (ptype pe1') as [|()] eqn: Hpe1; result_eq_inv;
      destruct (ptype pe2') as [|()] eqn: Hpe2; result_eq_inv;
      intros Ha1' Ha2'.
      * (* int, int *)
        subst. eapply NB. eauto.
        econstructor (eauto).
        destruct v; try exact B.
        destruct w; try exact B.
        elim (Ha2'); discriminate.
        elim (Ha1'); discriminate.
      * (* ptr, int *)
        with_ne_of_pe result_eq_inv. intros (Hpe2' & NB2) (Hpe1' & NB1).
        with_is_null result_eq_inv.
        generalize (convert_wf q1). rewrite LP1. intros K1. specialize (K1 _ PE1).
        generalize (convert_wf q2). rewrite LP2. intros K2. specialize (K2 _ PE2).
        pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K2 _ Hpe2' NB2) as Q2.
        pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K1 _ Hpe1' NB1) as Q1.
        pose proof (block_pexpr_def Q1 EV1) as Def1.
        pose proof (block_pexpr_def Q2 EV2) as Def2.
        specialize (Ha1' Def1). specialize (Ha2' Def2).
        AbTy.gamma_inv. hsplit. subst.
        generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ EV2 _ Hpe2' Def2).
        intros (n2 & Hn2 & Evn2 & Ht2). inv Hn2. specialize (H1 _ Evn2). subst.
        inv B; try intuition (hsplit; intuition discriminate).
        clear -H4; intuition; eauto.
        destruct c; result_eq_inv; intuition.
      * (* ptr, ptr *)
        with_ne_of_pe result_eq_inv.
        intros (Ha2n2 & NB2) (Ha1n1 & NB1).
        destruct (pt_eval_pexpr ge μ pt pe1') as [|xb]eqn: Hpt1; result_eq_inv.
        destruct (pt_eval_pexpr ge μ pt pe2') as [|yb]eqn: Hpt2; result_eq_inv.
        generalize (convert_wf q1). rewrite LP1. intros K1. specialize (K1 _ PE1).
        generalize (convert_wf q2). rewrite LP2. intros K2. specialize (K2 _ PE2).
        pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K2 _ Ha2n2 NB2) as Q2.
        pose proof (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc _ K1 _ Ha1n1 NB1) as Q1.
        pose proof (block_pexpr_def Q1 EV1) as Def1.
        pose proof (block_pexpr_def Q2 EV2) as Def2.
        specialize (Ha1' Def1). specialize (Ha2' Def2).
        AbTy.gamma_inv. hsplit. subst.
        generalize (pt_eval_pexpr_correct PT EV1). rewrite Hpt1. intros (ab1 & Habx & Hab1).
        generalize (pt_eval_pexpr_correct PT EV2). rewrite Hpt2. intros (ab2 & Haby & Hab2).
        generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ EV1 _ Ha1n1 Def1).
        intros (n1 & Hn1 & Evn1 & _). inv Hn1.
        generalize (N.nconvert_correct _ _ _ _ _ _ _ TP _ Nc _ _ EV2 _ Ha2n2 Def2).
        intros (n2 & Hn2 & Evn2 & _). inv Hn2.
        generalize (@cmp_blockset_correct xb yb). destruct (cmp_blockset xb yb); result_eq_inv.
        destruct (is_valid_ptr  _ xb _) eqn: Xvalid; simpl in H; result_eq_inv.
        destruct (is_valid_ptr  _ yb _) eqn: Yvalid; simpl in H; result_eq_inv.
        subst. intros K. specialize (K _ _ Habx Haby). hsplit. subst.
        assert (b0 = b) by congruence. subst.
        inv B; try intuition (hsplit; intuition discriminate).
        clear -H4; intuition; eauto.
        fold (Mem.weak_valid_pointer m b (Int.unsigned i)) in H2.
        fold (Mem.weak_valid_pointer m b (Int.unsigned i0)) in H2.
        rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2) in H2. simpl in H2.
        intuition eq_some_inv.
        destruct (is_valid_ptr  _ xb _) eqn: Xvalid; simpl in H; result_eq_inv.
        destruct (is_valid_ptr  _ yb _) eqn: Yvalid; simpl in H; result_eq_inv.
        destruct c; result_eq_inv; subst; intros K;
        inv B; try intuition (hsplit; intuition discriminate).
        clear -H4; intuition; eauto.
        rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2) in H6.
        intuition eq_some_inv.
        rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2) in H2. simpl in H2.
        intuition eq_some_inv.
        clear -H4; intuition; eauto.
        rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2) in H6.
        intuition eq_some_inv.
        rewrite (is_valid_ptr_valid _ _ Xvalid _ Habx Hab1 Evn1), (is_valid_ptr_valid _ _ Yvalid _ Haby Hab2 Evn2) in H2. simpl in H2.
        intuition eq_some_inv.
        intros K. eelim K; eassumption.
      + apply IHq1; auto.
        intros pe PE B'.
        destruct (Y (pe, pe2)). apply in_map2; (assumption || vauto).
        clear Y. hsplit.
        destruct (ptype pe) as [|()]; result_eq_inv;
        destruct (ptype pe2) as [|()]; result_eq_inv.
        * (* int, int *) subst. eapply NB. eauto. vauto.
        * (* ptr, int *)
          clear NB.
          with_ne_of_pe result_eq_inv. intros _ (X & NB).
          clear H.
          generalize (convert_wf q1). rewrite LP1. intros K. specialize (K _ PE).
          exact (N.nconvert_def _ _ _ _ _ _ _ TP _ Nc _ K _ X NB B').
        * (* ptr, ptr *)
          clear NB.
          with_ne_of_pe result_eq_inv. intros _ (X & NB).
          clear H.
          generalize (convert_wf q1). rewrite LP1. intros K. specialize (K _ PE).
          exact (N.nconvert_def _ _ _ _ _ _ _ TP _ Nc _ K _ X NB B').
      + apply IHq2; auto.
      intros pe PE B'.
      destruct (Y (pe1, pe)). apply in_map2; (assumption || vauto).
      clear Y. hsplit.
      destruct (ptype pe1) as [|()]; result_eq_inv;
      destruct (ptype pe) as [|()]; result_eq_inv.
      subst. eapply NB. eauto. vauto.
      clear NB.
      with_ne_of_pe result_eq_inv. intros (X & NB) _.
      clear H.
      generalize (convert_wf q2). rewrite LP2. intros K. specialize (K _ PE).
      exact (N.nconvert_def _ _ _ _ _ _ _ TP _ Nc _ K _ X NB B').
      clear NB.
      with_ne_of_pe result_eq_inv. intros (X & NB) _.
      clear H.
      generalize (convert_wf q2). rewrite LP2. intros K. specialize (K _ PE).
      exact (N.nconvert_def _ _ _ _ _ _ _ TP _ Nc _ K _ X NB B').
    - rename m0 into κ. simpl.
      apply bm_bind_case. intros; exact I.
      intros lp LP. rewrite LP in IHq. simpl in IHq.
      apply bm_bind_case. intros; exact I.
      intros ptr' PTR.
      apply bm_bind_case. intros; exact I.
      intros κ' K. assert (Hκ:∃ Hκ, κ' = Some (exist _ κ Hκ)) by (destruct κ; inv K; eauto).
      destruct Hκ as [Hκ ->]. clear K.
      destruct ptr' as [|ptr']. exact I.
      apply bm_bind_case. intros; exact I.
      intros [|cells] Hcells. exact I.
      generalize (realize_inv (exist _ κ Hκ) lp ptr').
      rewrite Hcells.
      intros Q.
      simpl.
      with_bm_map'.
      simpl.
      intros (X & Y) NB.
      assert ( ∀ pe, In pe lp → ¬ block_pexpr ge m ρ e pe) as NB'.
      {
        intros pe PE B.
        specialize (Q _ PE). destruct Q as (Q & ne & NE & CI & CS).
        generalize (ne_of_pe_correct pe). rewrite NE. intros (NC & NEB).
        eapply (λ P, N.nconvert_def ge e _ m ρ ρn tp TP _ Nc pe P ne NC NEB).
        generalize (convert_wf q); rewrite LP. simpl. intuition. exact B.
      }
      specialize (IHq NB').
      intros B. apply block_expr_Load_inv in B.
      destruct B as [(v' & V' & B) | B]. 2: easy.
      generalize (proj1 (convert_correct q)). rewrite LP. simpl.
      intros Z. specialize (Z _ V'). destruct Z as (pe & PE & EV).
      specialize (Q pe PE). destruct Q as (TY & ne & NE & CA & CI & OFS).
      generalize (ne_of_pe_correct pe). rewrite NE. clear NE. intros (NE & NEB).
      generalize (λ P, block_pexpr_def (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc pe P ne NE NEB) EV).
      generalize (convert_wf q); rewrite LP. simpl. intros WF.
      intros Z. specialize (Z (WF _ PE)).
      generalize (@N.Eval.eval_expr_correct _ AbTy.adom _ _ ge m e tp ρ pe _ TP EV Z).
      rewrite TY.
      destruct v'; destruct 1. clear TY.
      clear Z.
      generalize (proj2 (convert_correct q)). rewrite PTR. simpl.
      intros Z. specialize (Z _ _ V'). destruct Z as (bs & BS & Hb).
      exploit @N.nconvert_correct; try eassumption.
      discriminate.
      intros (n & Z & EV' & ?); inv Z; try eassumption.
      assert (∃ v, Mem.load κ m b (IntSet.proj i) = Some v ∧ v ≠ Vundef) as (v & V & DEF).
      {
        set (c := cell_from bs (IntSet.proj i, exist _ κ Hκ)).
        destruct (Y c) as (pe' & PE' & IN).
        rewrite <- CellSet.elements_spec. subst c. eapply OFS; auto.
        eapply concretize_int_correct; eauto.
        generalize (TP c).
        destruct (ACTreeDom.get) as [|ty]. discriminate. inversion PE'. subst pe'. clear PE'.
        unfold γ, ρ. simpl.
        destruct bs as [ f r | ab ]; simpl in *; eq_some_inv.
        * destruct @eq_dec. destruct (e ! r) as [(?,?)|]; eq_some_inv. ssubst.
          destruct (Mem.load) as [v|].
          intros V. exists v. split. reflexivity. intros ->. destruct ty; destruct V.
          destruct ty; destruct 1.
          destruct assoc as [(_, e')|]. destruct (e' ! r) as [(?,?)|]; eq_some_inv. subst b0.
          destruct (Mem.load) as [v|].
          intros V. exists v. split. reflexivity. intros ->. destruct ty; destruct V.
          destruct ty; destruct 1. eq_some_inv.
        * destruct Genv.invert_symbol; eq_some_inv. subst c ab.
          destruct (Mem.load) as [v|].
          intros V. exists v. split. reflexivity. intros ->. destruct ty; destruct V.
          destruct ty; destruct 1.
      }
      inv B.
      eelim H0; reflexivity.
      unfold IntSet.proj in *. congruence.
      unfold IntSet.proj in *. congruence.
    Qed.

    Lemma realize_cells_correct κ q :
      match bm_get (realize_cells κ q) with
      | Some cells =>
        ¬ block_expr ge e tmp m q ∧
        ∀ v, eval_expr ge e tmp m q v →
        ∃ ab b i, v = Vptr b i ∧
                  In (cell_from ab (Int.unsigned i, κ)) cells ∧
                  block_of_ablock ge ((μ, (tmp, e)) :: stk) ab = Some b ∧
                  (align_chunk (proj1_sig κ) | Int.unsigned i)
      | None => True
      end.
    Proof.
      unfold realize_cells.
      apply bm_bind_case. intros; exact I.
      intros lpe LPE.
      apply bm_bind_case. intros; exact I.
      intros [|ptr'] PTR. exact I.
      apply bm_bind_case. intros; exact I.
      intros [|cells] Hcells. exact I.
      generalize (realize_inv κ lpe ptr').
      rewrite Hcells.
      intros Q.
      generalize (convert_correct q). rewrite LPE, PTR. simpl.
      intros (CC & CC').
      split.
    - generalize (convert_nonblock q). rewrite LPE. simpl. intros X. apply X.
      intros pe PE.
      destruct (Q _ PE) as (TY & ne & NE & ? & OFS). clear Q.
      generalize (ne_of_pe_correct pe). rewrite NE. clear NE. intros (NE & NEB).
      generalize (λ P, N.nconvert_def ge e _ m ρ ρn tp TP _ Nc pe P ne NE NEB).
      generalize (convert_wf q); rewrite LPE. simpl. intros WF.
      intros Y. specialize (Y (WF _ PE)).
      exact Y.
    - intros v EV.
      specialize (CC _ EV).
      destruct CC as (pe & PE & EV').
      specialize (Q pe PE). destruct Q as (TY & ne & NE & CA & ? & OFS).
      generalize (ne_of_pe_correct pe). rewrite NE. clear NE. intros (NE & NEB).
      generalize (λ P, block_pexpr_def (N.nconvert_def ge e _ m ρ ρn tp TP _ Nc pe P ne NE NEB) EV').
      generalize (convert_wf q); rewrite LPE. simpl. intros WF.
      intros Y. specialize (Y (WF _ PE)).
      generalize (@N.Eval.eval_expr_correct _ AbTy.adom _ _ ge m e tp ρ pe v TP EV' Y).
      rewrite TY.
      destruct v; destruct 1. clear Y.
      specialize (CC' _ _ EV). destruct CC' as (ab & AB & Hab).
      eexists ab, b, i.
      exploit @N.nconvert_correct; try eassumption.
      discriminate.
      intros (n & Z & EVn & ?); inv Z; try eassumption.
      repeat match goal with |- _ ∧ _ => split end.
      reflexivity. 2: assumption.
      apply CellSet.elements_spec.
      apply OFS. apply AB.
      eapply concretize_int_correct; eauto.
      rewrite (check_align_correct (align_chunk_4_cases _) CA EVn).
      rewrite Z.mul_comm. eexists. reflexivity.
    Qed.

    Lemma nconvert_sound (q: expr) :
      match bm_get (nconvert q) with
      | Some _ => ¬ block_expr ge e tmp m q
      | None => True
      end.
    Proof.
      unfold nconvert. apply bm_bind_case. intros; exact I.
      intros lpe LPE.
      generalize (convert_nonblock q). rewrite LPE. simpl. intros H.
      generalize (ne_of_pe_list_nonblock lpe).
      destruct @bm_get as [lne|]. 2: exact id.
      intros [NBE NE].
      apply H. clear H.
      generalize (convert_wf q). rewrite LPE. simpl.
      intros H pe PE.
      specialize (NE _ PE). destruct NE as (ne & NE & Hne).
      eapply N.nconvert_def; eauto.
    Qed.

    Lemma nconvert_ex (q: expr) :
      match bm_get (nconvert q) with
      | Some lne =>
        ∃ lpe, convert q = ret_bm lpe ∧
        ∀ v, eval_expr ge e tmp m q v →
             v ≠ Vundef →
             ∃ pe ne n,
               In pe lpe ∧
               eval_pexpr ge m ρ e pe v ∧
               N.nconvert _ tp (λ x, x) pe = Some ne ∧
               In ne lne ∧
               N.ncompat v n ∧
               N.wtype_num n (N.Eval.eval_expr AbTy.adom tp pe) ∧
               eval_nexpr ρn ne n
      | None => True
      end.
    Proof.
      unfold nconvert. apply bm_bind_case. intros; exact I.
      intros lpe LPE.
      generalize (ne_of_pe_list_nonblock lpe).
      destruct (ne_of_pe_list lpe) as (nle & [|]). 2: exact id.
      simpl.
      intros [NB PE]. exists lpe. split. exact LPE.
      intros v EV Def.
      generalize (proj1 (convert_correct q)). rewrite LPE. simpl.
      intros H. destruct (H v EV) as (pe & IN & EV'); clear H.
      destruct (PE pe IN) as (ne & IN' & NE).
      exploit (N.nconvert_correct ge); try eassumption.
      intros (n & N).
      exists pe, ne, n. intuition.
    Qed.

    Lemma convert_to_ne (q: expr) :
      match bm_get (convert q) with
      | Some lpe =>
        match bm_get (ne_of_pe_list lpe) with
        | Some lne =>
          ∃ v, eval_expr ge e tmp m q v ∧ v ≠ Vundef
             ∧ ∃ pe ne n,
                 In pe lpe ∧
                 eval_pexpr ge m ρ e pe v ∧
                 N.nconvert _ tp (λ x, x) pe = Some ne ∧
                 In ne lne ∧
                 N.ncompat v n ∧
                 N.wtype_num n (N.Eval.eval_expr AbTy.adom tp pe) ∧
                 eval_nexpr ρn ne n
        | None => True
        end
      | None => True
      end.
    Proof.
      generalize (nconvert_sound q).
      generalize (nconvert_ex q).
      unfold nconvert.
      destruct (convert q) as (lpe & [|]). 2: intros; exact I.
      simpl.
      destruct (ne_of_pe_list lpe) as (lne & [|]). 2: intros; exact I.
      simpl. intros (lpe' & Hlpe' & H) NB.
      unfold ret_bm in Hlpe'. eq_some_inv. subst lpe'.
      destruct (nonblock_eval_expr_def NB) as (v & Def & V).
      specialize (H v V). destruct H as (pe & ne & n & Hn & Hpn & Hvn & Htn & Vn).
      exact Def.
      exists v. split. exact V. split. exact Def.
      exists pe, ne, n. intuition.
    Qed.

    End CONVERT_CORRECT.

  End CONVERT.

  Definition t : Type := (stack+⊤ * (sizes * (points_to * tnum)))%type.

  Definition AbMem stk sz pt tp nm : t := (stk, (sz, (pt, (tp, nm)))).
  Definition ab_stk (ab: t) : stack+⊤ := fst ab.
  Definition ab_sz (ab: t) : sizes := fst (snd ab).
  Definition ab_pt (ab: t) : points_to := fst (snd (snd ab)).
  Definition ab_nm (ab: t) : tnum := snd (snd (snd ab)).
  Definition ab_tp (ab: t) : types := fst (snd (snd (snd ab))).
  Definition ab_num (ab: t) : num := snd (snd (snd (snd ab))).

  Arguments ab_stk ab /.
  Arguments ab_sz ab /.
  Arguments ab_pt ab /.
  Arguments ab_nm ab /.
  Arguments ab_tp ab /.
  Arguments ab_num ab /.

  Definition with_stk (f: stack+⊤ → stack+⊤) (ab: t) : t :=
    AbMem (f (ab_stk ab)) (ab_sz ab) (ab_pt ab) (ab_tp ab) (ab_num ab).

  Definition with_sz (f: sizes → sizes) (ab: t) : t :=
    AbMem (ab_stk ab) (f (ab_sz ab)) (ab_pt ab) (ab_tp ab) (ab_num ab).

  Definition with_pt (f: points_to → points_to) (ab: t) : t :=
    AbMem (ab_stk ab) (ab_sz ab) (f (ab_pt ab)) (ab_tp ab) (ab_num ab).

  Definition with_tp (f: types → types) (ab: t) : t :=
    AbMem (ab_stk ab) (ab_sz ab) (ab_pt ab) (f (ab_tp ab)) (ab_num ab).

  Definition with_num (f: num → num) (ab: t) : t :=
    AbMem (ab_stk ab) (ab_sz ab) (ab_pt ab) (ab_tp ab) (f (ab_num ab)).

  Definition dom {A} (x: positive) (m: PTree.t A) : bool :=
    match PTree.get x m with
    | Some _ => true
    | None => false
    end.

  Remark dom_set {A} x m y (a: A) : dom x (PTree.set y a m) = peq x y || dom x m.
  Proof.
    unfold dom. rewrite PTree.gsspec. destruct peq; reflexivity.
  Qed.

  (* les noms de fonction dans la pile coincident et l’ensemble des variables locales est correct. *)
  Instance stack_elt_gamma : gamma_op (ident * lvarset) (ident * (temp_env * env)) :=
    λ f_lv g_te_e,
    let (f, lv) := f_lv in
    let '(g, (te, e)) := g_te_e in
    g = f
  ∧ ∀ x, PSet.mem x lv = dom x e.

  Existing Instance list_gamma.

  Lemma stack_adom : adom (stack+⊤) _ _ (toplift_gamma (list_gamma stack_elt_gamma)).
  Proof.
    eapply adom_of_pre_lattice, list_plat_correct.
    split.
    - intros (μ & ε) (μ' & ε') LE (f & t & e) (-> & K).
      simpl in *. rewrite andb_true_iff in LE. destruct LE as (LEμ & LEε).
      split. exact (Peqb_true_eq _ _ LEμ).
      intros x. rewrite <- (pset_beq_mem _ _ LEε). auto.
    - intros (μ & ε) (μ' & ε') (f & t & e) [(-> & K)|(-> & K)];
      simpl; destruct (μ =? μ')%positive eqn: Hμ; hnf; try exact I;
      destruct (PSet.beq ε ε') eqn: Hε; hnf; try exact I.
      split; easy.
      split. symmetry; apply Peqb_true_eq; assumption.
      intros x. erewrite pset_beq_mem; eauto.
  Qed.
  Hint Resolve stack_adom.

  Record gamma (ab: t) (cs: concrete_state) : Prop :=
  { _ : fst cs ∈ γ(ab_stk ab)
  ; _ : ab_stk ab ≠ All → invariant cs
  ; _ : snd cs ∈ sizes_gamma ge (fst cs) (ab_sz ab)
  ; _ : mem_as_fun ge cs ∈ (points_to_gamma ge (fst cs) (ab_pt ab) ∩ γ(ab_nm ab))
  }.
  Instance t_gamma : gamma_op t concrete_state := gamma.

  Ltac ung :=
  repeat
  match goal with
  | G : γ (gamma_op := t_gamma) ?a ?b |- _ =>
    let STK := fresh "STK" in
    let INV := fresh "INV" in
    let SZ := fresh "SZ" in
    let PT := fresh "PT" in
    let TPNM := fresh "TPNM" in
    destruct G as [STK INV SZ (PT & TPNM)]
  | G : _ ≠ All → _ |- _ =>
    specialize (G (λ K, All_not_Just (eq_sym K) False))
  | G : γ (gamma_op := toplift_gamma _) (Just _) _ |- _ =>
    red in G; simpl in G
  | G : γ (gamma_op := lift_gamma _) (NotBot _) _ |- _ =>
    red in G; simpl in G
  | G : γ (gamma_op := stack_elt_gamma) (_, _) (_, (_, _)) |- _ =>
    destruct G
  | G : γ (gamma_op := stack_elt_gamma) ?a (_, (_, _)) |- _ =>
    destruct a as (? & ?);
    destruct G
  | G : γ (gamma_op := stack_elt_gamma) (_, _) ?a |- _ =>
    destruct a as (? & ? & ?);
    destruct G
  | G : γ (gamma_op := list_gamma _) _ (_ :: _) |- _ =>
    apply gamma_cons_inv in G;
    destruct G as (? & ? & ? & ? & G)
  | G : γ (gamma_op := list_gamma _) (_ :: _) _ |- _ =>
    apply gamma_cons_inv_r in G;
    destruct G as (? & ? & ? & ? & G)
  | G : γ (gamma_op := list_gamma _) nil _ |- _ =>
    apply gamma_nil_r in G
  | G : γ (gamma_op := list_gamma _) _ nil |- _ =>
    apply gamma_nil_l in G
  | G : γ (gamma_op := tnum_gamma) _ _ |- _ =>
    let TP := fresh "TP" in
    let ρn := fresh "ρn" in
    let CPT := fresh "CPT" in
    let NM := fresh "NM" in
    destruct G as (TP & ρn & CPT & NM)
  end.

  Lemma stack_elt_gamma_in : ∀ s σ,
    σ ∈ γ(s) → ∀ f e,
    In (f, e) σ → ∃ ε, In (f, ε) s.
  Proof.
    induction s as [|(μ, lv) s IH]; intros σ H; ung; subst.
    intros ? ? ().
    intros ? ? [X | X]. eq_some_inv. subst. vauto.
    edestruct IH; eauto. vauto.
  Qed.

  Definition split_t (ab:t) : option (ident * lvarset * stack * sizes * points_to * types * num) :=
    match ab_stk ab with
    | All | Just nil => None
    | Just ((μ, ε) :: σ) =>
      let '(sz, (pt, (tp, nm))) := (snd ab) in
      Some (μ, ε, σ, sz, pt, tp, nm)
    end.

  Instance TToString : ToString t :=
    (λ x,
    "Mem { " ++ "stack : " ++ to_string (ab_stk x) ++ new_line
             ++ "blocks: " ++ to_string (ab_sz x) ++ new_line
             ++ "points: " ++ to_string (ab_pt x) ++ new_line
             ++ "types : " ++ to_string (ab_tp x) ++ new_line
             ++ "num   : " ++ to_string (ab_num x) ++ new_line
             ++ "}" ++ new_line).

  Lemma t_adom : adom t concrete_state _ t_gamma.
  Proof.
    constructor.
  - intros (stk & sz & pt & tpnm) (stk' & sz' & pt' & tpnm') LE (STK & MEM) G. ung.
    simpl in *. repeat rewrite andb_true_iff in LE. destruct LE as (LEstk & LEsz & LEpt & LEnm).
    constructor.
    simpl. eapply gamma_monotone; eauto.
    simpl. destruct stk' as [|stk']. intuition. intros _. apply INV.
    destruct stk. simpl in LEstk. eq_some_inv. exact (λ K, All_not_Just (eq_sym K) False).
    eapply gamma_monotone; eauto.
    simpl. intuition; eapply gamma_monotone; eauto. exact tnum_adom.
  - intros cc.
    constructor.
    apply gamma_top; eauto.
    exact (λ K, False_ind _ (K eq_refl)).
    apply gamma_top, sizes_adom.
    split.
    intros c. simpl. unfold ACTreeDom.get. rewrite ACTree.gempty. exact I.
    split.
    apply gamma_top; eauto.
    exists (compat_fun ∘ (extend (read_cell ge cc))). split. intros c. split.
    apply N.ncompat_compat_fun. simpl. unfold N.Eval.TDom.get. rewrite ACTree.gempty. auto.
    apply gamma_top; eauto.
  - intros (stk & sz & pt & nm) (stk' & sz' & pt' & nm') cc [H|H]; ung;
    (split;[| | |split]); auto;
    try (apply join_sound; eauto;
         exact tnum_adom).
    (destruct stk; [simpl; intuition| intros _; apply INV; discriminate]).
    destruct stk. simpl; intuition.
    destruct stk'. simpl; intuition.
    intros _; apply INV. exact (λ K, All_not_Just (eq_sym K) False).
  Qed.

  Definition nonblock (exp: expr) (ab: t) : block_mon unit :=
    match split_t ab with
    | Some (μ, ε, σ, sz, pt, tp, nm) =>
      do_bm _ <- nconvert μ ε sz pt tp nm exp;
      ret_bm tt
    | None => do_bm _ <- err_bm "anomaly (nonblock)"; ret_bm tt
    end.

  Definition forget_cell (ab: t) (c: cell) : t+⊥ :=
    do_bot nm' <- AbMachineEnv.forget c (ab_num ab);
    NotBot (with_pt (λ pt, ACTreeDom.set pt c All)
           (with_tp (λ tp, ACTreeDom.set tp c (⊤))
           (with_num (λ _, nm') ab))).

  Lemma forget_all_correct
        {A V: Type} (P: A → (cell → V) → Prop)
        (P_ext: ∀ ρ ρ', (∀ x, ρ(x) = ρ'(x)) → ∀ m, ρ ∈ P m → ρ' ∈ P m)
        (F: A → cell → A)
        (PF: ∀ m ρ ρ' x, ρ ∈ P m → (∀ y, y ≠ x → ρ'(y) = ρ(y)) → ρ' ∈ P (F m x))
        l :
    ∀ m : A,
    ∀ ρ ρ' : cell → V,
      ρ ∈ P m →
      (∀ c, ρ c = ρ' c ∨ In c l) →
      ρ' ∈ P (fold_left F l m).
  Proof.
    induction l as [|i l IH] using rev_ind.
  - intros m ρ ρ' H K.
    apply (P_ext ρ). intros c; destruct (K c) as [ ? | () ]; assumption.
    assumption.
  - intros m ρ ρ' H K. rewrite fold_left_app. simpl.
    apply PF with (λ c, if eq_dec c i then ρ c else ρ' c).
    eapply IH. eassumption.
    intros c. destruct (eq_dec c i). left; reflexivity.
    destruct (K c) as [X | X]. left; exact X.
    rewrite in_app in X. destruct X as [ X | [ X | () ] ].
    right; exact X. intuition.
    intros y H0. rewrite dec_eq_false; auto.
  Qed.

  Definition ct_forget_all {A: Type} := fold_left (λ (m: ACTreeDom.t A) c, ACTreeDom.set m c All).
  Definition nm_forget_all :=
    fold_left (λ m c, botbind (AbMachineEnv.forget c) m).

  Lemma ct_forget_all_get {A} l : ∀ m c,
    ACTreeDom.get c (@ct_forget_all A l m) =
    if in_dec eq_dec c l
    then All
    else ACTreeDom.get c m.
  Proof.
    Opaque ACTreeDom.set.
    unfold ct_forget_all.
    induction l as [|j l IH] using rev_ind; intros m c. reflexivity.
    rewrite fold_left_app. simpl. rewrite ACTreeDom.gsspec.
    destruct (ACTree.elt_eq c j).
    rewrite (pred_dec_true (P:=In c _)). reflexivity.
    rewrite in_app. right; left; symmetry; assumption.
    rewrite IH.
    destruct in_dec. rewrite pred_dec_true. reflexivity. rewrite in_app. auto.
    rewrite pred_dec_false. reflexivity. rewrite in_app. simpl. intuition.
  Qed.

    Lemma pt_forget_all_correct stk l :
      ∀ m : points_to,
      ∀ ρ ρ' : cell → val,
        ρ ∈ points_to_gamma ge stk m →
        (∀ c, ρ c = ρ' c ∨ In c l) →
        ρ' ∈ points_to_gamma ge stk (ct_forget_all l m).
    Proof.
      intros m ρ ρ' H H0.
      eapply forget_all_correct; eauto.
      intros ρ0 ρ'0 H1 m0 H2 c. rewrite <- H1. auto.
      intros m0 ρ0 ρ'0 x H1 H2 c. rewrite ACTreeDom.gsspec.
      destruct (ACTree.elt_eq). exact I. rewrite H2; auto.
    Qed.

    Lemma tp_forget_all_correct l :
      ∀ m : types,
      ∀ ρ ρ' : cell → val,
        ρ ∈ N.tree_gamma m →
        (∀ c, ρ c = ρ' c ∨ In c l) →
        ρ' ∈ N.tree_gamma (ct_forget_all l m).
    Proof.
      intros m ρ ρ' H H0.
      eapply forget_all_correct; eauto.
      intros ρ0 ρ'0 H1 m0 H2 c. rewrite <- H1. auto.
      intros m0 ρ0 ρ'0 x H1 H2 c. rewrite ACTreeDom.gsspec.
      destruct (ACTree.elt_eq). exact I. rewrite H2; auto.
    Qed.
  
  Definition ct_upd {A} (m:ACTreeDom.t A) (c:cell) (a:A+⊤) : ACTreeDom.t A :=
    ACTreeDom.set (ct_forget_all (overlap c) m) c a.

  Definition nm_upd (nm:num) (c:cell) (e:(nexpr cell)) : num+⊥ :=
    let nm' := AbMachineEnv.assign c e nm in
    nm_forget_all (overlap c) nm'.

  Definition loose_upd_spec {B} (f: cell → B) (x: cell) (v: B) (f': cell → B) : Prop :=
    f'(x) = v ∧ ∀ y, y ∈ cells_disjoint(x) → f'(y) = f(y).

  Lemma nm_upd_correct nm c e n (ρ ρ': cell → num_val) :
    ρ ∈ γ(nm) →
    n ∈ eval_nexpr ρ e →
    ρ' ∈ loose_upd_spec ρ c n →
    ρ' ∈ γ (nm_upd nm c e).
  Proof.
    intros NM EV UPD.
    unfold nm_upd.
    match goal with
    | |- context [ @AbMachineEnv.assign ?var ?t ?H ?E ?v ?ne ?ab ] =>
      generalize (@AbMachineEnv.assign_correct var t H E v ne ρ n ab NM EV);
      destruct (@AbMachineEnv.assign var t H E v ne ab)
    end.
    intros ().
    intros NM'.
    unfold nm_forget_all.
    eapply forget_all_correct.
    exact gamma_num_ext.
    intros [|m] ρ0 ρ'0 x0. intros ().
    eapply (forget_correct' NumDom).
    exact NM'.
    intros c'.
    destruct (cells_disjoint_dec c c') as [D | OV ].
    left. rewrite upd_o. exact (eq_sym (proj2 UPD _ D)).
    exact (not_eq_sym (cells_disjoint_not_eq D)).
    destruct (eq_dec c c'). left. subst. rewrite upd_s. exact (eq_sym (proj1 UPD)).
    right. apply overlap_correct'; auto.
  Qed.

  Lemma nm_upd_not_bot nm c e n (ρ: cell → num_val) :
    ρ ∈ γ(nm) →
    n ∈ eval_nexpr ρ e →
    nm_upd nm c e ≠ Bot.
  Proof.
    intros H H0 K.
    change (upd ρ c n ∈ γ(@Bot num)). rewrite <- K.
    eapply nm_upd_correct; try eassumption.
    split. apply upd_s.
    intros c' D. rewrite upd_o. reflexivity. exact (sym_not_eq (cells_disjoint_not_eq D)).
  Qed.

  Definition pt_upd (pt:points_to) (l:list cell) (s:BlockSet.t+⊤) : points_to :=
    union_list_map (λ c, ct_upd pt c s) pt join l.

  Lemma pt_upd_correct stk pt ρ cells ptr c v :
    v ∈ toplift_gamma (ptset_gamma ge stk) ptr →
    In c cells →
    points_to_gamma ge stk pt ρ →
    loose_upd_spec ρ c v ⊆ points_to_gamma ge stk (pt_upd pt cells ptr).
  Proof.
    intros Hptr Hc G ρ' UPD c'.
    unfold pt_upd. rewrite union_list_map_correct.
    refine ((λ x ρ H G, union_list_correct pt join _ (join_sound _ _ _ _ (points_to_adom ge stk))
                                              (map (λ c, ct_upd pt c ptr) cells)
                                              x ρ H G c') _ _ _ _).
    rewrite in_map_iff. exists c. split. reflexivity. exact Hc.
    clear c'. intros c'.
    unfold ct_upd. rewrite ACTreeDom.gsspec. destruct ACTree.elt_eq. subst.
    rewrite (proj1 UPD). auto.
    rewrite ct_forget_all_get.
    destruct in_dec as [IN|NI]. exact I.
    rewrite (proj2 UPD). auto.
    destruct (overlap_correct _ _ NI); intuition.
  Qed.

  Definition tp_upd_cells (tp:types) (l:list cell) (tt:AbTy.t+⊤) : types :=
    union_list_map (λ c, ct_upd tp c tt) tp join l.

  Lemma tp_upd_correct (val:Type) (Γ: gamma_op AbTy.t val) (PL: pre_lattice_spec AbTy.plat Γ) tp ρ cells ty c v :
    v ∈ γ(ty) →
    In c cells →
    ρ ∈ γ (gamma_op := ACTreeDom.gamma Γ) (tp) →
    loose_upd_spec ρ c v ⊆ γ (gamma_op := ACTreeDom.gamma Γ) (tp_upd_cells tp cells ty).
  Proof.
    intros Hty Hc G ρ' UPD c'.
    unfold tp_upd_cells. rewrite union_list_map_correct.
    refine ((λ x ρ H G, union_list_correct tp join (ACTreeDom.gamma Γ) (ACTreeDom.join_sound PL _)
                                           (map (λ c, ct_upd tp c ty) cells)
                                           x ρ H G c') _ _ _ _).
    rewrite in_map_iff. exists c. split. reflexivity. exact Hc.
    clear c'. intros c'.
    unfold ct_upd. rewrite ACTreeDom.gsspec. destruct ACTree.elt_eq. subst.
    rewrite (proj1 UPD); auto.
    rewrite ct_forget_all_get.
    destruct in_dec as [IN|NI]. exact I.
    rewrite (proj2 UPD). auto.
    destruct (overlap_correct _ _ NI); intuition.
  Qed.

  Definition nm_upd_cells (nm:num) (l:list cell) (e:(nexpr cell)) : num+⊥ :=
    union_list_map (λ c, nm_upd nm c e) (NotBot nm) join l.

  Lemma nm_upd_cells_not_bot nm l e n (ρ: cell → num_val) :
    ρ ∈ γ(nm) →
    n ∈ eval_nexpr ρ e →
    nm_upd_cells nm l e ≠ Bot.
  Proof.
    intros H H0.
    unfold nm_upd_cells. rewrite union_list_map_correct.
    match goal with
    | |- context[ @union_list (?A+⊥) (NotBot ?z) ?j ?l ] =>
      intros K;
      generalize (@union_list_bot num _ _ _ id_adom z l K)
    end.
    intros (N & L).
    destruct l. exact (N eq_refl). clear N.
    specialize (L _ (or_introl eq_refl)).
    exact (nm_upd_not_bot nm _ H H0 L).
  Qed.

  Definition ab_eval_expr (ab: t) (a: expr)
  : block_mon ((PTSet * AbTy.t+⊤ * list (nexpr cell)) +⊥) :=
    match split_t ab with
    | Some (μ, ε, σ, sz, pt, tp, nm) =>
      do_bm lp <- convert μ ε sz pt tp nm a;
      do_bm ln <- ne_of_pe_list tp nm lp;
      match lp with nil => ret_bm Bot | _ =>
      do_bm a' <- eval_expr_ptr μ ε sz pt tp nm a;
      let tt := union_list_map (N.Eval.eval_expr AbTy.adom tp) All join lp in
      ret_bm (NotBot (a', tt, ln))
      end
    | None => do_bm _ <- err_bm "ab_eval_expr failed"; ret_bm Bot (* unsound *)
    end.

  Definition ab_eval_exprlist (ab: t) (l: list expr) : block_mon (list (_+⊥)) :=
    do_bm l' <- bm_rev_map (ab_eval_expr ab) l;
    ret_bm (rev' l').

  Definition chunk_type (κ: typed_memory_chunk) (ty: AbTy.t+⊤) : bool :=
    match ty with
    | All => false
    | Just ty' =>
      match κ, ty' with
      | exist (Mint8signed | Mint8unsigned | Mint16signed | Mint16unsigned) _, AbTy.VI
      | exist Mint32 _, (AbTy.VI | AbTy.VP | AbTy.VIP)
      | exist Mint64 _, AbTy.VL
      | exist Mfloat64 _, AbTy.VF
      | exist Mfloat32 _, AbTy.VS
        => true
      | _, _ => false
      end
    end.

  Definition assign_cells (κ: option typed_memory_chunk) (c: list cell) (a:expr) (ab:t) : block_mon (t+⊥) :=
    do_bm res <- ab_eval_expr ab a;
    match res with
    | NotBot (ptr, ty, ln) =>
        do_bm _ <- bm_assert match κ with Some κ => chunk_type κ ty | None => true end
                             ("bad type: " ++ to_string ty);
        ret_bm (
        do_bot nm' <- union_list_map (nm_upd_cells (ab_num ab) c) (NotBot (ab_num ab)) join ln;
        let tp' := tp_upd_cells (ab_tp ab) c ty in
        NotBot (AbMem (ab_stk ab)
                      (ab_sz ab)
                      (pt_upd (ab_pt ab) c ptr)
                      tp' nm')
        )
    | Bot => ret_bm Bot
    end.

  Definition assign_in (μ: fname) (x: ident) (a: expr) (ab: t) : block_mon (t+⊥) :=
    (* print ("assign_in(" ++ to_string μ ++ ") " ++ to_string x
         ++ " := " ++ to_string a) *)
    (assign_cells None (ACtemp μ x :: nil) a ab).

  Definition assign (x:ident) (a:expr) (ab:t) : block_mon (t+⊥) :=
    match ab_stk ab with
    | Just ((μ,_)::_) => assign_in μ x a ab
    | _ => do_bm _ <- err_bm "assign failed"; ret_bm Bot (* unsound *)
    end.

  Definition assign_any (x: ident) (ty: AbTy.t) (ab: t) : block_mon (t+⊥) :=
    match split_t ab with
    | Some (μ, ε, σ, sz, pt, tp, nm) =>
      let c := ACtemp μ x in
      do_bm nm' <-
            match ty with
            | AbTy.VI => ret_bm (AbMachineEnv.assign c (NEconst _ (NVIint All)) nm)
            | AbTy.VL => ret_bm (AbMachineEnv.assign c (NEconst _ (NVIint64 All)) nm)
            | AbTy.VF => ret_bm (AbMachineEnv.assign c (NEconst _ (NVIfloat All)) nm)
            | AbTy.VS => ret_bm (AbMachineEnv.assign c (NEconst _ (NVIsingle All)) nm)
            | _ => do_bm _ <- err_bm ("assign_any: bad type " ++ to_string ty); ret_bm Bot (* unsound *)
            end;
        ret_bm (
        do_bot nm' <- nm';
        NotBot (with_num (λ _, nm')
                         (with_pt (λ pt, ACTreeDom.set pt c (Just BlockSet.empty))
                                  (with_tp (λ tp, ACTreeDom.set tp c (Just ty)) ab))))
    | None => do_bm _ <- err_bm "assign_any failed"; ret_bm Bot (* unsound *)
    end.

  Definition cell_in_bounds (κ:typed_memory_chunk) (ab: t) (c: cell) : block_mon unit :=
    let b := ablock_of_cell c in
    let '(ofs, _) := range_of_cell c in
    match ABTreeDom.get b (ab_sz ab) with
    | All => do_bm _ <- err_bm ("block of unknown size: " ++ to_string b); ret_bm tt
    | Just (hi, rw) =>
      if rw then
      if zle 0 ofs && zle (ofs + size_chunk (proj1_sig κ)) hi
      then ret_bm tt
      else do_bm _ <- err_bm ("cell out of range("++ to_string hi ++"): " ++ to_string c); ret_bm tt
      else do_bm _ <- err_bm ("read-only cell: " ++ to_string c); ret_bm tt
    end.

  Lemma cell_in_bounds_range_perm κ ab c:
    bm_get (cell_in_bounds κ ab c) = Some tt →
    ∀ m σ b,
      m ∈ sizes_gamma ge σ (ab_sz ab) →
      block_of_ablock ge σ (ablock_of_cell c) = Some b →
      Mem.range_perm m b (fst (range_of_cell c)) (fst (range_of_cell c) + size_chunk (proj1_sig κ)) Cur Writable.
  Proof.
    unfold cell_in_bounds.
    set (b := ablock_of_cell c).
    destruct (range_of_cell c) as (ofs, κ').
    destruct (ABTreeDom.get b (ab_sz ab)) as [|(hi, rw)] eqn: Hhi.
    simpl. intros ?; eq_some_inv.
    destruct rw. 2: simpl; intro; eq_some_inv.
    destruct zle as [L|]. 2: simpl; intro; eq_some_inv.
    destruct zle as [R|]. 2: simpl; intro; eq_some_inv.
    intros _.
    intros m σ b' H Hb o K.
    specialize (H _ _ _ _ Hhi Hb).
    eapply Mem.perm_implies. apply H. simpl in *. Psatz.lia. destruct b; vauto.
  Qed.
  Arguments cell_in_bounds_range_perm [κ] [ab] [c] _ [m] [σ] [b] _ _ _ _.
  Opaque cell_in_bounds.

  Definition expr_has_cast_for_chunk (e: expr) (κ: typed_memory_chunk) : bool :=
    match κ, e with
    | exist Mint8signed _, Eunop Ocast8signed _
    | exist Mint8unsigned _, Eunop Ocast8unsigned _
    | exist Mint16signed _, Eunop Ocast16signed _
    | exist Mint16unsigned _, Eunop Ocast16unsigned _
    | exist Mint32 _, _
    | exist Mint64 _, _
    | exist Mfloat32 _, _
    | exist Mfloat64 _, _
      => true
    | _, _ => false
    end.

  Definition store (κ: memory_chunk) (a e: expr) (ab: t) : block_mon (t+⊥) :=
    let κ : option typed_memory_chunk :=
        match κ with
          | Many32 | Many64 => None
          | κ => Some (exist _ κ I)
        end
    in
    match κ with
      | None =>
        do_bm _ <- err_bm "invalid store memory chunk"; ret_bm Bot (* unsound *)
      | Some κ =>
        match split_t ab with
          | Some (μ, ε, σ, sz, pt, tp, nm) =>
            do_bm _ <- bm_assert (expr_has_cast_for_chunk e κ)
                  ("Uncasted expr for " ++ to_string (proj1_sig κ) ++ ": " ++ to_string e);
            do_bm cells <- realize_cells μ ε sz pt tp nm κ a;
            match cells with nil => ret_bm Bot | _ =>
              do_bm _ <- bm_rev_map (cell_in_bounds κ ab) cells;
              assign_cells (Some κ) cells e ab
            end
          | None => do_bm _ <- err_bm "store failed"; ret_bm Bot (* unsound *)
        end
    end.

  Definition nassume b nm (F: nexpr cell → nexpr cell) e : num+⊥ :=
    union_list_map (λ ne, AbMachineEnv.assume (F ne) b nm) (NotBot nm) join e.

  Lemma nassume_correct b nm F el :
    ∀ ρ,
      ρ ∈ γ(nm) →
      ∀ e, In e el → eval_nexpr ρ (F e) (of_bool b) →
      ρ ∈ γ (nassume b nm F el).
  Proof.
    unfold nassume.
    intros ρ G e E EV.
    rewrite union_list_map_correct.
    refine (union_list_correct _ _ _ _ _ (AbMachineEnv.assume (F e) b nm) _ _ _).
    apply join_sound; auto.
    refine (in_map _ el _ E).
    apply assume_correct; assumption.
  Qed.

  Definition assume (b: expr) (ab: t) : block_mon (t+⊥ * t+⊥) :=
    match split_t ab with
    | Some (μ, ε, σ, sz, pt, tp, nm) =>
      do_bm b' <- nconvert μ ε sz pt tp nm (as_boolean_expr b);
      ret_bm
        (do_bot nz <- nassume true nm id b'; NotBot (AbMem (Just ((μ,ε)::σ)) (ab_sz ab) pt tp nz),
         do_bot z  <- nassume false nm id b'; NotBot (AbMem (Just ((μ,ε)::σ)) (ab_sz ab) pt tp z))
    | None => do_bm _ <- err_bm "assume failed"; ret_bm (Bot, Bot) (* unsound *)
    end.

  Definition deref_fun (e: expr) (ab: t) : block_mon (list (ident * fundef)) :=
    match split_t ab with
    | Some (μ, ε, σ, sz, pt, tp, nm) =>
      do_bm cells <- realize_cells μ ε sz pt tp nm (exist _ Mint32 I) e;
      bm_map' (λ c,
              match c with
              | ACglobal b 0 (exist Mint32 _) =>
                match Genv.find_funct_ptr ge b with
                | Some f =>
                  (* TODO: maybe not efficient *)
                  match Genv.invert_symbol ge b with
                  | Some i => Result (i,f)
                  | None => Error None "unknown function"
                  end
                | None => Error None ("bad function pointer " ++ to_string b)
                end
              | _ => Error None ("bad function pointer (2)")
              end) cells
    | None => do_bm _ <- err_bm "anomaly: deref_fun";
              ret_bm nil (* unsound *)
    end.

  Lemma env_ok (ε: lvarset) (e: env) :
    (∀ x, PSet.mem x ε = dom x e) →
    ∀ i, is_local ε i = true ↔ e ! i ≠ None.
  Proof.
    unfold is_local, dom.
    intros H i. rewrite H. destruct (e ! i); intuition. eq_some_inv.
  Qed.

  Lemma ab_eval_expr_correct (ab: t) e :
    match bm_get (ab_eval_expr ab e) with
    | Some res =>
      match ab with
      | (Just ((μ, ε) :: σ), (sz, (pt, (tp, nm)))) =>
        ∀ env tmp stk m
          (INV: invariant ((μ, (tmp, env)) :: stk, m))
          (SZ : m ∈ sizes_gamma ge ((μ, (tmp, env)) :: stk) sz)
          (ρ : cell → val)
          (TP : γ tp ρ)
          (ρ' : ACTree.elt → num_val)
          (CPT : compat tp ρ ρ')
          (NM : γ nm ρ')
          (H: ∀ i : ident, is_local ε i = true ↔ env ! i ≠ None)
          (H' : γ σ stk)
          (MEM : ρ = mem_as_fun ge ((μ, (tmp, env)) :: stk, m))
          (PT : points_to_gamma ge ((μ, (tmp, env)) :: stk) pt ρ),
          match res with
          | NotBot (ptr, ty, ln) =>
            ∃ v, eval_expr ge env tmp m e v ∧ v ≠ Vundef ∧
                 v ∈ (toplift_gamma (ptset_gamma ge ((μ, (tmp, env))::stk)) ptr ∩ γ(ty)) ∧
                 ∃ ne n, In ne ln ∧ (* N.nconvert _ tp (λ x, x) pe = Some ne ∧ *)
                         N.ncompat v n ∧
                         (∀ x y : ident, ne_free_var ne (ACtemp x y) → In x (μ :: map fst stk)) ∧
                         (* N.wtype_num n (N.Eval.eval_expr AbTy.adom tp pe) ∧ *)
                         eval_nexpr ρ' ne n
          | Bot => False
          end
       | _ => False
       end
    | None => True
    end.
  Proof.
    unfold ab_eval_expr.
    destruct ab as ([|[|(μ, ε) σ]] & sz & pt & tp & nm); try exact I.
    Opaque join.
    simpl.
    apply bm_bind_case. vauto.
    intros lpe LPE.
    apply bm_bind_case. vauto.
    intros lne LNE.
    destruct lpe as [|pe₀ lpe].
    intros env tmp stk m INV SZ ρ TP ρ' CPT NM H H' -> PT.
    generalize (convert_to_ne ε nm H INV SZ TP PT NM CPT e).
    rewrite LPE. simpl. intros; hsplit. assumption.
    apply bm_bind_case. vauto.
    intros ptr' PTR.
    intros env tmp stk m INV SZ ρ TP ρ' CPT NM H H' -> PT.
    generalize (convert_to_ne ε nm H INV SZ TP PT NM CPT e).
    rewrite LPE. unfold bm_get. rewrite LNE.
    intros (v & EV & Def & pe & ne & n & Hpe & Hne & Hlne & Necpt & Hnty & EV').
    exists v. split. assumption. split. assumption. split. split.
    destruct (convert_correct ε nm H INV SZ TP PT NM CPT e) as [_ PTR'].
    rewrite PTR in PTR'. simpl in PTR'. destruct ptr' as [|ptr].
    exact I.
    destruct v; first [ exact I | exact (PTR' _ _ EV) ].
    rewrite union_list_map_correct.
    eapply union_list_correct. eapply join_sound. apply AbTy.dom.
    rewrite in_map_iff. vauto.
    eapply N.Eval.eval_expr_correct; eauto.
    exists ne, n. split. assumption. split. assumption. split. 2: intuition.
    intros x y FV.
    destruct (N.nconvert_fv' _ _ _ _ _ Hlne _ FV) as (? & FV' & <- ).
    generalize (N.nconvert_fv_have_type _ _ _ _ _ Hlne _ FV'). subst.
    unfold N.TDom.get. fold (ACTreeDom.get (ACtemp x y) tp).
    generalize (TP (ACtemp x y)). simpl.
    destruct @eq_dec. left; easy. destruct (assoc x stk) as [(t',?)|] eqn: Ha.
    intros _. right. apply (assoc_in) in Ha. apply in_map_iff. eexists. split. 2: eassumption. reflexivity.
    intros. AbTy.gamma_inv.
    easy.
  Qed.

  Remark bind_parameters_zip formals : ∀ args te,
    bind_parameters formals args te =
    fold_left2 (λ acc id v, do_opt te <- acc; Some (PTree.set id v te))
               (λ _ _ , None) (λ _ _, None)
               formals args (Some te).
  Proof.
    induction formals as [|id formals IH].
    reflexivity.
    destruct args as [|v args]. reflexivity.
    simpl. intros te. rewrite IH. reflexivity.
  Qed.

  Definition ab_bind_parameters (μ: ident)
             (formals: list ident) (args: list ((PTSet * AbTy.t+⊤ * list (nexpr cell)) +⊥))
             : block_mon (t+⊥) → block_mon (t+⊥) :=
    fold_left2
      (λ acc id v,
       do_bm ab' <- acc;
       ret_bm (
           do_bot ab' <- ab';
           do_bot v <- v;
           let c := ACtemp μ id in
           let '(ptr, ty, ln) := v in
           do_bot nm' <- union_list_map (λ ne, AbMachineEnv.assign c ne (ab_num ab')) (NotBot (ab_num ab')) join ln;
           NotBot (AbMem (ab_stk ab')
                         (ab_sz ab')
                         (ct_upd (ab_pt ab') c ptr)
                         (ct_upd (ab_tp ab') c ty)
                         nm'))
      )
      (λ _ _, do_bm _ <- err_bm "ab_bind_parameters: too many formals"; ret_bm Bot) (* unsound *)
      (λ _ _, do_bm _ <- err_bm "ab_bind_parameters: too many arguments"; ret_bm Bot) (* unsound *)
      formals args.

  Lemma ab_bind_parameters_mono μ₀ formals : ∀ args ab,
    bm_get ab = None → bm_get (ab_bind_parameters μ₀ formals args ab) = None.
  Proof.
    induction formals as [|id formals IH]; destruct args as [|v args]; try easy.
    intros (ab, [|x y]). simpl; intro; eq_some_inv.
    intros _. simpl.
    destruct ab as [|ab]. eauto.
    destruct v as [|((ptr, ty), ln)]. eauto.
    simpl. destruct (union_list_map _ _ _); eauto.
  Qed.

  Lemma ab_bind_parameters_bot μ₀ formals : ∀ args,
    match bm_get (ab_bind_parameters μ₀ formals args (ret_bm Bot)) with
    | Some (NotBot _) => False
    | Some Bot
    | None => True
    end.
  Proof.
    induction formals as [|id formals IH]; destruct args as [|v args]; try easy.
    simpl.
    specialize (IH args). auto.
  Qed.

  (* Adjust the sizes of blocks allocated to local variables *)
  Definition set_local_sizes μ vars : sizes → sizes :=
    fold_left (λ sz x_hi, ABTreeDom.set sz (ABlocal μ (fst x_hi)) (Just (snd x_hi, true)))
              vars.

  Lemma set_local_sizes_get μ vars sz b :
    list_norepet (map fst vars) →
    ABTreeDom.get b (set_local_sizes μ vars sz) =
    match b with
    | ABlocal μ' x => if peq μ' μ
                      then match assoc x vars with
                           | Some hi => Just (hi, true)
                           | None => ABTreeDom.get b sz
                           end
                      else ABTreeDom.get b sz
    | _ => ABTreeDom.get b sz
    end.
  Proof.
    revert sz.
    unfold set_local_sizes.
    induction vars as [|(x,hi) vars IH].
    simpl. destruct b as [μ' ?|]; auto. destruct peq; auto.
    simpl. intros sz NR. rewrite IH. clear IH.
    fold (ABTreeDom.set sz (ABlocal μ x) (Just (hi, true))). rewrite ABTreeDom.gsspec.
    destruct (ABTree.elt_eq b (ABlocal μ x)) as [EQ|NE].
  - subst. rewrite !peq_true.
    rewrite (proj2 (assoc_None x vars)). reflexivity.
    intros b K. inv NR. eapply H1. rewrite in_map_iff. eexists (x, _); vauto.
  - destruct b; auto.
    destruct peq; auto. subst.
    destruct @eq_dec; auto. subst.
    elim NE. reflexivity.
  - inv NR; auto.
  Qed.

  Definition push_frame (μ: fname) (μ_def: function) (args: list expr) (ab:t) : block_mon (t+⊥) :=
    (* old stack *)
    do_bm σ' <- match ab_stk ab with
                | All => do_bm _ <- err_bm "no stack"; ret_bm nil
                | Just σ' => ret_bm σ'
                end;
    match assoc μ σ' with
    | Some _ => do_bm _ <- err_bm "push_frame: recursive call"; ret_bm Bot (* unsound *)
    | None =>
      (* Evaluates the arguments *)
      do_bm eargs <- ab_eval_exprlist ab args;
      (* Remember the set of local variable names *)
      let ε := PSetOp.build (rev_map fst μ_def.(fn_vars)) in
      (* Push the current function name and environment on the stack *)
      let σ := (μ, ε) :: σ' in
      let ab' := with_stk (λ _, Just σ) (with_sz (set_local_sizes μ μ_def.(fn_vars)) ab) in
      ab_bind_parameters μ μ_def.(fn_params) eargs (ret_bm (NotBot ab'))
    end.

  (** Invalidate cells local to μ. *)
  Let is_local_cell (μ: fname) (c: cell) : bool :=
    match c with
    | AClocal μ' _ _ _
    | ACtemp μ' _ => eq_dec μ μ'
    | ACglobal _ _ _ => false
    end.

  Let is_local_block (μ: fname) (ab: ablock) : bool :=
    match ab with
    | ABlocal μ' _ => eq_dec μ μ'
    | ABglobal _ => false
    end.

  Definition pop_local_tp (μ: fname) : types → types :=
    ACTreeDom.SP.filter (λ c _, negb (is_local_cell μ c)).

  (** Invalidate pointers to cells local to μ. *)
  Definition pop_local_pt (μ: fname) : points_to → points_to :=
    ACTreeDom.SP.filter (λ c bs, negb (is_local_cell μ c) && negb (BSO.existsb (is_local_block μ) bs)).

  Lemma pop_local_tp_get μ tp c :
    ACTree.get c (pop_local_tp μ tp) = if is_local_cell μ c
                                       then None else ACTree.get c tp.
  Proof.
    unfold pop_local_tp.
    rewrite ACTreeDom.SP.gfilter. destruct ACTree.get, is_local_cell; auto.
  Qed.

  Lemma pop_local_pt_get μ pt c :
    ACTree.get c (pop_local_pt μ pt) = do_opt bs <- ACTree.get c pt;
    if negb (is_local_cell μ c) && negb (BSO.existsb (is_local_block μ) bs)
    then Some bs
    else None.
  Proof.
    unfold pop_local_pt.
    rewrite ACTreeDom.SP.gfilter. auto.
  Qed.

  Definition clear_local_sizes μ ε : sizes → sizes :=
    PSet.fold (λ x, ABTree.remove (ABlocal μ x)) ε.

  Lemma clear_local_sizes_get μ ε b sz :
    ABTreeDom.get b (clear_local_sizes μ ε sz) =
    match b with
    | ABlocal μ' x => if peq μ μ' && PSet.mem x ε then All else ABTreeDom.get b sz
    | _ => ABTreeDom.get b sz
    end.
  Proof.
    unfold clear_local_sizes.
    apply PSetOp.foldspec.
    intros s s' a b0 x H [REM REM'] H1.
    unfold ABTreeDom.get in *. rewrite ABTree.grspec.
    destruct (ABTree.elt_eq b (ABlocal μ x)) as [EQ|NE]. subst. destruct peq. 2: congruence.
    rewrite H. reflexivity.
    rewrite H1.
    destruct b as [μ' x'|b]; auto.
    destruct peq; simpl; auto.
    subst.
    specialize (REM x'). specialize (REM' x').
    destruct (PSet.mem x' s).
    specialize (REM eq_refl). destruct REM as [REM|REM].
    elim NE. congruence. rewrite REM. reflexivity.
    destruct (PSet.mem x' s'). 2:auto.
    destruct (REM' eq_refl) as [_ Y]. discriminate.
    intros s H.
    destruct b as [μ' x'|b]; auto.
    destruct peq; simpl; auto.
    subst. specialize (H x').
    destruct (PSet.mem); auto. intuition.
  Qed.

  Definition pop_frame (res: option (expr)) (temp : option (ident)) (ab: t) : block_mon (t+⊥) :=
    match ab_stk ab, res, temp with
    | Just ((μ₀,ε₀)::(μ,ε)::σ), Some ret, Some x =>
      (*print "ret" *)(
      do_bm ab' <- assign_in μ x ret ab;
      match ab' with
      | NotBot ab'' =>
        ret_bm (NotBot (with_sz (clear_local_sizes μ₀ ε₀)
                       (with_tp (pop_local_tp μ₀)
                       (with_pt (pop_local_pt μ₀)
                       (with_stk (λ _, Just ((μ,ε)::σ)) ab'')))))
      | Bot => ret_bm Bot
      end
      )
    | Just ((μ,ε)::nil), Some ret, Some 1%positive =>
      (* return of main *)
      do_bm ret' <- ab_eval_expr ab ret;
      match ret' with
      | Bot => ret_bm Bot
      | NotBot (_, ty, _) =>
        match ty with
        | Just AbTy.VI =>
          ret_bm (NotBot (with_sz (clear_local_sizes μ ε)
                         (with_tp (pop_local_tp μ)
                         (with_pt (pop_local_pt μ)
                         (with_stk (λ _, Just nil) ab)))))
        | _ => do_bm _ <- err_bm "main ret not an int"; ret_bm Bot
        end
      end
    | Just (_::nil), _, _ =>
      (* return of main (blocking) *)
      do_bm _ <- err_bm "pop_frame failed at main"; ret_bm Bot
    | Just ((μ,ε)::σ), None, None =>
      ret_bm (NotBot (with_sz (clear_local_sizes μ ε)
                     (with_tp (pop_local_tp μ)
                     (with_pt (pop_local_pt μ)
                     (with_stk (λ _, Just σ) ab)))))
    | Just ((μ,ε)::σ), Some ret, None =>
      do_bm _ <- nonblock ret ab;
      ret_bm (NotBot (with_sz (clear_local_sizes μ ε)
                     (with_tp (pop_local_tp μ)
                     (with_pt (pop_local_pt μ)
                     (with_stk (λ _, Just σ) ab)))))
    | _, _, _ => do_bm _ <- err_bm ("pop_frame failed: " ++ to_string temp ++ " := " ++ to_string res (* ++ " in " ++ to_string ab *)); ret_bm Bot
    end.

  Let bad_align {T} (r: T) := do_bm _ <- err_bm "misaligned init data"; ret_bm r.
  Instance InitDataToString : ToString init_data :=
    λ id,
    match id with
    | Init_int8 i => "int8: " ++ to_string i
    | Init_int16 i => "int16: " ++ to_string i
    | Init_int32 i => "int32: " ++ to_string i
    | Init_int64 i => "int64: " ++ to_string i
    | Init_float32 f => "float32: " ++ to_string f
    | Init_float64 f => "float64: " ++ to_string f
    | Init_space n => "space: " ++ to_string n
    | Init_addrof i o => "&" ++ to_string i ++ " + " ++ to_string o
    end.
 
  Definition type_of_chunk (κ: typed_memory_chunk) : AbTy.t :=
    match κ with
    | exist (Mint8signed | Mint8unsigned |
             Mint16signed | Mint16unsigned |
             Mint32) _
      => AbTy.VI
    | exist Mint64 _ => AbTy.VL
    | exist Mfloat64 _ => AbTy.VF
    | exist Mfloat32 _ => AbTy.VS
    | exist _ H => match H with end
    end.

  Definition zero_all_cells (b: block) (base: Z) (sz: Z) (ab: t): t+⊥ :=
    N.peano_rect
      (λ _, t+⊥)
      (NotBot ab)
      (λ ofs ab,
       memory_chunk_fold
         (λ ab κ,
          let ofs' := base + Z.of_N ofs in
          if Zdivide_dec (align_chunk (proj1_sig κ)) ofs' (align_chunk_pos (proj1_sig κ)) then
          if zle (Z.of_N ofs + size_chunk (proj1_sig κ)) sz then
            do_bot ab <- ab;
              let c := ACglobal b ofs' κ in
              do_bot nm <- AbMachineEnv.assign c
                (match κ with
                   | exist (Mint8signed | Mint8unsigned | Mint16signed | Mint16unsigned | Mint32) _
                     => ne_const_i _ 0
                   | exist Mint64 _ => ne_const_l _ 0
                   | exist (Mfloat32) _ => ne_const_s _ Float32.zero
                   | exist (Mfloat64) _ => ne_const_f _ Float.zero
                   | exist _ H => match H with end
                 end)
                (ab_num ab);
              NotBot
                (with_tp (ACTree.set c (type_of_chunk κ))
                (with_pt (ACTree.set c BlockSet.empty)
                (with_num (λ _, nm) ab)))
          else ab
          else ab
         ) ab)
      (Z.to_N sz).

  Definition ab_alloc_global rw b (gv: list init_data) ab : block_mon (t+⊥) :=
    do_bm sz_ab' <- bm_fold
      (λ acc id,
       let '(ofs, ab) := acc in
       let ofs' := ofs + Genv.init_data_size id in
       let c κ H := ACglobal b ofs (exist _ κ H) in
       if match id with
          | Init_space _
          | Init_int8 _ => true
          | Init_int16 _ => Zdivide_dec (align_chunk Mint16signed) ofs eq_refl
          | Init_int32 _
          | Init_addrof _ _ => Zdivide_dec (align_chunk Mint32) ofs eq_refl
          | Init_int64 _ => Zdivide_dec (align_chunk Mint64) ofs eq_refl
          | Init_float32 _ => Zdivide_dec (align_chunk Mfloat32) ofs eq_refl
          | Init_float64 _ => Zdivide_dec (align_chunk Mfloat64) ofs eq_refl
          end then
       match id with
       | Init_int32 n =>
          ret_bm (ofs',
                  do_bot ab' <- ab;
                  do_bot nm <- AbMachineEnv.assign (c Mint32 I) (ne_const_i _ (Int.signed n)) (ab_num ab');
                       NotBot
                         (with_tp (ACTree.set (c Mint32 I) AbTy.VI)
                         (with_pt (ACTree.set (c Mint32 I) BlockSet.empty)
                         (with_num (λ _, nm) ab'))))
       | Init_int64 n =>
          ret_bm (ofs',
                  do_bot ab' <- ab;
                  do_bot nm <- AbMachineEnv.assign (c Mint64 I) (ne_const_l _ (Int64.signed n)) (ab_num ab');
                       NotBot
                         (with_tp (ACTree.set (c Mint64 I) AbTy.VL)
                         (with_pt (ACTree.set (c Mint64 I) BlockSet.empty)
                         (with_num (λ _, nm) ab'))))
       | Init_addrof s o =>
         match Genv.find_symbol ge s with
         | Some b' => ret_bm (ofs',
                              do_bot ab' <- ab;
                              do_bot nm <- AbMachineEnv.assign (c Mint32 I) (ne_const_i _ (Int.signed o)) (ab_num ab');
                              NotBot
                                (with_tp (ACTree.set (c Mint32 I) AbTy.VP)
                                (with_pt (ACTree.set (c Mint32 I) (BSO.singleton (ABglobal b')))
                                (with_num (λ _, nm) ab'))))
         | None => do_bm _ <- err_bm ("ab_alloc_globals: not a symbol ("
                                        ++ to_string s ++ ")"); ret_bm (ofs', ab)
         end
       | Init_space s =>
         ret_bm (ofs', do_bot ab' <- ab; zero_all_cells b ofs s ab')
       | Init_float64 f =>
         ret_bm (ofs',
           if Fappli_IEEE.is_nan _ _ f
           then ab
           else
             do_bot ab' <- ab;
             do_bot nm <- AbMachineEnv.assign (c Mfloat64 I) (NEconst _ (NVIfloat (Just (FloatIntervals.ITV f f)))) (ab_num ab');
             NotBot
               (with_tp (ACTree.set (c Mfloat64 I) AbTy.VF)
               (with_pt (ACTree.set (c Mfloat64 I) BlockSet.empty)
               (with_num (λ _, nm) ab'))))
       | Init_float32 f =>
         ret_bm (ofs',
           if Fappli_IEEE.is_nan _ _ f
           then ab
           else
             do_bot ab' <- ab;
             let f := Float.of_single f in
             do_bot nm <- AbMachineEnv.assign (c Mfloat32 I) (NEconst _ (NVIsingle (Just (FloatIntervals.ITV f f)))) (ab_num ab');
             NotBot
               (with_tp (ACTree.set (c Mfloat32 I) AbTy.VS)
               (with_pt (ACTree.set (c Mfloat32 I) BlockSet.empty)
               (with_num (λ _, nm) ab'))))
       | Init_int8 i =>
         ret_bm (ofs',
                 do_bot ab' <- ab;
                 do_bot nm <- AbMachineEnv.assign (c Mint8unsigned I) (NEunop Ocast8unsigned (ne_const_i _ (Int.unsigned i))) (ab_num ab');
                 NotBot
                   (with_tp (ACTree.set (c Mint8unsigned I) AbTy.VI)
                   (with_pt (ACTree.set (c Mint8unsigned I) BlockSet.empty)
                   (with_num (λ _, nm) ab'))))
       | Init_int16 i =>
         ret_bm (ofs',
                 do_bot ab' <- ab;
                 do_bot nm <- AbMachineEnv.assign (c Mint16unsigned I) (NEunop Ocast16unsigned (ne_const_i _ (Int.unsigned i))) (ab_num ab');
                 NotBot
                   (with_tp (ACTree.set (c Mint16unsigned I) AbTy.VI)
                   (with_pt (ACTree.set (c Mint16unsigned I) BlockSet.empty)
                   (with_num (λ _, nm) ab'))))
       end
       else bad_align acc)
      gv
      (0, ab);
    let '(sz, ab') := sz_ab' in
    ret_bm (do_bot ab'' <- ab'; NotBot (with_sz (ABTree.set (ABglobal b) (sz, rw)) ab'')).

  Definition ab_alloc_globals (gl: list (ident * globdef fundef unit)) : block_mon (block * t+⊥) :=
    (bm_fold
       (λ acc ig,
        let '(b, ab) := acc in
        let b' := Pos.succ b in
        match ig with
        | (_, Gfun _) => ret_bm (b', botlift_fun1 (with_sz (ABTree.set (ABglobal b) (0, false))) ab)
        | (_, Gvar gv) =>
          if gv.(gvar_volatile)
          then do_bm _ <- err_bm "skipped global volatile"; ret_bm (b', ab)
          else
            do_bm ab' <- ab_alloc_global (negb gv.(gvar_readonly)) b gv.(gvar_init) ab;
            ret_bm (b', ab')
        end)
       gl
       (xH, NotBot (Just nil, ⊤))).

  Definition init_mem (defs:list (ident * globdef fundef unit)) : block_mon (t+⊥) :=
    do_bm _ <- bm_assert (norepet_map fst defs) ("duplicate name in globals");
    do_bm ab <- ab_alloc_globals defs;
    ret_bm (snd ab).

  Lemma ab_alloc_global_ex:
    ∀ ab gv rw id m ab',
      ab_alloc_global rw (Mem.nextblock m) gv.(gvar_init) ab = (ab', nil) ->
      ∃ m', Genv.alloc_global ge m (id, Gvar gv) = Some m'.
  Proof.
    intros * EQ_AB. simpl.
    destruct (Mem.alloc m 0 (Genv.init_data_list_size (gvar_init gv))) eqn:EQ_ALLOC.
    assert (STORE:=EQ_ALLOC). eapply alloc_can_store in STORE. 2:apply init_data_list_size_pos.
    destruct STORE as [m' STORE]. rewrite STORE.
    erewrite <- Mem.alloc_result in EQ_AB by eauto.
    assert (PERM:=Mem.perm_alloc_2 _ _ _ _ _ EQ_ALLOC).
    setoid_rewrite (Genv.store_zeros_perm _ _ _ _ _ _ _ _ STORE) in PERM.
    destruct (Genv.store_init_data_list ge m' b 0 (gvar_init gv)) as [m''|] eqn:STORELST.
    { setoid_rewrite (Genv.store_init_data_list_perm _ _ _ _ _ _ _ _ _ STORELST) in PERM.
      destruct (Mem.range_perm_drop_2 m'' b 0 (Genv.init_data_list_size (gvar_init gv)) 
                                      (Genv.perm_globvar gv)).
      intro. apply PERM. eauto. }
    exfalso.
    revert ab EQ_AB PERM STORELST. generalize (gvar_init gv). clear.
    intros l ab. unfold ab_alloc_global.
    change (Genv.init_data_list_size l) with (0 + Genv.init_data_list_size l).
    apply bm_bind_case. discriminate. intros. clear EQ_AB rw ab'.
    revert ab m' x H PERM STORELST.
    generalize 0 at 7 8 9 10.
    induction l as [|id  idl].
    - simpl. discriminate.
    - simpl.
      change (id::idl) with ((id::nil)++idl)%list. intros*. rewrite bm_fold_app.
      apply bm_bind_case. discriminate. intros.
      unfold bm_fold in H. simpl in H. destruct x0.
      destruct (Genv.store_init_data ge m' b z id) eqn:EQSTORE.
      + assert (z0 = z + Genv.init_data_size id).
        { clear H0.
          match type of H with (if ?b then _ else _) = _ => destruct b end.
          destruct id; try now inv H.
          destruct (Genv.find_symbol ge i); inv H; auto.
          inv H. }
        subst z0.
        eapply IHidl. 3:eauto.
        apply H0. intros.
        assert (PERM':Mem.perm m' b ofs k Freeable).
        { apply PERM. pose proof Genv.init_data_size_pos id. omega. }
        clear - EQSTORE PERM'. destruct id; simpl in EQSTORE;
        eauto using Mem.perm_store_1.
        inv EQSTORE. auto.
        destruct Genv.find_symbol. eauto using Mem.perm_store_1. discriminate.
      + clear IHidl. clear H0.
        pose proof (init_data_list_size_pos idl).
        destruct id; simpl in EQSTORE; [..|discriminate|destruct (Genv.find_symbol ge i)];
        match type of EQSTORE with
          | Mem.store ?c m' b z ?v = None =>
            destruct (Mem.valid_access_store m' c b z v); [
              split;[eapply Mem.range_perm_implies; [
                        repeat intro; eapply PERM;
                        unfold size_chunk, Genv.init_data_size in *;
                        omega
                       |constructor]
                |apply Z.divide_1_l || (destruct Zdivide_dec; [auto|discriminate])]
              |congruence]
          | _ => idtac
        end.
        destruct Zdivide_dec; discriminate.
  Qed.

  Lemma ab_alloc_globals_ex:
    ∀ gl b ab, ab_alloc_globals gl = ((b, ab), nil) ->
    ∃ m, Genv.alloc_globals ge Mem.empty gl = Some m /\
         Mem.nextblock m = b.
  Proof.
    induction gl using rev_ind.
    - simpl. intros. inv H. eauto.
    - intros ab. unfold ab_alloc_globals.
      rewrite bm_fold_app.
      apply bm_bind_case. discriminate.
      intros [b ab'] Hbab'. destruct (IHgl _ _ Hbab') as (m & Hm & Hbm). subst b.
      erewrite <- Genv.alloc_globals_app; eauto. simpl.
      unfold bm_fold. simpl.
      pose proof Genv.alloc_global_nextblock ge x m.
      destruct x as [id [gf|gv]]. simpl in *.
      + intros.
        destruct (Mem.alloc m 0 1) as [m' b0] eqn:EQm'.
        edestruct alloc_can_drop as [m'' Hm'']. eauto. rewrite Hm''.
        eexists. split. eauto. rewrite H. inv H0. auto. auto.
      + destruct (gvar_volatile gv) eqn:VOL. discriminate.
        apply bm_bind_case. discriminate. intros ab'' Hab'' b' eq. inv eq.
        edestruct ab_alloc_global_ex as [m' Hm']; eauto.
        rewrite Hm'. eauto.
  Qed.

  Lemma ab_alloc_global_init_mem_ex:
    ∀ prog, ge = Genv.globalenv prog ->
    ∀ b ab, ab_alloc_globals prog.(prog_defs) = ((b, ab), nil) ->
    ∃ m, Genv.init_mem prog = Some m /\
         Mem.nextblock m = b.
  Proof.
    intros. apply ab_alloc_globals_ex in H0. rewrite H in H0. auto.
  Qed.

  Lemma ab_alloc_global_ok:
    ∀ prog m, ge = Genv.globalenv prog ->
              Genv.init_mem prog = Some m ->
    ∀ rw b gv ab ab' id,
      Genv.find_var_info ge b = Some gv ->
      Genv.invert_symbol ge b = Some id ->
      gv.(gvar_volatile) = false ->
      ab_alloc_global rw b (gvar_init gv) (NotBot ab) = (ab', nil) ->
    Genv.perm_globvar gv = (if rw then Writable else Readable) →
    m ∈ sizes_gamma ge nil (ab_sz ab) ->
    ab_stk ab = Just nil ->
    mem_as_fun ge (nil, m) ∈ (points_to_gamma ge nil (ab_pt ab) ∩ γ(ab_nm ab)) ->
    match ab' with
      | Bot => False
      | NotBot ab' =>
        m ∈ sizes_gamma ge nil (ab_sz ab') /\
        ab_stk ab' = Just nil /\
        mem_as_fun ge (nil, m) ∈ (points_to_gamma ge nil (ab_pt ab') ∩ γ(ab_nm ab'))
    end.
  Proof.
    intros.
    eapply Genv.init_mem_characterization in H0. 2:rewrite <- H; apply H1. clear H1.
    destruct H0 as (PERM & _ & INIT). specialize (INIT H3). clear H3.
    rewrite <- H in INIT. clear H prog.
    rewrite H5 in PERM. clear H5.
    revert H4. unfold ab_alloc_global.
    apply bm_bind_case. discriminate.
    intros [sz ab''] Hab'' eq. inv eq.
    revert ab PERM INIT H6 H7 H8 Hab''.
    change (Genv.init_data_list_size (gvar_init gv))
    with (0+Genv.init_data_list_size (gvar_init gv)).
    generalize 0 at 2 3 10 as offs. induction (gvar_init gv).
    - unfold bm_fold. simpl.
      intros. inv Hab''. simpl. rewrite <- surjective_pairing. split. 2:eauto.
      unfold sizes_gamma in *. intros.
      unfold ABTreeDom.get in H. rewrite ABTree.gsspec in H. destruct ABTree.elt_eq.
      + inv H. rewrite Z.add_0_r in PERM.
        simpl in H0. rewrite H2 in H0. inv H0. apply PERM.
      + auto.
    - change (a::l) with ((a::nil)++l)%list. intros *. rewrite bm_fold_app.
      apply bm_bind_case. discriminate.
      destruct ab as (abstk & absz & abpt & (abty & abnm)).
      intros [offs' ab'] Ha PERM INIT SZ STK (PT & TY & ρ' & COMPAT & NM) Hl.
      unfold bm_fold in Ha. simpl in Ha.
      match type of Ha with
        | (if ?b then _ else _) = _ => destruct b; [|discriminate]
      end.
      cut (match ab' with
             | Bot => False
             | NotBot ab' =>
               sizes_gamma ge nil (ab_sz ab') m
               ∧ ab_stk ab' = Just nil
               ∧ points_to_gamma ge nil (ab_pt ab')
                                 (mem_as_fun ge (nil, m))
               ∧ γ (ab_nm ab') (mem_as_fun ge (nil, m))
           end).
      { intro. destruct ab'. contradiction. destruct H as (? & ? & ? & ?).
        destruct a, INIT; [..|destruct Genv.find_symbol; [|discriminate]];
        try now (inv Ha; apply IHl in Hl; auto; rewrite <- Z.add_assoc; apply PERM). }
      clear Hl IHl.
      destruct a, INIT; [..|destruct (Genv.find_symbol ge i) eqn:EQSYMB; [|discriminate]];
      inv Ha;
      try match goal with
       | |- context [Fappli_IEEE.is_nan ?a ?b ?f] =>
          destruct (Fappli_IEEE.is_nan a b f) eqn:NAN; [repeat (split; eauto)|]
      end;
      try match goal with
       | |- context [AbMachineEnv.assign ?c ?e ?ab] =>
          refine (let ASS := AbMachineEnv.assign_correct c (e:=e) _ NM _ in _);
            [repeat econstructor;
              auto using Zle_refl, Fleb_refl, floatofsingle_nan|
             destruct AbMachineEnv.assign as [|abnm']; [contradiction|simpl]]
      end.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2, Int.repr_unsigned. repeat econstructor.
          apply COMPAT.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2, Int.repr_unsigned. repeat econstructor.
          apply COMPAT.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2, Int.repr_signed. repeat econstructor.
          apply COMPAT.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2, Int64.repr_signed. repeat econstructor.
          apply COMPAT.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2. repeat econstructor.
          apply COMPAT.
      + split. auto. split. auto. split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2. repeat econstructor.
          apply COMPAT.
      + unfold zero_all_cells.
        revert abstk absz abpt abty abnm ρ' SZ STK PT NM TY COMPAT.
        induction (Z.to_N z) using N.peano_ind.
        * repeat (split; eauto).
        * setoid_rewrite N.peano_rect_succ. intros. eapply memory_chunk_fold_rec. 2:eapply IHn; now eauto.
          { clear abstk absz abpt abty abnm ρ' H1 H3 H4 H5 H6 H7.
            intros [|(abstk & absz & abpt & (abty & abnm))] c. contradiction.
            intros (SZ & STK & PT & TY & ρ' & COMPAT & NM).
            destruct zle, Zdivide_dec; try now repeat (split; eauto). simpl.
            match goal with
              | |- context [AbMachineEnv.assign ?ce ?e ?ab] =>
                refine (let ASS := AbMachineEnv.assign_correct ce (e:=e)
                            (n:=match c with exist (Many32|Many64) H => match H with end
                                          | _ => _ end) _ NM _
                        in _)
            end.
            - destruct c as [[] []]; repeat constructor;
              auto using Zle_refl, Fleb_refl.
            - destruct AbMachineEnv.assign as [|abnm']; [contradiction|simpl].
              refine (let H':= H (proj1_sig c) (offs + Z.of_N n) _ _ _ in _).
              zify; omega. omega. auto.
              split. auto. split. auto. split. 2:split.
              + intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
                subst. simpl. rewrite H', H2. destruct(proj1_sig c); constructor.
                apply PT.
              + intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
                subst. simpl. rewrite H', H2. destruct c as [[] []]; constructor.
                apply TY.
              + eexists. split. 2:apply ASS. intro. simpl.
                unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
                destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
                subst c0. simpl. rewrite H', H2. destruct c as [[] []]; repeat econstructor.
                apply COMPAT. }
      + split. auto. split. auto.
        destruct H as (b' & eq & H). inv eq.
        split. 2:split.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2.
          eexists. split. apply BSO.mem_singleton. eauto. simpl. erewrite Genv.find_invert_symbol; eauto.
          apply PT.
        * intro. unfold ACTreeDom.get. rewrite ACTree.gsspec. destruct ACTree.elt_eq.
          subst. simpl. rewrite H, H2. constructor.
          apply TY.
        * eexists. split. 2:apply ASS. intro. simpl.
          unfold upd, N.Eval.TDom.get. rewrite ACTree.gsspec.
          destruct (@eq_dec cell), ACTree.elt_eq; try congruence.
          subst c. simpl. rewrite H, H2, Int.repr_signed. repeat econstructor.
          apply COMPAT.
  Qed.

  Lemma nth_def_block :
    ∀ prog, ge = Genv.globalenv prog ->
            list_norepet (map fst prog.(prog_defs)) ->
    ∀ idx id gv,
      nth idx (List.map Some prog.(prog_defs)) None = Some (id, Gvar gv) ->
      Genv.find_var_info ge (Pos.of_succ_nat idx) = Some gv /\
      Genv.find_symbol ge id = Some (Pos.of_succ_nat idx).
  Proof.
    unfold Genv.globalenv. intros*.
    generalize ge.
    induction (prog_defs prog) using rev_ind.
    - destruct idx; discriminate.
    - intros. subst.
      rewrite Genv.add_globals_app. simpl.
      rewrite map_app in H0, H1. rewrite list_norepet_app in H0.
      destruct H0 as (? & _ & ?). simpl in H0.
      assert (~ In (fst x) (map fst l)).
      { intro. eapply H0; simpl; eauto. }
      clear H0.
      unfold Genv.add_global, Genv.find_var_info, Genv.find_symbol in *. simpl.
      assert (Pos.of_succ_nat (Datatypes.length l) =
              Genv.genv_next (Genv.add_globals (Genv.empty_genv fundef unit) l)).
      { clear IHl H H1 H2.
        induction l using rev_ind. reflexivity.
        rewrite app_length, Genv.add_globals_app. simpl. rewrite <- IHl.
        rewrite !Pos.of_nat_succ, Pos.succ_of_nat, Pos.of_nat_succ by discriminate.
        f_equal. Psatz.lia. }
      destruct (le_lt_dec (length (map Some l)) idx).
      + rewrite app_nth2 in H1 by auto.
        destruct (idx - Datatypes.length (map Some l))%nat as [|[]] eqn:EQ; inv H1.
        replace idx with (Datatypes.length (map Some l)) in * by Psatz.lia.
        rewrite list_length_map.
        simpl; split; rewrite H0, PTree.gss; auto.
      + rewrite app_nth1 in H1 by auto.
        edestruct IHl; eauto.
        split.
        * destruct x as [? []]; simpl. auto. rewrite PTree.gso. auto.
          rewrite <- H0. intro. apply SuccNat2Pos.inj in H5.
          rewrite map_length in l0. Psatz.lia.
        * rewrite PTree.gso. auto.
          intro. subst.
          edestruct list_in_map_inv as (? & ? & ?). eapply nth_In with (d:=None). eauto.
          rewrite H1 in H5. inv H5.
          apply in_map with (f:=fst), H2 in H6. auto.
  Qed.

  Lemma ab_alloc_globals_ok:
    ∀ prog m, ge = Genv.globalenv prog ->
              Genv.init_mem prog = Some m ->
      list_norepet (map fst prog.(prog_defs)) ->
    ∀ b ab,
      ab_alloc_globals prog.(prog_defs) = ((b, ab), nil) ->
      b = Pos.of_succ_nat (length prog.(prog_defs)) /\
    match ab with
      | Bot => False
      | NotBot ab =>
        m ∈ sizes_gamma ge nil (ab_sz ab) /\
        ab_stk ab = Just nil /\
        mem_as_fun ge (nil, m) ∈ (points_to_gamma ge nil (ab_pt ab) ∩ γ(ab_nm ab))
    end.
  Proof.
    intros. generalize (nth_def_block _ H H1).
    revert ab b H2. clear H1.
    induction (prog_defs prog) using rev_ind.
    - intros. inv H2. simpl. split. auto. split. 2:split; [easy|split; [|split]].
      + red. unfold ABTreeDom.get. setoid_rewrite ABTree.gempty. discriminate.
      + red. red. intros. unfold ACTreeDom.get. rewrite ACTree.gempty. constructor.
      + apply gamma_top, N.tree_adom.
      + eexists. split.
        constructor. apply N.ncompat_compat_fun.
        simpl. unfold N.Eval.TDom.get. rewrite ACTree.gempty. apply N.wtype_num_top.
        apply gamma_top, id_adom.
    - intros.
      unfold ab_alloc_globals in H2. rewrite bm_fold_app in H2. revert H2.
      apply bm_bind_case. discriminate. intros [b' ab'] Hb'ab'.
      edestruct IHl.
      + apply Hb'ab'.
      + intros. apply H1. rewrite map_app.
        destruct (le_lt_dec (length (map Some l)) idx).
        rewrite nth_overflow in H2 by auto. discriminate.
        rewrite app_nth1 by auto. auto.
      + subst b'. clear Hb'ab'.
        unfold bm_fold. simpl. destruct x as [id [gf|gv]].
        * intros eq. inv eq. split.
          rewrite app_length, !Pos.of_nat_succ, Pos.succ_of_nat, Pos.of_nat_succ.
          f_equal. simpl. Psatz.lia. discriminate.
          destruct ab' as [|(? & ? & ? & ? & ?)]. contradiction.
          destruct H3 as (? & ? & ? & ?). simpl.
          split; [|split; [|split]]; auto.
          red. intros*. unfold ABTreeDom.get. setoid_rewrite ABTree.gsspec.
          destruct ABTree.elt_eq. 2:apply H2.
          subst b. intro eq. inversion eq. clear eq. subst hi rw.
          repeat intro. Psatz.lia.
        * destruct (gvar_volatile gv) eqn:?. discriminate.
          eapply bm_bind_case. discriminate. intros x AAG eq.
          inversion eq. subst x b. clear eq. split.
          rewrite app_length, !Pos.of_nat_succ, Pos.succ_of_nat, Pos.of_nat_succ.
          f_equal. simpl. Psatz.lia. discriminate.
          destruct ab' as [|ab']. contradiction.
          destruct H3 as (? & ? & ? & ?).
          edestruct H1.
          rewrite map_app, app_nth2, map_length, minus_diag. reflexivity.
          rewrite map_length. auto.
          eapply ab_alloc_global_ok; eauto.
          erewrite Genv.find_invert_symbol; eauto.
          unfold Genv.perm_globvar. rewrite Heqb0. destruct gvar_readonly; reflexivity.
  Qed.

  Instance mem_cshm_dom : mem_dom t :=
    {| assign := assign
     ; assign_any := assign_any
     ; store := store
     ; assume := assume

     ; nonblock := nonblock
     ; deref_fun := deref_fun

     ; push_frame := push_frame
     ; pop_frame  := pop_frame

     ; ab_malloc sz dst ab := do_bm _ <- err_bm "malloc"; ret_bm (⊤)
     ; ab_free e ab := do_bm _ <- err_bm "free"; ret_bm (⊤)
     ; ab_memcpy sz al dst src ab :=
         do_bm _ <- err_bm ("memcpy(" ++ to_string sz ++ ", " ++ to_string al ++ ", " ++ to_string dst ++ ", " ++ to_string src ++ ")");
         ret_bm (⊤)

     ; init_mem := init_mem
    |}.

  Lemma nonblock_correct e ab :
    match bm_get (nonblock e ab) with
    | Some tt => ∀ f tmp env stk m, ((f, (tmp, env)) :: stk, m) ∈ γ(ab) →
                                    ∃ v, eval_expr ge env tmp m e v ∧ v ≠ Vundef
    | None => True
    end.
  Proof.
    destruct ab as ([|[|(μ',ε') σ']], (sz, (pt, (tp, nm)))); try exact I.
    unfold nonblock. simpl.
    apply bm_bind_case. intros; exact I.
    simpl. intros ne NE f tmp env stk m H. ung. simpl in *. ung. eq_some_inv. subst.
    generalize (nconvert_sound _ nm (env_ok _ _ H1) INV SZ TP PT NM CPT e).
    rewrite NE. intros NB.
    destruct (nonblock_eval_expr_def NB) as (v & DEF & EV).
    vauto.
  Qed.

  Existing Instance UnitGamma.
  Lemma nonblock_sound: ∀ e ab, Nonblock ge e (γ ab) ⊆ γ (nonblock e ab).
  Proof.
    intros q ab.
    generalize (nonblock_correct q ab). destruct (nonblock q ab) as ((), [|]). 2: exact (λ _ _ _, I).
    intros X a [f t e s m H Nb]. specialize (X f t e s m H).
    destruct X as (v & V & Def).
    specialize (Nb _ V Def). rewrite Nb; reflexivity.
  Qed.

  Existing Instance boolean_partitionning.
  Lemma assume_sound :
   ∀ (e : expr) (ab : t) (a : option (bool * concrete_state)),
   Assume ge e (γ ab) a → γ (assume e ab) a.
  Proof.
    intros e ([|[|(μ,ε) σ]] & sz & pt & (tp & nm)) cs ASSUME; try exact I.
    destruct ASSUME as [f t' e' s m H K].
    ung. simpl in *. ung. eq_some_inv. subst x0 i l f.
    unfold assume. simpl.
    apply bm_bind_case. intros; exact I.
    intros nc NC.
    generalize (nconvert_sound _ nm (env_ok _ _ H1) INV SZ TP PT NM CPT (as_boolean_expr e)).
    rewrite NC. intros NB.
    destruct (nonblock_eval_expr_def NB) as (v & DEF & EV).
    destruct (as_boolean_expr_eval _ EV DEF) as
        (v' & V' & b & -> & Hb).
    specialize (K v' b Hb V').
    rewrite K.
    set (f b := do_bot z <- nassume b nm id nc; NotBot (AbMem (Just ((μ, ε) :: σ)) sz pt tp z)).
    fold (f true). fold (f false).
    apply boolean_partition. unfold f. clear f.
    generalize (nconvert_ex ε nm (env_ok _ _ H1) INV SZ TP PT NM CPT (as_boolean_expr e)).
    rewrite NC. simpl. intros (lpe & Hlpe & X).
    specialize (X _ EV). destruct X as (pe & ne & n & INp & Hpe & Hpn & INn & Hn & Hnty & NE).
    exact DEF.
    refine ((λ X Y, botbind_spec _ _ _ Y _ X) _ _).
    eapply nassume_correct; try eassumption.
    destruct b; inv Hn; exact NE.
    intros a' A'. constructor. red. simpl. econstructor; auto. split. reflexivity. auto.
    subst; auto.
    simpl. auto.
    simpl. auto.
    split. eauto. split. eauto. exists ρn. auto.
  Qed.

  Lemma cpt_wt ty v n : v ∈ γ(ty) → n ∈ N.ncompat v → N.wtype_num n ty.
  Proof. destruct ty as [|ty]; auto; intros T C; inv C;
         AbTy.gamma_inv; repeat destruct T as [T|T]; inversion T; vauto.
  Qed.
  Hint Resolve cpt_wt.

  Lemma assign_in_correct (μ: fname) (x: ident) (a: expr) (ab: t) :
    match bm_get (assign_in μ x a ab) with
    | Some ab' =>
      ∀ stk m,
        (stk, m) ∈ γ(ab) →
        match ab' with
        | Bot => False
        | NotBot ab' =>
          ab_stk ab = ab_stk ab' ∧
          ab_sz ab = ab_sz ab' ∧
          ∃ μ₀ env tmp stk',
          stk = (μ₀, (tmp, env)) :: stk' ∧
          match assoc_dep μ stk with
          | inleft (exist (pre, (tmp', env'), post) _) =>
            ∃ v, eval_expr ge env tmp m a v ∧ v ≠ Vundef ∧
                 (pre ++ (μ, (PTree.set x v tmp', env')) :: post, m)%list ∈ γ(ab')
          | inright _ => True
          end
        end
    | None => True
    end.
  Proof.
    unfold assign_in, assign_cells.
    apply bm_bind_case. intros; exact I.
    intros res Hres.
    generalize (ab_eval_expr_correct ab a). rewrite Hres.
    destruct ab as ([|[|(μ₀, ε₀) σ]] & sz & pt & tp & nm).
    intros (). intros ().
    intros H.
    destruct res as [|((ptr & ty) & ln)]; simpl.
    intros stk m G.
    ung. simpl in *. ung. subst.
    exact (H _ _ _ m INV SZ _ TP ρn CPT NM (env_ok _ _ H2) STK eq_refl PT).
    intros stk m G.
    ung. simpl in *. ung. subst.
    specialize (H _ _ _ m INV SZ _ TP ρn CPT NM (env_ok _ _ H2) STK eq_refl PT).
    destruct H as (v & V & Def & (Hvptr & Hty) & ne & n & INne & Cptn & _ & EVn).
    simpl.
    rewrite union_list_map_correct.
    unfold nm_upd_cells. simpl.
    generalize (union_list_bot (AD := id_adom) nm (map (nm_upd nm (ACtemp μ x)) ln)).
    generalize (union_list_correct (NotBot nm) join (lift_gamma (@id_gamma _ _ _ NumDom))
                                   (@join_sound _ _ _ _ (lift_bot id_adom))
                                   (map (nm_upd nm (ACtemp μ x)) ln)).
    destruct (union_list (NotBot nm) join (map _ ln)) as [|nm'].
    - intros _ X.
      simpl. specialize (X eq_refl). destruct X as [X Y].
      setoid_rewrite in_map_iff in Y. specialize (Y _ (ex_intro _ _ (conj eq_refl INne))).
      unfold nm_upd, nm_forget_all in Y.
      generalize (AbMachineEnv.assign_correct (ACtemp μ x) nm NM EVn).
      destruct AbMachineEnv.assign as [|nm']. intros ().
      intros NM'. exfalso.
      match goal with
      | H : context[fold_left ?F ?l ?m] |- _ =>
        generalize (λ E A ρ, forget_all_correct γ E F A l m _ ρ NM')
      end.
      unfold ACTree.elt in *.
      rewrite Y. intros K; eapply K.
      exact gamma_num_ext.
      { clear.
        intros [|m]. intros ? ? ? ().
        intros ρ ρ' x G E. simpl. eapply forget_correct'; eauto.
      }
      { intros c. left. reflexivity. }
    - intros X _.
      setoid_rewrite in_map_iff in X.
      specialize (λ m, X _ m (ex_intro _ ne (conj eq_refl INne))).
      simpl. split. reflexivity. split. reflexivity. eexists μ₀, _, _, _. split. reflexivity.
      destruct assoc_dep as [IN|]. 2: easy.
      destruct IN as (((pre & (tmp', env')) & post), IN). simpl in IN.
      destruct IN as ( STK' & IN).
      exists v. split. easy. split. easy.
      assert (∀ q, q ≠ ACtemp μ x → mem_as_fun ge ((μ₀, (t0, e)) :: x1, m) q = mem_as_fun ge (pre ++ (μ, (PTree.set x v tmp', env')) :: post, m) q)%list as Hupd.
      { intros q H. rewrite STK'.
        destruct q as [f y|f y|b]. simpl.
        rewrite !assoc_app. destruct (assoc f pre) as [(t',e')|]. reflexivity.
        simpl. destruct (eq_dec μ f). reflexivity. reflexivity.
        simpl. rewrite !assoc_app. destruct (assoc f pre) as [(t',e')|]. reflexivity.
        simpl. destruct (eq_dec μ f). rewrite PTree.gso. reflexivity. congruence. reflexivity.
        reflexivity. }
      assert (mem_as_fun ge (pre ++ (μ, (PTree.set x v tmp', env')) :: post ,m) (ACtemp μ x) = v)%list as Hupd'.
      { simpl. rewrite assoc_app. rewrite (proj2 (assoc_None _ _)). simpl.
        rewrite dec_eq_true, PTree.gss. reflexivity. auto. }
      split;[| | |split;[|split]].
      destruct pre as [|(?, (?, ?)) pre].
      + simpl in *; eq_some_inv; ssubst; vauto.
      + simpl in *. eq_some_inv. ssubst. constructor. vauto.
        apply Forall2_app_inv_r in STK. hsplit.
        apply list_append_map_inv in H0. hsplit. ssubst.
        fold (map γ (l0 ++ l3)).
        rewrite map_app.
        apply Forall2_app. auto.
        inversion H; clear H. ssubst.
        symmetry in H0; apply map_cons_inv in H0. hsplit.
        ssubst. vauto.
      + intros _.
        eapply invariant_tmp. 2: eassumption.
        unfold stacks_tmp. rewrite STK', !map_app. reflexivity.
      + simpl. eapply (sizes_gamma_tmp). 2: eassumption.
        unfold stacks_tmp. rewrite STK', !map_app. reflexivity.
      + assert (∀ b, ¬ In (μ, b) pre) as Hμ.
        { intros b K. generalize INV.(noRec). rewrite STK'. simpl. rewrite map_app. simpl.
          rewrite list_norepet_app. intros [_ [_ H]]. refine (H μ _ _ _ eq_refl).
          rewrite in_map_iff. exists (μ, b); vauto. vauto. }
        simpl. intros c.
        eapply ACTreeDom.gamma_set.
        eapply pt_forget_all_correct.
        eapply points_to_gamma_tmp. 2: eassumption.
        unfold stacks_tmp. rewrite STK', !map_app. reflexivity.
        intros. left. reflexivity.
        simpl. rewrite assoc_app. simpl. rewrite (proj2 (assoc_None _ _)), dec_eq_true, PTree.gss.
        destruct ptr as [|ptr]. exact I.
        eapply ptset_gamma_tmp. 2: eassumption.
        unfold stacks_tmp. rewrite STK', !map_app. reflexivity.
        exact Hμ. exact Hupd.
      + eapply ACTreeDom.gamma_set.
        eapply tp_forget_all_correct.
        eassumption.
        intros. left. reflexivity.
        destruct ty as [|ty]. exact I. rewrite Hupd'. auto. auto.
      + exists (upd ρn (ACtemp μ x) n). split.
        eapply N.compat_upd; eauto.
        intros q. destruct (eq_dec q) as [ -> | NE]. auto. symmetry; apply Hupd; exact NE.
        apply X.
        eapply forget_all_correct.
        exact gamma_num_ext.
        { clear. intros [|m]. intros ? ? ? (). simpl.
          intros ρ ρ' x H H0. eapply forget_correct'; eauto. }
        eapply AbMachineEnv.assign_correct; try eassumption.
        intuition.
  Qed.

  Transparent assoc_dep.
  Lemma assign_sound :
   ∀ (x : ident) (e : expr) (ab : t) (a : option concrete_state),
   Assign ge x e (γ ab) a → γ (assign x e ab) a.
  Proof.
    intros x e ab a H.
    unfold assign.
    destruct ab as ([|[|(μ,ε) σ]] & sz & pt & tp & nm); try (exact I).
    simpl.
    generalize (assign_in_correct μ x e (Just ((μ, ε) :: σ), (sz, (pt, (tp, nm))))).
    fold stack. fold tnum.
    destruct assign_in as (ab', [|]). 2: exact id. simpl.
    destruct H as [f tmp en stk m H Ha].
    intros H'. specialize (H' _ _ H).
    destruct ab' as [|ab']. destruct H'.
    hsplit. ung. eq_some_inv. subst. simpl in *. ung. eq_some_inv. subst i l x1 μ₀.
    destruct (eq_dec μ μ) as [_|K]. 2: elim (K eq_refl).
    hsplit.  simpl in *. erewrite Ha; eauto.
    eassumption.
  Qed.
  Opaque assoc_dep.

  Lemma assign_any_sound :
   ∀ (x : ident) (ty: AbTy.t) (ab : t) (a : option concrete_state),
   AssignAny x ty (γ ab) a → γ (assign_any x ty ab) a.
  Proof.
    intros x ty ab cs [f t e s m G (v & Hty' & ->)].
    destruct ab as ([|[|(μ, ε) stk]] & sz & pt & tp & nm); try exact I.
    unfold assign_any.
    ung. simpl in *. ung. eq_some_inv. subst i f l x1.
    assert (mem_as_fun ge ((μ, (PTree.set x v t, e)) :: s, m) ∈ upd_spec (mem_as_fun ge ((μ, (t, e)) :: s, m)) (ACtemp μ x) v) as Hupd.
    {
      intros c.
      destruct @eq_dec. subst. simpl. rewrite dec_eq_true, PTree.gss. reflexivity.
      destruct c; simpl; auto;
      destruct @eq_dec; auto. rewrite PTree.gso; auto. congruence.
    }
    assert (stacks_tmp ((μ, (t, e)) :: s) ((μ, (PTree.set x v t, e)) :: s)) by reflexivity.
    assert (ty <> AbTy.VP -> ty <> AbTy.VIP ->
        points_to_gamma ge ((μ, (PTree.set x v t, e)) :: s) (ACTreeDom.set pt (ACtemp μ x) (Just BlockSet.empty))
                            (mem_as_fun ge ((μ, (PTree.set x v t, e)) :: s, m))) as PT'.
    {
      intros.
      eapply points_to_gamma_tmp; eauto.
      intros c. generalize (PT c). rewrite (Hupd c), ACTreeDom.gsspec.
      unfold eq_dec.
      destruct (ACTree.elt_eq c), (ACellDec c (ACtemp μ x)); try congruence.
      destruct ty, v, Hty'; try congruence; exact (fun _ => I).
    }
    assert (v ∈ γ(Just ty)) as Hty by exact Hty'.
    destruct ty; try exact I;
    AbTy.gamma_inv; hsplit;
    match goal with |- appcontext[ AbMachineEnv.assign ?c ?e ?nm ] =>
                    cut (compat_fun v ∈ eval_nexpr ρn e);[
                        intros X;
                        generalize (AbMachineEnv.assign_correct c nm NM X);
                        destruct (AbMachineEnv.assign c e nm) as [|nm'];
                        [exact id|simpl; intros Hnm; split; simpl; vauto]
                       |]
    end;
    repeat
    match goal with
    | |- sizes_gamma _ _ _ _ => eapply sizes_gamma_tmp; eauto
    | |- appcontext[invariant] => intros _; eapply invariant_tmp; eauto
    | |- appcontext[@γ tnum] => unfold γ; simpl
    | |- _ ∧ _ => split
    | |- appcontext[points_to_gamma] => apply PT'; discriminate
    | |- appcontext[@γ types] => eapply ACTreeDom.gamma_set; eauto;
                     [rewrite Hupd, dec_eq_true; subst v; vauto
                     |intros; rewrite Hupd, dec_eq_false;easy]
    | |- ∃ _, _ ∧ γ _ _ =>
      eexists; split;[|eassumption];
      (eapply N.compat_upd; eauto;[eapply N.ncompat_compat_fun|subst; exact I])
    end.
    subst; constructor; rewrite <- (Int.repr_unsigned i); vauto.
    subst; constructor; rewrite <- (Int64.repr_unsigned l); vauto.
    subst; vauto.
    subst; vauto.
  Qed.

  Lemma store_sound :
   ∀ (κ : memory_chunk) (dst src : expr) (ab : t)
   (a : option concrete_state),
   Store ge κ dst src (γ ab) a
   → γ (store κ dst src ab) a.
  Proof.
    intros κ dst src ab a [f t e s m G H].
    unfold store.
    match goal with
        |- γ (match ?κ0 with _ => _ end) _ =>
        remember κ0 as κ'
    end.
    destruct κ' as [κ''|]. 2:constructor.
    assert (κ = proj1_sig κ'').
    { destruct κ; inv Heqκ'; reflexivity. }
    subst κ. clear Heqκ'.
    destruct ab as ([|[|(μ,ε) σ]] & sz & pt & tp & nm); try (exact I).
    simpl. apply bm_assert_spec. intros; exact I.
    intros Hcast.
    ung. simpl in *. ung. eq_some_inv. subst i l x0 f.
    pose proof (realize_cells_correct _ nm (env_ok _ _ H2) INV SZ TP PT NM CPT κ'' dst) as REALIZE.
    apply bm_bind_case. easy.
    intros cells Hcells. rewrite Hcells in REALIZE.
    destruct REALIZE as (NB & REALIZE).
    apply nonblock_eval_expr in NB. destruct NB as (v & Hv).
    specialize (REALIZE _ Hv). destruct REALIZE as (ab & b & i & -> & Habi & Hab & Hal).
    specialize (λ v m' S, H v _ m' S Hv).
    destruct cells as [|c₀ cells]. elim Habi; fail.
    match goal with
    | |- appcontext[ bm_rev_map ?f ?l ] =>
      pose proof (bm_rev_map_correct f l) as CIB
    end.
    apply bm_bind_case. easy. intros TT Hcib.
    rewrite Hcib in CIB. destruct CIB as [_ CIB'].
    assert (∀ c, In c (c₀ :: cells) → ∀ b, block_of_ablock ge ((μ, (t, e)) :: s) (ablock_of_cell c) = Some b →
                                   Mem.range_perm m b (fst (range_of_cell c)) (fst (range_of_cell c) + size_chunk (proj1_sig κ'')) Cur Writable) as CIB.
    {
      intros c IN b' Hb'.
      specialize (CIB' _ IN). destruct CIB' as (() & CIB' & _).
      apply (cell_in_bounds_range_perm CIB' SZ Hb').
    } clear CIB'.
    unfold assign_cells.
    match goal with
    | |- appcontext[ ab_eval_expr ?ab ?e ] =>
      pose proof (ab_eval_expr_correct ab e) as Hsrc
    end.
    apply bm_bind_case. intros; exact I.
    intros q Hq. rewrite Hq in Hsrc.
    specialize (Hsrc _ _ _ _ INV SZ _ TP _ CPT NM (env_ok _ _ H2) STK eq_refl PT).
    destruct q as [|((ptr & ty) & ln)]. destruct Hsrc.
    Opaque pt_upd tp_upd_cells. simpl.
    apply bm_assert_spec. intros; exact I.
    intros Htypes.
    destruct Hsrc as (v & Hsrc & Def & (Hvpt & Hvty) & ne & n & INne & Hvn & _ & Hn).
    specialize (λ m', H v m' Hsrc).
    destruct (Mem.valid_access_store m (proj1_sig κ'') b (Int.unsigned i) v) as (m' & Hm').
    { split. simpl.
      specialize (CIB _ Habi b). rewrite ablock_of_cell_from in CIB.
      specialize (CIB Hab).
      destruct ab; try exact CIB. simpl in Hab. eq_some_inv.
      exact Hal.
    } specialize (H m' Hm'). subst a.
    rewrite union_list_map_correct.
    set (l := map (nm_upd_cells nm (c₀ :: cells)) ln).
    generalize (union_list_bot (AD := id_adom) nm l).
    generalize (union_list_correct (NotBot nm) join _ (join_sound _ _ _ _ (lift_bot id_adom)) l).
    destruct (union_list (NotBot nm) join l) as [|nm'].
    intros _ X. specialize (X eq_refl).
    destruct X as (X & X').
    elim (nm_upd_cells_not_bot nm (c₀ :: cells) NM Hn). apply X'. subst l. rewrite in_map_iff. vauto.
    intros X _. simpl.
    assert (Val.load_result (proj1_sig κ'') v = v) as Hlr.
    {
      clear -Htypes Hvty Hcast Hsrc.
      destruct ty as [|()];
      simpl in *; eq_some_inv;
      AbTy.gamma_inv;
      destruct κ'' as [[] []]; eq_some_inv;
      intuition; hsplit; subst; try reflexivity;
      simpl; f_equal;
      destruct src; simpl in *; eq_some_inv;
      destruct u; eq_some_inv; eval_expr_inv; simpl in Hsrc; eq_some_inv;
      repeat
      match goal with
      | H : appcontext[Val.sign_ext _ ?a] |- _ => destruct a; inversion H; clear H;
                                                  apply Int.sign_ext_widen; intuition
      | H : appcontext[Val.zero_ext _ ?a] |- _ => destruct a; inversion H; clear H;
                                                  apply Int.zero_ext_widen; intuition
      end.
    }
    assert (loose_upd_spec
              (mem_as_fun ge ((μ, (t, e)) :: s, m))
              (cell_from ab (Int.unsigned i, κ''))
              v
              (mem_as_fun ge ((μ, (t, e)) :: s, m'))) as Hupd.
    {
      split. 
      + destruct ab as [f x|b']; simpl in *.
        destruct (eq_dec μ f).
        destruct (e ! x) as [(b' & _) | ]; eq_some_inv; subst b'.
        rewrite (Mem.load_store_same _ _ _ _ _ _ Hm'); auto.
        destruct (assoc f s) as [(_ & e')|]; eq_some_inv.
        destruct (e' ! x) as [(b' & _) | ]; eq_some_inv; subst b'.
        rewrite (Mem.load_store_same _ _ _ _ _ _ Hm'); auto.
        eq_some_inv.
        destruct (Genv.invert_symbol ge b'); eq_some_inv; subst b'.
        rewrite (Mem.load_store_same _ _ _ _ _ _ Hm'); auto.
      +
        pose proof (Mem.store_valid_access_3 _ _ _ _ _ _ Hm') as [_ AL].
        intros [f x i' κ'|f x|b' i' κ'] DIS; simpl.
        * destruct (eq_dec μ f).
          destruct (e ! x) as [(b' & s') | ] eqn: Hb'; auto; eq_some_inv.
          destruct (Zdivide_dec (align_chunk (proj1_sig κ')) i') as [AL'|NA]. apply align_chunk_pos.
          2: rewrite !(misaligned_load NA); reflexivity.
          rewrite (Mem.load_store_other _ _ _ _ _ _ Hm'). reflexivity.
          {
            destruct ab as [f' x'|b'']; simpl in Hab; eq_some_inv.
            destruct (eq_dec b' b); auto.
            right.
            assert ((f, x) = (f', x')).
            {
              subst μ b'. pose proof (INV.(@localInj _) (f, x) (f', x') b) as LOCALS. simpl in LOCALS.
              rewrite dec_eq_true in LOCALS. simpl in LOCALS. rewrite Hb' in LOCALS.
              specialize (LOCALS eq_refl).
              destruct (eq_dec f f'). simpl in LOCALS.
              destruct (e ! x') as [(b', s'')|]; eq_some_inv. subst b'.
              specialize (LOCALS eq_refl). congruence.
              destruct (assoc f' s) as [(?, e')|]; eq_some_inv. simpl in LOCALS.
              destruct (e' ! x') as [(b', s'')|]; eq_some_inv. subst b'.
              specialize (LOCALS eq_refl). congruence.
            }
            eq_some_inv. subst f' x'.
            destruct DIS as [DIS|DIS]. simpl in DIS. unfold is_true in DIS; hsplit; eq_some_inv.
            simpl in DIS. clear -DIS. intuition.
            apply do_opt_some_inv in Hab. destruct Hab as (sym & Hsym & Hab). eq_some_inv. subst b''.
            pose proof (INV.(localNotGlobal) _ _ _ (or_introl eq_refl)) as K.
            left. apply not_eq_sym. eapply K. apply Genv.invert_find_symbol, Hsym. exact Hb'.
          }
          destruct (assoc f s) as [(t' & e')|] eqn: He'; auto; eq_some_inv.
          destruct (e' ! x) as [(b' & s') | ] eqn: Hb'; auto; eq_some_inv.
          destruct (Zdivide_dec (align_chunk (proj1_sig κ')) i') as [AL'|NA]. apply align_chunk_pos.
          2: rewrite !(misaligned_load NA); reflexivity.
          rewrite (Mem.load_store_other _ _ _ _ _ _ Hm'). reflexivity.
          {
            destruct ab as [f' x'|b'']; simpl in Hab; eq_some_inv.
            destruct (eq_dec b' b); auto.
            right.
            assert ((f, x) = (f', x')).
            {
              subst b'. pose proof (INV.(@localInj _) (f, x) (f', x') b) as LOCALS. simpl in LOCALS.
              rewrite dec_eq_false in LOCALS; auto. simpl in LOCALS. rewrite He' in LOCALS.
              simpl in LOCALS. rewrite Hb' in LOCALS.
              specialize (LOCALS eq_refl).
              destruct (eq_dec μ f'); simpl in LOCALS.
              destruct (e ! x') as [(b', s'')|]; eq_some_inv. subst b'.
              specialize (LOCALS eq_refl). congruence.
              destruct (assoc f' s) as [(?, e'')|]; eq_some_inv. simpl in LOCALS.
              destruct (e'' ! x') as [(b', s'')|]; eq_some_inv. subst b'.
              specialize (LOCALS eq_refl). congruence.
            }
            eq_some_inv. subst f' x'.
            destruct DIS as [DIS|DIS]. simpl in DIS; unfold is_true in DIS; hsplit; eq_some_inv.
            simpl in DIS. clear -DIS. intuition.
            apply do_opt_some_inv in Hab. destruct Hab as (sym & Hsym & Hab). eq_some_inv. subst b''.
            left. apply not_eq_sym. eapply INV.(localNotGlobal).
            right. eapply assoc_in, He'.
            apply Genv.invert_find_symbol, Hsym. exact Hb'.
          }
        * reflexivity.
        * destruct (Genv.invert_symbol ge b') as [sym'|] eqn: Hsym'; auto; eq_some_inv.
          destruct (Zdivide_dec (align_chunk (proj1_sig κ')) i') as [AL'|NA]. apply align_chunk_pos.
          2: rewrite !(misaligned_load NA); reflexivity.
          rewrite (Mem.load_store_other _ _ _ _ _ _ Hm'). reflexivity.
          {
            destruct ab as [f' x'|b'']; simpl in Hab; eq_some_inv.
            left. intros ->.
            destruct (eq_dec μ f'). subst μ.
            pose proof (INV.(localNotGlobal) _ _ _ (or_introl eq_refl)) as GE.
            destruct (e ! x') as [(?,?)|] eqn: Hex'; eq_some_inv.
            eapply GE. eapply Genv.invert_find_symbol. eassumption. exact Hex'.
            auto.
            destruct (assoc f' s) as [(t',e')|] eqn: Hf's.
            destruct (e' ! x') as [(?,?)|] eqn: He'x'; eq_some_inv.
            eapply INV.(localNotGlobal).
            right. eapply assoc_in. exact Hf's.
            eapply Genv.invert_find_symbol. eassumption. exact He'x'.
            auto. eq_some_inv.
            apply do_opt_some_inv in Hab. destruct Hab as (sym & Hsym & Hab). eq_some_inv. subst b''.
            destruct (eq_dec b' b); auto. right. subst b'.
            clear - AL AL' DIS.
            destruct DIS as [DIS|DIS]; simpl in DIS.
            unfold is_true in DIS; hsplit; eq_some_inv.
            intuition.
          }
    }
    split;[| | |split;[|split]].
  - simpl. vauto.
  - intros _. split.
    + intros f t' e' H'.
      intros b0 hi H o' Ho'.
      eapply Mem.perm_store_1. eassumption.
      eapply INV; eassumption.
    + eapply INV; eauto.
    + eapply INV.(localInj).
    + eapply INV.
    + simpl.
      intros x b0 H. eapply (Mem.store_valid_block_1) . eassumption. eapply INV; eauto.
    + intros f t' e' H' x b' z H. eapply Mem.store_valid_block_1; try eassumption.
      eapply INV.(localValid); eauto.
  - simpl.
    intros b' hi rw b'' Hhi Hb'' o' Ho'.
    eapply Mem.perm_store_1. eassumption.
    eapply SZ; eassumption.
  - refine (pt_upd_correct (c₀ :: cells) ptr _ Habi PT Hupd).
    destruct ptr as [|bs]. exact I. simpl.
    destruct v; try exact I;
    try (elim (Def eq_refl));
    AbTy.gamma_inv;
    destruct κ'' as [[][]]; intuition (subst ty);
    simpl in Htypes; eq_some_inv.
  - refine (tp_upd_correct AbTy.plat_correct (c₀ :: cells) ty _ Habi TP Hupd).
    destruct ty as [|()]; try exact I;
    destruct κ'' as [[][]]; simpl in Htypes;
    eq_some_inv; AbTy.gamma_inv; intuition; hsplit; subst;
    vauto.
  - 
    set (Γ := (λ ty n, N.wtype_num n (Just ty))).
    assert (pre_lattice_spec AbTy.plat Γ) as PL.
    {
      split. intros () (); simpl; intros ? ? ?; eq_some_inv; auto.
      intros () () n' [H|H]; simpl; auto; try exact I.
    }
    assert (γ (gamma_op := toplift_gamma Γ) ty n) as Q.
    { destruct ty. exact I. exact (cpt_wt _ Hvty Hvn). }
    assert (γ (gamma_op := ACTreeDom.gamma Γ) tp ρn) as Q'.
    { intros c. destruct (CPT c) as [_ K]. simpl in *.
      unfold N.Eval.TDom.get in K. unfold ACTreeDom.get.
      destruct (ACTree.get c tp). exact K. exact I. }
    set (ρn' := λ c, if cells_disjoint_dec (cell_from ab (Int.unsigned i, κ'')) c
                     then ρn(c) else if eq_dec (cell_from ab (Int.unsigned i, κ'')) c
                                     then n
                                     else compat_fun (mem_as_fun ge ((μ, (t, e)) :: s, m') c)).
    assert (loose_upd_spec ρn (cell_from ab (Int.unsigned i, κ'')) n ρn') as H'.
    { clear. unfold ρn'. split.  destruct cells_disjoint_dec as [K|K].
      elim (cells_disjoint_not_eq K eq_refl).
      rewrite dec_eq_true. reflexivity.
      intros y H. destruct cells_disjoint_dec. reflexivity.
      eapply cells_disjoint_overlap; eauto. }
    exists ρn'.
    split.

  + unfold ρn'.
    intros c. destruct (CPT c) as [K L]. simpl in *. split.
    destruct (cells_disjoint_dec).
    generalize (proj2 Hupd c). simpl. intros Y. rewrite Y. auto. auto.
    destruct (eq_dec (cell_from _ _) c). subst c. generalize (proj1 Hupd). simpl. intros ->. assumption.
    apply N.ncompat_compat_fun.
    generalize (@tp_upd_correct num_val Γ PL tp ρn (c₀ :: cells) ty (cell_from ab (Int.unsigned i, κ'')) n Q Habi Q' ρn' H' c).
    unfold ACTreeDom.get, N.Eval.TDom.get.
    destruct (ACTree.get c) as [ty'|]. intros M. eapply M. auto.

  + subst ρn'.
    refine (X _ _ _ _).
    subst l. rewrite in_map_iff. exists ne. split. reflexivity. exact INne.
    unfold nm_upd_cells. rewrite union_list_map_correct.
    eapply union_list_correct; eauto. apply join_sound; auto.
    rewrite in_map_iff. vauto.
    eapply nm_upd_correct; eassumption.
  Qed.

  Definition alloc_variables' e m v :=
    List.fold_left (λ e_m id_sz,
                    let '(e, m) := e_m in
                    let '(id, sz) := id_sz in
                    let '(m₁, b₁) := Mem.alloc m 0 sz in
                    (PTree.set id (b₁, sz) e, m₁)) v (e, m).

  Lemma alloc_variables_alt v:
    ∀ e m e' m',
      alloc_variables e m v e' m' ↔ alloc_variables' e m v = (e', m').
  Proof.
    induction v as [|(id, sz) v IH].
    - intros e m e' m'. split; intros H.
      inv H. reflexivity.
      compute in H. eq_some_inv. subst. vauto.
    - intros e m e' m'.
      unfold alloc_variables'. simpl.
      destruct (Mem.alloc m 0 sz) as (m₁, b₁) eqn: H.
      change (alloc_variables e m ((id, sz) :: v) e' m' ↔ alloc_variables' (PTree.set id (b₁, sz) e) m₁ v = (e', m')).
      etransitivity. 2: apply IH.
      split; intros H'. inv H'. congruence.
      vauto.
  Qed.

  Definition alloc_var := alloc_variables' empty_env.
  Lemma alloc_var_alt m v e' m' :
    alloc_variables empty_env m v e' m' ↔ alloc_var m v = (e', m').
  Proof. apply alloc_variables_alt. Qed.

  Lemma alloc_var_snoc m v id sz :
    alloc_var m (v ++ (id, sz) :: nil)%list =
    let '(e, m') := alloc_var m v in
    let '(m₁, b₁) :=  Mem.alloc m' 0 sz in
    (PTree.set id (b₁, sz) e, m₁).
  Proof. apply fold_left_app. Qed.

  Lemma alloc_var_load m v:
    ∀ e' m',
    alloc_var m v = (e', m') →
    ∀ b,
      Mem.valid_block m b →
      Mem.valid_block m' b ∧
      ∀ κ o,
        Mem.load κ m' b o = Mem.load κ m b o.
  Proof.
    induction v as [|(id, sz) v IH] using rev_ind.
    intros e' m' H b H0. compute in H. eq_some_inv. subst. auto.
    intros e' m' H b H0. rewrite alloc_var_snoc in H.
    destruct (alloc_var m v) as (e₀, m₀).
    destruct (Mem.alloc m₀ 0 sz) as (m₁, b₁) eqn: H₁. eq_some_inv. subst.
    specialize (IH _ _ eq_refl _ H0). destruct IH as [V IH].
    split. apply (Mem.valid_block_alloc _ _ _ _ _ H₁). exact V.
    intros κ o. etransitivity. 2: apply IH.
    exact (Mem.load_alloc_unchanged _ _ _ _ _ H₁ _ _ _ V).
  Qed.

  Lemma alloc_variables_dom e m v e' m' :
    alloc_variables e m v e' m' →
    ∀ x, dom x e' = dom x e || in_dec peq x (List.map fst v).
  Proof.
    clear.
    induction 1. intros; simpl; rewrite orb_false_r; reflexivity.
    intros x. simpl. rewrite (IHalloc_variables x). clear.
    rewrite dom_set.
    destruct peq. destruct peq. rewrite orb_true_r. reflexivity. intuition.
    simpl. destruct (dom x e). reflexivity. simpl.
    destruct peq. intuition. destruct (in_dec); reflexivity.
  Qed.

  Lemma in_elements_set {A} e x y (v: A):
    In x (PTree.elements (PTree.set y v e)) →
    x = (y, v) ∨ In x (PTree.elements e) ∧ fst x ≠ y.
  Proof.
    destruct x as (x & v').
    intros H.
    apply PTree.elements_complete in H. rewrite PTree.gsspec in H. destruct peq. eq_some_inv. left. f_equal; auto.
    right. split. apply PTree.elements_correct; auto. easy.
  Qed.

  Lemma alloc_variables_perm_mono m e v e' m':
    alloc_variables e m v e' m' →
    ∀ b o p,
      Mem.perm m b o Cur p → Mem.perm m' b o Cur p.
  Proof.
    induction 1; auto.
    intros b o p H1. eapply IHalloc_variables.
    eapply Mem.perm_alloc_1; eauto.
  Qed.

  Lemma alloc_variables_perm' e m v e' m':
    alloc_variables e m v e' m' →
  ∀ b hi, In (b, 0, hi) (blocks_of_env e') →
          In (b, 0, hi) (blocks_of_env e) ∨ Mem.range_perm m' b 0 hi Cur Freeable.
  Proof.
    induction 1; intros b hi IN. left; exact IN.
    destruct (IHalloc_variables _ _ IN) as [IN'|R]; auto.
    clear IHalloc_variables.
    unfold blocks_of_env in IN'.
    rewrite in_map_iff in IN'. destruct IN' as (q & Hq & IN').
    apply in_elements_set in IN'. destruct IN' as [ -> | (IN' & Hq') ].
    2: left; apply in_map_iff; vauto.
    simpl in Hq. eq_some_inv. subst. right.
    intros o Ho. eapply Mem.perm_alloc_2 in H; eauto.
    eapply alloc_variables_perm_mono; eauto.
  Qed.

  Corollary alloc_variables_perm m v e m':
    alloc_variables empty_env m v e m' →
    ∀ b hi, In (b, 0, hi) (blocks_of_env e) →
            Mem.range_perm m' b 0 hi Cur Freeable.
  Proof.
    intros H b hi H0.
    destruct (alloc_variables_perm' H b hi H0) as [K|K]. elim K. exact K.
  Qed.

  Lemma alloc_variables_env' e m v e' m':
    alloc_variables e m v e' m' →
    list_norepet (List.map fst v) →
    ∀ x y, PTree.get x e' = Some y →
           PTree.get x e = Some y ∨ assoc x v = Some (snd y).
  Proof.
    induction 1. auto.
    intros NR x y H1. inv NR. destruct (IHalloc_variables H5 _ _ H1) as [K|K].
    rewrite PTree.gsspec in K. destruct peq. eq_some_inv. subst. right. simpl. rewrite dec_eq_true. reflexivity.
    left. exact K. simpl. rewrite dec_eq_false. auto.
    intro; apply H4. apply assoc_in in K. rewrite in_map_iff. subst. eexists; split; eauto; reflexivity.
  Qed.

  Corollary alloc_variables_env m v e m':
    alloc_variables empty_env m v e m' →
    list_norepet (List.map fst v) →
    ∀ x y, PTree.get x e = Some y →
           assoc x v = Some (snd y).
  Proof.
    intros H H0 x y H1. destruct (alloc_variables_env' H H0 _ H1) as [X|X]; eauto. rewrite PTree.gempty in X. eq_some_inv.
  Qed.

  Lemma alloc_variables_load' e m v :
    ∀ e' m',
      alloc_variables e m v e' m' →
      ∀ x y,
        e' ! x = Some y →
        e ! x = Some y ∨ Mem.valid_block m' (fst y) ∧ ∀ κ o, extend (Mem.load κ m' (fst y)) o = Vundef.
  Proof.
    setoid_rewrite alloc_variables_alt.
    induction v as [|(id, sz) v IH] using rev_ind.
    intros e' m' H x y H0. compute in H. eq_some_inv. left. congruence.
    intros e' m' H x y H0.
    unfold alloc_variables' in H. rewrite fold_left_app in H.
    assert ((let '(e, m) := alloc_variables' e m v in let '(m₁, b₁) := Mem.alloc m 0 sz in (PTree.set id (b₁, sz) e, m₁)) = (e', m')) as H' by
    exact H.
    clear H.
    destruct (alloc_variables' e m v) as (ee, mm). specialize (IH _ _ eq_refl).
    destruct (Mem.alloc mm 0 sz) as (m₁, b₁) eqn: H₁. eq_some_inv. subst.
    rewrite PTree.gsspec in H0.
    destruct peq. eq_some_inv. subst.
    right. split. apply (Mem.valid_new_block _ _ _ _ _ H₁).
    intros κ o.
    generalize (Mem.load_alloc_same _ _ _ _ _ H₁ κ o). simpl.
    destruct Mem.load; auto.
    specialize (IH _ _ H0). destruct IH as [IH | [V IH]].
    left; exact IH.
    right. split. exact (Mem.valid_block_alloc _ _ _ _ _ H₁ _ V). intros κ o.
    specialize (IH κ o). simpl in *.
    rewrite (Mem.load_alloc_unchanged _ _ _ _ _ H₁ κ (fst y) o). exact IH. exact V.
  Qed.

  Corollary alloc_variables_load m v e m':
    alloc_variables empty_env m v e m' →
    ∀ x y, e ! x = Some y →
           Mem.valid_block m' (fst y) ∧ ∀ κ o, extend (Mem.load κ m' (fst y)) o = Vundef.
  Proof.
    intros H x y H0. destruct (alloc_variables_load' H _ H0); auto. rewrite PTree.gempty in *. eq_some_inv.
  Qed.

  Lemma alloc_var_inj m v :
    ∀ e' m', alloc_var m v = (e', m') →
             (Mem.valid_block m ⊆ Mem.valid_block m') ∧
             (∀ x y, e' ! x = Some y → Mem.valid_block m' (fst y) ∧ ¬ Mem.valid_block m (fst y)) ∧
             injective_map (λ x, option_lift fst (e' ! x)).
  Proof.
    induction v as [|(id, sz) v IH] using rev_ind.
    intros e' m' H. compute in H. eq_some_inv.  subst. split. easy.
    split; repeat intro; rewrite PTree.gempty in *; simpl in *; eq_some_inv.
    intros e' m' H. rewrite alloc_var_snoc in H.
    destruct (alloc_var m v) as (e₀, m₀). specialize (IH _ _ eq_refl).
    destruct (Mem.alloc m₀ 0 sz) as (m₁, b₁) eqn: H₁. eq_some_inv. subst.
    destruct IH as (M & V & IH).
    split. intros b B. eapply Mem.valid_block_alloc; eauto.
    split. intros x y H. rewrite PTree.gsspec in H. destruct peq. eq_some_inv. subst.
    split. exact (Mem.valid_new_block _ _ _ _ _ H₁). intros X.
    apply (Mem.fresh_block_alloc _ _ _ _ _ H₁), M, X.
    destruct (V _ _ H); intuition.
    apply (Mem.valid_block_alloc _ _ _ _ _ H₁). eauto.
    intros x x' a Ha Ha'.
    rewrite PTree.gsspec in *. destruct peq. simpl in *. eq_some_inv. subst id a.
    destruct peq. easy.
    destruct (e₀ ! x') as [(?,?)|] eqn: Hx'; simpl in *; eq_some_inv; subst.
    exfalso. generalize (proj1 (V _ _ Hx')). exact (Mem.fresh_block_alloc _ _ _ _ _ H₁).
    destruct peq. simpl in *. eq_some_inv. subst.
    destruct (e₀ ! x) as [(?,?)|] eqn: Hx; simpl in *; eq_some_inv; subst.
    exfalso. generalize (proj1 (V _ _ Hx)). exact (Mem.fresh_block_alloc _ _ _ _ _ H₁).
    exact (IH _ _ _ Ha Ha').
  Qed.

  Lemma alloc_variables_build vars m m' :
    ∀e',
      alloc_variables empty_env m vars e' m' →
      ∀ x : PSet.elt,
        PSet.mem x (PSetOp.build (rev_map fst vars)) = dom x e'.
  Proof.
    rewrite rev_map_correct.
    induction vars as [|v vars IH];
      intros e1 E1; inv E1.
    clear. intros x. unfold dom, empty_env. rewrite PTree.gempty.
    apply not_true_is_false. rewrite (PSetOp.mem_build x nil). auto.
    intros x. rewrite (alloc_variables_dom H6), dom_set, eq_iff_eq_true.
    setoid_rewrite PSetOp.mem_build. simpl.
    unfold dom, empty_env.
    rewrite !Bool.orb_true_iff, in_app, <- in_rev, PTree.gempty.
    split.
    - intros [?|[?|[]]].
      destruct in_dec; intuition.
      destruct peq; intuition.
    - simpl. intros [[?|?]|?].
      destruct peq; intuition. discriminate. destruct in_dec; intuition.
  Qed.

  Remark create_undef_temps_undef l x :
    extend (λ x, (create_undef_temps l) ! x) x = Vundef.
  Proof.
    simpl. generalize (create_undef_temps_get l x).
    destruct ((create_undef_temps l) ! x); intuition.
  Qed.

  Remark bm_get_some {T} (x: block_mon T) (y: T) :
    bm_get x = Some y → x = ret_bm y.
  Proof. destruct x as (x, [|]); simpl; intro; eq_some_inv. subst. reflexivity. Qed.

  Lemma push_frame_invariant μ vars :
    ∀ σ m tmp e m',
      alloc_variables empty_env m vars e m' →
      (∀ b,  ¬ In (μ, b) σ) →
      invariant (σ, m) →
      invariant ((μ, (tmp, e)) :: σ, m').
  Proof.
    intros σ m tmp e m' Hal Nr Inv. split.
    + (* local freeable *)
      intros f t0 e0 [H|H] b hi H0.
      eq_some_inv. subst f t0 e0.

      eapply (alloc_variables_perm); eauto.
      intros o Ho. eapply (alloc_variables_perm_mono); eauto.
      eapply Inv.(localFreeable); eauto.
    + (* local not global *)
      intros f t0 e0 [H|H] x b y b' z H0 H1.
      2: exact (Inv.(localNotGlobal) _ _ _ H _ _ H0 H1).
      eq_some_inv. subst f t0 e0.
      pose proof (Inv.(globalValid) _ H0) as X.
      pose proof (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) Hal)) as (_ & Y & _).
      specialize (Y _ _ H1). destruct Y as [_ Y]. simpl in *. intros ->. exact (Y X).
    + (* local injective *)
      intros (f, x) (f', x') a. simpl.
      destruct (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) Hal)) as (_ & Q & H).
      destruct (eq_dec μ f).
      destruct (eq_dec μ f'). subst f f'. simpl.
      intros H0 H1. f_equal. exact (H _ _ _ H0 H1).
      subst f. simpl.
      intros H0 H1. apply do_opt_some_inv in H0. destruct H0 as ((a' & al) & H0 & ?). simpl in *. eq_some_inv. subst a'.
      specialize (Q _ _ H0). simpl in Q.
      apply do_opt_some_inv in H1. destruct H1 as (tee & Htee & H1).
      apply do_opt_some_inv in H1. destruct H1 as ((a' & al') & H1 & ?). simpl in *. eq_some_inv. subst a'.
      apply assoc_in in Htee. destruct tee as (te' & e'). simpl in *.
      assert (In (a, 0, al') (blocks_of_env e')) as Ha. unfold blocks_of_env. rewrite in_map_iff.
      eexists. split. 2: apply PTree.elements_correct, H1. reflexivity.
      elim (proj2 Q). eapply Inv.(localValid); eauto.
      destruct (eq_dec μ f'). subst f'. simpl.
      intros H0 H1.
      apply do_opt_some_inv in H0. destruct H0 as ((te', e') & Htee & H0).
      apply do_opt_some_inv in H0. destruct H0 as ((a', al) & Ha & ?). simpl in *. eq_some_inv. subst a'.
      apply do_opt_some_inv in H1. destruct H1 as ((a', al') & Ha' & ?). simpl in *. eq_some_inv. subst a'.
      specialize (Q _ _ Ha'). simpl in Q. elim (proj2 Q). clear Q.
      apply assoc_in in Htee. eapply (Inv.(localValid)); eauto.
      exact (Inv.(@localInj _) (f, x) (f', x') a).
    + (* no rec *)
      simpl.
      constructor. intros K. apply in_map_iff in K. destruct K as ((i & ?) & ? & K). subst. exact (Nr _ K).
      exact Inv.(noRec).
    + (* global valid *)
      simpl. intros x b H.
      destruct (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) Hal)) as (Q & _). apply Q.
      eapply Inv.(globalValid); eauto.
    + (* local valid *)
      simpl. intros f t' e' [H|H] x b z H'. eq_some_inv. subst. apply (proj1 (alloc_variables_load Hal _ H')).
      destruct (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) Hal)) as (Q & _). apply Q.
      eapply Inv.(localValid); eauto.
  Qed.

  Lemma push_frame_sizes σ sz μ vars m t' e' m' :
    list_norepet (map fst vars) →
    alloc_variables empty_env m vars e' m' →
    m ∈ sizes_gamma ge σ sz →
    m' ∈ sizes_gamma ge ((μ, (t', e')) :: σ) (set_local_sizes μ vars sz).
  Proof.
    intros NR AL SZ.
    intros b hi rw b' H H'. rewrite set_local_sizes_get in H; auto.
    destruct b as [μ' x|b]; simpl in H'; eq_some_inv.
  - destruct (eq_dec μ μ'). subst μ'. rewrite peq_true in H.
    apply (alloc_variables_perm AL). unfold blocks_of_env. rewrite in_map_iff.
    destruct (e' ! x) as [(q, s)|] eqn: Hx; eq_some_inv. subst q.
    eexists . split. 2: apply PTree.elements_correct, Hx. simpl.
    erewrite (alloc_variables_env AL NR) in H; eauto. simpl in H. congruence.
    rewrite peq_false in H; auto.
    intros o Ho. apply (alloc_variables_perm_mono AL).
    exact (SZ _ hi rw b' H H' _ Ho).
  - generalize (SZ _ hi rw b' H H').
    apply do_opt_some_inv in H'. hsplit. eq_some_inv. subst b'.
    intros X ofs Ho. eapply alloc_variables_perm_mono; eauto.
  Qed.

  Lemma push_frame_pt σ μ₀ tmp e' (Nr: ¬ In μ₀ (map fst σ)) bs :
    ptset_gamma ge σ bs ⊆ ptset_gamma ge ((μ₀, (tmp, e')) :: σ) bs.
  Proof.
    intros (); simpl; try easy.
    intros b _ (b' & Hb' & Hb). exists b'. split. exact Hb'.
    destruct b' as [f' x|b']; simpl in *;
    try apply do_opt_some_inv in Hb.
    destruct Hb as ((t'', e'') & H & Hb).
    apply do_opt_some_inv in Hb. destruct Hb as ((b', sz') & Hb & ?). eq_some_inv. subst b'.
    rewrite H. rewrite dec_eq_false, Hb. reflexivity.
    intros <-. apply Nr. apply in_map_iff. eexists. split. 2: eapply assoc_in; eassumption. reflexivity.
    destruct Hb as (q & Hq & Hb). eq_some_inv. subst b'. rewrite Hq. reflexivity.
  Qed.

  Section ARGS_MATCH.
    Local Set Elimination Schemes.
    Variables (f: ident) (en : env) (tmp: temp_env) (σ: list (ident * (temp_env * env))) (m: mem) (ρn: cell → num_val).
    Inductive args_match : list expr → list val → list ((PTSet * AbTy.t+⊤ * list (nexpr cell))+⊥) → Prop :=
    | AMnil : args_match nil nil nil
    | AMcons eargs vargs aargs e v ptr ty ln ne n :
        v ≠ Vundef →
        v ∈ eval_expr ge en tmp m e →
        v ∈ toplift_gamma (ptset_gamma ge ((f, (tmp, en)) :: σ)) ptr →
        v ∈ γ(ty) →
        In ne ln →
        n ∈ N.ncompat v → n ∈ eval_nexpr ρn ne →
        (∀ x y, ACtemp x y ∈ ne_free_var ne → In x (f :: map fst σ)) →
        args_match eargs vargs aargs →
        args_match (e :: eargs) (v :: vargs) (NotBot (ptr, ty, ln) :: aargs)
    .

    Lemma am_a_cons_inv eargs vargs a aargs :
      args_match eargs vargs (a :: aargs) →
      ∃ e eargs' v vargs' ptr ty ln ne n,
        eargs = e :: eargs' ∧
        vargs = v :: vargs' ∧
        a = NotBot (ptr, ty, ln) ∧
        v ≠ Vundef ∧
        v ∈ eval_expr ge en tmp m e ∧
        v ∈ toplift_gamma (ptset_gamma ge ((f, (tmp, en)) :: σ)) ptr ∧
        v ∈ γ(ty) ∧
        In ne ln ∧
        n ∈ (N.ncompat v ∩ eval_nexpr ρn ne) ∧
        (∀ x y, ACtemp x y ∈ ne_free_var ne → In x (f :: map fst σ)) ∧
        args_match eargs' vargs' aargs.
    Proof.
      inversion 1. repeat (econstructor (eauto)).
    Qed.
  End ARGS_MATCH.

  Lemma read_cell_upd f t e σ m x v c:
    c ≠ ACtemp f x →
    read_cell ge ((f, (PTree.set x v t, e)) :: σ, m) c =
    read_cell ge ((f, (t, e)) :: σ, m) c.
  Proof.
    destruct c as [f' x'|f' x'|b]; simpl; auto.
    intros _. destruct @eq_dec. reflexivity.
    destruct assoc. reflexivity. reflexivity.
    intros NE. destruct @eq_dec. rewrite PTree.gso. reflexivity. congruence.
    reflexivity.
  Qed.

  Lemma ab_bind_parameters_sound μ₀ f e t σ m vars e' m' :
    invariant ((f, (t, e)) :: σ, m) →
    ¬ In μ₀ (f :: map fst σ) →
    alloc_variables empty_env m vars e' m' →
    ∀ params eargs aargs vargs ρn,
    args_match f e t σ m ρn eargs vargs aargs →
    ∀ tmp tmp',
    bind_parameters params vargs tmp = Some tmp' →
    ∀ ab ab',
    ab_bind_parameters μ₀ params aargs (ret_bm (NotBot ab)) = ret_bm ab' →
    mem_as_fun ge ((μ₀, (tmp, e')) :: (f, (t, e)) :: σ, m) ∈ points_to_gamma ge ((μ₀, (tmp, e')) :: (f, (t, e)) :: σ) (ab_pt ab) →
    mem_as_fun ge ((μ₀, (tmp, e')) :: (f, (t, e)) :: σ, m) ∈ γ(ab_tp ab) →
    ρn ∈ (compat (ab_tp ab) (mem_as_fun ge ((μ₀, (tmp, e')) :: (f, (t, e)) :: σ, m)) ∩ γ(ab_num ab)) →
    ∃ q,
      ab' = NotBot q ∧
      ab_stk q = ab_stk ab ∧
      ab_sz q = ab_sz ab ∧
      mem_as_fun ge ((μ₀, (tmp', e')) :: (f, (t, e)) :: σ, m') ∈ (points_to_gamma ge ((μ₀, (tmp', e')) :: (f, (t, e)) :: σ) (ab_pt q) ∩ γ(ab_tp q)) ∧
      ∃ ρn', ρn' ∈ (compat (ab_tp q) (mem_as_fun ge ((μ₀, (tmp', e')) :: (f, (t, e)) :: σ, m')) ∩ γ (ab_num q)).
  Proof.
    intros INV Nr Hm'.
    induction params as [|id params IH];
    intros eargs aargs vargs ρn AM tmp tmp' BIND ab ab' AB_BIND PT TP (CPT & NM);
    destruct ab as (stk & sz & pt & tp & nm).
    + destruct aargs; simpl in AB_BIND, BIND; unfold ret_bm in AB_BIND; eq_some_inv. subst ab'.
      destruct vargs; eq_some_inv. subst tmp'.
      exists (stk, (sz, (pt, (tp, nm)))). split. reflexivity. split. reflexivity. split. reflexivity. split;[split|exists ρn;split]; auto.
    - intros c. generalize (PT c). simpl.
      destruct (ACTreeDom.get c pt) as [|bs]. exact id.
      destruct c as [f' x ofs|f' x|b ofs]; simpl.
      * destruct @eq_dec. subst f'. destruct (e' ! x) as [(b, h)|] eqn: Hx.
        pose proof (proj2 (alloc_variables_load Hm' _ Hx) (proj1_sig κ) ofs) as X. simpl in X. rewrite X. exact (λ _, I). exact (λ _, I).
        destruct @eq_dec. subst f'. destruct (e ! x) as [(b, h)|] eqn: Hx.
        assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). left. reflexivity. exact Hx.
        rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)).
        exact id. exact id.
        destruct (assoc f' σ) as [(t'', e'')|] eqn: Hf'.
        destruct (e'' ! x) as [(b, h)|] eqn: Hx.
        assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). right. apply (assoc_in _ _ Hf'). exact Hx.
        rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)).
        exact id. exact id. exact id.
      * exact id.
      * destruct (Genv.invert_symbol ge b) eqn: B; auto.
        rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
        exact id.
    - intros c. generalize (TP c). simpl.
      destruct (ACTreeDom.get c tp) as [|ty]. exact id.
      destruct c as [f' x ofs|f' x|b ofs]; simpl; intros X. AbTy.gamma_inv;  try discriminate.
      destruct (eq_dec μ₀ f'). subst f'. destruct (e' ! x) as [(b,h)|] eqn: Hx.
      pose proof (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) Hm')) as (_ & Y & _).
      specialize (Y _ _ Hx). destruct Y as [_ Y]. simpl in Y.
      rewrite (not_valid_block_load_None Y) in X. AbTy.gamma_inv. discriminate.
      exact X.
      destruct @eq_dec. subst f'. destruct (e ! x) as [(b, h)|] eqn: Hx.
      assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). left. reflexivity. exact Hx.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)). exact X. exact X.
      destruct (assoc f' σ) as [(t'', e'')|] eqn: Hf'.
      destruct (e'' ! x) as [(b, h)|] eqn: Hx.
      assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). right. apply (assoc_in _ _ Hf'). exact Hx.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)). exact X. exact X. exact X. exact X.
      destruct (Genv.invert_symbol ge b) eqn: B; auto.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
      exact X.
    - intros c. split. 2: exact (proj2 (CPT c)). generalize (proj1 (CPT c)). simpl.
      destruct c as [f' x ofs|f' x|b ofs]; simpl; intros X.
      destruct (eq_dec μ₀ f'). subst f'.
      destruct (e' ! x) as [(b,s)|] eqn: Hx. clear X.
      pose proof (proj2 (alloc_variables_load Hm' _ Hx) (proj1_sig κ) ofs) as X. simpl in X. rewrite X. vauto. exact X.
      destruct (eq_dec f f'). subst f'.
      destruct (e ! x) as [(b,s)|] eqn: Hx.
      assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). left. reflexivity. exact Hx.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)).
      exact X. exact X.
      destruct (assoc f' σ) as [(t'', e'')|] eqn: Hf'.
      destruct (e'' ! x) as [(b, h)|] eqn: Hx.
      assert (Mem.valid_block m b) as Hb. eapply INV.(localValid). right. apply (assoc_in _ _ Hf'). exact Hx.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') Hb)). exact X. exact X. exact X. exact X.
      destruct (Genv.invert_symbol ge b) eqn: B; auto.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) Hm') (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
      exact X.

    + destruct aargs as [|a aargs']. simpl in AB_BIND; unfold ret_bm in AB_BIND; eq_some_inv.
      destruct (am_a_cons_inv AM) as
          (ex & eargs' & v & vargs' & ptr & ty & ln & ne & n & -> & -> & -> & Def & EV & Vpt & Vtp & Hne & (Ncpt & EVn) & Hfv & AM').
      assert (args_match f e t σ m (upd ρn (ACtemp μ₀ id) n) eargs' vargs' aargs') as AM''.
      {
        clear -AM' Hfv Nr.
        induction AM'. constructor.
        econstructor; eauto. eapply eval_nexpr_fv. 2: eassumption.
        intros x H7. unfold upd. destruct (eq_dec x); auto.
        elim Nr. subst. eauto.
      }
      specialize (IH _ _ _ _ AM'' _ _ BIND).
      simpl in AB_BIND.
      rewrite union_list_map_correct in AB_BIND.
      destruct (union_list (NotBot nm) join (map (λ ne, AbMachineEnv.assign (ACtemp μ₀ id) ne nm) ln)) as [|nm'] eqn: NM'.
    - apply (union_list_bot (AD:=id_adom)) in NM'.
      destruct NM' as [_ NM'].
      assert (∀ ne, In ne ln → AbMachineEnv.assign (ACtemp μ₀ id) ne nm = Bot) as NM''.
      { intros ? ?. apply NM'. rewrite in_map_iff. vauto. }
      clear NM'.
      specialize (NM'' _ Hne).
      generalize (AbMachineEnv.assign_correct (ACtemp μ₀ id) nm NM EVn). rewrite NM''. intros ().
    - specialize (IH _ _ AB_BIND).
      destruct IH as ((stk' & sz' & pt' & tp' & nm'') & -> & Hstk' & Hsz' & (Hpt' & Htp') & Hq).
      simpl.
      * intros c. unfold ct_upd. rewrite ACTreeDom.gsspec. rewrite ct_forget_all_get.
        destruct (ACTree.elt_eq c). subst c. simpl. rewrite dec_eq_true, PTree.gss.
        destruct ptr. exact I. apply push_frame_pt; eauto.
        destruct in_dec. exact I.
        simpl. rewrite read_cell_upd.
        generalize (PT c). simpl. destruct (ACTreeDom.get). exact (λ _, I).
        apply ptset_gamma_tmp. reflexivity. assumption.
      * intros c. simpl. unfold ct_upd. rewrite ACTreeDom.gsspec. rewrite ct_forget_all_get.
        destruct (ACTree.elt_eq c). subst c. simpl. rewrite dec_eq_true, PTree.gss. exact Vtp.
        destruct in_dec. exact I. rewrite read_cell_upd; auto.
        exact (TP c).
      * split. simpl. intros c. generalize (CPT c).
        unfold N.Eval.TDom.get, ct_upd.
       (* fold (ACTreeDom.get c (ACTreeDom.set (ct_forget_all (overlap (ACtemp μ₀ id)) tp) (ACtemp μ₀ id) ty)). *)
        simpl. rewrite ACTreeDom.gsspec.
        destruct ACTree.elt_eq. intros _. subst c. rewrite upd_s. split.
        simpl. rewrite dec_eq_true, PTree.gss. exact Ncpt. eauto.
        rewrite upd_o; auto. rewrite ct_forget_all_get. destruct in_dec. intuition. intros [A B]. split; auto.
        rewrite read_cell_upd; auto.
        generalize (union_list_correct (NotBot nm) join (lift_gamma (id_gamma)) (@join_sound _ _ _ _ (lift_bot id_adom)) (map (λ ne, AbMachineEnv.assign (ACtemp μ₀ id) ne nm) ln)).
        rewrite NM'. simpl.
        intros X. eapply X. apply in_map. eassumption.
        apply AbMachineEnv.assign_correct; eauto.
      * simpl in *. subst. eexists (stk, (sz, (pt', (tp', nm'')))). vauto.
  Qed.

  Lemma push_frame_sound :
   ∀ (fi : ident) (f : function) (args : list expr)
   (ab : t) (a : option concrete_state),
   PushFrame ge fi f args (γ ab) a → γ (push_frame fi f args ab) a.
  Proof.
    intros fi fn args.
    intros (stk & sz & pt & (tp & nm)) cs (σ & m & H & NR & PF).
    ung. simpl in *.
    unfold push_frame. simpl.
    destruct stk as [|σ']. simpl. apply bm_bind_case. easy.
    intros ? ?. destruct ab_bind_parameters; easy.
    generalize (proj1 (assoc_None fi σ')). simpl.
    destruct assoc. intros; exact I.
    intros NOREC. specialize (NOREC eq_refl).
    assert ( ∃ m1 e1, alloc_variables empty_env m fn.(fn_vars) e1 m1) as (m1 & e1 & H1).
    {
      generalize fn.(fn_vars). intros l.
      generalize empty_env, m.
      induction l as [|(v,sz') l IH]. vauto2.
      intros e m'.
      destruct (Mem.alloc m' 0 sz') as [m'' b] eqn:AL.
      destruct (IH (PTree.set v (b, sz') e) m'') as (m1 & e1 & H1).
      vauto2.
    }
    specialize (λ vargs Hvargs, PF vargs Hvargs m1 e1 H1).
    unfold ab_eval_exprlist.
    generalize (bm_rev_map_inv(ab_eval_expr (Just σ', (sz, (pt, (tp, nm))))) args).
    generalize (bm_rev_map_correct (ab_eval_expr (Just σ', (sz, (pt, (tp, nm))))) args).
    unfold tnum.
    destruct (bm_rev_map (ab_eval_expr (Just σ', (sz, (pt, (tp, nm))))) args) as
    (eargs, [|]).
    2: intros _ _; simpl; destruct ab_bind_parameters; exact I.
    intros [EARGS EARGS'] ->.
    assert (∀ {A B}, ∀ ab : A * B, (let '(a, b) := ab in (a, b)) = ab) as zeta
    by (intros A B (a, b); reflexivity).
    repeat setoid_rewrite zeta.
    clear zeta.
    unfold rev'. simpl. rewrite <- rev_alt, rev_involutive.
    destruct σ as [|(μ, (t, e)) σ].

  - (* empty stack: initial call to main *)
    specialize (PF _ eq_refl).
    ung. subst σ'.
    clear NOREC.
    generalize dependent (fn_params fn). intros [|] PF.
    { 
    clear EARGS EARGS'.
    specialize (PF _ eq_refl).
    subst cs.
    unfold ab_bind_parameters.
    destruct args. 2: exact I.
    simpl.
    split; [| | |split;[|split]].
    + hnf. simpl. constructor. 2: vauto.
      split. reflexivity.
      exact (alloc_variables_build H1).
    + simpl. intros _. eapply push_frame_invariant; eauto.
    + simpl. eapply push_frame_sizes; eauto.
    + intros c. generalize (PT c). simpl.
      destruct (ACTreeDom.get c pt) as [|bs]. exact id.
      destruct c as [f x ofs|f x|b ofs]; simpl.
      destruct @eq_dec. subst. destruct (e1 ! x) as [(b, h)|] eqn: Hx; auto.
      pose proof (proj2 (alloc_variables_load H1 _ Hx) (proj1_sig κ) ofs) as X. simpl in X. rewrite X. exact id. exact id.
      destruct @eq_dec; auto. subst.
      generalize (create_undef_temps_get (fn_temps fn) x).
      destruct ((create_undef_temps (fn_temps fn)) ! x); intuition. subst; auto.
      destruct (Genv.invert_symbol ge b) eqn: B; auto.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) H1) (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
      destruct (Mem.load (proj1_sig κ) m b ofs) as [()|]; try exact id.
      intros (q & Q & Q'). exists q. split. exact Q. destruct q; simpl in *; eq_some_inv. exact Q'.
    + intros c. generalize (TP c). simpl.
      destruct (ACTreeDom.get c tp) as [|ty]. exact id.
      destruct c as [f x ofs|f x|b ofs]; simpl; intros X; AbTy.gamma_inv;  try discriminate.
      destruct (Genv.invert_symbol ge b) eqn: B; auto.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) H1) (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
      exact X.
    + exists ρn. split; auto.
      intros c. split. 2: exact (proj2 (CPT c)). generalize (proj1 (CPT c)). simpl.
      destruct c as [f x ofs|f x|b ofs]; simpl; intros X.
      destruct @eq_dec.
      destruct (e1 ! x) as [(b,s)|] eqn: Hx.
      generalize (proj2 (alloc_variables_load H1 _ Hx) (proj1_sig κ) ofs). simpl. intros Y. rewrite Y. exact X.
      exact X. exact X.
      destruct @eq_dec.
      generalize (create_undef_temps_get (fn_temps fn) x).
      destruct ((create_undef_temps (fn_temps fn)) ! x); intuition. subst; auto. auto.
      destruct (Genv.invert_symbol ge b) eqn: B; auto.
      rewrite (proj2 (alloc_var_load _ (proj1 (alloc_var_alt _ _ _ _) H1) (INV.(globalValid) _ (Genv.invert_find_symbol _ _ B)))).
      exact X.
    }

    destruct args.
    exact I. specialize (EARGS Bot). rewrite <- in_rev in EARGS.
    specialize (EARGS (or_introl eq_refl)).
    destruct EARGS as (? & HE & INe).
    simpl in HE. eq_some_inv.

  - (* non empty stack. *)
    assert (∀ x, In x args → ∃ y, bm_get (ab_eval_expr (Just σ', (sz, (pt, (tp, nm)))) x) = Some y) as EARG.
    intros x H. destruct (EARGS' x H) as (y & ? & ?); vauto.
    clear EARGS EARGS'.
    assert (∃ vargs, eval_exprlist ge e t m args vargs ∧
                     Forall2
                       (λ x v,
                        ∃ ptr ty ln, ab_eval_expr (Just σ', (sz, (pt, (tp, nm)))) x = ret_bm (NotBot ((ptr, ty), ln)) ∧
                                     v ∈ eval_expr ge e t m x ∧ v ≠ Vundef ∧
                                     v ∈ toplift_gamma (ptset_gamma ge ((μ, (t, e)) :: σ)) ptr ∧
                                     v ∈ γ(ty) ∧
                                     (∃ ne n, In ne ln ∧ N.ncompat v n ∧ eval_nexpr ρn ne n ∧
                                              ∀ x y, ACtemp x y ∈ ne_free_var ne → In x (μ :: map fst σ)))
                       args vargs
           ) as VARGS.
    {
      clear PF.
      induction args as [|a args IH]. vauto2.
      destruct IH as (vargs & VARGS).
      intros x H. destruct (EARG _ (or_intror H)) as (y & Hy). vauto.
      destruct (EARG _ (or_introl eq_refl)) as (y & Hy).
      generalize (ab_eval_expr_correct (Just σ', (sz, (pt, (tp, nm)))) a).
      rewrite Hy. destruct σ' as [|(?, ε) ?]. intros ().
      ung. eq_some_inv. subst i i0 l x0.
      intros H. specialize (H e t σ m INV SZ _ TP _ CPT NM (env_ok _ _ H2) STK eq_refl PT).
      destruct y as [|((ptr, ty), ln)]. easy.
      hsplit.
      eexists. split. vauto.
      constructor; auto.
      apply bm_get_some in Hy. rewrite Hy.
      repeat eexists; eauto.
    }
    destruct VARGS as (vargs & VARGS & EARGS). specialize (PF _ VARGS).
    clear EARG.
    clear VARGS.
    unfold with_stk. simpl.
    match goal with |- appcontext[ ab_bind_parameters ?μ ?f ?a (ret_bm (NotBot ?t))] => set (ab_args := a); set (ab' := t) end.
    assert (((∀ t0, ∃ t1, bind_parameters fn.(fn_params) vargs t0 = Some t1) ∧ ∃ ab'', bm_get (ab_bind_parameters fi fn.(fn_params) ab_args (ret_bm (NotBot ab'))) = Some ab'')
            ∨ bm_get (ab_bind_parameters fi fn.(fn_params) ab_args (ret_bm (NotBot ab'))) = None) as T1.
    {
      clear PF.
      setoid_rewrite bind_parameters_zip.
      generalize (NotBot ab'). clear ab'.
      generalize dependent vargs.
      generalize dependent args.
      generalize dependent fn.(fn_params).
      intros formals.
      induction formals as [|id formals IH]; destruct args as [|q args]; intros ab_args vargs EARGS; simpl; try vauto2.
      inv EARGS. vauto.
      destruct vargs as [|v vargs]; inv EARGS.
      intros ab'.
      edestruct (IH args vargs). eauto.
      hsplit. vauto.
      hsplit. vauto.
    }
    destruct T1 as [(T1 & Hab'')|T1].
    2: destruct ab_bind_parameters as (?, [|]); simpl in T1; eq_some_inv; exact I.
    specialize (T1 (create_undef_temps fn.(fn_temps))). hsplit. specialize (PF t1 T1).
    rewrite PF; clear cs PF.
    apply bm_get_some in Hab''. rewrite Hab''.
    red. red. simpl. match goal with |- ?x ∈ γ _ => set (cs := x) end.
    assert (γ(ab_stk ab') (fst cs)) as G1.
    { subst cs ab' ab_args. ung. subst. simpl. constructor. constructor. reflexivity.
      exact (alloc_variables_build H1). vauto. }
    assert (invariant cs) as INV'.
    {
      subst cs. refine (push_frame_invariant _ _ H1 _ (INV _)).
      subst ab' ab_args. simpl in *. ung. subst.
      intros b [Hb|Hb]. eq_some_inv. subst. subst. exact (NOREC _ (or_introl eq_refl)).
      eq_some_inv. subst. destruct (stack_elt_gamma_in STK _ _ Hb) as (q & Q). exact (NOREC _ (or_intror Q)).
      easy.
    }
    red. red.
    assert (Just σ' ≠ All) as Q by discriminate. specialize (INV Q); clear Q.
    edestruct (λ K, @ab_bind_parameters_sound fi μ e t σ m fn.(fn_vars) _ _ INV K H1 fn.(fn_params) args ab_args vargs ρn)
      as ((stk' & sz' & pt' & tp' & nm') & -> & Hstk & Hsz & (Hpt & Htp) & ρn' & Hnm);
      eauto.
    {
      clear -NOREC STK.
      ung. subst. intros [ -> | X ]. exact (NOREC _ (or_introl eq_refl)).
      apply in_map_iff in X. destruct X as ((q & ?) & <- & X).
      destruct (stack_elt_gamma_in STK _ _ X) as (x & Y). exact (NOREC _ (or_intror Y)).
    }
    {
      subst ab_args.
      clear -EARGS.
      induction EARGS. constructor.
      simpl. hsplit. rewrite H. econstructor (eauto).
    }
    {
      subst ab' ab_args cs. simpl.
      assert (¬ (μ = fi ∨ In fi (map fst σ))) as Hμ.
      {
        intros [ -> | K ]. generalize INV'.(noRec).  inversion 1. intuition.
        apply in_map_iff in K. destruct K as ((?, ?) & ? & K). simpl in *. subst.
        ung. subst. eq_some_inv. subst.
        eapply stack_elt_gamma_in in K; eauto. hsplit. eapply NOREC; right; eassumption.
      }
      intros c. pose proof (PT c) as X. simpl. destruct (ACTreeDom.get c) as [|ty]. exact X.
      destruct c as [f' x'|f' x'|b]; simpl.
      destruct (eq_dec fi f'). subst f'. clear X.
      destruct (e1 ! x') as [(b, h)|] eqn: Hx.
      pose proof (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) H1)) as (_ & Y & _).
      specialize (Y _ _ Hx). destruct Y as [_ Y]. simpl in Y.
      rewrite (not_valid_block_load_None Y). exact I. exact I.
      apply push_frame_pt; auto.
      destruct (eq_dec fi f'). generalize (create_undef_temps_get fn.(fn_temps) x'). destruct ((create_undef_temps _) ! x').
      intros; hsplit; subst; exact I. intros _; exact I.
      apply push_frame_pt; auto.
      apply push_frame_pt; auto.
    }
    {
      subst ab' ab_args cs. simpl. intros c. pose proof (TP c) as X.
      simpl. destruct (ACTreeDom.get c tp) as [|ty]. exact X.
      destruct c as [f' x'|f' x'|b]; simpl.
      rewrite (@dec_eq_false _ _ _ fi f'). exact X.
      simpl in X. intro; subst f'. rewrite dec_eq_false in X. rewrite (proj2 (assoc_None _ _)) in X. AbTy.gamma_inv. discriminate.
      intros b K. ung. simpl in *. ung. subst. eq_some_inv. subst. eapply stack_elt_gamma_in in K; eauto. hsplit. eapply (NOREC); right; eassumption.
      intros <-. generalize INV'.(noRec). inversion 1; intuition.
      rewrite (@dec_eq_false _ _ _ fi f'). exact X.
      simpl in X. intro; subst f'. rewrite dec_eq_false in X. rewrite (proj2 (assoc_None _ _)) in X. AbTy.gamma_inv. discriminate.
      intros b K. ung. simpl in *. ung. subst. eq_some_inv. subst. eapply stack_elt_gamma_in in K; eauto. hsplit. eapply (NOREC); right; eassumption.
      intros <-. generalize INV'.(noRec). inversion 1; intuition.
      exact X.
    }
    {
      subst ab' ab_args cs. simpl.
      split; auto.
      intros c; destruct (CPT c) as [X Y]. split. 2: exact Y. clear Y.
      destruct c as [f' x'|f' x'|b]; simpl in *.
      destruct (eq_dec fi f'). subst f'.
      destruct (e1 ! x') as [(b, h)|] eqn: Hx.
      pose proof (alloc_var_inj _ _ (proj1 (alloc_var_alt _ _ _ _) H1)) as (_ & Y & _).
      specialize (Y _ _ Hx). destruct Y as [_ Y]. simpl in Y.
      rewrite (not_valid_block_load_None Y). vauto. vauto.
      exact X.
      destruct (eq_dec fi f'). generalize (create_undef_temps_get fn.(fn_temps) x'). destruct ((create_undef_temps _) ! x').
      intros; hsplit; subst; vauto. intros _; vauto.
      exact X. exact X.
    }
    split;[| | |split].
    + rewrite Hstk. subst ab'. econstructor; eauto. split. reflexivity.
      eapply alloc_variables_build; eauto.
    + intros _. exact INV'.
    + rewrite Hsz. simpl. eapply push_frame_sizes; eauto.
    + exact Hpt.
    + split. exact Htp. exists ρn'. eauto.
  Qed.

  (** A (concrete) block is valid given a stack iff in this block is allocated
    * either a global symbol s
    * or a local variable x of a function f *)
  Definition valid_block (stk: list (ident * (temp_env * env))) b : Prop :=
     (∃ s, Genv.invert_symbol ge b = Some s)
   ∨ (∃ f t e x sz, In (f, (t, e)) stk ∧ e ! x = Some (b, sz)).

  Lemma valid_block_tmp stk stk' :
    stacks_tmp stk stk' →
    valid_block stk' ⊆ valid_block stk.
  Proof.
    intros H b [Hb | Hb];[left; exact Hb|right].
    hsplit. apply (stacks_tmp_in H) in Hb.
    hsplit. vauto.
  Qed.

  Lemma pop_frame_pt μ₀ τ₀ ε₀ σ m m' (pt: points_to) c:
    (∀ b, b ∈ valid_block σ →
          (∀ κ o, Mem.load κ m b o = Mem.load κ m' b o)
    ) →
    mem_as_fun ge ((μ₀, (τ₀, ε₀)) :: σ, m) c ∈
               γ (gamma_op := toplift_gamma (ptset_gamma ge ((μ₀, (τ₀, ε₀)) :: σ))) (ACTreeDom.get c pt) →
    mem_as_fun ge (σ, m') c ∈
               γ (gamma_op := toplift_gamma (ptset_gamma ge σ)) (ACTreeDom.get c (pop_local_pt μ₀ pt)).
  Proof.
    intros Hb.
    unfold ACTreeDom.get. rewrite pop_local_pt_get.
    destruct (ACTree.get c pt) as [bs|]. 2: exact id. simpl.
    destruct (is_local_cell μ₀ c) eqn: GL. exact (λ _, I). simpl.
    destruct (BSO.existsb (is_local_block μ₀) bs) eqn: EX. exact (λ _, I). simpl.
    destruct c as [f x ofs|f x|b ofs].
    - simpl in *. destruct (eq_dec μ₀). simpl in *. eq_some_inv.
      destruct (assoc f σ) as [(t, e)|] eqn: Hv. 2: exact id.
      destruct (e ! x) as [(b, s)|] eqn: Hv'. 2: exact id. rewrite Hb.
      destruct (Mem.load (proj1_sig κ) m' b ofs) as [()|]; simpl; try exact id.
      intros (b' & Hb' & Hb''). exists b'. split. exact Hb'.
      destruct b' as [g y|b''']; simpl in *; eq_some_inv.
      2: exact Hb''.
      rewrite dec_eq_false in Hb''. exact Hb''.
      rewrite BSO.existsb_forall in EX. specialize (EX _ Hb'). simpl in EX.
      destruct (eq_dec μ₀); simpl in EX; eq_some_inv. assumption.
      right. repeat eexists. exact (assoc_in _ _ Hv). exact Hv'.
    - simpl in *. destruct (eq_dec μ₀). simpl in GL; eq_some_inv.
      destruct (assoc f σ) as [(t, e)|]. 2: exact id.
      destruct (t ! x) as [()|]; try exact id.
      intros (b' & Hb' & Hb''). exists b'. split. exact Hb'.
      destruct b' as [g y|b''']; simpl in *; eq_some_inv.
      2: exact Hb''.
      rewrite dec_eq_false in Hb''. exact Hb''.
      rewrite BSO.existsb_forall in EX. specialize (EX _ Hb'). simpl in EX.
      destruct (eq_dec μ₀); simpl in EX; eq_some_inv. assumption.
    - simpl. destruct (Genv.invert_symbol ge b) as [s|] eqn: Hs. 2: reflexivity.
      rewrite Hb.
      destruct (Mem.load (proj1_sig κ) m' b ofs) as [()|]; simpl; try exact id.
      intros (b' & Hb' & Hb''). exists b'. split. exact Hb'.
      destruct b' as [g y|b''']; simpl in *; eq_some_inv.
      2: exact Hb''.
      rewrite dec_eq_false in Hb''. exact Hb''.
      rewrite BSO.existsb_forall in EX. specialize (EX _ Hb'). simpl in EX.
      destruct (eq_dec μ₀); simpl in EX; eq_some_inv. assumption.
      left. vauto.
  Qed.

  Lemma pop_frame_tp μ₀ τ₀ ε₀ σ m m' (tp: types) c:
    (∀ b, b ∈ valid_block σ →
          (∀ κ o, Mem.load κ m b o = Mem.load κ m' b o)
    ) →
    mem_as_fun ge ((μ₀, (τ₀, ε₀)) :: σ, m) c ∈
               γ(ACTreeDom.get c tp) →
    mem_as_fun ge (σ, m') c ∈
               γ (ACTreeDom.get c (pop_local_tp μ₀ tp)).
  Proof.
    intros Hb.
    unfold ACTreeDom.get. rewrite pop_local_tp_get.
    destruct (is_local_cell μ₀ c) eqn: GL. exact (λ _, I).
    destruct (ACTree.get c tp) as [ty|]. 2: exact id.
    destruct c as [f x ofs|f x|b ofs]; simpl in *.
    - destruct (eq_dec μ₀). simpl in GL. eq_some_inv.
      destruct (assoc f σ) as [(t, e)|] eqn: Hv. 2: exact id.
      destruct (e ! x) as [(b, s)|] eqn: Hv'. 2: exact id. rewrite Hb.
      destruct (Mem.load (proj1_sig κ) m' b ofs) as [()|]; simpl; try exact id.
      right. repeat eexists. exact (assoc_in _ _ Hv). exact Hv'.
    - destruct (eq_dec μ₀). simpl in GL; eq_some_inv. exact id.
    - destruct (Genv.invert_symbol ge b) as [s|] eqn: Hs. 2: exact id.
      rewrite Hb. exact id. left. vauto.
  Qed.

  Lemma pop_frame_compat μ₀ τ₀ ε₀ σ m m' (tp: types):
    list_norepet (List.map fst ((μ₀, (τ₀, ε₀)) :: σ)) →
    (∀ b, b ∈ valid_block σ →
          (∀ κ o, Mem.load κ m b o = Mem.load κ m' b o)
    ) →
    compat tp (mem_as_fun ge ((μ₀, (τ₀, ε₀)) :: σ, m)) ⊆
    compat (pop_local_tp μ₀ tp) (mem_as_fun ge (σ, m')).
  Proof.
    intros NR Hb ρ H c. split.
    - generalize (proj1 (H c)). destruct c as [f x|f x|b]; simpl.
      + destruct (eq_dec μ₀) as [->|NE].
        * intros _. rewrite (proj2 (assoc_None f σ)). vauto.
          clear -NR. intros b K. inv NR. elim H1. rewrite in_map_iff. exists (f, b). auto.
        * destruct (assoc f σ) as [(t,e)|] eqn: Hv. 2: exact id.
          destruct (e ! x) as [(b, s)|] eqn: Hv'. 2: exact id.
          rewrite Hb. exact id.
          right. repeat eexists; eauto. exact (assoc_in _ _ Hv).
      + destruct (eq_dec μ₀) as [->|NE].
        * intros _. rewrite (proj2 (assoc_None f σ)). vauto.
          clear -NR. intros b K. inv NR. elim H1. rewrite in_map_iff. exists (f, b). auto.
        * destruct (assoc f σ) as [(t,e)|]. 2: exact id.
          destruct (e ! x) as [(b, s)|]; exact id.
      + destruct (Genv.invert_symbol ge b) eqn: Hv. 2: exact id.
          rewrite Hb. exact id.
          left. vauto.
    - generalize (proj2 (H c)). simpl. unfold N.Eval.TDom.get. rewrite pop_local_tp_get.
      bif. auto.
  Qed.

  Lemma pop_frame_invariant μ₀ τ₀ ε₀ σ m m':
    invariant ((μ₀, (τ₀, ε₀)) :: σ, m) →
    (∀ b, b ∈ valid_block σ →
          (∀ lo hi p, Mem.range_perm m b lo hi Cur p → Mem.range_perm m' b lo hi Cur p) ∧
          (b ∈ Mem.valid_block m → b ∈ Mem.valid_block m')
    ) →
    invariant (σ, m').
  Proof.
    intros INV HB.
    split.
    - intros f t0 e H b hi H0. simpl. apply HB. simpl in H. right.
      unfold blocks_of_env in H0. rewrite in_map_iff in H0. destruct H0 as ((x & b' & s') & Hb' & Hx).
      simpl in Hb'. eq_some_inv. subst.
      apply PTree.elements_complete in Hx. repeat econstructor; eauto.
      eapply INV. right. exact H. easy.
    - intros f t0 e H. eapply INV. right. exact H.
    - intros x x' a H H'.
      pose proof (INV.(noRec)) as NR.
      generalize (INV.(@localInj _) x x' a).
      destruct x as (f, x), x' as (f', x'). simpl in *.
      destruct (eq_dec μ₀ f). rewrite (proj2 (assoc_None _ _)) in H. eq_some_inv.
      intros b K. inv NR. apply H2. rewrite in_map_iff. eexists (_, _).  vauto.
      destruct (eq_dec μ₀ f'). rewrite (proj2 (assoc_None _ _)) in H'. eq_some_inv.
      intros b K. inv NR. apply H2. rewrite in_map_iff. eexists (_, _).  vauto.
      eauto.
    - simpl. pose proof INV.(noRec) as NR. inv NR. easy.
    - intros x b H. generalize (INV.(globalValid) _ H). simpl.
      refine (proj2 (HB _ _)).
      left. apply Genv.find_invert_symbol in H. vauto.
    - intros f t0 e H x b z H0. apply HB. vauto2. eapply INV.(localValid); eauto. right. eassumption.
  Qed.

  Lemma pop_frame_sizes μ₀ τ₀ ε₀ σ m m' (sz: sizes) ε':
    invariant ((μ₀, (τ₀, ε₀)) :: σ, m) →
    (∀ b, b ∈ valid_block σ →
          (∀ lo hi p, Mem.range_perm m b lo hi Cur p → Mem.range_perm m' b lo hi Cur p)
    ) →
    m ∈ sizes_gamma ge ((μ₀, (τ₀, ε₀)) :: σ) sz →
    m' ∈ sizes_gamma ge σ (clear_local_sizes μ₀ ε' sz).
  Proof.
    intros INV HB SZ.
    intros b hi rw b'.
    rewrite clear_local_sizes_get.
    intros Hhi Hb.
    eapply HB;[|eapply (SZ b hi rw b')];
    destruct b as [f x|b];
    simpl in Hb; eq_some_inv; auto.
    destruct (assoc f σ) as [(t,e)|] eqn: Hv; eq_some_inv.
    destruct (e ! x) as [(b,s)|] eqn: Hv'; eq_some_inv. right. subst. repeat econstructor. exact (assoc_in _ _ Hv). exact Hv'.
    left. destruct (Genv.invert_symbol ge b) eqn: Hs; eq_some_inv; subst. vauto.
    revert Hhi. bif.
    simpl. destruct (eq_dec μ₀ f). 2: exact Hb.
    rewrite (proj2 (assoc_None _ _)) in Hb. eq_some_inv.
    pose proof INV.(noRec) as NR.
    intros b K. inv NR. apply H1. rewrite in_map_iff. eexists (_, _).  vauto.
  Qed.

  Lemma free_list_ex μ₀ tmp env stk m :
    invariant ((μ₀, (tmp, env)) :: stk, m) →
    ∃ m', Mem.free_list m (blocks_of_env env) = Some m' ∧
          ∀ b, valid_block stk b →
               (∀ κ ofs, Mem.load κ m b ofs = Mem.load κ m' b ofs) ∧
               (∀ lo hi p, Mem.range_perm m b lo hi Cur p → Mem.range_perm m' b lo hi Cur p) ∧
               (b ∈ Mem.valid_block m → b ∈ Mem.valid_block m').
  Proof.
    intros INV.
    pose proof INV.(noRec) as NR.
    simpl in *.
    assert (∀ x x' b, option_lift fst env ! x = Some b → option_lift fst env ! x' = Some b → x = x') as INJ.
    { intros x x' b H H0.
      pose proof (INV.(@localInj _) (μ₀, x) (μ₀, x') b) as INJ'.
      simpl in INJ'. rewrite !dec_eq_true in INJ'. simpl in INJ'.
      specialize (INJ' H H0). eq_some_inv. assumption.
    }
    assert (∀ x f' x' b b', option_lift fst env ! x = Some b →
                            (do_opt t_e <- assoc f' stk;
                            option_lift fst (snd t_e) ! x') = Some b' →
                            b ≠ b') as INJ₀.
    { intros x f' x' b b' H H0 <-.
      assert (μ₀ ≠ f'). intros ->.
      generalize (assoc_in f' stk).
      destruct (assoc f' stk). intros X. specialize (X _ eq_refl).
      inv NR. apply H3. rewrite in_map_iff. eexists (f', _). vauto.
      eq_some_inv.
      pose proof (INV.(@localInj _) (μ₀, x) (f', x') b) as INJ'.
      simpl in INJ'. rewrite dec_eq_true in INJ'.
      rewrite dec_eq_false in INJ'. simpl in INJ'.
      specialize (INJ' H H0). eq_some_inv. auto. assumption.
    }
    assert (injective_map (λ x_v : ident * positive,
            do_opt t_e <- assoc (fst x_v) stk;
            option_lift fst (snd t_e) ! (snd x_v))) as INJ₁.
    { intros (f, x) (f', x'). simpl. intros a H H0.
      assert (μ₀ ≠ f). intros ->.
      generalize (assoc_in f stk).
      destruct (assoc f stk). intros X. specialize (X _ eq_refl).
      inv NR. apply H3. rewrite in_map_iff. eexists (f, _). vauto.
      eq_some_inv.
      assert (μ₀ ≠ f'). intros ->.
      generalize (assoc_in f' stk).
      destruct (assoc f' stk). intros X. specialize (X _ eq_refl).
      inv NR. apply H4. rewrite in_map_iff. eexists (f', _). vauto.
      eq_some_inv.
      pose proof (INV.(@localInj _) (f, x) (f', x') a) as INJ'. simpl in *.
      rewrite !dec_eq_false in INJ'; auto.
    }
    generalize (INV.(localFreeable) _ _ _ (or_introl eq_refl)).
    generalize (INV.(localNotGlobal) _ _ _ (or_introl eq_refl)).
    setoid_rewrite (ptree_get_assoc env).
    generalize (λ f t e H, INV.(localFreeable) f t e (or_intror H)).
    generalize (λ f t e H, INV.(localNotGlobal) f t e (or_intror H)).
    setoid_rewrite (ptree_get_assoc env) in INJ.
    setoid_rewrite (ptree_get_assoc env) in INJ₀.
    generalize (PTree.elements_keys_norepet env).
    unfold blocks_of_env in *.
    generalize dependent (PTree.elements env).
    simpl.
    clear INV env.
    intros l.
    revert m.
    induction l as [|(x, (b, hi)) l IH]. intros m _ _ _ _. simpl. vauto.
    intros m INJ INJ₀ NR' LG LF LG' LF'.
    simpl in *. inv NR'. inv NR.
    destruct (Mem.free m b 0 hi) as [m₁|] eqn: M₁.
    destruct (IH m₁) as (m' & M' & IH'); clear IH.
  + intros x0 x' b0 H H0.
    specialize (INJ x0 x' b0).
    assert (x ≠ x0).
    { intros ->.
      generalize (assoc_in x0 l). destruct (assoc x0 l).
      intros X. specialize (X _ eq_refl). apply H1.
      rewrite in_map_iff. eexists (x0, _). vauto.
      simpl in *. eq_some_inv.
    }
    assert (x ≠ x').
    { intros ->.
      generalize (assoc_in x' l). destruct (assoc x' l).
      intros X. specialize (X _ eq_refl). apply H1.
      rewrite in_map_iff. eexists (x', _). vauto.
      simpl in *. eq_some_inv.
    }
    rewrite !dec_eq_false in INJ; auto.
  + intros x0 f' x' b0 b' H H0.
    specialize (INJ₀ x0 f' x' b0 b').
    rewrite dec_eq_false in INJ₀.
    eauto. intros ->.
    generalize (assoc_in x0 l). destruct (assoc x0 l).
    intros X. specialize (X _ eq_refl). apply H1.
    rewrite in_map_iff. eexists (x0, _). vauto.
    simpl in *. eq_some_inv.
  + eauto.
  + eauto.
  + intros f t0 e H b0 hi0 H0 ofs X.
    specialize (LF _ _ _ H _ _ H0 ofs X).
    eapply Mem.perm_free_1; eauto. left.
    rewrite in_map_iff in H0. destruct H0 as ((x' & b' & sz) & ? & H').
    simpl in *. eq_some_inv. subst.
    cut (b ≠ b0). auto.
    apply (INJ₀ x f x'). rewrite dec_eq_true. reflexivity.
    rewrite (In_norepet_assoc _ _ _ H4 H). simpl.
    rewrite ptree_get_assoc.
    apply (In_norepet_assoc _ _ _ (PTree.elements_keys_norepet _)) in H'.
     unfold PTree.elt in *. rewrite H'. reflexivity.
  + intros x0 b0 y b' z H H0.
    eapply (LG' _ _ y); try eassumption.
    rewrite dec_eq_false. eassumption.
    intros ->. apply H1.
    apply assoc_in in H0. rewrite in_map_iff.
    eexists (y, _). vauto.
  + intros b0 hi0 H ofs Hofs.
    eapply Mem.perm_free_1; eauto. left.
    rewrite in_map_iff in H. destruct H as ((x' & b' & sz) & H & H').
    simpl in *. eq_some_inv. intro; subst.
    destruct (eq_dec x x'). subst.
    apply H1. rewrite in_map_iff. eexists (x', _). vauto.
    specialize (INJ x x' b).
    apply (In_norepet_assoc _ _ _ H2) in H'.
    rewrite dec_eq_true, dec_eq_false in INJ; auto.
    unfold PTree.elt in *.
    rewrite H' in INJ.
    specialize (INJ eq_refl eq_refl).
    auto.
    eapply LF'. vauto. eauto.
  + exists m'. split. easy.
    intros b0 [(s & H0)|(f & t & e & y & sz & LOC & H0)].
    - assert (b0 ≠ b).
      { eapply LG'; eauto. apply Genv.invert_find_symbol; eauto.
        rewrite dec_eq_true. reflexivity. }
      split;[|split].
    * intros κ ofs. etransitivity. 2: eapply IH'; vauto. clear IH'.
      symmetry. eapply Mem.load_free; eauto.
    * intros lo0 hi0 p H'.
      eapply IH'. vauto.
      intros ofs Hofs; eapply Mem.perm_free_1; eauto.
    * intros H5. eapply IH'. vauto. eapply Mem.valid_block_free_1; eauto.
    - assert (b0 ≠ b).
      { generalize (INJ₀ x f y b b0). rewrite dec_eq_true. intros X.
        specialize (X eq_refl).
        apply (In_norepet_assoc _ _ _ H4) in LOC. rewrite LOC in X.
        simpl in X. rewrite H0 in X.
        specialize (X eq_refl).
        auto. }
      split;[|split].
    * etransitivity. 2: eapply IH'; eauto. clear IH'.
      symmetry. eapply Mem.load_free; eauto.
      vauto2.
    * intros lo0 hi0 p H5.
      eapply IH'. vauto2.
      intros ofs Hofs; eapply Mem.perm_free_1; eauto.
    * intros H5. eapply IH'. vauto2. eapply Mem.valid_block_free_1; eauto.
  +
    Transparent Mem.free.
    unfold Mem.free in M₁.
    Opaque Mem.free.
    destruct Mem.range_perm_dec; eq_some_inv. intuition.
  Qed.

  Lemma pop_frame_sound :
   ∀ (ret : option expr) (rettemp : option ident) (ab : t)
   (a : option concrete_state),
   PopFrame ge ret rettemp (γ ab) a
   → γ (AbMemSignatureCsharpminor.pop_frame ret rettemp ab) a.
  Proof.
    Hint Resolve env_ok.
    intros ret rettemp (σ & sz & pt & tp & nm) a H.
    destruct H as [μ tmp env stk m H PF].
    destruct σ as [|[|(μ₀, ε₀) σ]]. exact I. exact I.
    ung. simpl in STK. ung. eq_some_inv. subst i l x0 μ.
    destruct σ as [|(μ₁, ε₁) σ].
  - (* return from main *)
    destruct ret as [ret|]. 2: exact I.
    destruct rettemp as [ [| |]|]. exact I. exact I. 2: exact I.
    ung. subst.
    simpl. unfold pop_frame; simpl.
    generalize (ab_eval_expr_correct (Just ((μ₀, ε₀) :: nil), (sz, (pt, (tp, nm)))) ret).
    destruct (ab_eval_expr(Just ((μ₀, ε₀) :: nil), (sz, (pt, (tp, nm)))) ret)
      as ([|((pt' & ty) & ln)], [|]) eqn: RET'.
    intros X; eelim X; eauto; vauto.
    intros _. setoid_rewrite RET'. exact I.
    intros X. simpl in X.
    setoid_rewrite RET'. simpl.
    destruct ty as [|()]; try exact I. simpl.
    clear RET'.
    edestruct X as (v & V & Def & (_ & TY) & _); try eassumption; clear X.
    eauto. vauto. reflexivity. AbTy.gamma_inv. hsplit. subst.
    specialize (λ m' M', PF m' M' _ V eq_refl _ eq_refl).
    destruct (free_list_ex INV) as (m' & Hm' & Hm''). specialize (PF _ Hm'). subst a.
    split;[| | |split;[|split]].
  + vauto.
  + intros _. split.
    intros ? ? ? ().
    intros ? ? ? ().
    simpl. repeat intro; eq_some_inv. vauto.
    intros x b H. refine (proj2 (proj2 (Hm'' _ _)) _). left. erewrite Genv.find_invert_symbol; eauto. apply (INV.(globalValid) _ H).
    intros ? ? ? ().
  + eapply pop_frame_sizes; eauto. simpl. eapply Hm''.
  + intros c. refine (pop_frame_pt μ₀ tmp env m m' pt c _ (PT c)). eapply Hm''.
  + intros c. refine (pop_frame_tp μ₀ tmp env m m' tp c _ (TP c)). eapply Hm''.
  + exists ρn. split; auto. simpl.
    refine (pop_frame_compat m' _ _ CPT). eapply INV. eapply Hm''.
  + intros _. setoid_rewrite RET'. destruct ty as [|()]; exact I.
  - (* regular case *)
    simpl. unfold pop_frame. simpl.
    destruct (free_list_ex INV) as (m' & Hm' & Hm''). specialize (PF _ Hm').
    destruct ret as [ret|].
    destruct rettemp as [rettemp|].
    + apply bm_bind_case. intros; exact I.
      intros ab' Hab'.
      generalize (assign_in_correct μ₁ rettemp ret (Just ((μ₀, ε₀) :: (μ₁, ε₁) :: σ), (sz, (pt, (tp, nm))))).
      fold stack; fold tnum.
      rewrite Hab'. simpl.
      intros X. specialize (X ((μ₀, (tmp, env)) :: stk) m).
      exploit X; clear X. econstructor; vauto. eauto. eauto.
      destruct ab' as [|(σ' & sz' & pt' & tp' & nm')]. intros ().
      intros. hsplit. simpl in *. subst sz' σ'. eq_some_inv. subst env0 tmp0 μ₀0 stk'. ung. subst i stk.
      destruct assoc_dep as [IN|NI].
      2: eelim NI; right; left; reflexivity.
      destruct IN as (((pre & (tmp', env')) & post) & EQ & IN).
      hsplit. simpl in *. destruct pre as [|? [|? ?]]; simpl in *;
      eq_some_inv. subst. pose proof INV.(noRec) as NR. inv NR. exfalso. intuition.
      2: subst; eelim IN; right; left; reflexivity.
      subst. clear STK SZ PT TP CPT NM ρn. ung. simpl in *. ung. eq_some_inv. subst l i x0 i0. eq_some_inv. subst x1 l0.
      erewrite PF; eauto. clear PF.
      assert (∀ b : block, valid_block ((μ₁, (PTree.set rettemp v tmp', env')) :: post) b →
                           ∀ (κ : memory_chunk) (o : Z), Mem.load κ m b o = Mem.load κ m' b o) as Hb'.
      { intros b ?. refine (proj1 (Hm'' b _)). eapply valid_block_tmp; eauto. reflexivity. }
      assert (∀ b : block, valid_block ((μ₁, (PTree.set rettemp v tmp', env')) :: post) b →
                           (∀ (lo hi : Z) p, Mem.range_perm m b lo hi Cur p → Mem.range_perm m' b lo hi Cur p) ∧
                           (b ∈ Mem.valid_block m → b ∈ Mem.valid_block m')) as Hb''.
      { intros b ?. eapply Hm''; eauto. eapply valid_block_tmp; eauto. reflexivity. }
      constructor; vauto;[| |split;[|split]].
      * intros _. eapply pop_frame_invariant; eauto.
      * eapply pop_frame_sizes; eauto. intros b Q. exact (proj1 (Hb'' _ Q)).
      * intros c. generalize (PT c). simpl. apply pop_frame_pt. exact Hb'.
      * intros c. generalize (TP c). simpl. apply pop_frame_tp. exact Hb'.
      * exists ρn. split. 2: exact NM.
        generalize CPT. apply pop_frame_compat. eapply INV.(noRec). exact Hb'.
    + match goal with |- context[ nonblock ?e ?ab ] => generalize (nonblock_correct e ab) end.
      destruct nonblock as ((), [|]).
      2: exact (λ _, I).
      simpl in *. intros H. edestruct H as (v & V & Def). econstructor; simpl; eauto. vauto.
      intuition eauto. vauto. clear H.
      specialize (PF _ V).
      ung. subst. subst.
      constructor; vauto;[| |split;[|split]].
      * intros _. eapply pop_frame_invariant; eauto. eapply Hm''.
      * eapply pop_frame_sizes; eauto. simpl. eapply Hm''.
      * intros c. generalize (PT c). simpl. apply pop_frame_pt. eapply Hm''.
      * intros c. generalize (TP c). simpl. apply pop_frame_tp. eapply Hm''.
      * exists ρn. split. 2: exact NM.
        generalize CPT. apply pop_frame_compat. eapply INV. eapply Hm''.
    + destruct rettemp. exact I.
      specialize (PF _ eq_refl). ung. subst. subst. simpl in *.
      constructor; vauto;[| |split;[|split]].
      * intros _. eapply pop_frame_invariant; eauto. eapply Hm''.
      * eapply pop_frame_sizes; eauto. eapply Hm''.
      * intros c. generalize (PT c). simpl. apply pop_frame_pt. eapply Hm''.
      * intros c. generalize (TP c). simpl. apply pop_frame_tp. eapply Hm''.
      * exists ρn. split. 2: exact NM.
        generalize CPT. apply pop_frame_compat. eapply INV. eapply Hm''.
  Qed.

  Existing Instance ListIncl.
  Lemma deref_fun_sound :
   ∀ (e : expr) (ab : t) (a : option (ident * fundef)),
   DerefFun ge e (γ ab) a → γ (deref_fun e ab) a.
  Proof.
    intros e ([|[|(μ,ε) σ]] & sz & pt & (tp & nm)) fifd DF; try exact I.
    destruct DF as [f tmp en stk m H DF].
    ung. simpl in *. ung. eq_some_inv. subst i l f.
    unfold deref_fun. simpl.
    apply bm_bind_case. exact (λ _ _ _, I).
    intros cells Cells.
    generalize (λ U PT, @realize_cells_correct μ ε sz pt tp nm en tmp _ stk (env_ok ε en U) INV SZ TP PT ρn NM CPT (exist _ Mint32 I) e).
    rewrite Cells. simpl.
    intros X. edestruct X as (NB & Y); auto.
    destruct (nonblock_eval_expr_def NB) as (v & DEF & EV).
    destruct (Y v EV) as (ab & b & i & Hv & Habi & Hab & Hal). clear X Y.
    with_bm_map'.
    intros (_ & L).
    destruct (L (cell_from ab (Int.unsigned i, exist _ Mint32 I)) Habi) as ((f, fd) & Hf & K); clear L.
    destruct ab as [|b']; try discriminate.
    simpl in Hf.
    simpl in Hab. destruct (Genv.invert_symbol ge b') eqn: IS; eq_some_inv. subst.
    destruct (Int.unsigned i) eqn: Hi; try discriminate.
    specialize (DF b f).
    destruct (Genv.find_funct_ptr ge b). 2: discriminate.
    inv Hf.
    apply (f_equal Int.repr) in Hi. rewrite Int.repr_unsigned in Hi. subst i.
    rewrite (DF _ EV eq_refl). exact K.
    apply Genv.invert_find_symbol, IS.
  Qed.

  Lemma init_mem_sound :
   ∀ (defs : list (ident * globdef fundef unit)) (a : option concrete_state),
   Init_Mem ge defs a → γ (AbMemSignatureCsharpminor.init_mem defs) a.
  Proof.
    intros defs a (prog & Hge & Hdefs & H). simpl. subst defs.
    unfold init_mem.
    eapply bm_assert_spec. constructor. intro.
    apply bm_bind_case. constructor. intros [b ab] AAG. simpl.
    edestruct ab_alloc_globals_ex as (m & EQm & NEXT). eauto.
    erewrite H by eauto.
    rewrite Hge in EQm. apply norepet_map_iff in H0.
    eapply ab_alloc_globals_ok in AAG; eauto.
    destruct AAG.
    destruct ab as [|ab]. contradiction.
    destruct ab as (stk, (sz, (pt, (tp, nm)))).
    destruct H2 as (? & ? & ? & ? & ?).
    simpl in *. subst stk.
    split;[| | |split;[|split]]; auto.
  + vauto.
  + intros _. split; try easy. vauto. simpl.
    intros x b0 Hb0. subst ge.
    eapply Genv.find_symbol_not_fresh; eauto.
  Qed.

  Lemma mem_cshm_dom_correct : mem_dom_spec ge mem_cshm_dom t_gamma.
  Proof.
    constructor.
  - exact t_adom.

  - (* assign *)
    exact assign_sound.

  - exact assign_any_sound.

  - (* store *)
    exact store_sound.

  - (* assume *)
    exact assume_sound.

  - (* non block *)
    exact nonblock_sound.
  - (* deref fun *)
    exact deref_fun_sound.

  - (* pop frame *)
    exact pop_frame_sound.
  - (* push frame *)
    exact push_frame_sound.
  - (* malloc *)
    intros; exact I.
  - (* free *)
    intros; exact I.
  - (* memcpy *)
    intros; exact I.
  - (* init_mem *)
    exact init_mem_sound.
  Qed.

End num.
